/*
 * Created on 25 juin 2003
 *
 */
package com.rational.apidoc.chart.internal;

import java.text.DecimalFormat;
import java.util.Vector;

import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.graphics.FontMetrics;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.SWT;
import com.rational.apidoc.core.*;
import com.rational.apidoc.chart.IChartLocation;
import com.rational.apidoc.chart.IChartRender;
import com.rational.apidoc.drawutil.HSV;
import com.rational.apidoc.drawutil.DrawUtil;
import com.rational.apidoc.drawutil.Rect;
import com.rational.apidoc.drawutil.Point2D;

/**
 * ChartRender is made to 
 * - draw DGraphic item in an area
 * - locate a point on a DGraphic area.
 */
public class DefaultChartRender implements IChartRender
{
  /** margin for render */
  protected int margin_; 
  /** Default style to draw */
  protected DChartStyle default_;
 
  /** temporary data during draw/locate */
  protected Data d_ ; 
  /** used to dispose font/color from styles (MUST BE disposed for windows systems)
   *  (?huh? work'in java and take care of windows ... smell bad (but smell'SWT!) */
  protected Vector styleof_;

  /** Create a chart render */
  public DefaultChartRender()
  {
    margin_=10;
    default_ = new DChartStyle();
    d_=null;
  }
  
  //paire courbe / style
  private static class CurveStyle
  {
    public DCurve curve_;
    public DChartStyle style_;
    public CurveStyle( DCurve _c, DChartStyle _s )
    {
      curve_=_c; style_=_s;
    }
  }
  
  private static class Data
  {
     
    public boolean drawing() { return draw_; }
    public boolean locating() { return !draw_; }
    
    public boolean draw_;
    public GC gc_; //mode drawing
    public Font default_font_;
    public int lx_, ly_; //mode locating
    public DGraphic G_; 
    public Rect R_;
    public Vector curves_; //list of CurveStyle
    
    public Data( boolean _draw, DGraphic _g, Rect _R, GC _gc )
    {
      gc_=_gc;
      draw_=_draw;
      G_=_g;
      R_=_R;
      lx_=ly_=-1;
      curves_ = new Vector();
      default_font_ = _gc.getFont();
    }        

    static boolean horizontal3DContains( int _lx, int _ly, int _w, int _3d, int _px, int _py )
    {
      //? si (_px,_py) est contenus dans le rectangle cisaille en X:
      //    _lx+_3d
      //       +---------+ _ly-_3d
      //      /         /
      // _ly +---------+
      //    _lx     _lx+_w
      //!! les Y ecrans sont inverses
      if( _py < (_ly-_3d) ) return false;
      if( _py >  _ly      ) return false;
      if( _py==_ly )
      {
        return (_px>=_lx)&&(_px<=(_lx+_w));
      }
      if( _py==(_ly-_3d))
      {
        return (_px>=(_lx+_3d))&&(_px<=(_lx+_w+_3d));
      }
      //dans le rectangle cisaille en X
      int dx = _ly-_py; //OK: car proj. 3d en pente 45 deg.
      return (_px>=(_lx+dx))&&(_px<=(_lx+dx+_w));
    }
    
    static boolean vertical3DContains( int _lx, int _ly, int _h, int _3d, int _px, int _py )
    {
      //? si (_px,_py) est contenus dans le rectangle cisaille en Y:
      //       _lx+_3d
      //          + _ly-_3d
      //         /|
      //    _ly + |
      //        | |
      //        | +
      //        |/
      // _ly+_h +
      //      _lx
      if( _px <  _lx     ) return false;
      if( _px > (_lx+_3d)) return false;
      if( _px == _lx      ) return (_py>=_ly)&&(_py<=(_ly+_h));
      ////if( _px == (_lx+_3d)) return (_py>=(_ly-_3d))&&(_py<=(_ly-_3d+_h));
      //!! les Y ecrans sont inverses
      //dans le rectangle cisaille en Y
      int dy = _px-_lx; //OK: car proj. 3d en pente 45 deg.
      int npy = _py+dy;
     return (npy>=_ly)&&(npy<=(_ly+_h));
    }

  }
  
  protected static class ChartLocation extends Throwable
  {
    public IChartLocation location_;
    public ChartLocation( IChartLocation _l )
    {
      location_=_l;
    }
    public ChartLocation( String _id, IDItem _item, Rect _rect )
    {
      location_=new DChartLocation( _id, _item, _rect );
    }
    public ChartLocation( String _id, IDItem _item, Rect _rect, String _dtxt )
    {
      location_=new DChartLocation( _id, _item, _rect, _dtxt );
    }
  }
  
  /** @return true since default render can draw anything */
  public boolean canRender( String _graph_type )
  {
    return true;
  }
  
  /** Draw graphic _g on rectangle _R using GC _gc. */
  public void draw( GC _gc, Rectangle _R, DGraphic _g )
  {
    if( _g==null ) return ;
    d_ = new Data( true, _g, new Rect(_R), _gc );
    try {
      render();
    }
    catch( ChartLocation _l) {}//musn't appears
    d_ = null;
  }
  
  /** Return a location for the given point, GC must be provided to compute text extends.
   * can return null if no location is found.
   */
  public IChartLocation locate( int _x, int _y, GC _gc, Rectangle _R, DGraphic _G )
  {
    if(_G==null) return null;
    if(!_R.contains(_x,_y) ) return null;

    d_ = new Data( false, _G, new Rect(_R), _gc );
    d_.lx_ = _x;
    d_.ly_ = _y;
    
    try {
      render();
    }
    catch( ChartLocation _l )
    {
      return _l.location_;
    }
    finally
    {
      d_=null;
    }
    return null;
  }
  
  /** Return style for item, walk upstair in tree to try to complete style. */
  protected DChartStyle styleOf( IDItem _item )
  {
    IDStyle apidoc_style=null;
    IDFont  apidoc_font =null;
    IDColor apidoc_color=null;
    DChartStyle style=new DChartStyle();
    char missing = 0x07; //font|fore|back

    for( IDItem item=_item; item!=null; item=item.getParent() )
    {
      if( (apidoc_style=item.getStyle())!=null )
      {
        //font
        if ( (missing & 0x01)!=0 && (apidoc_font=apidoc_style.getFont())!=null ) 
        {          
          style.font_ = DrawUtil.getFont( apidoc_font ); 
          missing &= ~0x01 ;       
        }
        //fore color
        if ( (missing & 0x02)!=0 && (apidoc_color=apidoc_style.getForeColor())!=null )
        {
          style.fore_ = DrawUtil.getColor( apidoc_color );
          missing &= ~0x02 ;
        }
        //back color
        if ( (missing & 0x04)!=0 && (apidoc_color=apidoc_style.getBackColor())!=null )
        {
          style.back_ = DrawUtil.getColor( apidoc_color );
          missing &= ~0x04 ;
        }
        //style completed.
        if( missing == 0x00 ) break ;
      }
    }//for

    if( styleof_==null ) styleof_ = new Vector();
    styleof_.add( style );
    return style;
  }
  
  protected void disposeStyleOf()
  {
    if( styleof_==null ) return ;
    for( int i=0; i<styleof_.size(); ++i )
    {
      DChartStyle s = (DChartStyle)styleof_.get(i);
      s.dispose();
    }
    styleof_.clear();
  }
  
  /** print error occured while rendering chart. */
  void error( String _message )
  {
//TBD:juste pour mise au point: affichage de ce qui a merde
    if( d_.drawing() )
    {
      Font font = d_.gc_.getFont();
      FontData fd[] = font.getFontData();
      FontData nfd = new FontData( fd[0].getName(), fd[0].getHeight(), fd[0].getStyle()|SWT.BOLD );
      Font nf = new Font( null, nfd );
      
      d_.gc_.setFont( nf );
      d_.gc_.setForeground( getRed() );
      d_.gc_.drawText( _message, d_.R_.x()+5, d_.R_.y()+5+12 );

      nf.dispose();
    }
  }
  
  /** draw or locate point using d_ Data */
  protected void render() throws ChartLocation
  {
    try
    {    
      DChartStyle g_style = styleOf( d_.G_ );

      //prepare curve list, and their styles.
      d_.curves_.clear();
      IDItem  item=null;
      DCurve curve=null;
      for( item=d_.G_.getFirstChild(); item!=null; item=item.getNext() )
      {
        if( item instanceof DCurve )
        {
          curve = (DCurve)item;
          DChartStyle cs = styleOf( curve );                
          d_.curves_.add( new CurveStyle( curve, cs ) );
        }
      }
    
      //graphics background
      if( d_.drawing() )
      {
        d_.gc_.setBackground( g_style.back_ );
        d_.gc_.fillRectangle( d_.R_.x(), d_.R_.y(), d_.R_.w(), d_.R_.h() );
      }

      //graphic title
      int title_h = renderGraphicTitle( g_style );

      //legend on the right side
      int legend_width = renderLegend( title_h );

      //the rest is available to render graphic
      Rect rg = new Rect( margin_,
                          title_h+margin_,
                          d_.R_.w()-legend_width-2*margin_,
                          d_.R_.h()-title_h-2*margin_ );
  
      //graphic is rendered depending on it type:
           if( d_.G_.isGraphicType( DGraphic.T_SECTORS      )) renderSector( rg, false );
      else if ( d_.G_.isGraphicType( DGraphic.T_SECTORS3D    )) renderSector( rg, true  );
      else if ( d_.G_.isGraphicType( DGraphic.T_HISTOGRAM    )) renderHistogram( rg, false, false );
      else if ( d_.G_.isGraphicType( DGraphic.T_HISTOGRAM3D  )) renderHistogram( rg, true, false );
      else if ( d_.G_.isGraphicType( DGraphic.T_HISTOGRAM_BS )) renderHistogram( rg, false, true );
      else if ( d_.G_.isGraphicType( DGraphic.T_XY           )) renderXY( rg );
      else if ( d_.G_.isGraphicType( DGraphic.T_XYZ          )) renderXYZ( rg );
      else if ( d_.G_.isGraphicType( DGraphic.T_SURFACE      )) renderXYZ( rg );
/*
      else if ( d_.G_.isGraphicType( DGraphic.T_HISTORS )) renderHistors( rg )
*/
      else
      {
        error( "Unknwon Graphic type '"+d_.G_.getGraphicType()+"'" );
        return ;
      }
    
      g_style.dispose();
    
      if( d_.locating() && rg.contains( d_.lx_, d_.ly_ ) )
      {
        throw new ChartLocation( DLocated.InGraphicArea, d_.G_, rg );
      }
    }
    finally
    {
       //dispose styles, must be done for poor platform like windows.
       disposeStyleOf();
       d_.curves_.clear();
       d_.curves_=null; 
    }
  }
  
  
  /** render graphic title.
   * @return the height of the title, 0 if title isn't displayed
   */
  protected int renderGraphicTitle( DChartStyle _g_style ) throws ChartLocation
  {
    int title_h = 0;

    boolean display_title = d_.G_.getProperties().get(DGraphic.P_SHOW_TITLE, true);
    String title = d_.G_.getTitle();  
    
    if( !display_title || DrawUtil.isEmpty( title ) ) return title_h;
    
    d_.gc_.setFont( _g_style.font_ );
    FontMetrics fm = d_.gc_.getFontMetrics();      
    title_h = fm.getHeight();
  
    String txt = DrawUtil.truncateTextH( d_.gc_, title, d_.R_.w() );
    int    aln = DrawUtil.HCENTER | DrawUtil.TOP;
    Rect r_ttl = new Rect( d_.R_.x(), d_.R_.y(), d_.R_.w(), title_h );

    if( d_.drawing() )
    {
      d_.gc_.setForeground( _g_style.fore_ );
      DrawUtil.drawText( d_.gc_, r_ttl, aln, txt);
    }
    else
    {
      //locating
      Rect br = DrawUtil.boundingRect( d_.gc_, r_ttl, aln, txt );
      br.union( r_ttl );

      if( br.contains( d_.lx_, d_.ly_ ) )
      {
        throw new ChartLocation( DLocated.InGraphicTitleText, d_.G_, br, txt );
      }
      if( r_ttl.contains( d_.lx_, d_.ly_ ) )
      {
        throw new ChartLocation( DLocated.InGraphicTitleArea, d_.G_, r_ttl );
      }
    }
 
    return title_h;
  }
  
  /** render legend and return the width, 0 if none */
  protected int renderLegend( int _title_h )  throws ChartLocation 
  {
    int legend_width = 0;

    //display legend ?
    boolean display_legend = d_.G_.getProperties().get(DGraphic.P_SHOW_LEGEND, true);
    if( !display_legend ) return 0;
    
    CurveStyle ccurve=null;

    int legend_height = 0;
    int legend_height_min =100000;
    for( int i=0; i<d_.curves_.size(); ++i )
    {
      ccurve = (CurveStyle)d_.curves_.get(i);
      d_.gc_.setFont( ccurve.style_.font_ );
      String name = ccurve.curve_.getName();
      if( !DrawUtil.isEmpty( name ) )
      {
        Point sz = d_.gc_.textExtent( name );
        if ( legend_width < sz.x ) legend_width=sz.x;
        if ( legend_height_min > sz.y ) legend_height_min=sz.y;
        legend_height += sz.y;
      }
    }

    int legend_rect=(legend_height_min-2);
    if(legend_rect>10) legend_rect=10; //limit to 10x10 pixel
    legend_width += legend_rect+margin_ ;
    //limit legend width to 1/3 of graphic
    if ( legend_width > (d_.R_.w()/3) ) legend_width=d_.R_.w()/3;
    int legend_x = d_.R_.w() - legend_width;
    int legend_H = d_.curves_.size() * 4 + legend_height;
    int legend_y = _title_h+(d_.R_.h()-legend_H)/2+2;

    Rect lgd_area= new Rect( d_.R_.right()-legend_width, d_.R_.top()+_title_h,
                             legend_width, d_.R_.h()-_title_h );
    //shortcut: on on y est pas, on passe la main ...
    //shortcut for locating: outside ?
    if( d_.locating() && !lgd_area.contains( d_.lx_, d_.ly_ ) )
    {
      return legend_width ;
    }
    
    Rect r_clr=new Rect();
    Rect r_lgd_area=new Rect();
    Rectangle r = new Rectangle(0,0,0,0);
    
    for( int i=0; i<d_.curves_.size(); ++i )
    {
      ccurve = (CurveStyle)d_.curves_.get(i);
      String name = ccurve.curve_.getName();
      if( DrawUtil.isEmpty( name )) continue;
      //text is choosen black, because fore color is curve color
      d_.gc_.setFont( ccurve.style_.font_ );
      int legend_h = d_.gc_.textExtent( name ).y;
      r_clr.setRect( legend_x,legend_y+(legend_h-legend_rect)/2, 
                     legend_rect, legend_rect );

      int available_width = legend_width-legend_rect-5;
      String txt = DrawUtil.truncateTextH( d_.gc_, name, available_width );

      if( d_.drawing() )
      {
        d_.gc_.setForeground( getBlack() );
        d_.gc_.setBackground( ccurve.style_.back_ );
        r_clr.toRectangle( r );
        d_.gc_.fillRectangle( r ); //fill it
        d_.gc_.drawRectangle( r ); //draw black square around it
      }
      else //locating
      {
        //inside color rectangle
        if( r_clr.contains( d_.lx_, d_.ly_ ) )
        {
          throw new ChartLocation( DLocated.LegendColor, ccurve.curve_, r_clr );
        }
      }
    
      r_lgd_area.setRect( legend_x+legend_rect+5, legend_y,
                          available_width,        legend_h );

      renderText( txt, r_lgd_area, DrawUtil.LEFT|DrawUtil.VCENTER,
                   ccurve.style_, DLocated.LegendText, ccurve.curve_ );

      legend_y += legend_h+4;
    }//for

    //dans la zone de la legende ??
    if( d_.locating() && lgd_area.contains( d_.lx_, d_.ly_ ) )
    {
      throw new ChartLocation( DLocated.LegendArea, null, lgd_area );
    }
  
    return legend_width;
  }    
  
  protected static class TruncateAxisTextData
  {
    public String sttl_ ; //string title
    public String sunit_; //string for unit
    public Rect   rttl_ ; //rectangle for title
    public Rect   runit_; //rectangle for unit
    public Rect   rtu_  ; //rectangle for both title and unit    
  }
  
  /** truncate unit and/or title text for an "horizontal" axis,
   * return also bounding rectangles.
   */
  protected TruncateAxisTextData truncateHAxisText(
      DAxis _axis, DChartStyle _axis_style,
      int _available_width, int _available_height
    )
  {
    Rect   rttl;
    Rect   runit;
    Rect   rtu;

    //reduction de la chaine de titre ou units..

    //title et unit sont geree dans des chaines/rect different cela va permettre
    //de les affichers avec des styles differents ... un jours prochains.
    String axis_title = _axis.getTitle();
    String axis_unit  = _axis.getUnit();
//TODO: mieux a faire, SWT ne supporte pas textExtent(null)    
    if( axis_title == null ) axis_title="";
    if( axis_unit  == null ) axis_unit="";

    //titre ( unit ) axe des Y
    d_.gc_.setFont( _axis_style.font_ );
    Point ext = d_.gc_.textExtent( axis_title );
    rttl = new Rect( 0,0, ext.x+2, _available_height );
    //unit a droite du titre
    ext = d_.gc_.textExtent( axis_unit );
    runit = new Rect( rttl.w()+5, 0, ext.x+2, _available_height );
    rtu = Rect.Union( rttl, runit ); //rectangle englobant titre et unite 

    //reduire d'abord le titre
    String n_axis_title = axis_title;
    while( rtu.w() > _available_width ) 
    {
      try {
        axis_title = axis_title.substring( axis_title.length()-2, axis_title.length()-1 );
      }
      catch( StringIndexOutOfBoundsException e )
      {
        n_axis_title="";
        break;
      }
      if( axis_title.length()==0 ) { n_axis_title=null; break; }
      n_axis_title = axis_title+"...";
      ext = d_.gc_.textExtent( n_axis_title );
      rttl.setWidth( ext.x+2 );
      runit.setX( rttl.w()+5 );
      // rtu = Rect.Union( rttl, runit ) but without allocation
      rtu.setRect( rttl );
      rtu.union( runit );
    }
    axis_title = n_axis_title ;
    
    //puis l'unite
    String n_axis_unit = axis_unit;
    while( rtu.w() > _available_width ) 
    {
      try {
        axis_unit = axis_unit.substring( axis_unit.length()-2, axis_unit.length()-1 );
      }
      catch( StringIndexOutOfBoundsException e )
      {
        n_axis_unit="";
        break;
      }
      if( axis_unit.length()==0 ) { n_axis_unit=null; break; }
      n_axis_unit = axis_unit+"...";
      ext = d_.gc_.textExtent( n_axis_unit );
      runit.setWidth( ext.x+2 );
      rtu.setRect( rttl );
      rtu.union( runit );
    }
    axis_unit = n_axis_unit;

    TruncateAxisTextData data = new TruncateAxisTextData();
    data.sttl_ = axis_title;
    data.sunit_= axis_unit;
    data.rttl_ = rttl;
    data.runit_= runit;
    data.rtu_  = rtu ;
    
    return data;
  }

    /** truncate unit and/or title text for an "vertical" axis,
     * return also bounding rectangles.
     */
    protected TruncateAxisTextData truncateVAxisText(
        DAxis _axis, DChartStyle _axis_style,
        int _available_width, int _available_height
      )
    {
      Rect   rttl;
      Rect   runit;
      Rect   rtu;
  
      //reduction de la chaine de titre ou units..
  
      //title et unit sont geree dans des chaines/rect different cela va permettre
      //de les affichers avec des styles differents ... un jours prochains.
      String axis_title = _axis.getTitle();
      String axis_unit  = _axis.getUnit();
  //TODO: mieux a faire, SWT ne supporte pas textExtent(null)    
      if( axis_title == null ) axis_title="";
      if( axis_unit  == null ) axis_unit="";
  
      //titre ( unit ) axe des Y
      d_.gc_.setFont( _axis_style.font_ );
      Point ext = DrawUtil.verticalTextExtent( d_.gc_, axis_title );
      rttl = new Rect( 0,0, _available_width, ext.y+2 );
      //unit a droite du titre
      ext = DrawUtil.verticalTextExtent( d_.gc_, axis_unit );
      runit = new Rect( 0, rttl.h()+5, _available_width, ext.y+2 );
      rtu = Rect.Union( rttl, runit ); //rectangle englobant titre et unite 
  
      //reduire d'abord le titre
      String n_axis_title = axis_title;
      while( rtu.w() > _available_height ) 
      {
        try {
          axis_title = axis_title.substring( axis_title.length()-2, axis_title.length()-1 );
        }
        catch( StringIndexOutOfBoundsException e )
        {
          n_axis_title="";
          break;
        }
        if( axis_title.length()==0 ) { n_axis_title=null; break; }
        //no sens for vertical text/ n_axis_title = axis_title+"...";
        ext = DrawUtil.verticalTextExtent( d_.gc_, n_axis_title );
        rttl.setHeight( ext.y+2 );
        runit.setY( rttl.h()+5 );
        // rtu = Rect.Union( rttl, runit ) but without allocation
        rtu.setRect( rttl );
        rtu.union( runit );
      }
      axis_title = n_axis_title ;
      
      //puis l'unite
      String n_axis_unit = axis_unit;
      while( rtu.w() > _available_height ) 
      {
        try {
          axis_unit = axis_unit.substring( axis_unit.length()-2, axis_unit.length()-1 );
        }
        catch( StringIndexOutOfBoundsException e )
        {
          n_axis_unit="";
          break;
        }
        if( axis_unit.length()==0 ) { n_axis_unit=null; break; }
        //no sense for vertical/ n_axis_unit = axis_unit+"...";
        ext = DrawUtil.verticalTextExtent( d_.gc_, n_axis_unit );
        runit.setHeight( ext.y+2 );
        rtu.setRect( rttl );
        rtu.union( runit );
      }
      axis_unit = n_axis_unit;
  
      TruncateAxisTextData data = new TruncateAxisTextData();
      data.sttl_ = axis_title;
      data.sunit_= axis_unit;
      data.rttl_ = rttl;
      data.runit_= runit;
      data.rtu_  = rtu ;
      
      return data;
    }
  
  /** Render property on Y axis, support for "3D-like".
   * @param _get_step_unit_width true to compute step unit width on axis, 
   *                             false to locate/draw.
   * @param _R rectangle to render axis.
   * @param _axis axis to render.
   * @param _property property identifier to render: STEP_LINE,STEP_UNIT,STEP_DOT.
   * @param _3d_line_x simulated 3d on x axis
   * @param _3d_line_y simulated 3d on y axis
   */
  protected int onYAxis( 
        boolean _get_step_unit_width, Rect _R,
        DAxis _axis, String _property,
        int _3d_line_x, int _3d_line_y
      )
    throws ChartLocation  
  {
    int max_step_text_width=0;

    if( !_axis.getProperties().hasKey( _property ) ) return 0;

    float value = _axis.getProperties().get( _property, 0.0f );
    if( value == 0.0f ) return 0;
     
    float min     = _axis.getProperties().get( DAxis.P_MIN, -1.0f );
    float max     = _axis.getProperties().get( DAxis.P_MAX, +1.0f );
    float max_min = max-min ;

    //prevent div/0
    if ( max_min == 0.0f ) return 0;
    
    float wsu = _R.h()*value/max_min;
    int xd = _R.x()-margin_;
    //la numerotation, les lignes ... se fait en modulo value
    float vdn = (float)Math.ceil(min/value);
    if ( Math.abs(vdn)<1e-6 ) vdn=0.0f; 
    vdn *= value;
    float yd = _R.bottom() - _R.h()*(vdn-min)/max_min;

DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);

    for( float vd = vdn; vd <= max; vd += value )
    {
       if( DAxis.P_STEP_UNIT.equals( _property ) )
       {
              if ( Math.abs(vd)<1e-6 ) vd=0.0f;
              String t=df.format(vd);
              org.eclipse.swt.graphics.Point ext = d_.gc_.textExtent( t );
              int wt = ext.x;
              if( wt>max_step_text_width) max_step_text_width=wt;
              Rect r_txt=new Rect( xd-wt, (int)yd-ext.y/2, wt, ext.y );
              if( d_.drawing() )
              {
                if( !_get_step_unit_width ) //mode calcul largeur ?
                {
                  DrawUtil.drawText( d_.gc_, r_txt, DrawUtil.CENTER, t );
                  d_.gc_.drawLine( xd+margin_-2, (int)yd, xd+margin_, (int)yd );
                }
              }
              else //locating
              {
                if ( r_txt.contains( d_.lx_, d_.ly_ ) )
                {
                  throw new ChartLocation( DLocated.InAxisYStepUnitText, _axis, r_txt, t );
                }
              }
       }
       else if ( DAxis.P_STEP_DOT.equals( _property ) )
       {
              if(d_.drawing())
               d_.gc_.drawLine( xd+margin_-2, (int)yd, xd+margin_, (int)yd );
       }
       else if ( DAxis.P_STEP_LINE.equals( _property ) )
       {
//TBD:locating: detection des lignes, mais apres les courbes, alors que le dessin se fait avant ...
            if(d_.drawing()) 
            {
              d_.gc_.drawLine( _R.left(), (int)yd, _R.right(), (int)yd );
              if( _3d_line_x!=0 && _3d_line_y!=0 )
              {
                d_.gc_.drawLine( _R.x(), (int)yd, _R.x()-_3d_line_x, (int)yd+_3d_line_y );
              }
            }
       }
       yd -= wsu;
    }

    return max_step_text_width;
  }
  
  /** Render property on X axis, no support for "3D-like".
   * @param _R rectangle to render axis.
   * @param _axis axis to render.
   * @param _property property identifier to render: STEP_LINE,STEP_UNIT,STEP_DOT.
   */
  protected void onXAxis( Rect _R, DAxis _axis, String _property )
   throws ChartLocation
  {
    if( !_axis.getProperties().hasKey( _property ) ) return ;
    float value = _axis.getProperties().get( _property, 0.0f );
    if( value==0.0f ) return ;
    float min     = _axis.getProperties().get( DAxis.P_MIN, -1.0f);
    float max     = _axis.getProperties().get( DAxis.P_MAX, +1.0f );
    float max_min = max - min ;

    if ( max_min == 0.0f ) return; //sinon div/0
    
    float wsu = (_R.w()*value/max_min);
    int yd = _R.y()+_R.h()+margin_;
    //la numerotation, les lignes ... se fait en modulo prop_todo
    float vdn = (float)Math.ceil(min/value) ;
    if ( Math.abs(vdn)<1e-6 ) vdn=0.0f; // x*0 = 0
    vdn = vdn * value ;
    float xd = _R.left()+_R.w()*(vdn-min)/max_min;
    float vd;
    Rect r_txt=new Rect();
DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);    

    for( vd  = vdn;//_axis->property(DAxis::MIN);
         vd <= max;
         vd += value
       )
    {
      if( DAxis.P_STEP_UNIT.equals( _property ) )
      {
          if ( Math.abs(vd)<1e-6 ) vd=0.0f;        
          String t=df.format( vd );
          Point ext = d_.gc_.textExtent( t );
          r_txt.setRect( (int)xd-ext.x/2, yd, ext.x, ext.y );
          if(d_.drawing())
          {
            DrawUtil.drawText( d_.gc_, r_txt, DrawUtil.CENTER, t );
            d_.gc_.drawLine( (int)xd, yd-margin_, (int)xd, yd-margin_+2 );
          }
          else //locating
          {
            if ( r_txt.contains( d_.lx_, d_.ly_ ) )
            {
              throw new ChartLocation( DLocated.InAxisXStepUnitText, _axis, r_txt, t );
            }
          }
      }
      else if ( DAxis.P_STEP_DOT.equals( _property ) )
      {
          d_.gc_.drawLine( (int)xd, yd-margin_, (int)xd, yd-margin_+2 );
      }
      else if ( DAxis.P_STEP_LINE.equals( _property ) )
      {
          d_.gc_.drawLine( (int)xd, _R.top(), (int)xd, _R.bottom() );
//TBD:locating: la ligne, mais apres la detection des courbes ...
      }
      xd += wsu;
    }//for
  }
  
  /** draw or locate for an horizontal/vertical text.
   * @param _text text to render.
   * @param _rect rectangle to draw text
   * @param _align LEFT/RIGHT/HCENTER/... flag
   * @param _style style to paint
   * @param _locid identifier for location
   * @param _item item for location
   */
  public void renderText( String _text, Rect _rect, int _align, DChartStyle _style,
                          String _locid, IDItem _item )
    throws ChartLocation                           
  {
    if( d_.drawing() )
    {
      //la valeur est affichee dans le style de la courbe
      d_.gc_.setForeground( _style.fore_ );
      d_.gc_.setFont( _style.font_ );
      DrawUtil.drawText( d_.gc_, _rect, _align, _text );
    }
    else //locating
    {
      Rect nrtxt = DrawUtil.boundingRect( d_.gc_, _rect, _align, _text );
      //parce que si le texte deborde de _rect, .. ne le clippe pas
      //nrtxt might be bigger than _rect, clip it
      nrtxt.intersection( _rect );

      if( (nrtxt.w()>0) && nrtxt.contains( d_.lx_, d_.ly_ ) )
      {
        throw new ChartLocation( _locid, _item, nrtxt, _text );
      }
    }
  }
  
  /** Render Histogram graphic family */
  protected void renderHistogram( Rect _RG, boolean _3d, boolean _bs )
    throws ChartLocation
  {
    //histogram must have only one axis.
    DAxis axis=null;
    for( IDItem i=d_.G_.getFirstChild(); i!=null; i=i.getNext() )
    {
      if( i instanceof DAxis )
      {
        axis=(DAxis)i;
        break;
      }
    }

    if ( axis==null )
    {
      error( "Error: Histogram without axis");
      return ;
    }

    DChartStyle axis_style = styleOf( axis );
    d_.gc_.setFont( axis_style.font_ );
    int axis_font_h = d_.gc_.getFontMetrics().getHeight();

    int max_h_title_unit = axis_font_h + 2 ;

    //title et unit sont geree dans des chaines/rect different cela va permettre
    //de les afficher avec des styles differents ... un jours prochains.
    TruncateAxisTextData tatd =null;
 /*   String axis_title ;
    String axis_unit  ;
    Rectangle rttly; //rect du titre de l'axe
    Rectangle runity;//rect de l'unit de l'axe
    Rectangle rtuy; //rect englobant les deux premiers.
*/
    //reduction de la chaine de titre ou units..
//XXXX    //available_width est un height a cause de la rotation de 90
    tatd = truncateVAxisText( axis, axis_style, _RG.h(), max_h_title_unit );

    int yg = _RG.y()+ axis_font_h/2; //numerotation de l'axe
    int hg = _RG.h()
           - axis_font_h/2  // 1/2 hauteur aff. mesure axe Y (au cas ou)
           - margin_
           ;

    //mesure sur l'axe des Y (avec marge):
    int wuy=0;
    hg -= axis_font_h+margin_ ;
    wuy = onYAxis( true, new Rect(), axis, DAxis.P_STEP_UNIT, 0, 0 );
  
    int wg = _RG.w() ;
    int xg = _RG.x() ;

    //display Y axis title
    {
      int axis_title_w = 0;
      //largeur du titre + unite => hauteur max. de police (tourne a 90 degre)
      if( !DrawUtil.isEmpty( tatd.sttl_ )
       || !DrawUtil.isEmpty( tatd.sunit_ ) ) axis_title_w = max_h_title_unit;

      //titre de l'axe + unite en vertical a droite de l'axe des Y
      int decoration_w = wuy //aff. mesure axe des Y (max.)
                       + margin_ //entre wuy et max_h_title 
                       + axis_title_w // titre et/ou unite
                       + margin_ //entre max_h_... et bord 
                       ;
      wg -= decoration_w ;
      xg += decoration_w ;
    }

//TBD: 20 a afficher comme valeur de 3D
    int  _3d_z    = 20 ;
    if ( _3d )
    {
      wg -= _3d_z;
      hg -= _3d_z;
      yg += _3d_z;
    }


    Rect rg = new Rect(xg,yg,wg,hg);

    //lost of Y axis enlightment
    {
      //rect global de placement titre/unit Y
      //a note: a cause de la rotation on utilise la translation du QPainter
      // pour le placement correct, les calculs sont fait a origine 0,0
      Rect grtuy= new Rect( 0, 0, _RG.h(), max_h_title_unit );
      //centrer le rect titre+unit Y
      tatd.rtu_.moveCenter( grtuy.centerX(), grtuy.centerY() );
      //replacer rect title et unite
      tatd.rttl_.moveBy(  tatd.rtu_.x(), tatd.rtu_.y() );
      tatd.runit_.moveBy(  tatd.rtu_.x(), tatd.rtu_.y() );

      if( d_.drawing() )
      {
//        d_.gc_.save();
//TODO: he les gars SWT on fait ca comment ????
//        d_.gc_.translate( _RG.x, _RG.y+_RG.height );
//        d_.gc_.rotate( -90.0f );
//ye! SWT tu fait comment la rotation ???        
        d_.gc_.setForeground( axis_style.fore_ );
        d_.gc_.setFont( axis_style.font_ );
      
        DrawUtil.drawText( d_.gc_, tatd.rttl_, DrawUtil.CENTER, tatd.sttl_ );
        DrawUtil.drawText( d_.gc_, tatd.runit_, DrawUtil.CENTER, tatd.sunit_ );
       
   //     d_.gc_.restore();
      }
      else //locating
      {
        //on deplace le pt en (0,0) sur grtuy et on invers x/y:
        //pour prendre en compte la rotation
        int lx=d_.lx_;
        int ly=d_.ly_;
        //rotation di -90: on inverse x/-y w/h
        grtuy.setRect(  -grtuy.y(), grtuy.x(), grtuy.h(), +grtuy.w() );
        tatd.rttl_.setRect(  -tatd.rttl_.y(), tatd.rttl_.x(), tatd.rttl_.h(), +tatd.rttl_.w());
        tatd.runit_.setRect(  -tatd.runit_.y(), tatd.runit_.x(), tatd.runit_.h(), +tatd.runit_.w() );
        //translation:
        grtuy.moveBy(   _RG.x(), _RG.y() /*+_RG.height() */);
        tatd.rttl_.moveBy(  _RG.x(), _RG.y() /*+_RG.height() */);
        tatd.runit_.moveBy(  _RG.x(), _RG.y() /*+_RG.height() */);
        if( grtuy.contains( lx, ly ) )
        {
          //dans la zone de l'axe
          if( !DrawUtil.isEmpty(tatd.sttl_) && tatd.rttl_.contains( lx, ly ) )
          {
            //dans la zone de titre
            throw new ChartLocation( DLocated.InAxisYTitle, axis, tatd.rttl_, tatd.sttl_ );                    
          }
          if( !DrawUtil.isEmpty(tatd.sunit_) && tatd.runit_.contains( lx, ly ) )
          {
            //dans la zone de titre
            throw new ChartLocation( DLocated.InAxisYUnit, axis, tatd.runit_, tatd.sunit_);
          }
          //donc dans la zone de l'axe Y
          throw new ChartLocation( DLocated.InAxisYArea, axis, grtuy );
        }
      }

      //lines, unit, 'n dot sur l'axe des Y
      Rect rg2 = new Rect(rg);
      if(_3d) rg2.moveBy( _3d_z,-_3d_z );

      //drawing et:ou locating:
      if(d_.drawing())
      {
        d_.gc_.setForeground( getLightGray() );
      }
      if(_3d)
      {
        onYAxis( false, rg2, axis, DAxis.P_STEP_LINE, _3d_z, _3d_z );
      } else {
        onYAxis( false, rg2, axis, DAxis.P_STEP_LINE, 0, 0 );
      }
      if(d_.drawing())
      {
        d_.gc_.setForeground( getBlack() ); //black
      }
      onYAxis( false, rg, axis, DAxis.P_STEP_DOT, 0,0  );
      if(d_.drawing())
      {
        d_.gc_.setForeground( axis_style.fore_ );
        d_.gc_.setFont( axis_style.font_ );
      }
      onYAxis( false, rg, axis, DAxis.P_STEP_UNIT, 0,0 );
    }
  
    if(d_.drawing())
    {
      d_.gc_.setForeground( axis_style.fore_ );
      d_.gc_.drawLine( xg, yg,    xg,    yg+hg ); //axe Y
      d_.gc_.drawLine( xg, yg+hg, xg+wg, yg+hg ); //axe X
    }

    //dessin des courbes dans l'ordre
    float ay_min=axis.getProperties().get( DAxis.P_MIN,  0.0f),
          ay_max=axis.getProperties().get( DAxis.P_MAX, +1.123412345f);
    float ky = ay_max - ay_min;
System.err.println("HISTO axis["+axis+"].MIN="+ay_min);    
System.err.println("HISTO axis["+axis+"].MAX="+ay_max);
System.out.println("  PROP=["+DAxis.P_MAX+"]");    
    

    //if min==max, force a delta of 10%
    if( Math.abs(ky)<1e-6 )
    {
      ay_max = (ay_max>0.0f ? 1.05f : 0.95f )* ay_max;
      ay_min = (ay_min>0.0f ? 0.95f : 1.05f )* ay_min;
      ky = ay_max - ay_min;     
      if( Math.abs(ky)<1e-6)
      {
        //smell like update document problem (MIN/MAX=0.0f)
        error("Error: can't draw document not updated");
        return ;
      }
    }

    ky = hg/ky;
    int nbar=d_.curves_.size();

    float wbar=rg.w()/(float)nbar;
    float wbar_bs = (wbar+1)/3.0f; // _BS

    float xbar=rg.x();
    int   y_bar_zero;
    float bar_zero;
    if ( ay_min > 0.0f ) {
      y_bar_zero=rg.bottom();
        bar_zero=ay_min;
    } else if ( ay_max < 0.0f ) {
      y_bar_zero=rg.y();
        bar_zero=ay_max;
    }else {
      y_bar_zero=rg.bottom()-(int)(ky*(-ay_min)); //pos de 0.0
        bar_zero=0.0f;
    }

    IDItem item=null;
    ChartLocation loc=null;

DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits( 2 );

    for( int icrv=0; icrv<d_.curves_.size(); ++icrv )
    {
      CurveStyle ccurve = (CurveStyle)d_.curves_.get(icrv);    
      for( item=ccurve.curve_.getFirstChild(); item!=null; item=item.getNext() )
      {
        //les courbes n'ont qu'un point
//TBD:on pourrait imaginer des histogram multi point en superposant les bar.???
        if( item instanceof DPoint )
        {
          DPoint point = (DPoint)item;
          float vy;
          DCoord coord=point.getCoord( axis );          
          if ( coord !=null )
          {
            vy = coord.getValue();
            //transformation.
            float bh = vy-bar_zero;
            int  ibh = (int)(ky*(bh));
            if (ibh>0)ibh+=1;else if (ibh<0) ibh-=1;        
            //curve==BAR forcement.
            Rect rbar;
            if( _bs )  // bandes separees
            {
              int dx = (int)(xbar+0.5f*(wbar-wbar_bs));
              rbar = new Rect( dx, y_bar_zero, (int)wbar_bs, -ibh );
            } else {
              rbar = new Rect( (int)xbar, y_bar_zero, (int)wbar+1, -ibh );
            }
            rbar.normalize();

            //barre principale
            if(d_.drawing())
            {
              d_.gc_.setForeground( getBlack() ); //black ://trait exterieur en noir
              d_.gc_.setBackground( ccurve.style_.back_ );
              d_.gc_.fillRectangle( rbar.x(), rbar.y(), rbar.w(), rbar.h() );
              d_.gc_.drawRectangle( rbar.x(), rbar.y(), rbar.w(), rbar.h() );
            } 
            else //locating
            {
              if( rbar.contains( d_.lx_, d_.ly_ ) )
              {
                throw new ChartLocation( DLocated.CurveBar, point,  rbar );
              }
            }

//TBD: a revoir parce que les barre value>0 sont mal affiches en 3d
            if ( _3d )
            {
//TBD:20: c'est le Z 3d
              int ar1[] = new int[8],
                  ar2[] = new int[8];
                    
              ar1[0] = rbar.x();                ar1[1] = rbar.y();
              ar1[2] = rbar.x()+20;             ar1[3] = rbar.y()-20;
              ar1[4] = (int)(rbar.x()+wbar+20); ar1[5] = rbar.y()-20;
              ar1[6] = (int)(rbar.x()+wbar+1);  ar1[7] = rbar.y();
    
              ar2[0] = (int)(rbar.x()+wbar);    ar2[1] = rbar.y();
              ar2[2] = ar1[4];                ar2[3] = ar1[5]; 
              ar2[4] = (int)(rbar.x()+wbar+20); ar2[5] = rbar.y()+rbar.h()-20;
              ar2[6] = (int)(rbar.x()+wbar);    ar2[7] = rbar.y()+rbar.h();

              if(d_.drawing())
              {
                Color clr   = ccurve.style_.back_;
                Color dark1 = DrawUtil.lighter( clr, 0.50f );
                Color dark2 = DrawUtil.lighter( clr, 0.75f );
                d_.gc_.setBackground( dark1 );
                d_.gc_.fillPolygon( ar1 );
                d_.gc_.drawPolygon( ar1 );
                d_.gc_.setBackground( dark2 );
                d_.gc_.fillPolygon( ar2 );
                d_.gc_.drawPolygon( ar2 );
              }
              else //locating
              //Il faudrait parcourir les barres a l'envers pour ne pas 
              //se tromper lorsque les barres 3D se recouvrent ...
              //MAIS, je preferre y aller comme le dessin et choisir
              //la derniere zone 3d designee par la souris.
              {
//evidement:20 c'est la simili projection 3d
                //l'ideal c'est que l'on est un ar1.contains( ... )
                boolean in_ar1 = Data.horizontal3DContains( rbar.x(), rbar.y(),
                                          rbar.w(), 20, d_.lx_,d_.ly_ );
                boolean in_ar2 = Data.vertical3DContains( rbar.right(), rbar.y(),
                                         rbar.h(), 20, d_.lx_,d_.ly_ );
                //pas encore vu ...
                if( in_ar1 )
                {
                  //memo...
                  Rect br = new Rect( rbar.x(), rbar.y()-20, rbar.w()+20, 20 );
                  loc = new ChartLocation( DLocated.CurveBar, point, br );
                }
                else if( in_ar2 )
                {
                  Rect br = new Rect( rbar.right(), rbar.y()-20, 20, rbar.h()+20 );
                  loc = new ChartLocation( DLocated.CurveBar, point, br );
                }
                //sinon si on en avait trouve, c'est la derniere...
                else if( loc!=null )
                {
                  throw loc;
                }
              }
            }

            //value is displayed below the bar.
            {
              String full_txt = df.format( vy );
              d_.gc_.setFont( ccurve.style_.font_ );
              String txt = DrawUtil.truncateTextH( d_.gc_, full_txt, (int)wbar-2 );
              if( !DrawUtil.isEmpty( txt ) )
              {
                int font_h = d_.gc_.getFontMetrics().getHeight();
                Rect rtxt = new Rect( (int)xbar, rg.bottom()+margin_,
                                                (int)wbar-2, font_h+2 );
                //drawn using curve style
                renderText( txt, rtxt, DrawUtil.CENTER, ccurve.style_,
                             DLocated.CurvePointText, point );
              }
            }
          }
          else
          {
            error( "Error: Bar without value." );
          }

          //si plusieurs valeur... xbar += wbar;
        }//if 
      }//for 
      xbar += wbar;
    }//for
 
    //pour la derniere zone 3d vue...
    if( d_.locating() && loc!=null )
    {
      throw loc;
    }
  }
  
  /** Render sector-ish graphic */
  protected   void renderSector( Rect _RG, boolean _3d ) throws ChartLocation
  {
    boolean display_values = true; //true to display value of sector
    int d = _RG.w();
    //DR2358: je fais ce que je peux...avec le temps que j'ai.
    // Comme ca bave un peu au dessus surl le titre du graphique,
    // je vais reserver une zone de 2*20 pixels (dessus/dessous)
    // pour le texte (si display_decoration).
    // Ok, 20 c'est sorti de la cuisse de Jupiter, car a ce stade
    // on ne sait pas si il y aura vraiment du texte dessus/dessous
    // (il peut etre un peu decale), ni si ce texte a un style dont
    // la police est <= 20 pixels...
    // Vous aurez compris que c'est totalement du pif !
    // En attendant ca calme M'sieu Telep, en attendant qu'il trouve 
    // que les % se chevauchent quand le graphique est petit...
    int h_for_text = display_values ? 40 : 0;
    if ( d > _RG.h()-h_for_text ) d=_RG.h()-h_for_text;

    //si on affiche les texte:
    if ( display_values )
    {
//TBD:il y a surement un calcul plus serieux que 3/4 !!!
      d = 3*d/4; //les secteurs occupent 3/4 de la place (prevoir pr texte)
    }

    float v_sum = 0.0f;

    CurveStyle ccurve=null;
    IDItem item=null;
    
    DAxis axis=null;
    for( item=d_.G_.getFirstChild(); item!=null; item=item.getNext() )
    {
      if( item instanceof DAxis ) axis=(DAxis)item;
    }
    if( axis==null )
    {
      error("Error: missing Axis in graphic");
      return;
    }
    String axis_unit = axis.getUnit();
    if( axis_unit==null ) axis_unit="";

    //cumul des valeurs des courbes (pour correspondance au 100% du cercle)
    int n_points=0;
    for( int i=0; i<d_.curves_.size(); ++i )
    {
      ccurve = (CurveStyle)d_.curves_.get(i);
      //courbes: plusieurs point a une seule coord possibles
      for( item=ccurve.curve_.getFirstChild(); item!=null; item=item.getNext())
      {
        if ( item instanceof DPoint )
        {
          DCoord coord= ((DPoint)item).getCoord( axis );
          if ( coord!=null )
          {
            v_sum += coord.getValue();
            n_points++;
          }
        } 
      }
    }

    //v_sum represente 100% du secteur.
    if ( v_sum == 0.0f )
    {
      error( "Error: can't print sectors (sum:0)" );
      return ;
    }

    int x3d = 0, y3d = 0, r3d=0, a3d=0;

    if ( _3d )
    {
      //parce que c'est a peut pres correct
      if( d >= 48 ) //ombre 4 fois moins importante que le pie chart
      {
        x3d = - 8;
        y3d = +12;
      } else {
        y3d = d / 4;
        x3d = -(int)(2.0f*y3d/3.0f);
        // on voit rien: circulez y'a rien a voir
        if( Math.abs(x3d) < 2 ) //au moins 2 pixels d'ombrage
        {
          _3d = false;
          x3d = y3d = 0;
        }
      }
      if( _3d )
      {
        r3d = (int)Math.sqrt( (double)(x3d*x3d + y3d*y3d) );
        float a = (float)Math.atan2( (float)y3d, (float)x3d ); //-PI..PI
        if( a < 0.0f ) a += 2*Math.PI ; 
        //a est dans 0..2PI 
        a3d = (int)(180*a/(Math.PI));
      }
    }

    LineAlg l3d = new LineAlg( 0, 0, x3d, y3d );

    //locating: les sector sont detectes dans l'ordre de dessin,
    //donc on les teste tous, et c'est le dernier qui gagne
    ChartLocation loc_sector=null;

DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);

    //because line separating two sector must be drawn after sector
    //otherwise fill next sector will overwrite line.
    //this array store x,y coordinates of point to draw joined to center or pie.
    int between_sectors[] = new int[n_points*2];
    int wbs=0;
    Rect r = new Rect();
    
    while( l3d.nextPoint() )
    {
      int dx3d = l3d.x();
      int dy3d = l3d.y();
      boolean top=(dx3d==x3d)&&(dy3d==y3d);

      int angle=0;
      int ray = d >> 1; 

      r.setRect( _RG.x()+(_RG.w() -d)/2 +dx3d,
                 _RG.y()+(_RG.h() -d)/2 -dy3d, d,d ); 
      if( r.w()==0 || r.h() == 0 ) continue;

      //locating: (ici pour optimiser le calcul)
      float loc_dist=-1.0f;
      int loc_angle=-1; //en 16eme de degre
      if(d_.locating())
      {
        Point rc=r.center();
        int dx=d_.lx_-rc.x, dy=d_.ly_-rc.y;
        loc_dist=(float)Math.sqrt( dx*dx + dy*dy );
        if(loc_dist<=ray)
        {
          float a = (float)Math.atan2( (float)-dy, (float)dx ); //-PI..PI
          if( a < 0.0f ) a += 2*Math.PI ; 
          //a est dans 0..2PI 
          loc_angle = (int)(360*a/(2*Math.PI));
        }
      }
      
      int ipt=0;
      for( int i=0; i<d_.curves_.size(); ++i )
      {
        ccurve = (CurveStyle)d_.curves_.get(i);
        //Courbes: plusieurs points possibles
        for( item=ccurve.curve_.getFirstChild(); item!=null; item=item.getNext() )
        {
          if ( item instanceof DPoint )
          {
            DPoint point = (DPoint)item;
            DCoord coord = point.getCoord( axis );
            if ( coord !=null )
            {
              ipt++;
              float v = coord.getValue()/v_sum; //% du disque
              int end_angle = angle+(int)(360*v); //degree;
              if( ipt >= n_points ) //last sector ?
              {
                end_angle=360;
              }

              if( d_.drawing() )
              {
                if(top)
                {
                  d_.gc_.setForeground( getBlack() ); //black
                  Color b = ccurve.style_.back_;
                  d_.gc_.setBackground( b );
                  d_.gc_.fillArc( r.x(),r.y(),r.w(),r.h(), angle, end_angle-angle );
                  d_.gc_.drawArc( r.x(),r.y(),r.w(),r.h(), angle, end_angle-angle );
                  int cx = r.centerX(), cy=r.centerY();
                  double a = (Math.PI*angle)/180.0f;
                  between_sectors[wbs++]= (int)( cx+r.w()*Math.cos( a )/2.0f );
                  between_sectors[wbs++]= (int)( cy-r.h()*Math.sin( a )/2.0f );                  
                } else {
                  Color dark = DrawUtil.lighter( ccurve.style_.back_, 0.75f );
                  d_.gc_.setForeground( dark );
                  d_.gc_.setBackground( dark );
                  d_.gc_.fillArc( r.x(),r.y(),r.w(),r.h(), angle, end_angle-angle );
                }
              }
              else //locating
              {
                if( (loc_dist <= ray)
                    &&(loc_angle>=angle)&&(loc_angle<=end_angle) )
                {
                  //memo pour connaitre le don secteur a la fin ..
                  loc_sector = new ChartLocation( DLocated.CurveSector, point, r );
                  ((DChartLocation)loc_sector.location_).setAngles( angle, end_angle );
                }
              }
              

              //affichage de la valeur + unite de l'axe
              if( top && display_values )
              {
                int atxt = angle+(end_angle-angle)/2;
                //calculer la compensation sur le rayon en mode _3d pour
                //que le texte ne recouvre pas l'ombre:
                int r3=0;
                if ( _3d )
                {
                  /* aller, j'en dis un peu plus:
                   * en 3d on va ajouter un rayon pour ne pas afficher le texte
                   * sur l'ombre, le + c'est que ce rayon depend de l'angle
                   * du texte ET de l'angle choisit pour l'ombrage, eh, eh, eh!
                   * C'est pas beau ca !
                   */
                  int da = (int)Math.abs( atxt-a3d-180 );
                  //compenser que dans la zone de l'ombre:
                  if( da < 90 ) //PI/2
                  {
                    //approx du 1er degre:
                    float k = 1.0f - da/90.f;
                    r3 = (int)(k * r3d);
                  }
                }
                float arad=(float)( 2*Math.PI-(2*Math.PI*atxt)/360 );
                int sector_ray = ray+margin_+r3 ;
                int tx = (int)(r.x()+(r.w() >>1)+sector_ray*Math.cos(arad));
                int ty = (int)(r.y()+(r.h()>>1)+sector_ray*Math.sin(arad));

                String txt = df.format( coord.getValue() ) + axis_unit;

                Point te = d_.gc_.textExtent( txt );
                int tw = te.x;
                int th = te.y;

                //on essaye de placer le texte au mieux
                     if ( atxt < 90  ) {}/*raf*/
                else if ( atxt < 180 ) { tx -= tw; }
                else if ( atxt < 270 ) { tx -= tw; ty+=th; }
                else                    {          ty+=th; };

                Rect rtxt = new Rect( tx, ty-th, tw, th );

                renderText( txt, rtxt,  DrawUtil.LEFT|DrawUtil.BOTTOM,
                             ccurve.style_, DLocated.CurvePointText,
                             point );
                if( loc_sector!=null ) throw loc_sector;
              }

              angle = end_angle;
            }
          }//if
        }//item
      }//curve
    }//3d
    
    if( d_.drawing() )
    {
      d_.gc_.setForeground( getBlack() ); //black
      int cx = r.centerX(), cy=r.centerY();
      for( int i=0; i<between_sectors.length; )
      {
        int x = between_sectors[i++];
        int y = between_sectors[i++];
        d_.gc_.drawLine( cx,cy, x,y );
      }
    }
    else //locating
    {
      if( loc_sector!=null ) throw loc_sector;
    }
  }

  /** Render sector-ish graphic */
  protected   void renderXY( Rect _RG ) throws ChartLocation
  {
    //decorations means axis titles/units/...
    boolean display_decorations = true;
    
    //search for axis "x" and "y", must exists
    DAxis x_axis=null, y_axis=null;
    for( IDItem item=d_.G_.getFirstChild(); item!=null; item=item.getNext() )
    {
      if( item instanceof DAxis )
      {
        DAxis axis = (DAxis)item;
        if( x_axis==null && "x".equals( axis.getName() ) )
        {
          x_axis = axis;
        }
        else if ( y_axis==null && "y".equals( axis.getName() ))
        {
          y_axis = axis;
        }
        if( x_axis!=null && y_axis!=null ) break;
      }
    }

    if ( x_axis==null || y_axis==null )
    {
      String txt;
      if( x_axis==null )
      {
        txt = "Error: missing 'x' axis";
      } else {
        txt = "Error: missing 'y' axis";
      }
      error( txt );
      return ;
    }

    DChartStyle x_axis_style = styleOf( x_axis );
    DChartStyle y_axis_style = styleOf( y_axis );

    d_.gc_.setFont( x_axis_style.font_ );
    int x_axis_style_h = d_.gc_.textExtent( "X" ).y+2;
    d_.gc_.setFont( y_axis_style.font_ );
    int y_axis_style_h = d_.gc_.textExtent( "X" ).y+2;
//evidement c'est sorti de la cuisse de jupiter...    

    //metrics titre ( unit ) axe des X pour les chaines/rect
    int xg,yg,hg,wg;

    xg = _RG.x();
    yg = _RG.y();
    hg = _RG.h();
    wg = _RG.w();

DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);
    if ( display_decorations )
    {  
      d_.gc_.setFont( y_axis_style.font_ );
      //compute steps width ont Y Axis...
      int wuy= onYAxis( true,  new Rect(), y_axis, DAxis.P_STEP_UNIT, 0,0 );

      //it's a worth measure.
      int wux=0;
      if ( x_axis.getProperties().hasValue(DAxis.P_STEP_UNIT))
      {
        String s = df.format( x_axis.getProperties().get(DAxis.P_MAX, +1.0f) );
        d_.gc_.setFont( x_axis_style.font_ );
        wux = d_.gc_.textExtent( s ).x;
      }

//Note: title and unit of Y axis are drawn verticaly
      int y_axis_title_unit_w = 0;
      if ( !DrawUtil.isEmpty( y_axis.getTitle())
        || !DrawUtil.isEmpty( y_axis.getUnit()) )
      {
//TODO: SWT sait pas faire le 90 !      
//TODO:pas de rotation:           y_axis_title_unit_w = y_axis_style_h ; //texte tourne a 90 degre.
        y_axis_title_unit_w = d_.gc_.textExtent( "X" ).x+2;           
      }

      wg -= y_axis_title_unit_w // titre ( unite ) axe des Y ecris en vertical
          + margin_         // marge en titre et mesure axe des Y
          + wuy             // aff. mesure axe des Y (maximum)
          + margin_         // entre aff mesure axe des Y et l'axe des Y
          + wux/2           // au pire on doit afficher la mesure maxi (axe X)
          ;
     
      xg += y_axis_title_unit_w
         + margin_
          + wuy
          + margin_
          ;

      int x_axis_title_unit_h = 0;
      if ( !DrawUtil.isEmpty( x_axis.getTitle() )
        || !DrawUtil.isEmpty( x_axis.getUnit() ) )
           x_axis_title_unit_h = y_axis_style_h;

      hg -= y_axis_style_h/2  // 1/2 hauteur aff. mesure axe Y (au cas ou)
          + margin_           // marge en axe X et affichage mesure des X
          + x_axis_style_h    // mesure des X  
          + margin_           // entre mesure des X et titre axe X
          + x_axis_title_unit_h // titre (unite) axe des X
          ;

      yg += y_axis_style_h/2 ; //1/2 nombre possible en haut axe des Y

      Rect rg = new Rect(xg,yg,wg,hg);

      //X Axis:
      TruncateAxisTextData tatdx
            = truncateHAxisText( x_axis, x_axis_style, wg, x_axis_style_h );

      //rect global de placement titre/unit X pour recenter les rect...
      Rect grtux=new Rect( xg, _RG.bottom()-margin_-x_axis_style_h, wg, x_axis_style_h );
      //centrer dans grtux
      tatdx.rtu_.moveCenter( grtux.center() );
      //donc:
      tatdx.rttl_.moveBy( tatdx.rtu_.x(), tatdx.rtu_.y() );
      tatdx.runit_.moveBy( tatdx.rtu_.x(), tatdx.rtu_.y() );

      d_.gc_.setFont( x_axis_style.font_ );
      if( d_.drawing() )
      {
        d_.gc_.setForeground( x_axis_style.fore_ );
     
        DrawUtil.drawText( d_.gc_, tatdx.rttl_, DrawUtil.CENTER, tatdx.sttl_ );
        DrawUtil.drawText( d_.gc_, tatdx.runit_, DrawUtil.CENTER, tatdx.sunit_ );

        //lines, unit, 'n dot sur l'axe des X
        d_.gc_.setForeground( getLightGray() ); 
//TBD:render
        onXAxis( rg, x_axis, DAxis.P_STEP_LINE );
        d_.gc_.setForeground( getBlack() );
        onXAxis( rg, x_axis, DAxis.P_STEP_DOT  );
        d_.gc_.setForeground ( x_axis_style.fore_  );
        d_.gc_.setFont( x_axis_style.font_ );
        onXAxis( rg, x_axis, DAxis.P_STEP_UNIT );
      }
      else //locating
      {
        if( grtux.contains( d_.lx_, d_.ly_ ) )
        {
          //inside axis area...
          if( !DrawUtil.isEmpty( tatdx.sttl_ ) && tatdx.rttl_.contains( d_.lx_, d_.ly_ ) )
          {
            //inside title area
            throw new ChartLocation( DLocated.InAxisXTitle, x_axis, tatdx.rttl_, tatdx.sttl_ );
          }
          if( !DrawUtil.isEmpty( tatdx.sunit_ ) && tatdx.runit_.contains( d_.lx_, d_.ly_ ) )
          {
            //dans la zone de titre
            throw new ChartLocation( DLocated.InAxisXUnit, x_axis,
                                      tatdx.runit_, tatdx.sunit_ );
          }
          //so, in axis area.
          throw new ChartLocation( DLocated.InAxisXArea, x_axis, 
                                    grtux, tatdx.sttl_+tatdx.sunit_ );
        }

        //locating unit step text ...
        onXAxis( rg, x_axis, DAxis.P_STEP_UNIT );
      }

      // Axe Y:
      TruncateAxisTextData tatdy
        = truncateVAxisText( y_axis, y_axis_style, xg-_RG.x(), hg );

      //rect global de placement titre/unit Y
//TODO: pas de text verticaux dans SWT.      
      //a note: a cause de la rotation on utilise la translation du QPainter
      // pour le placement correct, les calculs sont fait a origine 0,0
 //     Rect grtuy=new Rect( yg, 0, hg, y_axis_style_h );
 Rect grtuy = new Rect( 0, yg, xg-_RG.x(), hg ); //y_axis_style_h, hg );
      //centrer le rect titre+unit Y
      tatdy.rtu_.moveCenter( grtuy.center() );

      //replace title and unit rectangle (remember: drawn verticaly)
      tatdy.rttl_ .setLocation( tatdy.rtu_.x(), tatdy.rtu_.y() );
      tatdy.runit_.setLocation( tatdy.rtu_.x(), tatdy.rtu_.bottom()-tatdy.runit_.h() );

      d_.gc_.setFont( y_axis_style.font_ );
      if( d_.drawing() )
      {
//TODO: pas de texte verticaux...        
/*        d_->P_->save();
        d_->P_->translate( _RG.left(), _RG.top()+_RG.height() );
        d_->P_->rotate( -90.0f );
*/        
        d_.gc_.setForeground( y_axis_style.fore_ );

        DrawUtil.drawText( d_.gc_, tatdy.rttl_,  DrawUtil.CENTER|DrawUtil.VERTICAL, tatdy.sttl_ );
        DrawUtil.drawText( d_.gc_, tatdy.runit_, DrawUtil.CENTER|DrawUtil.VERTICAL, tatdy.sunit_  );
//        d_->P_->restore();

        //lines, unit, 'n dot sur l'axe des Y
        d_.gc_.setForeground( getLightGray() ); 
        onYAxis( false, rg, y_axis, DAxis.P_STEP_LINE, 0,0 );
        d_.gc_.setForeground( getBlack() );
        onYAxis( false, rg, y_axis, DAxis.P_STEP_DOT, 0,0  );
        d_.gc_.setForeground( y_axis_style.fore_  );
        d_.gc_.setFont( y_axis_style.font_  );
        onYAxis( false, rg, y_axis, DAxis.P_STEP_UNIT, 0,0 );
      } 
      else //locating
      {
//TODO: pas de texte verticaux, locating faux        
        //on deplace le pt en (0,0) sur grtuy et on invers x/y:
        //pour prendre en compte la rotation
        int lx=d_.lx_;
        int ly=d_.ly_;
        //rotation di -90: on inverse x/-y w/h
        grtuy.setRect( -grtuy.y(), grtuy.x(), grtuy.h(), +grtuy.w() );
        tatdy.rttl_.setRect( -tatdy.rttl_.y(), tatdy.rttl_.x(), tatdy.rttl_.h(), +tatdy.rttl_.w() );
        tatdy.runit_.setRect( -tatdy.runit_.y(), tatdy.runit_.x(), tatdy.runit_.h(), +tatdy.runit_.w() );
        //translation:
        grtuy.moveBy(  _RG.left(), _RG.top() /*+_RG.height() */);
        tatdy.rttl_.moveBy(  _RG.left(), _RG.top() /*+_RG.height() */);
        tatdy.runit_.moveBy(  _RG.left(), _RG.top() /*+_RG.height() */);
        if( grtuy.contains( lx, ly ) )
        {
          //in axis area... 
          if( !DrawUtil.isEmpty( tatdy.sttl_ ) && tatdy.rttl_.contains( lx, ly ) )
          {
            //in axis title area
            throw new ChartLocation( DLocated.InAxisYTitle, y_axis,
                                      tatdy.rttl_, tatdy.sttl_ );
          }
          if( !DrawUtil.isEmpty( tatdy.sunit_ ) && tatdy.runit_.contains( lx, ly ) )
          {
            //in unit text area
            throw new ChartLocation( DLocated.InAxisYUnit, y_axis,
                                      tatdy.runit_, tatdy.sunit_ );
          }
          //so in axis area...
          throw new ChartLocation( DLocated.InAxisYArea, y_axis, 
                                    grtuy, tatdy.sttl_+tatdy.sunit_ );
        }

        //locating unit step text ...
        onYAxis( false, rg, y_axis, DAxis.P_STEP_UNIT, 0,0 );
      }
    }

    //graph rect without axis
    Rect rg = new Rect(xg,yg,wg,hg);

    if( d_.drawing() )
    {
      d_.gc_.setForeground( getBlack() );
      d_.gc_.drawLine( xg, yg+hg, xg+wg, yg+hg ); // X axis
      d_.gc_.drawLine( xg, yg,    xg,    yg+hg ); // Y axis
    }

    //draw curves in ordre.
    float x_min = x_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
          x_max = x_axis.getProperties().get( DAxis.P_MAX, -1.0f ),
          y_min = y_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
          y_max = y_axis.getProperties().get( DAxis.P_MAX, +1.0f );
    float kx = x_max - x_min;
    float ky = y_max - y_min;
    if ( (kx==0.0f)||(ky==0.0f) )
    {
      error( "Error: can't draw curves (divison by 0)" );
      return;
    }
  
    kx = wg/kx;
    ky = hg/ky;
    //basis for histogram bar (xy curve might be bars)
    int y_bar_zero ;
    float bar_zero;

    //except if we haven't 0
    if ( y_min > 0.0f ) {
      y_bar_zero = rg.bottom(); //DAxis.MIN
        bar_zero = y_min;
    } else if ( y_max < 0.0f ) {
      y_bar_zero = rg.top(); //DAxis.MAX
        bar_zero = y_max;
    } else {
      //position of 0on Y axis:
      y_bar_zero = rg.bottom()-(int)(ky*(-y_min));
        bar_zero = 0.0f;
    }
     
    int last_x=-1, last_y=-1; 
    float  last_vy=-1;
    CurveStyle ccurve=null;
    Rect rp=new Rect();
    Rect r_bar=new Rect();

    //for locating mode, curve are checking in reverse order to handle covering.
    int ic_inc = d_.locating() ? -1 : +1;
    int ic_size = d_.curves_.size();
    for( int ic=d_.locating() ? ic_size-1 : 0;
         d_.locating()? (ic>=0) : (ic<ic_size);
         ic += ic_inc )
    {
      ccurve = (CurveStyle)d_.curves_.get(ic);      
      boolean first=true;
      IDItem last_item=null;
      DPoint point=null;
      for( IDItem item=ccurve.curve_.getFirstChild(); item!=null; item=item.getNext() )
      {
        if ( item instanceof DPoint )
        {
          point = (DPoint)item;
          DCoord cx = point.getCoord( x_axis );
          DCoord cy = point.getCoord( y_axis );
          if ( cx!=null && cy!=null )
          {
            float vx = cx.getValue();
            float vy = cy.getValue();
            //transformation.
            int px = xg+(int)(kx*(vx-x_min));
            int py = yg+hg-(int)(ky*(vy-y_min));
            if ((!first)|| ccurve.curve_.isType( DCurve.T_POINTS ) )
            {
              if( ccurve.curve_.isType( DCurve.T_LINE ) )
              {
                if( d_.drawing() )
                {
                  d_.gc_.setForeground( ccurve.style_.back_ );
                  //d_->P_->setBrush(Qt::NoBrush);
                  d_.gc_.drawLine( last_x, last_y, px,py );
                }
                else //locating
                {
                  if( DrawUtil.segmentContains( last_x, last_y, px, py, d_.lx_, d_.ly_ ) )
                  {
                    throw new ChartLocation( DLocated.CurveLine, point, new Rect(last_x,last_y,px-last_x,py-last_y));
                  }
                }
              }
              else if ( ccurve.curve_.isType( DCurve.T_POINTS ))
              {
//TBD:choix de 6x6 pour la dimension des points
                rp.setRect( px-3, py-3, 6, 6 );
                if( d_.drawing() )
                {
                  d_.gc_.setForeground( getBlack() );
                  d_.gc_.setBackground( ccurve.style_.back_ );
                  d_.gc_.fillRectangle( rp.x(), rp.y(), 6,6 );
                  d_.gc_.drawRectangle( rp.x(), rp.y(), 6,6 );
                }
                else //locating
                {
                  if( rp.contains( d_.lx_, d_.ly_ ) )
                  {
                     throw new ChartLocation( DLocated.CurvePoint, point, rp );
                  }
                }
              }
              else if ( ccurve.curve_.isType( DCurve.T_BAR ))
              {   
                float bh=last_vy-bar_zero;
                int  ibh=(int)(ky*(bh));
                if (ibh>0)ibh+=1;else if (ibh<0) ibh-=1;
                r_bar.setRect( last_x, y_bar_zero, px-last_x+1, -ibh );
                r_bar.normalize();
/* cf ci-dessus               if(r_bar.h()<0) 
                {
                     //redresser le rectangle sinon contains marche pas
                     r_bar.setRect( r_bar.x(), r_bar.y()+r_bar.height(),
                                    r_bar.width(), -r_bar.height() );
                   }*/
                if( d_.drawing() )
                {
                  d_.gc_.setForeground( getBlack() );
                  d_.gc_.setBackground( ccurve.style_.back_ );
                  d_.gc_.fillRectangle( r_bar.x(), r_bar.y(), r_bar.w(), r_bar.h() );
                  d_.gc_.drawRectangle( r_bar.x(), r_bar.y(), r_bar.w(), r_bar.h() );
                }
                else //locating
                {
                  //on s'arrete ici parce que l'on recouvre tout...
                  if( r_bar.contains( d_.lx_, d_.ly_ ) )
                  {
                    throw new ChartLocation( DLocated.CurveBar, last_item, r_bar );
                  }
                }
              }
            } 
            else
            {
              first=false;
            }
           last_x = px;
           last_y = py;
           last_vy= vy;
         }
         else
         {
error( "Warning:Point without (x,y) coord.");
          }
       }
 
       last_item=item;
     }//for 

      //last histogram bar
//TBD: c'est pas terrible, mais sinon ca veut dire que la derniere
// valeur d'une courbe histo d'a de sens que pour X (termine la valeur precedente).
     if ( !first && (ccurve.curve_.isType( DCurve.T_BAR ) ))
     {
       int px=rg.right(); //on force jusqu'a la fin du graphe
       r_bar.setRect( last_x-1, last_y, px-last_x+1, rg.bottom()-last_y+1 );
       r_bar.normalize();
       if( d_.drawing() )
       {
         d_.gc_.setForeground( getBlack() );
         d_.gc_.setBackground( ccurve.style_.back_);
         d_.gc_.fillRectangle( r_bar.x(), r_bar.y(), r_bar.w(), r_bar.h() );
         d_.gc_.drawRectangle( r_bar.x(), r_bar.y(), r_bar.w(), r_bar.h() );
       }
       else //locating
       {
         if( r_bar.contains( d_.lx_, d_.ly_ ) )
         {
           throw new ChartLocation( DLocated.CurveBar, point, r_bar );
         }
       }
     }

    }//for
  }
  
  private static Color c_black_=null;
  protected Color getBlack()
  {
    if( c_black_==null)
    {
      c_black_=new Color(null,0,0,0);
    }
    return c_black_;
  }
  
  private static Color c_light_gray_=null;
  protected Color getLightGray()
  {
    if( c_light_gray_==null )
    {
      c_light_gray_ = new Color( null, 160,160,160);
    }
    return c_light_gray_;
  }
  
  private static Color c_red_ = null;
  protected Color getRed()
  {
    if( c_red_==null )
    {
      c_red_ = new Color( null, 255,0,0);
    }
    return c_red_;
  }
  
  private static class ForceCoord implements ICoordTransformation
  {
    public float value_ = 0.0f;
    public float apply( float _f ) { return value_; }
  }
  private static ForceCoord force_coord = new ForceCoord();
  
  protected void drawBox3D( View3D _view, Box3D _bb )
  {
    Point2D p[] = new Point2D[8];
    float xn = _bb.getXMin(), xx = _bb.getXMax();
    float yn = _bb.getYMin(), yx = _bb.getYMax();
    float zn = _bb.getZMin(), zx = _bb.getZMax();
    p[0] = _view.projection( xn,yn,zn, new Point2D() );
    p[1] = _view.projection( xn,yn,zx, new Point2D() );
    p[2] = _view.projection( xx,yn,zx, new Point2D() );
    p[3] = _view.projection( xx,yn,zn, new Point2D() );
    p[4] = _view.projection( xn,yx,zx, new Point2D() );
    p[5] = _view.projection( xx,yx,zx, new Point2D() );
    p[6] = _view.projection( xx,yx,zn, new Point2D() );
    p[7] = _view.projection( xn,yx,zn, new Point2D() );

    d_.gc_.drawLine( (int)p[0].getX(),(int)p[0].getY(), (int)p[1].getX(),(int)p[1].getY());
    d_.gc_.drawLine( (int)p[0].getX(),(int)p[0].getY(), (int)p[3].getX(),(int)p[3].getY());
    d_.gc_.drawLine( (int)p[0].getX(),(int)p[0].getY(), (int)p[7].getX(),(int)p[7].getY());
    d_.gc_.drawLine( (int)p[2].getX(),(int)p[2].getY(), (int)p[1].getX(),(int)p[1].getY());
    d_.gc_.drawLine( (int)p[2].getX(),(int)p[2].getY(), (int)p[3].getX(),(int)p[3].getY());
    d_.gc_.drawLine( (int)p[2].getX(),(int)p[2].getY(), (int)p[5].getX(),(int)p[5].getY());
    d_.gc_.drawLine( (int)p[4].getX(),(int)p[4].getY(), (int)p[1].getX(),(int)p[1].getY());
    d_.gc_.drawLine( (int)p[4].getX(),(int)p[4].getY(), (int)p[5].getX(),(int)p[5].getY());
    d_.gc_.drawLine( (int)p[4].getX(),(int)p[4].getY(), (int)p[7].getX(),(int)p[7].getY());
    d_.gc_.drawLine( (int)p[6].getX(),(int)p[6].getY(), (int)p[3].getX(),(int)p[3].getY());
    d_.gc_.drawLine( (int)p[6].getX(),(int)p[6].getY(), (int)p[5].getX(),(int)p[5].getY());
    d_.gc_.drawLine( (int)p[6].getX(),(int)p[6].getY(), (int)p[7].getX(),(int)p[7].getY());
  }
  
  //identifiers for axis X,Y,Z...
  protected static final int AxisX = 1;
  protected static final int AxisY = 2;
  protected static final int AxisZ = 3;
  
  protected String axis3DName( DAxis _axis )
  {
    String title = _axis.getTitle();

    if( title != null )
    {
      title = title + _axis.getUnit();
    } else {
      title = _axis.getUnit();
    }
  
    return title ;
  }
  
  protected void renderAnAxis3D( 
      View3D _view, Box3D _bb, DAxis _axis, DChartStyle _axis_style, int _axis_type
    )
    throws ChartLocation
  {
    float x_proj = _bb.getXMin(), x_proj2 = _bb.getXMax(),
          y_proj = _bb.getYMin(), y_proj2 = _bb.getYMax(),
          z_proj = _bb.getZMin(), z_proj2 = _bb.getZMax();

    float coord_min, coord_max;
    switch( _axis_type )
    {
      case AxisX: coord_min = _bb.getXMin(); coord_max = _bb.getXMax(); break;
      case AxisY: coord_min = _bb.getYMin(); coord_max = _bb.getYMax(); break;
      case AxisZ: coord_min = _bb.getZMin(); coord_max = _bb.getZMax(); break;
      default: return ;
    }

    Point2D p0 = new Point2D(), 
            p1 = new Point2D(),
            p2 = new Point2D();

    /* 
     * Lignes des reperes
     */
    if ( d_.drawing() && _axis.getProperties().hasValue( DAxis.P_STEP_LINE ) )
    {
      float coord_step = _axis.getProperties().get( DAxis.P_STEP_LINE, 1.0f );
      if ( coord_step != 0.0f )
      {
        float coord;
        float coord_start = (float)( coord_step*Math.ceil(coord_min/coord_step) );

        d_.gc_.setForeground( getLightGray() );

                
        //ca a l'air plus complique, mais comme ca on fait le switch une
        //seule fois par for...
        switch( _axis_type )
        {
          case AxisX: for( coord = coord_start; coord < coord_max; coord += coord_step )
                      {
                        _view.projection( coord, y_proj2, z_proj, p0 );
                        _view.projection( coord, y_proj, z_proj, p1 );
                        _view.projection( coord, y_proj, z_proj2, p2 );
          
                        d_.gc_.drawLine( (int)p0.getX(), (int)p0.getY(), (int)p1.getX(), (int)p1.getY() );
                        d_.gc_.drawLine( (int)p2.getX(), (int)p2.getY(), (int)p1.getX(), (int)p1.getY() );
                      }
                      break;
          case AxisY: for( coord = coord_start; coord < coord_max; coord += coord_step )
                      {
                        _view.projection( x_proj2, coord, z_proj, p0 );
                        _view.projection( x_proj,  coord, z_proj, p1 );
                        _view.projection( x_proj,  coord, z_proj2, p2 );
          
                        d_.gc_.drawLine( (int)p0.getX(), (int)p0.getY(), (int)p1.getX(), (int)p1.getY() );
                        d_.gc_.drawLine( (int)p2.getX(), (int)p2.getY(), (int)p1.getX(), (int)p1.getY() );
                      }
                      break;
          case AxisZ: for( coord = coord_start; coord < coord_max; coord += coord_step )
                      {
                        _view.projection( x_proj2, y_proj,  coord, p0 );
                        _view.projection( x_proj,  y_proj,  coord, p1 );
                        _view.projection( x_proj,  y_proj2, coord, p2 );
          
                        d_.gc_.drawLine( (int)p0.getX(), (int)p0.getY(), (int)p1.getX(), (int)p1.getY() );
                        d_.gc_.drawLine( (int)p2.getX(), (int)p2.getY(), (int)p1.getX(), (int)p1.getY() );
                      }
                      break;
        }
      }
    }

    /*
     * Graduation (texte) de l'axe 
     */
    String axis_title = axis3DName( _axis );
    boolean have_step_unit = false;
    
    if ( _axis.getProperties().hasValue( DAxis.P_STEP_UNIT )
      || !DrawUtil.isEmpty( axis_title )
       )
    {
      have_step_unit = true;
      float d_angle = 0.0f;
      switch( _axis_type )
      {
        case AxisX:
            _view.projection( _bb.getXMin(),_bb.getYMin(),_bb.getZMin(), p0 );
            _view.projection( _bb.getXMax(),_bb.getYMin(),_bb.getZMin(), p1 );
            d_angle = 180;
          break;
        case AxisY:
             _view.projection( _bb.getXMin(),_bb.getYMin(),_bb.getZMin(), p0 );
             _view.projection( _bb.getXMin(),_bb.getYMax(),_bb.getZMin(), p1 );
             d_angle = 0;
           break;
        case AxisZ:
             _view.projection( _bb.getXMin(),_bb.getYMin(),_bb.getZMin(), p0 );
             _view.projection( _bb.getXMin(),_bb.getYMin(),_bb.getZMax(), p1 );
             d_angle = 180;
           break;
      }
       
      float angle = (float)Math.atan2( p1.getY()-p0.getY(), p1.getX()-p0.getX() ); //rad.
      angle = (float)(180*angle/Math.PI + d_angle); //degree

//SWT en est loin de ca...      QWMatrix m ;

      if( d_.drawing() )
      {
        d_.gc_.setForeground(  _axis_style.fore_ );
        d_.gc_.setFont( _axis_style.font_ );
      }
DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);      
      if ( _axis.getProperties().hasValue( DAxis.P_STEP_UNIT ) )
      {
        float coord_step = _axis.getProperties().get( DAxis.P_STEP_UNIT, +1.0f );
        if (  coord_step != 0.0f )
        {
          float coord;
//          float *cx, *cy, *cz;
          //optim: fait le switch une seule fois par for...
          /*
          switch( _axis_type )
          { 
            case AxisX:
          cx = &coord;  cy = &y_proj;  cz = &z_proj2 ; break;
            case AxisY:
          cx = &x_proj; cy = &coord;   cz = &z_proj2 ; break;
            case AxisZ:
          cx = &x_proj; cy = &y_proj2; cz = &coord ; break;
          }*/
 Rect rtxt = new Rect();
 int fh = d_.gc_.getFontMetrics().getHeight();
          for( coord = (float)(coord_step*Math.ceil(coord_min/coord_step));
               coord < coord_max;
               coord += coord_step )
          {
            switch( _axis_type )
            { 
              case AxisX: _view.projection(  coord,  y_proj,  z_proj2, p2 ); break;
              case AxisY: _view.projection(  x_proj, coord,   z_proj2, p2 ); break;
              case AxisZ: _view.projection(  x_proj, y_proj2, coord,   p2 ); break;
            }
            
//SWT?? tu fais quoi avec ca ???            
   /*         m.reset();
            m.translate( px[2], py[2] );
            m.rotate( angle );*/

            String txt = df.format( coord );
            int wt = d_.gc_.textExtent( txt ).x;
            rtxt.setRect( -wt/2, -3-fh, wt, fh );
rtxt.moveBy( (int)p2.getX(), (int)p2.getY() );            

            if( d_.drawing() )
            {
//SWT???
/*              
                d_->P_->save();
                d_->P_->setWorldMatrix( m );
                d_->P_->drawText( rtxt, Qt::AlignCenter, txt );
                d_->P_->restore();*/
                DrawUtil.drawText( d_.gc_, rtxt, DrawUtil.CENTER, txt );
              }
              else //locating
              {
/*SWT*declare forfait*                QRegion reg( rtxt );
                reg = m * reg;
                if ( reg.contains( _pt_to_locate ) )
                {
                  DChartLocation::Location location;
                  switch( _axis_type )
                  {
                    case AxisX:
                      location = DChartLocation::AxisXStepUnitText; break;
                    case AxisY:
                      location = DChartLocation::AxisYStepUnitText; break;
                    case AxisZ:
                      location = DChartLocation::AxisZStepUnitText; break;
                  }
                  d_->location_= DChartLocation( location, _axis, reg.boundingRect(), txt );
                  return ;
                }
                */
              }
            }
          }
        }
      }

      /*
       * Axis title.
       */
      if ( !DrawUtil.isEmpty( axis_title ) )
      {
        String location=null;
 Rect rtxt = new Rect();
 int fh = d_.gc_.getFontMetrics().getHeight();       
        switch( _axis_type )
        {
          case AxisX:
                _view.projection( _bb.getXMin(),y_proj,z_proj2, p0 );
                _view.projection( _bb.getXMax(),y_proj,z_proj2, p1 );
               location = DLocated.InAxisXTitle;
            break;
          case AxisY:
                _view.projection( x_proj,_bb.getYMin(),z_proj2, p0 );
                _view.projection( x_proj,_bb.getYMax(),z_proj2, p1 );
                location = DLocated.InAxisYTitle; 
            break;

          case AxisZ:
                 _view.projection( x_proj,y_proj2,_bb.getZMin(), p0 );
                 _view.projection( x_proj,y_proj2,_bb.getZMax(), p1 );
                 location = DLocated.InAxisZTitle; 
            break;
        }

        float center_x=(p0.getX()+p1.getX())/2,
              center_y=(p0.getY()+p1.getY())/2;
/*SWT*? toujours pas ?
          m.reset();
          m.translate( center_x, center_y );
          m.rotate( angle );
*/
          int wt = d_.gc_.textExtent( axis_title ).x;
          rtxt.setRect( -wt/2, -6-2*fh, wt, fh );

//TODO: pas de rotation ... donc "a plat"
//-X-8 au dessus des unites et marge...
rtxt.moveCenter( (int)center_x, (int)center_y - (have_step_unit?(fh+4):0) -4 );

          if( d_.drawing() )
          {
/*            
            d_->P_->save();
            d_->P_->setWorldMatrix( m );
            d_->P_->drawText( rtxt, Qt::AlignCenter, axis_title );
      d_->P_->restore();
*/
  DrawUtil.drawText( d_.gc_, rtxt, DrawUtil.CENTER, axis_title );      
          }
          else //locating
          {
  if( rtxt.contains( d_.lx_, d_.ly_ ))
  {
    throw new ChartLocation( location, _axis, rtxt ); 
  }
/*SWT...            
            QRegion reg( rtxt );
            reg = m * reg ;
            if( reg.contains( _pt_to_locate ) )
            {
              d_->location_= DChartLocation( location, _axis, reg.boundingRect(),
                                    axis_title );
              return;
            }
*/            
          }
        }
      }
    
   protected static final int XProj=1;
   protected static final int YProj=2;
   protected static final int ZProj=3;
   protected static final int Curve=4;
   protected static final int Finished=5;

  
  /** render XYZ curves */
  void renderXYZ( Rect _RG ) throws ChartLocation
  {
    //decorations means axis title/units...
    boolean display_decorations = true;
    boolean is_surface = d_.G_.isGraphicType( DGraphic.T_SURFACE );

    DAxis x_axis = null;
    DAxis y_axis = null;
    DAxis z_axis = null;
    int   axis_missed = 3;
    for( IDItem item = d_.G_.getFirstChild(); item!=null; item=item.getNext() )
    {
      if( item instanceof DAxis )
      {
        DAxis axis = (DAxis)item;
        if( x_axis==null && "x".equals( axis.getName() ))
        {
          x_axis = axis; axis_missed--;
        }
        else if ( y_axis==null & "y".equals( axis.getName() ))
        {
          y_axis = axis; axis_missed--;
        }
        else if ( z_axis==null & "z".equals( axis.getName() ))
        {
          z_axis = axis; axis_missed--;
        }
        if( axis_missed==0 ) break;
      }
    }

    if ( axis_missed > 0 )
    {
      String txt;
      if(x_axis==null) {
        txt = "Error: missing 'x' axis";
      } else 
      if(y_axis==null) {
        txt = "Error: missing 'y' axis";
      } else {
        txt = "Error: missing 'z' axis";
      }
      error( txt );
      return ;
    }

    DChartStyle x_axis_style = styleOf( x_axis );
    DChartStyle y_axis_style = styleOf( y_axis );
    DChartStyle z_axis_style = styleOf( z_axis );

    //garder un peu d'espace pour les texte sur les axes 
    //c'est completement choisi au hazard (ou a peu pres!).
    int max_font_h  = 0 ;
    if ( display_decorations )
    {
      d_.gc_.setFont( x_axis_style.font_ );
      max_font_h = d_.gc_.getFontMetrics().getHeight();
      d_.gc_.setFont( y_axis_style.font_ );
      int hf = d_.gc_.getFontMetrics().getHeight();
      if ( max_font_h < hf ) max_font_h = hf;
      d_.gc_.setFont( z_axis_style.font_ );
      hf = d_.gc_.getFontMetrics().getHeight();
      if ( max_font_h < hf ) max_font_h = hf;
      //prevoir le titre + les unites sur l'axe
      max_font_h = 2*max_font_h +6;
    }

    Rect gR = new Rect( _RG.left()+max_font_h, _RG.top()+max_font_h,
                        _RG.w()-2*max_font_h,  _RG.h()-max_font_h );

    //boite englobante des courbes a afficher.
    Box3D bb = new Box3D( x_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
                          x_axis.getProperties().get( DAxis.P_MAX, +1.0f ),
                          y_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
                          y_axis.getProperties().get( DAxis.P_MAX, +1.0f ),
                          z_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
                          z_axis.getProperties().get( DAxis.P_MAX, +1.0f ) );

    float phi   = d_.G_.getProperties().get( DGraphic.P_XYZ_PHI,   0.3f );
    float theta = d_.G_.getProperties().get( DGraphic.P_XYZ_THETA, 0.2f );

    View3D view = new View3D( phi, theta, 0, 1.0f, bb );

    //mise a l'echelle de la vue 3D en fonction du rectangle de dessin
    view.scaleTo( gR.x(), gR.y(), gR.w(), gR.h() );

    int min_w_h = gR.w() ;
    if (min_w_h > gR.h() ) min_w_h = gR.h();

    //bounding Box
    if ( display_decorations )
    {
      if( d_.drawing() )
      {
        d_.gc_.setForeground( getLightGray() );
        drawBox3D( view, bb );
      }

      // render des axes...
      renderAnAxis3D( view, bb, x_axis, x_axis_style, AxisX );
//      if( d_->located() ) return ;
      renderAnAxis3D( view, bb, y_axis, y_axis_style, AxisY );
//      if( d_->located() ) return ;
      renderAnAxis3D( view, bb, z_axis, z_axis_style, AxisZ );
//      if( d_->located() ) return ;
    }

    //pour projection des courbes.
    float x_proj = bb.getXMin(), y_proj = bb.getYMin(), z_proj = bb.getZMin(); 
    CurveStyle ccurve =null;
    CurveStyle p_ccurve=null; //courbe precedente (connection plan)
    IDItem         item=null;
    IDItem       p_item=null;
    float       last_x=0, last_y=0;

    //memo h,s,v du fond pour les couleurs de projection des courbes...
    DChartStyle bgsty = styleOf( d_.G_ );
    Color background_color = bgsty.back_;
    HSV bhsv = new HSV( background_color );

    // check property status for curves projections
    boolean do_xproj = d_.G_.getProperties().get( DGraphic.P_XYZ_XPROJ, true);
    boolean do_yproj = d_.G_.getProperties().get( DGraphic.P_XYZ_YPROJ, true);
    boolean do_zproj = d_.G_.getProperties().get( DGraphic.P_XYZ_ZPROJ, true);

    Point2D p = new Point2D();
    
    for( int ic=0; ic<d_.curves_.size(); ++ic )
    {
      ccurve = (CurveStyle)d_.curves_.get(ic);
      //la courbe est dessinee 4 fois: 3 projections + la courbe
      int what=XProj;

      if( display_decorations
//       && d_.G_.display3DCurveProjection() 
       && d_.drawing() /*en locating on ne fait pas les projections*/
        )
      {
        //courbes projetees: couleur de la courbe en plus clair ...
        Color proj_color = DrawUtil.lighter( ccurve.style_.back_, 1.75f );
        HSV hsv = new HSV( proj_color );
        //sauf si ecart < 20, me demandez pas pourquoi, c'est une heuristique
        //finement elaboree...
        int ecart=(int)(Math.abs(bhsv.getV()-hsv.getV())+Math.abs(bhsv.getS()-hsv.getS()))>>1;
        if(ecart < 20 )
        {
          //couleur de projection trop proche du fond, on bascule
          //en couleur plus fonce..
          proj_color = DrawUtil.lighter( ccurve.style_.back_, 0.75f );
        }
        d_.gc_.setForeground( proj_color );
        what = XProj ;
      } else {
        what = Curve ;
      }
  for( ;what<Finished; )
  {
 
      //reset des projecitons
      view.resetCoordTransformation();
      switch( what )
      {
        case XProj: { if(!do_xproj) { what=YProj; continue; }
                      force_coord.value_ = x_proj ;
                      view.setXCoordTransformation( force_coord );
                      what = YProj;
                      break; }
        case YProj: { if(!do_yproj) { what=ZProj; continue; }
                      force_coord.value_ = y_proj ;
                      view.setYCoordTransformation( force_coord );
                      what = ZProj;
                      break; }
        case ZProj: { if(!do_zproj) { what=Curve; continue; }
                      force_coord.value_ = z_proj ;
                      view.setZCoordTransformation( force_coord );
                      what = Curve;
                      break; }
        case Curve: { if(d_.drawing()) d_.gc_.setForeground( ccurve.style_.back_ ); 
          what = Finished;
                      break; }
        case Finished:break;
      }//switch

      if ( is_surface )
      {
        p_item = p_ccurve!=null ? p_ccurve.curve_.getFirstChild(): null;
      }

      boolean first=true; 
      for( item = ccurve.curve_.getFirstChild(); item!=null; item=item.getNext() )
      {
        DPoint point   = (DPoint)item;
        DPoint p_point = p_item!=null ? (DPoint)p_item: null; 
        if( point!=null )
        {
          float v_x=0, v_y=0, v_z=0 ;
          DCoord c=null;
          boolean ok=true;
          if( (c=point.getCoord( x_axis))!=null ) { v_x=c.getValue(); } else { ok=false; }
          if( (c=point.getCoord( y_axis))!=null ) { v_y=c.getValue(); } else { ok=false; }
          if( (c=point.getCoord( z_axis))!=null ) { v_z=c.getValue(); } else { ok=false; }
  
          if( ok )
          {
            float p_x,p_y;
            view.projection( v_x,v_y,v_z, p );
            {
              p_x = p.getX();
              p_y = p.getY();
              if ( first )
              {
                first=false;
              }
              else
              {
                if(d_.drawing())
                {
                  d_.gc_.drawLine( (int)last_x, (int)last_y, (int)p_x, (int)p_y );
                }
                else //locating
                {
                  if( what==Finished ) //pas les projections
                  {
                    if( DrawUtil.segmentContains( (int)last_x, (int)last_y, (int)p_x, (int)p_y, d_.lx_,d_.ly_))
                    {
                      throw new ChartLocation( DLocated.CurveLine, item,
                             new Rect((int)p_x,(int)p_y, (int)(last_x-p_x),(int)(last_y-p_y) ));
                    }
                  }
                }
              }
            }
            last_x=p_x;
            last_y=p_y;

            //lien avec la courbe precedente (en mode surface)
            if ( p_point!=null )
            {
              float p_v_x=0,p_v_y=0,p_v_z=0;
              ok = true;
              ok = ok&&((c=p_point.getCoord( x_axis ))!=null); if(ok) {p_v_x=c.getValue(); }
              ok = ok&&((c=p_point.getCoord( y_axis ))!=null); if(ok) {p_v_y=c.getValue(); }
              ok = ok&&((c=p_point.getCoord( z_axis ))!=null); if(ok) {p_v_z=c.getValue(); }
              if( ok )
              {
                view.projection( p_v_x,p_v_y,p_v_z, p);
                {
                  float p_p_x = p.getX(),  p_p_y=p.getY();
                  if( d_.drawing() )
                  {
                    d_.gc_.drawLine( (int)p_x,(int)p_y, (int)p_p_x,(int)p_p_y);
                  } 
                  else // locating
                  {
                    if( what==Finished )
                    {
 //TBD en 3D(2D) plusieurs courbes peuvent se superposer...
                      if( DrawUtil.segmentContains( (int)p_x,(int)p_y, (int)p_p_x,(int)p_p_y, d_.lx_,d_.ly_ ))
                      {
                        throw new ChartLocation( DLocated.CurveLine, item, 
                            new Rect( (int)p_x,(int)p_y,(int)(p_p_x-p_x),(int)(p_p_y-p_y)));
                      } 
                    }
                  }
                }
              }
            }
          }
          if ( p_item!=null ) p_item = p_item.getNext();
        }
      }//for

  }//for:what

      p_ccurve=ccurve;
    }//for
  }

  
/*
class TESTRT_EXPORT DChartPainter
{
 protected:
  void renderHistor ( const QRect & );
  void renderBargraph ( const QRect & );

  QString valueWithUnit( float _value, DAxis *_axis ) const ;

*/
}
//Source file: C:\\Program Files\\eclipse21\\workspace\\apidoc\\com\\rational\\apidoc\\drivers\\builder\\TBuilder.java

//Source file: C:\\Program Files\\eclipse21\\workspace\\apidoc\\com\\rational\\apidoc\\drivers\\ui\\TBuilder.java

package com.rational.apidoc.drivers.builder;

import com.rational.apidoc.core.*;
import com.rational.apidoc.drawutil.DrawUtil;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.Device;
import java.util.Enumeration;
import java.util.Hashtable;
import com.rational.apidoc.parsers.DParser;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.FontMetrics;
import org.eclipse.swt.graphics.Rectangle;


public class TBuilder extends DParser 
{
   private Image brokenImage = null;
   private int alignment = DParagraph.LEFT;
   private int bottomMargin = 0;
   private int cellSpacing = 0;
   private TCell currentCell = null;
   private int currentHLine = 0;
   private int currentNLine = 0;
   private int currentNPage = 0;
   private TCell currentPage = null;
   private int currentPosition = 1;
   private int currentWLine = 0;
   private int currentXPos = 0;
   private int currentYPos = 0;
   private Font defaultFont = null;
   private Font previousFont = null;
   private Device device = null;
   private boolean drawPopup = true;
   private TCell firstCellLine = null;
   private boolean flatPopup = false;
   private Hashtable fonts = new Hashtable();
   private TCell footerCell = null;
   private GC gc = null;
   private TCell headerCell = null;
   private Hashtable imagePool = new Hashtable();
   private TCell lastCellLine = null;
   private int leftMargin = 0;
   private int lineSpacing = 0;
   private int listOffsetSpacing = 0;
   private int maxWidth = 100;
   private int minWidth = 20;
   private int pageBegin = 0;
   private int pageEnd = 0;
   private int pageHeight = 0;
   private int prefixSpacing = 0;
   private int rightMargin = 0;
   private int titleSpacing = 0;
   private int topMargin = 0;
   private boolean wordWrap = true;
   private int xoffset = 0;
   private DFooter footer = null;
   private DHeader header = null;
   private TCell firstCell = null;
   static int level = 0;
   
   private int currentNColumn = 0;
   private int currentTotalColumn = 0;
   private int currentNRow = 0;
   private DRow currentRow = null;
   private DTable currentTable = null;
   private int maxCellHeight = 0;
   
   
   private class CellInfo {
		  TCell cell = null;
		  int rowspan = 0;
   }
   
   private CellInfo[] currentRowSpan = null;
   
   private class Range 
   {
		public int start = 0;
		public int end = 0;
      
		/**
		 * @roseuid 3F0168C202A7
		 */
		public Range() 
		{
       
		}
   }
   
   /**
    * @param gc
    * @param d
    * @roseuid 3EF82BBF01A5
    */
   public TBuilder(GC gc, Device d) 
   {
      this.gc = gc;
	  this.device = d; 
            
      if (gc!=null)    
         defaultFont = gc.getFont();
      else
	     defaultFont = new Font(d, "courier", 12, SWT.NORMAL);   
   }
   
   
   private TBuilder _newBuilder(int size)
	  {
		 TBuilder b = new TBuilder(getGc(), getDevice());
		 b.setFlatPopup(getFlatPopup());
		 b.setCurrentPosition(getCurrentPosition());
		 b.setWordWrap(getWordWrap());
		 b.setMaxWidth(size);
		 b.setImagePool(getImagePool());
		 b.setFontPool(getImagePool());
		 b.setImageSearchPath(getImageSearchPath());
		 b.setCurrentNPage(getCurrentNPage());
		 return b;
	  }
   
	  private void _disposeBuilder(TBuilder b)
	  {
		 b.setImagePool(null);
		 b.setImageSearchPath(null);
		 b.setFontPool(null);
	  }
   
	  /**
	   * @param i
	   * @roseuid 3F004C9B00F9
	   */
	  private void _addListItemPrefix(IDItem i) 
	  {  
		 if (!(i.getParent() instanceof DList)) return;
   	
		 addListEntry();
		 String rval = getListStringPrefix(i.getParent());
      
		 if (rval != null)
		 {
			Point p = getGc().textExtent(rval);
			addXSpacing(-p.x);
			TCell c = addCell(p.x, p.y, i, rval.length());
			c.setText(rval);
			addXSpacing(getTitleSpacing());
		 }    
	  }
   
	  /**
	   * @param i
	   * @param text
	   * @param r
	   * @roseuid 3F0168EB0089
	   */
	  private void _buildPourTextBuffer(IDItem i, String text, Range r) 
	  {
		 if (text.length() == 0) return;
      
		 int w, wprev;
		 int count = r.end - r.start;
		 TCell item = null;
      
		 FontMetrics fm = getGc().getFontMetrics();
		 int lineSpacing = fm.getAscent()+fm.getDescent();
      
		 // rien a traiter
		 if (r.start >= r.end)
		 {
		   if (text.getBytes()[r.start] == '\n')
		   {
			   item = addCell(0, lineSpacing, i, 0);
		   }
		   else
		   {	
			   addYSpacing(lineSpacing);	
		   }
		   return;
		 }
      
		 String qs = null;
		 if (r.end == text.length())
			qs = text;
		 else
			qs = text.substring(r.start, r.end);
      
		 GC gc = getGc();
		 Point p = gc.textExtent(qs);
      
		 // Si on est en non WORD WRAP ou que le buffer rentre
		 // dans la ligne alors on ajoute ce dernier
		 // sur la ligne courante
		 if (!getWordWrap() 
			 || (getCurrentWLine() + p.x <= getMaxWidth()))
		 {
		   item = addCell(p.x, p.y, i, count);
		   item.setText(text, r.start, count); 
		   return;
		 }
         
		 // sinon on va couper le buffer sur un separateur de mot afin
		 // de la faire rentrer sur plusieurs lignes
		 wprev = 0;
      
		 Range r2 = new Range();
		 r2.start = r.start; // position dernier caractere de la sous chaine
      
      
		 while (r2.start < r.end)
		 {
			r2.end = 0;   // taille en pixel de la sous chaine
      	 
			_buildGetNextWord(qs, 
							  getMaxWidth() - getCurrentWLine(),
							  r2);
              
			count = r2.start - r.start;
			TCell cell = addCell(r2.end, lineSpacing, i, count);
			cell.setText(text, r.start, count);
			newLine();
         
			r.start = r2.start;
		 }    
	  }
   
	  /**
	   * @param b
	   * @param i
	   * @param text
	   * @roseuid 3F003ACA012D
	   */
	  private void _buildTextBuffer(IDItem i, String text) 
	  {
		 if (text.length()==0) return;
       
		 Range r = new Range();
		 r.start = 0;
		 r.end = text.indexOf("\n");
      
		 // tant que l'on a pas atteint la fin de la ligne
		 while (r.end != -1)
		 {
		   _buildPourTextBuffer(i, text, r);
      	
		   // ligne non adaptative car terminee par \n
		   // on ajoute donc une nouvelle ligne
		   newLine();
      	
		   r.start = r.end + 1;
      	
		   // recherche du \n suivant 
		   r.end = text.indexOf("\n", r.start);
		 }
      
		 if (r.start < text.length())
		 {
			r.end = text.length();
		   _buildPourTextBuffer(i, text, r);
		 }    
	  }
   
	  /**
	   * @param c
	   * @return int
	   * @roseuid 3F003121028F
	   */
	  private int _getColumnSizeForCell(DCell c) 
	  {
		  int size = 0;
		  if (currentTable.getAutoResize())  // toute la largeur du document 
		  {
				 float ws = (float)getMaxWidth() * currentTable.getWidthScale();
				 for (int i=0; i<c.getColSpan()+1; i = i + 1)
				 {
				   float factor = 1.0f;
				   int s = 0;
				   if ((currentTable.getColumnWidth() != null)
						&& (currentTable.getColumnWidth().length > currentNColumn + i))
				   {
					   factor = currentTable.getColumnWidth()[currentNColumn + i];
					   s = (int)(ws * factor);
				   }
				   else
				   {
					   s = (int)(ws / currentTotalColumn);
				   }
			  	
				   if (s < getMinWidth())
				   {
					 s = getMinWidth();
				   }
				   size = size + s;
				 }
		 }
		 else 
		 {
				size = getMaxWidth();
		 }
		 return size;    
	  }
   
	  /**
	   * @param b
	   * @param c
	   * @roseuid 3F0030E403A0
	   */
	  private void _setBuilderMarginForCell(TBuilder b) 
	  {
		  int m = getCellSpacing();
	   
	      b.setLeftMargin(m + currentTable.getBorder());	  
	      b.setTopMargin( m + currentTable.getBorder());   	  
	      b.setBottomMargin(m + currentTable.getBorder());
	      b.setRightMargin(m + currentTable.getBorder());
	  }
   
   
   /**
    * Access method for the alignment property.
    * 
    * @return   the current value of the alignment property
    */
   public int getAlignment() 
   {
      return alignment;    
   }
   
   /**
    * Sets the value of the alignment property.
    * 
    * @param aAlignment the new value of the alignment property
    */
   public void setAlignment(int aAlignment) 
   {
      alignment = aAlignment;    
   }
   
   /**
    * Access method for the bottomMargin property.
    * 
    * @return   the current value of the bottomMargin property
    */
   public int getBottomMargin() 
   {
      return bottomMargin;    
   }
   
   /**
    * Sets the value of the bottomMargin property.
    * 
    * @param aBottomMargin the new value of the bottomMargin property
    */
   public void setBottomMargin(int aBottomMargin) 
   {
      bottomMargin = aBottomMargin;    
   }
   
   /**
    * Access method for the cellSpacing property.
    * 
    * @return   the current value of the cellSpacing property
    */
   public int getCellSpacing() 
   {
      return cellSpacing;    
   }
   
   /**
    * Sets the value of the cellSpacing property.
    * 
    * @param aCellSpacing the new value of the cellSpacing property
    */
   public void setCellSpacing(int aCellSpacing) 
   {
      cellSpacing = aCellSpacing;    
   }
   
   /**
    * Access method for the currentCell property.
    * 
    * @return   the current value of the currentCell property
    */
   public TCell getCurrentCell() 
   {
      return currentCell;    
   }
   
   /**
    * Access method for the currentHLine property.
    * 
    * @return   the current value of the currentHLine property
    */
   public int getCurrentHLine() 
   {
      return currentHLine;    
   }
   
   /**
    * Access method for the currentNPage property.
    * 
    * @return   the current value of the currentNPage property
    */
   public int getCurrentNPage() 
   {
      return currentNPage;    
   }
   
   /**
    * Access method for the currentNPage property.
    * 
	* @param i
	*/
   public void setCurrentNPage(int i) {
	   currentNPage = i;
   }
   
   /**
    * Access method for the currentPage property.
    * 
    * @return   the current value of the currentPage property
    */
   public TCell getCurrentPage() 
   {
      return currentPage;    
   }
   
   /**
    * Access method for the currentPosition property.
    * 
    * @return   the current value of the currentPosition property
    */
   public int getCurrentPosition() 
   {
      return currentPosition;    
   }
   
   /**
    * Sets the value of the currentPosition property.
    * 
    * @param aCurrentPosition the new value of the currentPosition property
    */
   public void setCurrentPosition(int aCurrentPosition) 
   {
      currentPosition = aCurrentPosition;    
   }
   
   /**
    * Access method for the currentWLine property.
    * 
    * @return   the current value of the currentWLine property
    */
   public int getCurrentWLine() 
   {
      return currentWLine;    
   }
   
   /**
    * Access method for the currentXPos property.
    * 
    * @return   the current value of the currentXPos property
    */
   public int getCurrentXPos() 
   {
      return currentXPos;    
   }
   
   /**
    * Access method for the currentYPos property.
    * 
    * @return   the current value of the currentYPos property
    */
   public int getCurrentYPos() 
   {
      return currentYPos;    
   }
   
   /**
    * Access method for the defaultFont property.
    * 
    * @return   the current value of the defaultFont property
    */
   public Font getDefaultFont() 
   {
      return defaultFont;    
   }
   
   /**
    * Sets the value of the defaultFont property.
    * 
    * @param aDefaultFont the new value of the defaultFont property
    */
   public void setDefaultFont(Font aDefaultFont) 
   {
      defaultFont = aDefaultFont;    
   }
   
   /**
    * Access method for the device property.
    * 
    * @return   the current value of the device property
    */
   public Device getDevice() 
   {
      return device;    
   }
   
   /**
    * Determines if the drawPopup property is true.
    * 
    * @return   <code>true<code> if the drawPopup property is true
    */
   public boolean getDrawPopup() 
   {
      return drawPopup;    
   }
   
   /**
    * Sets the value of the drawPopup property.
    * 
    * @param aDrawPopup the new value of the drawPopup property
    */
   public void setDrawPopup(boolean aDrawPopup) 
   {
      drawPopup = aDrawPopup;    
   }
   
   /**
    * Access method for the firstCellLine property.
    * 
    * @return   the current value of the firstCellLine property
    */
   public TCell getFirstCellLine() 
   {
      return firstCellLine;    
   }
   
   /**
    * Determines if the flatPopup property is true.
    * 
    * @return   <code>true<code> if the flatPopup property is true
    */
   public boolean getFlatPopup() 
   {
      return flatPopup;    
   }
   
   /**
    * Sets the value of the flatPopup property.
    * 
    * @param aFlatPopup the new value of the flatPopup property
    */
   public void setFlatPopup(boolean aFlatPopup) 
   {
      flatPopup = aFlatPopup;    
   }
   
   /**
    * Access method for the footerCell property.
    * 
    * @return   the current value of the footerCell property
    */
   public TCell getFooterCell() 
   {
      return footerCell;    
   }
   
   /**
    * Sets the value of the footerCell property.
    * 
    * @param aFooterCell the new value of the footerCell property
    */
   public void setFooterCell(TCell aFooterCell) 
   {
      footerCell = aFooterCell;    
   }
   
   /**
    * Access method for the gc property.
    * 
    * @return   the current value of the gc property
    */
   public GC getGc() 
   {
      return gc;    
   }
   
   /**
    * Access method for the headerCell property.
    * 
    * @return   the current value of the headerCell property
    */
   public TCell getHeaderCell() 
   {
      return headerCell;    
   }
   
   /**
    * Sets the value of the headerCell property.
    * 
    * @param aHeaderCell the new value of the headerCell property
    */
   public void setHeaderCell(TCell aHeaderCell) 
   {
      headerCell = aHeaderCell;    
   }
   
   /**
    * Access method for the imagePool property.
    * 
    * @return   the current value of the imagePool property
    */
   public Hashtable getImagePool() 
   {
      return imagePool;    
   }
   
   /**
    * Sets the value of the imagePool property.
    * 
    * @param aImagePool the new value of the imagePool property
    */
   public void setImagePool(Hashtable aImagePool) 
   {
      imagePool = aImagePool;    
   }
   
   /**
	* Access method for the imagePool property.
	* 
	* @return   the current value of the imagePool property
	*/
   public Hashtable getFontPool() 
   {
		 return fonts;    
   } 
   
   /**
	* Sets the value of the imagePool property.
	* 
	* @param aImagePool the new value of the imagePool property
	*/
   public void setFontPool(Hashtable aFontPool) 
   {
		 fonts = aFontPool;    
   }
   
   /**
    * Access method for the leftMargin property.
    * 
    * @return   the current value of the leftMargin property
    */
   public int getLeftMargin() 
   {
      return leftMargin;    
   }
   
   /**
    * Sets the value of the leftMargin property.
    * 
    * @param aLeftMargin the new value of the leftMargin property
    */
   public void setLeftMargin(int aLeftMargin) 
   {
      leftMargin = aLeftMargin;    
   }
   
   /**
    * Access method for the lineSpacing property.
    * 
    * @return   the current value of the lineSpacing property
    */
   public int getLineSpacing() 
   {
      return lineSpacing;    
   }
   
   /**
    * Sets the value of the lineSpacing property.
    * 
    * @param aLineSpacing the new value of the lineSpacing property
    */
   public void setLineSpacing(int aLineSpacing) 
   {
      lineSpacing = aLineSpacing;    
   }
   
   /**
    * Access method for the listOffsetSpacing property.
    * 
    * @return   the current value of the listOffsetSpacing property
    */
   public int getListOffsetSpacing() 
   {
      return listOffsetSpacing;    
   }
   
   /**
    * Sets the value of the listOffsetSpacing property.
    * 
    * @param aListOffsetSpacing the new value of the listOffsetSpacing property
    */
   public void setListOffsetSpacing(int aListOffsetSpacing) 
   {
      listOffsetSpacing = aListOffsetSpacing;    
   }
   
   /**
    * Access method for the maxWidth property.
    * 
    * @return   the current value of the maxWidth property
    */
   public int getMaxWidth() 
   {
      return maxWidth;    
   }
   
   /**
    * Sets the value of the maxWidth property.
    * 
    * @param aMaxWidth the new value of the maxWidth property
    */
   public void setMaxWidth(int aMaxWidth) 
   {
	  maxWidth = Math.max(aMaxWidth, getMinWidth());
   }
   
   /**
    * Access method for the minWidth property.
    * 
    * @return   the current value of the minWidth property
    */
   public int getMinWidth() 
   {
      return minWidth;    
   }
   
   /**
    * Sets the value of the minWidth property.
    * 
    * @param aMinWidth the new value of the minWidth property
    */
   public void setMinWidth(int aMinWidth) 
   {
      minWidth = aMinWidth;    
   }
   
   /**
    * Determines if the modePage property is true.
    * 
    * @return   <code>true<code> if the modePage property is true
    */
   public boolean getModePage() 
   {
      return (getPageHeight()>0);    
   }
   
   /**
    * Access method for the pageBegin property.
    * 
    * @return   the current value of the pageBegin property
    */
   public int getPageBegin() 
   {
      return pageBegin;    
   }
   
   /**
    * Sets the value of the pageBegin property.
    * 
    * @param aPageBegin the new value of the pageBegin property
    */
   public void setPageBegin(int aPageBegin) 
   {
      pageBegin = aPageBegin;    
   }
   
   /**
    * Access method for the pageEnd property.
    * 
    * @return   the current value of the pageEnd property
    */
   public int getPageEnd() 
   {
      return pageEnd;    
   }
   
   /**
    * Sets the value of the pageEnd property.
    * 
    * @param aPageEnd the new value of the pageEnd property
    */
   public void setPageEnd(int aPageEnd) 
   {
      pageEnd = aPageEnd;    
   }
   
   /**
    * Access method for the pageHeight property.
    * 
    * @return   the current value of the pageHeight property
    */
   public int getPageHeight() 
   {
      return pageHeight;    
   }
   
   /**
    * Sets the value of the pageHeight property.
    * 
    * @param aPageHeight the new value of the pageHeight property
    */
   public void setPageHeight(int aPageHeight) 
   {
      pageHeight = aPageHeight;    
   }
   
   /**
    * Access method for the rightMargin property.
    * 
    * @return   the current value of the rightMargin property
    */
   public int getRightMargin() 
   {
      return rightMargin;    
   }
   
   /**
    * Sets the value of the rightMargin property.
    * 
    * @param aRightMargin the new value of the rightMargin property
    */
   public void setRightMargin(int aRightMargin) 
   {
      rightMargin = aRightMargin;    
   }
   
   /**
    * Access method for the titleSpacing property.
    * 
    * @return   the current value of the titleSpacing property
    */
   public int getTitleSpacing() 
   {
      return titleSpacing;    
   }
   
   /**
    * Sets the value of the titleSpacing property.
    * 
    * @param aTitleSpacing the new value of the titleSpacing property
    */
   public void setTitleSpacing(int aTitleSpacing) 
   {
      titleSpacing = aTitleSpacing;    
   }
   
   /**
    * Access method for the topMargin property.
    * 
    * @return   the current value of the topMargin property
    */
   public int getTopMargin() 
   {
      return topMargin;    
   }
   
   /**
    * Sets the value of the topMargin property.
    * 
    * @param aTopMargin the new value of the topMargin property
    */
   public void setTopMargin(int aTopMargin) 
   {
      topMargin = aTopMargin;    
   }
   
   /**
    * Determines if the wordWrap property is true.
    * 
    * @return   <code>true<code> if the wordWrap property is true
    */
   public boolean getWordWrap() 
   {
      return wordWrap;    
   }
   
   /**
    * Sets the value of the wordWrap property.
    * 
    * @param aWordWrap the new value of the wordWrap property
    */
   public void setWordWrap(boolean aWordWrap) 
   {
      wordWrap = aWordWrap;    
   }
   
   /**
    * Access method for the xoffset property.
    * 
    * @return   the current value of the xoffset property
    */
   public int getXoffset() 
   {
      return xoffset;    
   }
   
   /**
    * Sets the value of the xoffset property.
    * 
    * @param aXoffset the new value of the xoffset property
    */
   public void setXoffset(int aXoffset) 
   {
      xoffset = aXoffset;    
   }
   
   /**
    * Access method for the footer property.
    * 
    * @return   the current value of the footer property
    */
   public DFooter getFooter() 
   {
      return footer;    
   }
   
   /**
    * Sets the value of the footer property.
    * 
    * @param aFooter the new value of the footer property
    */
   public void setFooter(DFooter aFooter) 
   {
      footer = aFooter;    
   }
   
   /**
    * Access method for the header property.
    * 
    * @return   the current value of the header property
    */
   public DHeader getHeader() 
   {
      return header;    
   }
   
   /**
    * Sets the value of the header property.
    * 
    * @param aHeader the new value of the header property
    */
   public void setHeader(DHeader aHeader) 
   {
      header = aHeader;    
   }
   
   /**
    * Access method for the firstCell property.
    * 
    * @return   the current value of the firstCell property
    */
   public TCell getFirstCell() 
   {
      return firstCell;    
   }
   
   /**
    * Sets the value of the firstCell property.
    * 
    * @param aFirstCell the new value of the firstCell property
    */
   public void setFirstCell(TCell aFirstCell) 
   {
      firstCell = aFirstCell;    
   }
   
   /**
    * @param w
    * @param h
    * @param item
    * @param count
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF9510A011C
    */
   public TCell addCell(int w, int h, IDItem item, int count) 
   {
   	  TCell c = new TCell(); //currentPage, currentCell);
   	  addCell(w, h, item, count, c);
   	  return c;    
   }
   
   /**
    * @param w
    * @param h
    * @param item
    * @param count
    * @param c
    * @roseuid 3F00143402E1
    */
   public void addCell(int w, int h, IDItem item, int count, TCell c) 
   {   
   	   if (currentPage==null)
   	   {
		   TCell pageCell = new TCell();
		   if (firstCell == null)
		   {
				firstCell  = pageCell;
		   }
		   currentPage  = pageCell;
		   currentCell  = null;
		   currentYPos  = 0;
		   currentXPos = getXoffset();
   	   }
   	
   	   // accrochage dans l'arbre
   	   if (currentPage.getChild() == null)
	       currentPage.setChild(c);
       c.setParent(currentPage);
       c.setPrevious(currentCell);
       
       if (currentCell != null)
          currentCell.setNext(c);
       
	   if (firstCellLine == null)
	   {
			// new line
			c.setRect(currentXPos + getLeftMargin(), currentYPos + getTopMargin(), w, h);
			firstCellLine = c;
			lastCellLine = c;
   	  	
			currentWLine = currentWLine + w;
			currentHLine = h;
			currentXPos = w;
			currentNLine = currentNLine + 1;
	   }
	   else
	   {
			c.setRect(currentXPos + getLeftMargin(), currentYPos + getTopMargin(), w, h);
			currentHLine = Math.max(currentHLine, h);
			currentWLine = currentWLine + w;
			currentXPos = currentXPos + w;
	   }
       
       // position des caracteres
	   c.setPosition(currentPosition);
	   c.setNumberLine(currentNLine);
	   c.setItem(item);
	   currentPosition = currentPosition + count;   
	   
	   // mise a jour de la taille du parent
	   if (currentPage != null)
	   {
		  currentPage.getRect().height = getTopMargin()+ currentYPos + currentHLine + getBottomMargin();
	      currentPage.getRect().width  = Math.max(currentPage.getRect().width, getLeftMargin() + currentWLine + getRightMargin());
	   }
       
       currentCell = c;    
   }
   
   /**
    * @param s
    * @roseuid 3EFAF6900056
    */
   public void addXSpacing(int s) 
   {
      currentWLine = currentWLine + s;    
   }
   
   /**
    * @param s
    * @roseuid 3EFAF6BA00BA
    */
   public void addYSpacing(int s) 
   {
      currentHLine = currentHLine + s;    
   }
   
   /**
    * @param item
    * @roseuid 3EFAA5FD0218
    */
   protected void applyFont(IDItem item) 
   {
   	   Font font = null;
   	   previousFont = getGc().getFont();
	   	
	   if ((item.getStyle() != null)
			&& (item.getStyle().getFont() != null))
	   {
			IDFont f = item.getStyle().getFont();
			font = (Font)fonts.get(f.getKey());
			
			if (font == null)
			{
				font = DrawUtil.getFont(f);
				fonts.put(f.getKey(), font);
			}
	   }
	   else
	   {
	   	   font = getDefaultFont();
	   }   
	   
	   getGc().setFont(font);    
   }
   
   protected void restoreFont()
   {
   	  if (previousFont!=null)
   	     getGc().setFont(previousFont);
   }
   
   
   
   /**
    * @roseuid 3EF849AE0212
    */
   public void dispose() 
   {
   	  clear();
   	  
      // dispose all fonts
      if (fonts != null)
      {
	     for (Enumeration e = fonts.elements() ; e.hasMoreElements() ;) 
	     {
		    ((Font)e.nextElement()).dispose();
	     }
	     fonts.clear();
      }
	   
      //	reset all imagePool
      if (imagePool != null)
      {
	     for (Enumeration e = imagePool.elements() ; e.hasMoreElements() ;) 
	     {
		    ((Image)e.nextElement()).dispose();
	     }
	     imagePool.clear();
      }
   }
   
   
   /**
    * @return int
    * @roseuid 3EF8475300C0
    */
   public int getHeight() 
   {
	  int h = 0;
	  for (TCell c = getFirstCell(); c != null; c = c.getNext() )
	  {
		h = h +  c.getRect().height;
	  }
	  return h;    
   }
   
   /**
    * @param i
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EFACC7B0159
    */
   public TCell getPage(int i) 
   {
	 int j = 0;
     for (TCell c = getFirstCell(); c != null; c = c.getNext() )
	 {
	    if (i == j) return c;
		j = j + 1; 
	 }
     return  null;    
   }
   
   /**
    * @return int
    * @roseuid 3EFACC4101BA
    */
   public int getPageCount() 
   {
   	  int i = 0;
	  for (TCell c = getFirstCell(); c != null; c = c.getNext() )
	  {
		 i = i + 1;
	  }
	  return i;    
   }
   
   /**
    * @return int
    * @roseuid 3EF847650152
    */
   public int getWidth() 
   {
   	  int w = 0;
	  for (TCell c = getFirstCell(); c != null; c = c.getNext() )
	  {
		 w = Math.max(w, c.getRect().width);
	  }
      return w;    
   }
   
   /**
    * @param p
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF95FE600FE
    */
   public TCell mapPointToCell(Point p) 
   {
	  return mapPointToCell(getFirstCell(), p);    
   }
   
   /**
    * @param c
    * @param p
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF96C530167
    */
   protected TCell mapPointToCell(TCell c, Point p) 
   {
	 TCell i;
	 for (i = c.getChild(); i != null; i = i.getNext())
	 {
	   if ( ((i.getRect().y <= p.y) && (p.y <= i.getRect().y + i.getRect().height))
			&& ((i.getRect().x <= p.x) && (p.x <= i.getRect().x + i.getRect().width)))
	   {
				break;
	   }
	 }
	   
	 if (i==null) return null;
	   
	 // la cellule est trouvee si elle en contient d'autre on y va en 
	 // profondeur
	 if (i.getChild()!=null)
	 {
	 	 p.x = p.x - i.getRect().x;
	 	 p.y = p.y - i.getRect().y;
		 TCell cf = mapPointToCell(i, p);
	   	 
		 if (cf!=null) return cf;
	 }
	   
	 p.x = p.x - i.getRect().x;
	 p.y = p.y - i.getRect().y;
	 return  i;    
   }
   
   /**
    * @param pos
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF9610F0282
    */
   public TCell mapPositionToCell(int pos) 
   {
      return mapPositionToCell(getFirstCell(), pos);    
   }
   
   /**
    * @param c
    * @param pos
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF96C810159
    */
   protected TCell mapPositionToCell(TCell c, int pos) 
   {
	 TCell i;
     for (i = c.getChild(); i != null; i = i.getNext())
	 {
	 	if (i.getChild()!=null)
	 	{
	 		TCell j = mapPositionToCell(i, pos);
	 		if (j!=null) return j;
	 	}
	 	if ((i.getPosition() <= pos)
	 	     && (pos <= i.getPosition() + i.getLen()))
	 	{
	 		return i;
	 	}
	 }
	 return null;    
   }
   
   /**
    * @roseuid 3EF9A6E0028C
    */
   public void newLine() 
   {
      if (firstCellLine != null)
      {
      	int xoffset = getLeftMargin();
      	
      	if (getAlignment() == DParagraph.RIGHT)
      	       xoffset = xoffset + getWidth() - currentWLine;
      	else
      	if (getAlignment() == DParagraph.CENTER)
      	{
      	       xoffset = xoffset + (getWidth() - currentWLine)/2; 
      	}
      	
      	int maxDsc = 0;
      	TCell c;
      	for (c = firstCellLine; c!=null; c = c.getNext())
      	{
      		if (c.getItem() != null)
      		{
      			applyFont(c.getItem());
      			maxDsc=Math.max(maxDsc, getGc().getFontMetrics().getDescent());
				restoreFont();
      		}
      	}
      	
      	for (c = firstCellLine; c!=null; c = c.getNext())
      	{
      		int locDsc;
      		if (c.getItem()!=null)
      		{
				applyFont(c.getItem());
      			locDsc = getGc().getFontMetrics().getDescent();
				restoreFont();
      			c.setRect(xoffset + c.getRect().x, 
      			          c.getRect().y - (maxDsc-locDsc)+(currentHLine-c.getRect().height), 
      			          c.getRect().width, c.getRect().height);
      			//xoffset = xoffset + c.getRect().width;
      		}
      	}
      }
      
      if (getModePage())
      {
      	 int fh = 0;
      	 if (getFooterCell()!=null)
      	 {
			fh = fh + getFooterCell().getRect().height;
      	 }
      	 
      	 if ((currentYPos + currentHLine + fh) > getPageHeight())
      	 { 
      	 	newPage();
      	 }
      	 else
      	 {
			currentYPos = currentYPos + currentHLine;
      	 }
      }
      else
      {   
         currentYPos = currentYPos + currentHLine;
      }
      
	  currentXPos = getXoffset(); 
      
      currentHLine  = 0;
      currentWLine  = 0;
      firstCellLine = null;    
   }
   
   /**
    * @roseuid 3EFAF84300A1
    */
   public void newPage() 
   {
	  // creation de la page
      TCell pageCell = new TCell();
      if (firstCell == null)
	     firstCell  = pageCell;
      else
      {
		 currentPage.setNext(pageCell);
      }
         
      currentPage  = pageCell;
      currentNPage = currentNPage + 1;
      currentCell  = null;
      int hLine = 0;
	  
      // ajout de l'entete s'il y en a une
      if (getHeader()!=null)
      {
	    doItem(getHeader());
	    
	    if (getHeaderCell()!=null)
	    {
	       // retour a la ligne
		   currentYPos  = 0;
		   currentXPos  = getXoffset();
		   currentWLine = 0;
		   currentHLine = 0;
	    
	       addCell(getHeaderCell().getRect().width, 
						 getHeaderCell().getRect().height, 
						 getHeader(), 0, getHeaderCell());
						 
		   hLine = getHeaderCell().getRect().height;
	    }
      }   
      else
      {
	    if (firstCellLine != null)
	    {
		   firstCellLine.setPrevious(null);
		   pageCell.setChild(firstCellLine);
	    }
      }
      
      // ajout du bas de page en fin de la page
      if (getFooter()!=null)
      {
		doItem(getFooter());
      	
		if (getFooterCell()!=null)
		{
            //	retour a la ligne plus positionnement en bas
			currentYPos  = getPageHeight()-getFooterCell().getRect().height;
			currentXPos  = getXoffset();
			currentWLine = 0;
			currentHLine = 0;	
			 
			addCell(getFooterCell().getRect().width, 
					getFooterCell().getRect().height, 
					getFooter(), 0, getFooterCell());
		}     	
      }
         
      TCell c;
      for (c = firstCellLine; c!=null; c = c.getNext())
      {
	     c.setParent(pageCell);
      }    
      
	  currentYPos = hLine;
	  currentXPos = getXoffset();
      currentHLine  = 0;
      currentWLine  = 0;
      firstCellLine = null;    
   }
   
   /**
    * @roseuid 3EFAF25A020E
    */
   public void clear() 
   {
   	   // clears all pages
	   for (TCell c = getFirstCell(); c != null; )
	   {
			 TCell p = c.getNext();
			 c.clear();
			 c.setRect(0,0,0,0);
			 c.setPosition(1);
		 
			 c.setNext(null);
			 c = p;
	   }
	   
	   super.clear();
	   
       currentPosition = 1;
	   currentNLine = 0; 
	   currentXPos = 0;
	   currentYPos = 0;
	   firstCellLine = null;
	   lastCellLine = null;
	   currentCell = null;
	   currentHLine = 0;
	   currentWLine = 0;    
	   alignment = DParagraph.LEFT;
	   currentPage = null;    
	   firstCell = null;
	   currentNPage = 0;    
   }
   
   /**
    * @param sp
    * @param ep
    * @roseuid 3EF9A56C002E
    */
   public void setPageRange(int sp, int ep) 
   {
      setPageBegin(sp);
      setPageEnd(ep);    
   }
   
   /**
    * @roseuid 3F7294070388
    */
   public void dump() 
   {
   	  int i = 0;
   	  for (TCell c = firstCell; c != null; c = c.getNext())
   	  {
   	  	 i = i +1;
   	  	 System.out.println("------------------ Page (" + i + ")-------------------");
   	  	 c.dump(0);
   	  }
   }


   /**
    * @param i
    */
   public void setCurrentYPos(int i) 
   {
	   currentYPos = i;
   }

    /**
     * @return
     */
    public Image getBrokenImage() {
	    if(brokenImage == null){
		   brokenImage = new Image(getDevice(),30,30);
	    }
	    return brokenImage;
   }

   /**
     * @param image
    */
   public void setBrokenImage(Image image) {
	   brokenImage = image;
   }

   /**
	* @param c
	* @roseuid 3EFC03260070
	*/
   public void doItem(DCell c) 
   {
   	  applyFont(c);
   	
	  int size = _getColumnSizeForCell(c);
   	  
	  TBuilder b = _newBuilder(size);
	 	
	  _setBuilderMarginForCell(b);
	
	  b.doChildrenItem(c);
		
	  TCell cl = b.getFirstCell();
	  if (cl != null)
	  {
		if (currentTable.getAutoResize())
			addCell(b.getMaxWidth(), cl.getRect().height, c, 0, cl);
		else
			addCell(cl.getRect().width, cl.getRect().height, c, 0, cl);
	  	
		setCurrentPosition(b.getCurrentPosition());
	    
		// mise en place des marges
		if (currentNColumn > 0)
		{
			cl.getRect().x -= currentTable.getBorder()-1;
		}
		if (currentNRow > 0)
		{
			cl.getRect().y -= currentTable.getBorder()-1;
		}
	  }

	  // on met fin au constructeur 
	  _disposeBuilder(b);
	  
	  CellInfo cf = new CellInfo();
	  cf.rowspan = c.getRowSpan();
	  cf.cell = cl;
	  
	  if (c.getRowSpan()==0)
	  {
		maxCellHeight = Math.max(maxCellHeight, cl.getRect().height);
	  }
	  
	  currentRowSpan[currentNColumn] = cf;
	  currentNColumn = currentNColumn + c.getColSpan() + 1;
	  
	  restoreFont();    
   }
   
   /**
	* @param c
	* @roseuid 3EFC0332009F
	*/
   public void doItem(DCellText c) 
   {
   	  applyFont(c);
   	
	  int size = _getColumnSizeForCell(c);
	  
	  TBuilder b = _newBuilder (size);
	  _setBuilderMarginForCell(b);
	 
	  _buildTextBuffer(c, c.getText());	 
	 	
	  TCell cl = b.getFirstCell();
	   
	  addCell(cl.getRect().width, cl.getRect().height, c, 0, b.getFirstCell());
	  setCurrentPosition(b.getCurrentPosition());
	  
	  // on met fin au constructeur 
	  _disposeBuilder(b);	
	 
	  CellInfo cf = new CellInfo();
	  cf.rowspan = c.getRowSpan();
	  cf.cell = cl;
	  
	  if (c.getRowSpan()==0)
	  {
		  maxCellHeight = Math.max(maxCellHeight, cl.getRect().height);
	  }
	  
	  currentRowSpan[currentNColumn] = cf;
	  currentNColumn = currentNColumn + c.getColSpan() + 1;
	  
	  restoreFont();    
   }
   
   /**
	* @param d
	* @roseuid 3EFC0213029B
	*/
   public void doItem(DDocument d) 
   {
   	  applyFont(d);
   	
	  setAlignment(DParagraph.LEFT);
	  doChildrenItem(d);
	  
	  restoreFont();    
   }
   
   /**
	* @param g
	* @roseuid 3EFC03750268
	*/
   public void doItem(DGraphic g) 
   {
   	  applyFont(g);
   	
	  int w = (int)((float)getMaxWidth() * g.getWidthScale());
	  int h = (int)((float)getMaxWidth() * g.getHeightScale());
	  float ratio = g.getWidthScale()/g.getHeightScale();
	  if (ratio <= 1.0f)
	  {
		if (w < 100)
		{
			w = 100;
			h = (int)(w/ratio);
		}
		else
		if (h < 100)
		{
			h = 100;
			w = (int)(h*ratio);
      		
		}
	  }
      
	  addCell(w, h, g, 0);
	  
	  restoreFont();     
   }
   
   /**
	* @param i
	* @roseuid 3EFC02FC00CA
	*/
   public void doItem(DImage i) 
   {
	 _addListItemPrefix(i);
	  
	 // search image 
	 Image img = (Image)getImagePool().get(i.getName());
	 if (img == null)
	 {
		String imgfilename = searchImage(i.getName());
		if (imgfilename != null)
		{
		   img = new Image(getDevice(), i.getName());
		   getImagePool().put(i.getName(), img);
		}
		else
		{
		   img = (Image)getBrokenImage();
		}
	 }
  
	 Rectangle r = img.getBounds();
	 int w = (int)(r.width * i.getWidthScale());
	 int h = (int)(r.height * i.getHeightScale());
	      
	 if (getWordWrap() 
		 && (w + getCurrentWLine()) > getMaxWidth())
	 {
		 newLine();	
	 }
     
	 TCell c = addCell(w, h, i, 0);
	 c.setImage(img);    
   }
   
   
   /**
	* @param l
	* @roseuid 3EFC02AF03CC
	*/
   public void doItem(DList l) 
   {
   	  applyFont(l);
   	
	  newLine();

	  IDStyle s = l.getStyle();
	
	  addListLevel();
	  setXoffset(getXoffset() * getListOffsetSpacing());
		
	  doChildrenItem(l);
	
	  removeListLevel();
	  setXoffset(getXoffset() * getListOffsetSpacing());
	
	
	  newLine();   
	
	  restoreFont(); 
   }
   
   /**
	* @param l
	* @roseuid 3EFC02AF03CC
	*/
   public void doItem(DLine l) 
   {
	   newLine();
	
	   addCell(getMaxWidth(),
				l.getSize()+getCellSpacing()*2,l,0);
	
	   newLine();    
   }
   
   /**
	* @param pb
	* @roseuid 3F004C3A029E
	*/
   public void doItem(DPageBreak pb) 
   {    
	  if (getModePage())
		  newPage();
	  else
		  newLine();
   }
   
   /**
	* @param pc
	* @roseuid 3F004C20039B
	*/
   public void doItem(DPageCounter pc) 
   { 
   	  applyFont(pc);
   	
	  String s = new Integer(getCurrentNPage()).toString();
      
	  Point p = getGc().textExtent(s);
	  TCell c = addCell(p.x, p.y, pc, s.length());
	  c.setText(s); 
	  
	  restoreFont();   
   }
   
   /**
	* @param p
	* @roseuid 3EFC025D0388
	*/
   public void doItem(DParagraph p) 
   {
   	  applyFont(p);
   	
	  int al = getAlignment();
	  newLine();
	  setAlignment(p.getAlign());
	  
	  _addListItemPrefix(p);
	  
	  doChildrenItem(p);
	  
	  newLine();
	  setAlignment(al);
	  
	  restoreFont();    
   }
   
   /**
	* @param p
	* @roseuid 3EFC02DE0329
	*/
   public void doItem(DPopup p) 
   {
   	  applyFont(p);
   	
	  if (getDrawPopup())
	  {
		setAlignment(DParagraph.LEFT);
		doChildrenItem(p);
	  }    
	  
	  restoreFont();
   }
   
   /**
	* @param p
	* @roseuid 3EFC02CA0280
	*/
   public void doItem(DPopupLink p) 
   {
   	 applyFont(p);
   	
	 doChildrenItem(p);
	 if (!getFlatPopup()) return;
	 
	 DLinkResolver r = new DLinkResolver(p);
	 DPopup popup = (DPopup)r.getTaggedItem();
	 
	 if (popup!=null)
	 {
		TBuilder b = _newBuilder(getMaxWidth());
	 	
		int m = getCellSpacing();
		b.setTopMargin(m);
		b.setBottomMargin(m);
		b.setLeftMargin(m);
		b.setRightMargin(m);
		
		b.doChildrenItem(popup);
		
		TCell c = b.getFirstCell();
		
		addCell(c.getRect().width, c.getRect().height, popup, 0);   
		setCurrentPosition(b.getCurrentPosition());

		// on met fin au constructeur 
		_disposeBuilder(b);	
	 }    
	 
	 restoreFont();
   }
   
   /**
	* @param r
	* @roseuid 3EFC03190216
	*/
   public void doItem(DRow r) 
   {
   	  applyFont(r);
   	
	  currentNColumn = 0;
	  currentRow = r;
	  maxCellHeight = 0;
	  
	  // calcul numero de colonne de la premiere cellule de la liste
	  // a partir de la deuxieme ligne
	  if (currentNRow>0)
	  {
		for (currentNColumn=0; 
			   (currentRowSpan[currentNColumn].rowspan>0) 
				  && (currentNColumn<currentRowSpan.length); currentNColumn++);
	  }
	  
	  
	  doChildrenItem(r);
	  
	  if (currentNRow>0)
	  {
		int w = 0;
		for (int i=0; i<currentRowSpan.length; i++)
		{
		  if (currentRowSpan[i]==null) continue;
		  TCell c = currentRowSpan[i].cell;
		  
		  if (currentRowSpan[i].rowspan>0)
		  { 
			 w += c.getRect().width;
			 currentRowSpan[i].rowspan -= 1;
			 if (currentRowSpan[i].rowspan==0)
			 {
				c.getRect().height = (getCurrentYPos() + maxCellHeight) - c.getRect().y;
				currentRowSpan[i] = null;
			 }
		  }
		  else
		  {
			c.getRect().x = c.getRect().x + w;			
			c.getRect().height = maxCellHeight;
			w += c.getRect().width;
		  }
		}
	  }
	  
	  int currentYPos = getCurrentYPos();
	  newLine();
	  setCurrentYPos(maxCellHeight + currentYPos);
	  currentNRow = currentNRow + 1;
	  
	  restoreFont();    
   }
   
   /**
	* @param s
	* @roseuid 3EFC02410247
	*/
   public void doItem(DSection s) 
   {
   	  applyFont(s);
   	
	  setHeader(s.getHeader());
	  setFooter(s.getFooter());
	  
	  if (getModePage())
		  newPage();
	  else
		  newLine();
          
	  addCell(0,0,s,0);
      
	  doChildrenItem(s);
	  
	  restoreFont();    
   }
   
   /**
	* @param t
	* @roseuid 3EFC030901EB
	*/
   public void doItem(DTable t) 
   {
   	  applyFont(t);
   	
	  newLine();
	  currentTable = t;
	  currentNRow = 0;
	  currentTotalColumn = t.getColumnCount();
	  currentRowSpan = new CellInfo[t.getColumnCount()];
      
      
	  doChildrenItem(t);    
	  
	  currentTable = null;
	  currentRow = null;
	  currentRowSpan = null;
	  
	  restoreFont(); 
   }
   
   /**
	* @param t
	* @roseuid 3EFC03A001FC
	*/
   public void doItem(DTag t) 
   {
	  addCell(0,0,t,0);    
   }
   
   /**
	* @param t
	* @roseuid 3EFC02EE0142
	*/
   public void doItem(DText t) 
   {
   	  applyFont(t);
   	  
	  _addListItemPrefix(t);
	  _buildTextBuffer(t, t.getText());
	  
	  restoreFont();    
   }
   
   /**
	* @param t
	* @roseuid 3EFC02A00244
	*/
   public void doItem(DTitle t) 
   {
   	  applyFont(t);
   	
	  addTitleLevel(t.getLevel());
	  String st = getTitleStringPrefix(t);
	  
	  Point r = getGc().textExtent(st);
	  
	  addYSpacing(getTitleSpacing());
	  newLine();
	  
	  TCell c = addCell(r.x, r.y + getTitleSpacing(), t, st.length());
	  c.setText(st);
	  
	  doChildrenItem(t);
	  
	  newLine();   
	  
	  restoreFont(); 
   }
   
   /**
	* @param f
	* @roseuid 3F02AEBE025D
	*/
   public void doItem(DFooter f) 
   {
   	   applyFont(f);
   	
	   //	fabrication de l'entete et pied de base
	   TBuilder b = _newBuilder (getMaxWidth());
		
	   b.doChildrenItem(f);
		
	   setFooterCell(b.getFirstCell());
	   b.setFirstCell(null);
			
	   // on met fin au constructeur 
	   _disposeBuilder(b);
	   
	   restoreFont();    
   }
   
   /**
	* @param h
	* @roseuid 3F02AEAC03AC
	*/
   public void doItem(DHeader h) 
   {
   	   applyFont(h);
   	
	   // fabrication de l'entete et pied de base
	   TBuilder b = _newBuilder(getMaxWidth());
	 	
	   b.doChildrenItem(h);
		
	   setHeaderCell(b.getFirstCell());
	   b.setFirstCell(null);
			
	   // on met fin au constructeur 
	   _disposeBuilder(b);
	   
	   restoreFont();    
   }
   
   /**
	* @param text
	* @param w
	* @param r
	* @roseuid 3F02AF060329
	*/
   private void _buildGetNextWord(String text, int w, Range r) 
   {
	  String qs = text.substring(r.start);
	  int wt = getGc().textExtent(qs).x;
	  if (wt <= w)
	  {
		r.end = wt;
		r.start = text.length(); 
	  }
	  else
	  {
		int pos = ((w * qs.length()) / wt);
		for (int i=pos; i > 0; i = i -1)
		{
		   char c = qs.charAt(i);
		   if (!Character.isSpaceChar(c)
				&& (c!='.') && (c!=';') && (c!=',') && (c!='?')
				&& (c!=':') && (c!='(') && (c!=')') && (c!='/')
				&& (c!='[') && (c!=']') && (c!='\\'))
		   {
			  continue;
		   }

		   wt = getGc().textExtent(qs.substring(0, i)).x;
		   if (wt < w)
		   {
			  r.end = wt;
			  r.start = r.start + i;
			  return;
		   }
		}
		// on est pas arrive a decouper la chaine sur des separateurs
		// on va decouper les mots par dicotomie
		int a = qs.length() / 2;
		int b = qs.length();
		while (b-a>1)
		{
		   String s = qs.substring(0, a);
		   wt = getGc().textExtent(s).x;
		   if (wt < w)
		   {
			  a = a + (b-a)/2;
		   }
		   else
		   {
			  b = a;
			  a = a / 2;
		   }
		}
		r.end = wt;
		r.start = a + r.start;
	  }
   }

}



/*
 * Created on Sep 19, 2003
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package com.rational.apidoc.drivers.html;

import java.util.LinkedList;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;

import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.ImageLoader;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.SWTError;

import com.rational.apidoc.parsers.DParser;
import com.rational.apidoc.drivers.IWriter;
import com.rational.apidoc.core.*;
import com.rational.apidoc.examples.*;
import com.rational.apidoc.chart.ChartSnapShot;

/**
 * @author sduguet
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
public class DHtmlDriver extends DParser implements IWriter {

	private ConfParser confparser;
	private PrintStream out;

	private String dirname, filename;
	private DDocument document;
	
	private static Item default_hitem = new Item();

	/**
	 * 
	 */
	public DHtmlDriver() {
		super();
		confparser = new ConfParser();
		if (!confparser.parse()) {
			System.err.println("Html Configuration parsing failed");
			confparser = null;
			return;
		}
	}

	public void setDestDirAndFileName(String dname, String fname) {
		dirname = dname != null ? dname : ".";
		filename =
			dirname + "/" + fname + confparser.getHtmlFile().getExtension();
	}

	public PrintStream getStream() {
		return out;
	}

	public void htmlConvertDump(String string) {
		int space_count = 0;
		for (int i = 0; i < string.length(); i++) {
			char c = string.charAt(i);
			if (c == ' ') {
				space_count++;
				if (space_count == 1) {
					out.print(confparser.getConvertText().getFirstSpace());
				} else {
					out.print(confparser.getConvertText().getExtraSpace());
				}
			} else {
				space_count = 0;
				Character ch = new Character(c);
				String repl = confparser.getConvertText().getReplace(ch);
				String cstr = repl != null ? repl : ch.toString();
				try {
					byte a[] = cstr.getBytes("UTF-8");
					out.write(a, 0, a.length);
				} catch (UnsupportedEncodingException e) {
					out.print(cstr);
				}
			}
		}
	}

	public void htmlDump(String string) {
		try {
			byte a[] = string.getBytes("UTF-8");
			out.write(a, 0, a.length);
		} catch (UnsupportedEncodingException e) {
			out.print(string);
		}
	}

	public void htmlDump(int value) {
		out.print(value);
	}

	public void htmlDump(LinkedList list) {
		for (int i = 0; i < list.size(); i++) {
			String elt = (String) list.get(i);
			String value = confparser.getVariableValue(elt);
			if (value == "#") {
				htmlDump("${" + elt + "}");
			} else {
				htmlDump(value != null ? value : elt);
			}
		}
	}

	private void setTagVariable(String name, String tag, IDItem item) {
		confparser.setVariableValue(
			name,
			tag != null ? tag : "_" + item.hashCode(),
			item);
	}

	private String to2DigitsHexa(int color) {
		return Integer.toHexString((color % 256) / 16)
			+ Integer.toHexString(color % 16);
	}

	private String getColorHexa(IDColor color, String default_value) {
		if (color == null) {
			return default_value;
		}
		String result = "#";
		result += to2DigitsHexa(color.getRed());
		result += to2DigitsHexa(color.getGreen());
		result += to2DigitsHexa(color.getBlue());
		return result;
	}

	private String getFontAttrsBegin(IDStyle style) {
		IDFont font = style.getFont();
		IDColor color = style.getForeColor();
		if (font == null && color == null) {
			return "";
		}
		String result = "";
		if (font != null) {
			if (font.getBold()) {
				result += "<B>";
			}
			if (font.getUnderline() || font.getDashed()) {
				result += "<U>";
			}
			if (font.getItalic()) {
				result += "<I>";
			}
		}
		result += "<FONT COLOR=" + getColorHexa(color, "#000000");
		if (font != null && font.getSize() != -1) {
			int s = font.getSize() / 3; // size 12 -> 4 (normal size)
			if (s < 1) {
				s = 1; // 1 is the minimum
			}
			if (s > 7) {
				s = 7; // 7 is the maximum
			}
			result += " SIZE=\"" + Integer.toString(s) + "\"";
		}
		result += ">";
		if (font != null && confparser.isTTFont(font.getName())) {
			result += "<TT>";
		}

		return result;
	}

	private String getFontAttrsEnd(IDStyle style) {
		IDFont font = style.getFont();
		IDColor color = style.getForeColor();
		if (font == null && color == null) {
			return "";
		}
		String result = "";
		if (font != null && confparser.isTTFont(font.getName())) {
			result += "</TT>";
		}
		result += "</FONT>";
		if (font != null) {
			if (font.getBold()) {
				result += "</B>";
			}
			if (font.getUnderline() || font.getDashed()) {
				result += "</U>";
			}
			if (font.getItalic()) {
				result += "</I>";
			}
		}
		return result;
	}

	private void updateStyleVariables(IDItem item) {
		// PRE: item.getStyle() not null (to be checked by the caller)
		confparser.setVariableValue(
			"HTML_FONT_ATTRS_BEGIN",
			getFontAttrsBegin(item.getStyle()),
			item);
		confparser.setVariableValue(
			"HTML_FONT_ATTRS_END",
			getFontAttrsEnd(item.getStyle()),
			item);
		String fore_color =
			getColorHexa(item.getStyle().getForeColor(), "#000000");
		confparser.setVariableValue(
			"HTML_COLOR",
			"=\"" + fore_color + "\"",
			item);
		confparser.setVariableValue("HTML_FORE_COLOR", fore_color, item);
		String back_color =
			getColorHexa(item.getStyle().getBackColor(), "#ffffff");
		confparser.setVariableValue(
			"HTML_BGCOLOR",
			"=\"" + back_color + "\"",
			item);
		confparser.setVariableValue("HTML_BACK_COLOR", back_color, item);
	}

	public Item getItem(IDItem item) {
		// Perform some checks
		Item hitem = confparser.getItem(item.getClass());
		if (hitem == null) {
			System.err.println("No html parser found for " + item.getClass());
			hitem = default_hitem;
		}
		return hitem;
	}

	public void doOneItemAfter(IDItem item) {
		if (item instanceof DText) {
		} else if (item instanceof DCellText) {
		} else if (item instanceof DGraphic) {
			graphicDump((DGraphic) item);
		}
	}

	public void doItem(DDocument item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("DOCUMENT_TAG", item.getTag(), item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DSection item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("SECTION_TAG", ((DSection) item).getTag(), item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DHeader item) {
		// ignored
	}

	public void doItem(DFooter item) {
		// ignored
	}

	public void doItem(DParagraph item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"PARAGRAPH_ALIGN",
			((DParagraph) item).alignToString(((DParagraph) item).getAlign()),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DTag item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("TAG_NAME", ((DTag) item).getTag(), item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DText item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		htmlDump(hitem.getBegin());
		htmlConvertDump(item.getText());
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DImage item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"IMAGE_NAME",
			((DImage) item).getName(),
			item);
		htmlDump(hitem.getBegin());
		// no childs
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DList item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		if (((DList) item).getFormat() == DStyle.BULLET) {
			confparser.setVariableValue("LIST_TYPE", "BULLETED", item);
			confparser.setVariableValue("HTML_LIST_TYPE", "UL", item);
		} else if (((DList) item).getFormat() == DStyle.NUMBER) {
			confparser.setVariableValue("LIST_TYPE", "NUMBERED", item);
			confparser.setVariableValue("HTML_LIST_TYPE", "OL", item);
		} else {
			confparser.setVariableValue("LIST_TYPE", "", item);
			confparser.setVariableValue("HTML_LIST_TYPE", "UL", item);
		}
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DTable item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		if (((DTable) item).getAutoResize()) {
			float w = 0;
			float wc[] = ((DTable) item).getColumnWidth();
			if (wc == null || wc.length < ((DTable) item).getColumnCount()) {
				w = ((DTable) item).getWidthScale();
			} else {
				for (int i = 0; i < wc.length; i++) {
					w += wc[i];
				}
			}
			confparser.setVariableValue("TABLE_WIDTH", Float.toString(w), item);
			confparser.setVariableValue(
				"HTML_TABLE_WIDTH",
				Integer.toString((int) (w * 100.0f)),
				item);
		} else {
			float w = 0.0f;
			confparser.setVariableValue("TABLE_WIDTH", Float.toString(w), item);
			confparser.setVariableValue(
				"HTML_TABLE_WIDTH",
				Integer.toString((int) (w * 100.0f)),
				item);
		}
		confparser.setVariableValue(
			"TABLE_BORDER",
			Integer.toString(((DTable) item).getBorder()),
			item);
		confparser.setVariableValue(
			"TABLE_COLUMNS",
			Integer.toString(((DTable) item).getColumnCount()),
			item);
		confparser.setVariableValue(
			"TABLE_ROWS",
			Integer.toString(((DTable) item).getRowCount()),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DRow item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"ROW_CELLS",
			Integer.toString(((DRow) item).getCellCount()),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	private void setCellVariables(DCell item) {
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"CELL_COLSPAN",
			Integer.toString(((DCell) item).getColSpan()),
			item);
		confparser.setVariableValue(
			"HTML_CELL_COLSPAN",
			Integer.toString(((DCell) item).getColSpan() + 1),
			item);
		confparser.setVariableValue(
			"CELL_ROWSPAN",
			Integer.toString(((DCell) item).getRowSpan()),
			item);
		confparser.setVariableValue(
			"HTML_CELL_ROWSPAN",
			Integer.toString(((DCell) item).getRowSpan() + 1),
			item);
		confparser.setVariableValue(
			"CELL_COL",
			Integer.toString(((DCell) item).getColumnNumber()),
			item);
		confparser.setVariableValue(
			"CELL_ROW",
			Integer.toString(((DCell) item).getLineNumber()),
			item);
		DRow row = (DRow) item.getParent();
		DTable table = (DTable) row.getParent();
		if (table.getColumnWidth() != null
			&& table.getColumnCount() > ((DCell) item).getColumnNumber()) {
			float w = table.getColumnWidth()[((DCell) item).getColumnNumber()];
			confparser.setVariableValue("CELL_WIDTH", Float.toString(w), item);
			confparser.setVariableValue(
				"HTML_CELL_WIDTH",
				Integer.toString((int) (w * 100.0f)),
				item);
		}
	}

	public void doItem(DCell item) {
		Item hitem = getItem(item);
		setCellVariables(item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DCellText item) {
		Item hitem = getItem(item);
		setCellVariables(item);
		htmlDump(hitem.getBegin());
		htmlConvertDump(((DCellText) item).getText());
		// no childs
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DLink item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		String url =
			(((DLink) item).isInternal() ? "" : ((DLink) item).getMethod())
				+ ":"
				+ ((DLink) item).getPath();
		confparser.setVariableValue("LINK_TARGET", url, item);
		confparser.setVariableValue("HTML_LINK", url, item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DLine item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"HTML_LINE_SIZE",
			Integer.toString(((DLine) item).getSize()),
			item);
		htmlDump(hitem.getBegin());
		// no childs
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DPageBreak item) {
		// ignored
	}

	public void doItem(DPageCounter item) {
		// ignored
	}

	public void doItem(DPopup item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("POPUP_TAG", ((DPopup) item).getTag(), item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DPopupLink item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"POPUP_LINK_METHOD",
			((DPopupLink) item).typeToString(((DPopupLink) item).getType()),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DSummary item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DTitle item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("TITLE_TAG", ((DTitle) item).getTag(), item);
		addTitleLevel(((DTitle) item).getLevel());
		confparser.setVariableValue(
			"TITLE_LEVEL",
			Integer.toString(getTitleLevel()),
			item);
		confparser.setVariableValue(
			"HTML_TITLE_LEVEL",
			Integer.toString(getTitleLevel()),
			item);
		confparser.setVariableValue(
			"TITLE_FULL_NUMBER",
			getTitleStringPrefix(item),
			item);
		confparser.setVariableValue(
			"TITLE_NUMBER",
			Integer.toString(titleTab[getTitleLevel()]),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DIndex item) {
	}

	public void doItem(DIndexEntry item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DTimerItemContainer item) {
		// ignored
	}

	private void graphicDump(DGraphic graphic) {
		String image_filename =
			dirname
				+ "/"
				+ confparser.getVariableValue("HTML_GRAPHIC_FILE_NAME")
				+ ".jpg";
		int w = 500, h = w;
		w *= graphic.getWidthScale();
		h *= graphic.getHeightScale();
		Rectangle r = new Rectangle(0, 0, w, h);
		Image img = ChartSnapShot.Process(r, graphic);
		ImageLoader loader = new ImageLoader();
		if (loader.data == null) {
			loader.data = new ImageData[1];
		}
		loader.data[0] = img.getImageData();
		try {
			loader.save(image_filename, SWT.IMAGE_JPEG);
		} catch (SWTException _e) {
			System.err.println(
				"Error saving image:" + image_filename + ": " + _e);
		} catch (SWTError _r) {
			System.err.println(
				"Error saving image:" + image_filename + ": " + _r);
		}
		img.dispose();
	}

	public void doItem(DGraphic item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"GRAPHIC_TITLE",
			((DGraphic) item).getTitle(),
			item);
		confparser.setVariableValue(
			"GRAPHIC_WIDTH",
			Float.toString(((DGraphic) item).getWidthScale()),
			item);
		confparser.setVariableValue(
			"HTML_GRAPHIC_WIDTH",
			Integer.toString(
				(int) (((DGraphic) item).getWidthScale() * 100.0f)),
			item);
		confparser.setVariableValue(
			"GRAPHIC_HEIGHT",
			Float.toString(((DGraphic) item).getHeightScale()),
			item);
		confparser.setVariableValue(
			"HTML_GRAPHIC_FILE_NAME",
			"img" + Integer.toHexString(item.hashCode()),
			item);
		htmlDump(hitem.getBegin());
		graphicDump(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DAxis item) {
		// already made from the enclosing DGraphic
	}

	public void doItem(DCurve item) {
		// already made from the enclosing DGraphic
	}

	public void doItem(DPoint item) {
		// already made from the enclosing DGraphic
	}

	public void write(IDItem doc) {
		document = (DDocument) doc;

		if (confparser == null) {
			System.err.println("Html Configuration parsing failed");
			return;
		}
		try {
			out = new PrintStream(new FileOutputStream(filename));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return;
		}

		// Some variables at document level
		confparser.setVariableValue("HTML_FILE_NAME", filename, document);
		confparser.setVariableValue(
			"DOCUMENT_TITLE",
			document.getTitle(),
			document);
		confparser.setVariableValue("HTML_BGCOLOR", "=\"#ffffff\"", document);

		setTitleMaxLevel(confparser.getHtmlFile().getMaxTitleLevel() - 1);
		HtmlFile hf = confparser.getHtmlFile();
		htmlDump(hf.getBegin());
		doItem(document);
		htmlDump(hf.getEnd());

		out.close();
	}

	public static void main(String[] args) {
		System.out.println("initialize...");
		DHtmlDriver hd = new DHtmlDriver();

		System.out.println("create...");

//		Example1 ex1 = new Example1();
//		hd.setDestDirAndFileName(".", "file1");
//		System.out.println("translate...");
//		hd.write(ex1.doc);
//		hd.clear();
//
//		Example2 ex2 = new Example2();
//		hd.setDestDirAndFileName(".", "file2");
//		System.out.println("translate...");
//		hd.write(ex2.doc);
//		hd.clear();
//
//		Example3 ex3 = new Example3();
//		hd.setDestDirAndFileName(".", "file3");
//		System.out.println("translate...");
//		hd.write(ex3.doc);
//		hd.clear();
//
//		Example4 ex4 = new Example4();
//		hd.setDestDirAndFileName(".", "file4");
//		System.out.println("translate...");
//		hd.write(ex4.doc);
//		hd.clear();
//
//		Example5 ex5 = new Example5();
//		hd.setDestDirAndFileName(".", "file5");
//		System.out.println("translate...");
//		hd.write(ex5.doc);
//		hd.clear();
//
//		Example6 ex6 = new Example6();
//		hd.setDestDirAndFileName(".", "file6");
//		System.out.println("translate...");
//		hd.write(ex6.doc);
//		hd.clear();
//
		Source so = new Source();
		hd.setDestDirAndFileName(".", "source");
		System.out.println("translate...");
		hd.write(so.doc);
		hd.clear();
		
		System.out.println("done!");
	}
}
/*
 * Created on 26 juin 2003
 *
 */
package com.rational.apidoc.chart;

import org.eclipse.swt.SWT;

import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Layout;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Button;

import com.rational.apidoc.drawutil.*;
import com.rational.apidoc.core.* ;

/**
 * Example to create and render a new graphic type for ApiDoc.
 * The new graphic will draw polar curves.
 * 
 * To extends graphics type, follow these steps:
 * - define a new id for graphic type: see T_POLAR field.
 * - define an IChartRender that able to draw your graphic
 *   if render have support to location, extends IChartLocation 
 *   (thrown by your ChartRender).
 * 
 * - don't forget to register the new IChartRender using DChartRenderRegistry
 *   (see it in main() method)
 * 
 * - now you can build your document containing new graphic type, see this
 *   in BuildPolarGraphic() method.
 */
public class TestUserGraphicType
{  
   /** Define a new graphic type identifier: */   
   public static final String T_POLAR = "T_POLAR.sample.chart.apidoc.rational.com";
   
   /** First axis a Polar graphic must provide */
   public static final String AXIS_RAY = "ray-axis";
   /** Second axis a Polar graphic must provide */
   public static final String AXIS_ANGLE ="angle-axis";
   
   /** Class used to identify where is located a point on a polar graphic */
   public static class PolarLocation implements IChartLocation
   {
      /** Located on graphic title area */
      public static final String GRAPHIC_TITLE="graphic-title";
      /** Located on a segment of a curve (item is first point of segment) */
      public static final String CURVE_SEGMENT="curve-segment";
      /** located on a point of a curve (only for curve type DCurve.POINT) */
      public static final String CURVE_POINT  ="curve-point";
      
      private String id_;
      private IDItem item_;
      private Rectangle rect_;
      
      public PolarLocation( String _id, IDItem _item, Rectangle _r )
      {
        id_=_id; item_=_item; rect_=_r;
      }
      
      public String getId() { return id_; }
      public IDItem getItem() { return item_; }
      public Rectangle getRectangle() { return rect_; }
   }
   
   /** 
    * Implements an IChartRender for this new graphic type
    */   
   private static class PolarChartRender implements IChartRender
   {
     public boolean canRender(String _graphic_type)
     {
       if( T_POLAR.equals( _graphic_type )) return true;
       return false;
     }

     public void draw(GC _gc, Rectangle _R, DGraphic _g)
     {   
       render( true, 0,0, _gc, _R, _g );
     }
     
     public IChartLocation locate(int _x, int _y, GC _gc, Rectangle _R, DGraphic _G)
     {
       return render( false, _x, _y, _gc, _R, _G );
     }
     
     /** support for both draw and location modes */
     public IChartLocation render( boolean _draw, int _lx, int _ly, GC _gc, Rectangle _R, DGraphic _g)
     {   
       if( _draw )
       {
         _gc.setBackground( new Color( null, 255,255,255 ));
         _gc.fillRectangle( _R );                
       }
       else
       {
         if( !_R.contains( _lx, _ly ) ) return null; //outside graphic
       }
       
       //search axis
       DAxis r_axis = null;
       DAxis a_axis = null;
       for( IDItem item=_g.getFirstChild(); item!=null; item=item.getNext() )
       {
         if( item instanceof DAxis )
         {
           DAxis axis = (DAxis)item;
           if( AXIS_RAY.equals( axis.getName() ) ) r_axis=axis;
           else if ( AXIS_ANGLE.equals( axis.getName() ) ) a_axis=axis;
         }
       }
       if( r_axis==null || a_axis==null )
       {
         if( _draw )
         {
           Color red = new Color( null, 255,0,0 );
           String missed = (r_axis==null?"'AXIS_RAY'":"")+" "+(a_axis==null?"'AXIS_ANGLE'":"");
           _gc.setForeground( red );
           _gc.drawText( "Error: missing axis "+missed+"' in graphic item", 10, 30 );
         }
         return null;
       }
       
       final int margin = 10;
       int lx = _R.x+margin;
       int ty = _R.y+margin;
       int rx = _R.x+_R.width-margin;
       int by = _R.y+_R.height-margin;
       int cx = (lx+rx)/2;
       int cy = (ty+by)/2; 
       
       //render title if any
       String title = _g.getTitle();
       if( !DrawUtil.isEmpty( title )
         && _g.getProperties().get( DGraphic.P_SHOW_TITLE, true ) )
       {
         IDStyle style = _g.getStyle();
         Font save = _gc.getFont();
         Font font = DrawUtil.getFont( style.getFont() );
         int aln = DrawUtil.HCENTER|DrawUtil.VCENTER|DrawUtil.VERTICAL;
         if( font!=null ) _gc.setFont( font );
         int tw = DrawUtil.textExtent( _gc, title, aln ).x;
         if (_draw)
         {
           //gradient background
           _gc.setForeground( new Color(null,205,220,255));
           _gc.fillGradientRectangle( lx-1, ty, tw+3, by-ty, true );
           //title color and text
           _gc.setForeground( DrawUtil.getColor(style.getForeColor()) );
           DrawUtil.drawText(_gc, lx, ty, tw, by - ty, aln, title);
         } else {
           Rect r = DrawUtil.boundingRect( _gc, lx, ty, tw, by-ty, aln, title );
           if( r.contains( _lx,_ly ) )
           {
             return new PolarLocation( PolarLocation.GRAPHIC_TITLE, _g, r.toRectangle() );
           }
           if( _lx <= lx+tw )
           {
             //located on graphic title area (outside text),
             //no need to search in curve, but no id is defined for this...
             return null;
           }
         }

         _gc.setFont( save );
         
         lx+=tw+margin;
         cx = (lx+rx)/2;
       }
       
       //first search curve and points to compute min/max
       //an array of DCurve is stored to be able to iterate in reverse order
       //as location mode need.
       DCurve curves[] = new DCurve[ _g.getChildCount() ];
       int    n_curves=0; //because curves.length take into account DAxis too... ;-(
       float xmin=0.0f, xmax=0.0f, ymin=0.0f, ymax=0.0f;
       float ray_max=0.0f;
       boolean init=false;       
       
       for( IDItem item=_g.getFirstChild(); item!=null; item=item.getNext() )
       {
         if( !(item instanceof DCurve) ) continue;
         DCurve curve = (DCurve)item;
         curves[ n_curves++ ] = curve;
         for( IDItem item2=curve.getFirstChild(); item2!=null; item2=item2.getNext())
         {
           if(!(item2 instanceof DPoint)) continue;
           DPoint point = (DPoint)item2;
           DCoord c_angle = point.getCoord( a_axis );
           DCoord c_ray   = point.getCoord( r_axis );
           if( c_angle==null || c_ray==null )
           {
             //no error message here, skip point only
             continue;
           }
           float r = c_ray.getValue();
           float a = c_angle.getValue();
           if( r > ray_max ) ray_max = r;
           //convert to pixels
           float ptx = (float)( r*Math.cos(a) );
           float pty = (float)( r*Math.sin(a) );
           if( init )
           {
             if( ptx < xmin ) xmin=ptx; else if ( ptx > xmax ) xmax=ptx;
             if( pty < ymin ) ymin=pty; else if ( pty > ymax ) ymax=pty;
           }
           else
           {
             xmin=xmax=ptx; 
             ymin=ymax=pty;
             init=true;
           }
         }
       }
       
       float K=1.0f;
       float c_max ;
       {
         c_max = xmax;
         float x = Math.abs( xmin ); if( x > c_max) c_max=x;
         if( ymax > c_max ) c_max=ymax;
         x = Math.abs( ymin ); if ( x > c_max ) c_max=x;
         int pix_min = cx-lx < cy-ty ? cx-lx : cy-ty;
         if( c_max > 0.0f ) K = pix_min / c_max;
       }
       
       //draw axis lines
       if( _draw )
       {
         _gc.setForeground( new Color( null, 220,220,220 ));
         Rectangle r = _gc.getClipping();
         _gc.setClipping( lx, ty, rx-lx, by-ty );
         //Line Step property is set on 'ray' axis ?
         float ray_step_line = r_axis.getProperties().get( DAxis.P_STEP_LINE, 0.0f );
         if( ray_step_line >=0.0f )
         {
            for( float ray=ray_step_line; ray <= c_max; ray += ray_step_line )
            {
              int p_ray = (int)( K*ray );
              int p_ray2 = 2*p_ray;
              _gc.drawOval( cx-p_ray, cy-p_ray, p_ray2, p_ray2 );
            }
         }
         
         //Line Step property is set on 'angle' axis ?
         float ang_step_line = a_axis.getProperties().get( DAxis.P_STEP_LINE, 0.0f );
         if( ang_step_line >=0.0f )
         {
           float dpi = (float)( 2*Math.PI );
           for( float a = ang_step_line; a < dpi; a+=ang_step_line )
           {
             int px = (int)( cx + K*c_max*Math.cos(a) );
             int py = (int)( cy + K*c_max*Math.sin(a) );
             _gc.drawLine( cx, cy, px, py );
           }
         }
         //Dot Step property is set on 'angle' axis ?
         float ang_step_dot = a_axis.getProperties().get( DAxis.P_STEP_DOT, 0.0f );
         if( ang_step_dot >=0.0f )
         {
           float dpi = (float)( 2*Math.PI );
           _gc.setLineStyle( SWT.LINE_DOT );
           for( float a = ang_step_dot; a < dpi; a+=ang_step_dot )
           {
             int px = (int)( cx + K*c_max*Math.cos(a) );
             int py = (int)( cy + K*c_max*Math.sin(a) );
             _gc.drawLine( cx, cy, px, py );
           }
           _gc.setLineStyle( SWT.LINE_SOLID );
         }
         
         _gc.setForeground( new Color( null, 0,0,0 ));
         _gc.drawLine( lx, cy, rx, cy );
         _gc.drawLine( cx, ty, cx, by );
         

         _gc.setClipping( r );
       }

       
       Color def_clr = new Color( null, 0,150,0 );
       //render curves...
       // - draw in forward order
       // - locate in reverse order as the painter algorythm request.
       int crv_shift= _draw ? 0: n_curves-1 ;
       for( int icrv=0; icrv<n_curves; ++icrv )
       {
         DCurve curve = curves[ Math.abs( icrv - crv_shift) ];
         //curve type is dots or lines ?
         boolean dots = DCurve.T_POINTS.equals( curve.getType() );
         if( _draw )
         {        
           Color clr = def_clr;
           IDStyle style = curve.getStyle();
           if( style!=null )
           {
             clr = DrawUtil.getColor( style.getBackColor() );
             if( dots )
             {
               _gc.setBackground( clr );
               clr = DrawUtil.getColor( style.getForeColor() );
             }
           }
           _gc.setForeground( clr );
         }
         
         boolean draw_line=false;
         int ppx=0, ppy=0;
         DPoint last_point=null;
         for( IDItem item2=curve.getFirstChild(); item2!=null; item2=item2.getNext())
         {
           if(!(item2 instanceof DPoint)) continue;
           DPoint point = (DPoint)item2;
           DCoord c_angle = point.getCoord( a_axis );
           DCoord c_ray   = point.getCoord( r_axis );
           if( c_angle==null || c_ray==null )
           {
             //no error message here, skip point only
             draw_line=false;
             last_point=null;
             continue;
           }
           float r = c_ray.getValue();
           float a = c_angle.getValue();
           //convert to pixels
           int px = (int)( cx + K*r*Math.cos(a) );
           int py = (int)( cy + K*r*Math.sin(a) );
           if( dots )
           {
             Rectangle rect = new Rectangle( px-2, py-2, 4, 4 );
             if( _draw )
             {
               _gc.fillRectangle( rect );
               _gc.drawRectangle( rect );
             }
             else
             {
               //point located on the curve Points
               if( rect.contains(_lx, _ly ) )
               {
                 return new PolarLocation( PolarLocation.CURVE_POINT, point, rect );
               }
             }
           }
           else //other curves draw using lines
           {
             if( draw_line )
             {
               if( _draw )
               {
                 _gc.drawLine( ppx, ppy, px, py );
               }
               else
               {
                 //point near this segment ?
                 if( DrawUtil.segmentContains( ppx, ppy, px, py, _lx,_ly ) )
                 {
                   return new PolarLocation( PolarLocation.CURVE_SEGMENT, last_point,
                           new Rectangle( _lx-1,_ly-1,2,2) ); 
                 }
               }
             } else {
               draw_line=true;
             } 
             ppx=px;
             ppy=py;
             last_point=point;
           }
         }
       }  
       
       return null;     
     }

   }

/*---------------------------------------------------------------------------------------
 * Code from here up to the end of file are only made to build the test.
 * You can see how to use your graphic type in an apidoc document in the
 * BuildPolarGraphic method.
 *---------------------------------------------------------------------------------------
 */
 
   /** used only for test shell */
   public static interface IBuildGraphic
   {
     DGraphic buildGraphic();
   }   

  private static DColor black = new DColor(0,0,0);
  private static DColor blue1 = new DColor(0,0,255);
  private static DColor blue2 = new DColor(80,80,255);
  private static DColor blue3 = new DColor(150,150,255);
  private static DColor blue4 = new DColor(180,180,255);
  private static DColor green = new DColor(0,255,0);
  private static DColor red   = new DColor(255,0,0);
  private static DColor magenta = new DColor(255,0,255);
  private static DColor white = new DColor(255,255,255);

  private static DFont font1 = new DFont( "f1",null,null, 14, DFont.BOLD );
  private static DFont df    = new DFont( "df", "Arial",null,10,DFont.NORMAL );
  
  private static IDStyle sd = new DStyle( "sd", df, black, white );        
  private static IDStyle s0 = new DStyle( "s0", font1, green, white );
  private static IDStyle s1 = new DStyle( "s1", df, black, red );
  private static IDStyle s2 = new DStyle( "s2", df, black, blue2 );
  private static IDStyle s3 = new DStyle( "s3", df, black, blue4 );
  private static IDStyle s4 = new DStyle( "s4", df, black, green );
  private static IDStyle s5 = new DStyle( "s5", df, black, magenta );
  private static IDStyle s6 = new DStyle( "s6", df, black, green );
     
  private static DPoint newDPoint( DCurve _p, DAxis _x, DAxis _y, float _cx, float _cy )
  {
    DPoint pt = new DPoint();
    _p.addChild( pt );
    pt.add( new DCoord( _x, _cx ));
    pt.add( new DCoord( _y, _cy ));
    return pt;
  }
  private static DPoint newDPoint3( DCurve _p, DAxis _x, DAxis _y, DAxis _z, float _cx, float _cy, float _cz )
  {
    DPoint pt = new DPoint();
    _p.addChild( pt );
    pt.add( new DCoord( _x, _cx ));
    pt.add( new DCoord( _y, _cy ));
    pt.add( new DCoord( _z, _cz ));
    return pt;
  }
    /**
   * Graphic document builder for XY graphic 
   */
  private static class BuildXYGraphic implements IBuildGraphic
  {
    public DGraphic buildGraphic()
   {
     DGraphic  gph = new DGraphic();  
     gph.setGraphicType( DGraphic.T_XY );
     gph.setTitle( "Graphic XY" );
     gph.setStyle( s0 );
//gph.setProperties().store(DGraphic.P_SHOW_TITLE, false);     
//gph.setProperties().store(DGraphic.P_SHOW_LEGEND, false); 

     DAxis  x_axs = new DAxis("x");    gph.addChild( x_axs );
     DAxis  y_axs = new DAxis("y");    gph.addChild( y_axs );
     x_axs.setStyle( sd );
     y_axs.setStyle( sd );
     x_axs.getProperties().store( DAxis.P_STEP_UNIT, 2.0f );
     x_axs.getProperties().store( DAxis.P_STEP_DOT, 0.5f );
     y_axs.getProperties().store( DAxis.P_STEP_UNIT, 2.0f );
     y_axs.getProperties().store( DAxis.P_STEP_DOT, 0.5f );
     x_axs.setTitle( "X Axis");
     x_axs.setUnit("(x)");
     y_axs.setTitle( "Y Axis");
     y_axs.setUnit("(unit)");
     
     
     DCurve    cv0 = new DCurve();            gph.addChild( cv0 );
     cv0.setName( "Bars" );
     cv0.setType( DCurve.T_BAR );
     cv0.setStyle( s4 );
     newDPoint( cv0, x_axs, y_axs, 0, 0.9f );
     newDPoint( cv0, x_axs, y_axs, 1, 1.2f );
     newDPoint( cv0, x_axs, y_axs, 2, 1.6f );
     newDPoint( cv0, x_axs, y_axs, 3, 3.4f );
     newDPoint( cv0, x_axs, y_axs, 4, 4.6f );

     DCurve    cv1 = new DCurve();            gph.addChild( cv1 );
     cv1.setName("Line");
     cv1.setStyle( s1 );
     newDPoint( cv1, x_axs, y_axs, 0,0 );
     newDPoint( cv1, x_axs, y_axs, 1,1.2f );
     newDPoint( cv1, x_axs, y_axs, 2,3.2f );
     newDPoint( cv1, x_axs, y_axs, 3,2.4f );
     newDPoint( cv1, x_axs, y_axs, 4,0.6f );

     DCurve    cv2 = new DCurve();            gph.addChild( cv2 );
     cv2.setName( "Points" );
     cv2.setType( DCurve.T_POINTS );
     cv2.setStyle( s2 );
     newDPoint( cv2, x_axs, y_axs, 0, 1.9f );
     newDPoint( cv2, x_axs, y_axs, 1, 1.2f );
     newDPoint( cv2, x_axs, y_axs, 2, 1.1f );
     newDPoint( cv2, x_axs, y_axs, 3, 0.4f );
     newDPoint( cv2, x_axs, y_axs, 4, 0.0f );
     
     x_axs.getProperties().store( DAxis.P_MIN, 0.0f );
     x_axs.getProperties().store( DAxis.P_MAX, 5.0f );
     x_axs.getProperties().store( DAxis.P_STEP_LINE, 1.0f );
     y_axs.getProperties().store( DAxis.P_MIN, 0.0f );
     y_axs.getProperties().store( DAxis.P_MAX, 5.0f );
     y_axs.getProperties().store( DAxis.P_STEP_LINE, 1.0f );
     
     return gph;
   }
  }
  
  /**
   * Graphic document builder for Polar graphic 
   */
  private static class BuildPolarGraphic implements IBuildGraphic
  {
    public DGraphic buildGraphic()
   {
     DGraphic  gph = new DGraphic();  
     // Set my graphic type 
     gph.setGraphicType( T_POLAR );
     gph.setTitle( "Polar Graphic" );
     gph.setStyle( s0 );

     DAxis  ray_axs = new DAxis( AXIS_RAY );    gph.addChild( ray_axs );
     DAxis  angle_axs = new DAxis( AXIS_ANGLE );    gph.addChild( angle_axs );
     ray_axs.setStyle( sd );
     angle_axs.setStyle( sd );
     ray_axs.getProperties().store( DAxis.P_STEP_UNIT, 2.0f );
     ray_axs.getProperties().store( DAxis.P_STEP_DOT, 0.5f );
     angle_axs.getProperties().store( DAxis.P_STEP_UNIT, Math.PI/8 );
     angle_axs.getProperties().store( DAxis.P_STEP_DOT, Math.PI/16 );
     
     DCurve    cv0 = new DCurve();            gph.addChild( cv0 );
     cv0.setName( "Petals" );
     cv0.setType( DCurve.T_LINE );
     cv0.setStyle( s4 );
     DCurve    cv1 = new DCurve();            gph.addChild( cv1 );
     cv1.setName("Spiral");
     cv1.setType( DCurve.T_POINTS );
     cv1.setStyle( s1 );
     DCurve    cv2 = new DCurve();            gph.addChild( cv2 );
     cv2.setName("Curve 3");
     cv2.setStyle( s2 );
     DCurve    cv3 = new DCurve();            gph.addChild( cv3 );
     cv3.setName("Curve 4");
     cv3.setStyle( s5 );
     
     //populate curve using various function ( ray = f( angle ) )
     float da = (float)( Math.PI/100 );
     float r;
     float rmax=0.0f;
     int cnt=10;
     for( float a=0.0f; a<6*Math.PI; a+=da )
     {
       float ca = (float)Math.cos(a);
       float sa = (float)Math.sin(a);
       
       r = (float)( (a/(10*da))*ca*sa );
       if( r > rmax ) rmax=r;
       newDPoint( cv0, ray_axs, angle_axs, r, a );
       r = (float)( a );
       if( --cnt == 0 )
       {
         if( r > rmax ) rmax=r;
         newDPoint( cv1, ray_axs, angle_axs, r, a );
         cnt=10;
       }
       r = (float)(8*ca - 10.0f*sa/(1.1f+ca));
       if( r > rmax ) rmax=r;
       newDPoint( cv2, ray_axs, angle_axs, r, a );
       r = (float)(20+3*Math.cos(10*a)-20*Math.sin(5*a)*Math.cos(2*a));
       if( r > rmax ) rmax=r;
       newDPoint( cv3, ray_axs, angle_axs, r, a );
     }
   
     ray_axs.getProperties().store( DAxis.P_MIN, 0.0f );
     ray_axs.getProperties().store( DAxis.P_MAX, 1.1f*rmax );
     ray_axs.getProperties().store( DAxis.P_STEP_LINE, rmax/5 );
     angle_axs.getProperties().store( DAxis.P_MIN, 0.0f );
     angle_axs.getProperties().store( DAxis.P_MAX, 5.0f );
     angle_axs.getProperties().store( DAxis.P_STEP_LINE, Math.PI/8 );
     
     return gph;
   }
  }
  
  private static class RenderGraphic implements MouseListener, MouseMoveListener, PaintListener
  {
    private Canvas canvas_;
    private IChartRender render_;
    private DGraphic gph_;
    private boolean mouse_down_=false;
    private int mx_=0, my_=0;
    
    public RenderGraphic( Canvas _cvs )
    {
      canvas_=_cvs; 
      canvas_.addMouseListener( this );
      canvas_.addMouseMoveListener( this );
      canvas_.addPaintListener( this );
    }
    
    public void setGraphic( DGraphic _g )
    {
      gph_ = _g;
      render_ = DChartRenderRegistry.GetRender( _g.getGraphicType() );
      canvas_.redraw();
    }
    
    public void mouseMove(MouseEvent e)
    {
      if( render_==null || gph_==null ) return ;
      if( mouse_down_ &&( gph_.isGraphicType( DGraphic.T_XYZ )
                      || gph_.isGraphicType( DGraphic.T_SURFACE)) )
      {
        move3DView(e);
        return;
      }
          GC gc = new GC(canvas_);   
          Rectangle r = canvas_.getClientArea();
          IChartLocation loc = render_.locate( e.x,e.y, gc, r, gph_ );
//System.out.println("GOT LOC="+loc);          
          if( loc!=null && loc.getItem() instanceof DPoint )
          {
            DPoint pt = (DPoint)loc.getItem();
            DCurve crv= (DCurve)pt.getParent();
            DGraphic gph=(DGraphic)crv.getParent();
            String name = crv.getName();
            if( name==null ) name=""; else name=name+": ";
            for( IDItem item=gph.getFirstChild(); item!=null; item=item.getNext() )
            {
              if( item instanceof DAxis )
              {
                DAxis axis=(DAxis)item;
                DCoord crd= pt.getCoord( axis );
                if( crd!=null )
                {
                  name = name + Float.toString( crd.getValue() )+"  ";
                }
              }
            }
            canvas_.setToolTipText( name );
          } else {
            canvas_.setToolTipText( null );
          }
          gc.dispose();
          gc = null;
        }
    public void mouseDoubleClick(MouseEvent e){}

    public void mouseDown(MouseEvent e)
    {
      mouse_down_=true;
      mx_=e.x;
      my_=e.y;      
    }
    public void mouseUp(MouseEvent e)
    {
      mouse_down_=false;
    }     
    protected void move3DView(MouseEvent e)
    {  
      if( render_==null || gph_==null ) return ;

      float phi   = gph_.getProperties().get( DGraphic.P_XYZ_PHI, 0.3f);
      float theta = gph_.getProperties().get( DGraphic.P_XYZ_THETA, 0.2f);
      int dx = e.x -mx_;
      int dy = e.y -my_;
      
      phi   += 0.02f*dy;
      theta -= 0.02f*dx;

      gph_.getProperties().store( DGraphic.P_XYZ_PHI,  phi    );
      gph_.getProperties().store( DGraphic.P_XYZ_THETA, theta );
      mx_ = e.x;
      my_ = e.y;    
      canvas_.redraw();  
    }
    
    private Image img_;
    public void paintControl(PaintEvent e)
    {
       if( render_==null || gph_==null ) return ;
       Rectangle r = canvas_.getClientArea();
       if( img_==null || img_.getBounds().width!=r.width || img_.getBounds().height!=r.height)
       {
         if( img_!=null ) img_.dispose();
         img_ = new Image(null,r.width,r.height);
       }
       GC gc = new GC( img_ );
       render_.draw( gc, r, gph_ );
       e.gc.drawImage( img_, 0, 0 );
       gc.dispose();
    }
   }
   
   /* Layout horizontally children control, all width of them (but not the last one)
    * is limited (as the FillLayout doesn't do it).
    */
   private static class HLimLayout extends Layout
   {
     private int lim_;
     public HLimLayout(int _lim){lim_=_lim;}
     protected Point computeSize(Composite composite, int wHint, int hHint, boolean flushCache)
     {
       return new Point( 100, hHint );
     }

     protected void layout(Composite composite, boolean flushCache)
     {
       Rectangle r = composite.getClientArea();
       Control c[] = composite.getChildren();
       if( c==null ) return ;
       int x = r.x;
       for( int i=0; i<c.length; ++i )
       {
         if( (i+1)==c.length)
         {
           c[i].setBounds( x, r.y, r.x+r.width-x, r.height );
         } else {
           c[i].setBounds( r.x, r.y, lim_, r.height );
           x += lim_;
         }
       }
      
     }
   }
   
   public static void main( String[] argv )
   {
     //This is very important to register the render for new graphic type.
     DChartRenderRegistry.AddRender( new PolarChartRender() );
     
     Display disp = Display.getDefault();
     final Shell shell = new Shell(disp, SWT.DIALOG_TRIM|SWT.RESIZE );
     shell.setText( "Test ApiDoc: User Graphic" );
     shell.setLayout(new HLimLayout( 150 ));

     Composite cb = new Composite( shell, SWT.NONE );
     cb.setLayout( new FillLayout( SWT.VERTICAL ));
     Group gb = new Group( cb, SWT.NONE);
     gb.setLayout( new FillLayout(SWT.VERTICAL));
     gb.setText("Graphic Type:");
     final Button btn[] = new Button[2];
     int b=-1;

     btn[++b] = new Button( gb, SWT.RADIO ); btn[b].setText("XY (standard)");
     btn[b].setData( new BuildXYGraphic());     

     btn[++b] = new Button( gb, SWT.RADIO ); btn[b].setText("Polar (user defined)");
     btn[b].setData( new BuildPolarGraphic());     
       
     final Canvas canvas = new Canvas( shell, SWT.NO_BACKGROUND );
     final RenderGraphic rg = new RenderGraphic( canvas );
    
     //just on first button is enough
     SelectionListener listen_button= new SelectionListener(){
       public void widgetSelected(SelectionEvent e)
       {   
         for(int i=0; i<btn.length; ++i)
         {
           if( btn[i].getSelection())
           {
             DGraphic g = ((IBuildGraphic)btn[i].getData()).buildGraphic();
             rg.setGraphic( g );
             break;            
           }
         }
       }
 
       public void widgetDefaultSelected(SelectionEvent e){} //never called
     };
     
     //set selection listeners to all buttons...
     for( int i=0; i<btn.length; ++i )
     {
       btn[i].addSelectionListener( listen_button );
     }
         
     shell.setSize( 600, 500 );
     shell.open();
 
     
      while (!shell.isDisposed())
        if (!disp.readAndDispatch())
          disp.sleep();
   }
}
/*
 * Created on 25 juin 2003
 *
 */
package com.rational.apidoc.chart.internal;

import java.text.DecimalFormat;
import java.util.Vector;

import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.graphics.FontMetrics;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.SWT;
import com.rational.apidoc.core.*;
import com.rational.apidoc.chart.IChartLocation;
import com.rational.apidoc.chart.IChartRender;
import com.rational.apidoc.drawutil.HSV;
import com.rational.apidoc.drawutil.DrawUtil;
import com.rational.apidoc.drawutil.Rect;
import com.rational.apidoc.drawutil.Point2D;

/**
 * ChartRender is made to 
 * - draw DGraphic item in an area
 * - locate a point on a DGraphic area.
 */
public class DefaultChartRender implements IChartRender
{
  /** margin for render */
  protected int margin_; 
  /** Default style to draw */
  protected DChartStyle default_;
 
  /** temporary data during draw/locate */
  protected Data d_ ; 
  /** used to dispose font/color from styles (MUST BE disposed for windows systems)
   *  (?huh? work'in java and take care of windows ... smell bad (but smell'SWT!) */
  protected Vector styleof_;

  /** Create a chart render */
  public DefaultChartRender()
  {
    margin_=10;
    default_ = new DChartStyle();
    d_=null;
  }
  
  //paire courbe / style
  private static class CurveStyle
  {
    public DCurve curve_;
    public DChartStyle style_;
    public CurveStyle( DCurve _c, DChartStyle _s )
    {
      curve_=_c; style_=_s;
    }
  }
  
  private static class Data
  {
     
    public boolean drawing() { return draw_; }
    public boolean locating() { return !draw_; }
    
    public boolean draw_;
    public GC gc_; //mode drawing
    public Font default_font_;
    public int lx_, ly_; //mode locating
    public DGraphic G_; 
    public Rect R_;
    public Vector curves_; //list of CurveStyle
    
    public Data( boolean _draw, DGraphic _g, Rect _R, GC _gc )
    {
      gc_=_gc;
      draw_=_draw;
      G_=_g;
      R_=_R;
      lx_=ly_=-1;
      curves_ = new Vector();
      default_font_ = _gc.getFont();
    }        

    static boolean horizontal3DContains( int _lx, int _ly, int _w, int _3d, int _px, int _py )
    {
      //? si (_px,_py) est contenus dans le rectangle cisaille en X:
      //    _lx+_3d
      //       +---------+ _ly-_3d
      //      /         /
      // _ly +---------+
      //    _lx     _lx+_w
      //!! les Y ecrans sont inverses
      if( _py < (_ly-_3d) ) return false;
      if( _py >  _ly      ) return false;
      if( _py==_ly )
      {
        return (_px>=_lx)&&(_px<=(_lx+_w));
      }
      if( _py==(_ly-_3d))
      {
        return (_px>=(_lx+_3d))&&(_px<=(_lx+_w+_3d));
      }
      //dans le rectangle cisaille en X
      int dx = _ly-_py; //OK: car proj. 3d en pente 45 deg.
      return (_px>=(_lx+dx))&&(_px<=(_lx+dx+_w));
    }
    
    static boolean vertical3DContains( int _lx, int _ly, int _h, int _3d, int _px, int _py )
    {
      //? si (_px,_py) est contenus dans le rectangle cisaille en Y:
      //       _lx+_3d
      //          + _ly-_3d
      //         /|
      //    _ly + |
      //        | |
      //        | +
      //        |/
      // _ly+_h +
      //      _lx
      if( _px <  _lx     ) return false;
      if( _px > (_lx+_3d)) return false;
      if( _px == _lx      ) return (_py>=_ly)&&(_py<=(_ly+_h));
      ////if( _px == (_lx+_3d)) return (_py>=(_ly-_3d))&&(_py<=(_ly-_3d+_h));
      //!! les Y ecrans sont inverses
      //dans le rectangle cisaille en Y
      int dy = _px-_lx; //OK: car proj. 3d en pente 45 deg.
      int npy = _py+dy;
     return (npy>=_ly)&&(npy<=(_ly+_h));
    }

  }
  
  protected static class ChartLocation extends Throwable
  {
    public IChartLocation location_;
    public ChartLocation( IChartLocation _l )
    {
      location_=_l;
    }
    public ChartLocation( String _id, IDItem _item, Rect _rect )
    {
      location_=new DChartLocation( _id, _item, _rect );
    }
    public ChartLocation( String _id, IDItem _item, Rect _rect, String _dtxt )
    {
      location_=new DChartLocation( _id, _item, _rect, _dtxt );
    }
  }
  
  /** @return true since default render can draw anything */
  public boolean canRender( String _graph_type )
  {
    return true;
  }
  
  /** Draw graphic _g on rectangle _R using GC _gc. */
  public void draw( GC _gc, Rectangle _R, DGraphic _g )
  {
    if( _g==null ) return ;
    d_ = new Data( true, _g, new Rect(_R), _gc );
    try {
      render();
    }
    catch( ChartLocation _l) {}//musn't appears
    d_ = null;
  }
  
  /** Return a location for the given point, GC must be provided to compute text extends.
   * can return null if no location is found.
   */
  public IChartLocation locate( int _x, int _y, GC _gc, Rectangle _R, DGraphic _G )
  {
    if(_G==null) return null;
    if(!_R.contains(_x,_y) ) return null;

    d_ = new Data( false, _G, new Rect(_R), _gc );
    d_.lx_ = _x;
    d_.ly_ = _y;
    
    try {
      render();
    }
    catch( ChartLocation _l )
    {
      return _l.location_;
    }
    finally
    {
      d_=null;
    }
    return null;
  }
  
  /** Return style for item, walk upstair in tree to try to complete style. */
  protected DChartStyle styleOf( IDItem _item )
  {
    IDStyle apidoc_style=null;
    IDFont  apidoc_font =null;
    IDColor apidoc_color=null;
    DChartStyle style=new DChartStyle();
    char missing = 0x07; //font|fore|back

    for( IDItem item=_item; item!=null; item=item.getParent() )
    {
      if( (apidoc_style=item.getStyle())!=null )
      {
        //font
        if ( (missing & 0x01)!=0 && (apidoc_font=apidoc_style.getFont())!=null ) 
        {          
          style.font_ = DrawUtil.getFont( apidoc_font ); 
          missing &= ~0x01 ;       
        }
        //fore color
        if ( (missing & 0x02)!=0 && (apidoc_color=apidoc_style.getForeColor())!=null )
        {
          style.fore_ = DrawUtil.getColor( apidoc_color );
          missing &= ~0x02 ;
        }
        //back color
        if ( (missing & 0x04)!=0 && (apidoc_color=apidoc_style.getBackColor())!=null )
        {
          style.back_ = DrawUtil.getColor( apidoc_color );
          missing &= ~0x04 ;
        }
        //style completed.
        if( missing == 0x00 ) break ;
      }
    }//for

    if( styleof_==null ) styleof_ = new Vector();
    styleof_.add( style );
    return style;
  }
  
  protected void disposeStyleOf()
  {
    if( styleof_==null ) return ;
    for( int i=0; i<styleof_.size(); ++i )
    {
      DChartStyle s = (DChartStyle)styleof_.get(i);
      s.dispose();
    }
    styleof_.clear();
  }
  
  /** print error occured while rendering chart. */
  void error( String _message )
  {
//TBD:juste pour mise au point: affichage de ce qui a merde
    if( d_.drawing() )
    {
      Font font = d_.gc_.getFont();
      FontData fd[] = font.getFontData();
      FontData nfd = new FontData( fd[0].getName(), fd[0].getHeight(), fd[0].getStyle()|SWT.BOLD );
      Font nf = new Font( null, nfd );
      
      d_.gc_.setFont( nf );
      d_.gc_.setForeground( getRed() );
      d_.gc_.drawText( _message, d_.R_.x()+5, d_.R_.y()+5+12 );

      nf.dispose();
    }
  }
  
  /** draw or locate point using d_ Data */
  protected void render() throws ChartLocation
  {
    try
    {    
      DChartStyle g_style = styleOf( d_.G_ );

      //prepare curve list, and their styles.
      d_.curves_.clear();
      IDItem  item=null;
      DCurve curve=null;
      for( item=d_.G_.getFirstChild(); item!=null; item=item.getNext() )
      {
        if( item instanceof DCurve )
        {
          curve = (DCurve)item;
          DChartStyle cs = styleOf( curve );                
          d_.curves_.add( new CurveStyle( curve, cs ) );
        }
      }
    
      //graphics background
      if( d_.drawing() )
      {
        d_.gc_.setBackground( g_style.back_ );
        d_.gc_.fillRectangle( d_.R_.x(), d_.R_.y(), d_.R_.w(), d_.R_.h() );
      }

      //graphic title
      int title_h = renderGraphicTitle( g_style );

      //legend on the right side
      int legend_width = renderLegend( title_h );

      //the rest is available to render graphic
      Rect rg = new Rect( margin_,
                          title_h+margin_,
                          d_.R_.w()-legend_width-2*margin_,
                          d_.R_.h()-title_h-2*margin_ );
  
      //graphic is rendered depending on it type:
           if( d_.G_.isGraphicType( DGraphic.T_SECTORS      )) renderSector( rg, false );
      else if ( d_.G_.isGraphicType( DGraphic.T_SECTORS3D    )) renderSector( rg, true  );
      else if ( d_.G_.isGraphicType( DGraphic.T_HISTOGRAM    )) renderHistogram( rg, false, false );
      else if ( d_.G_.isGraphicType( DGraphic.T_HISTOGRAM3D  )) renderHistogram( rg, true, false );
      else if ( d_.G_.isGraphicType( DGraphic.T_HISTOGRAM_BS )) renderHistogram( rg, false, true );
      else if ( d_.G_.isGraphicType( DGraphic.T_XY           )) renderXY( rg );
      else if ( d_.G_.isGraphicType( DGraphic.T_XYZ          )) renderXYZ( rg );
      else if ( d_.G_.isGraphicType( DGraphic.T_SURFACE      )) renderXYZ( rg );
/*
      else if ( d_.G_.isGraphicType( DGraphic.T_HISTORS )) renderHistors( rg )
*/
      else
      {
        error( "Unknwon Graphic type '"+d_.G_.getGraphicType()+"'" );
        return ;
      }
    
      g_style.dispose();
    
      if( d_.locating() && rg.contains( d_.lx_, d_.ly_ ) )
      {
        throw new ChartLocation( DLocated.InGraphicArea, d_.G_, rg );
      }
    }
    finally
    {
       //dispose styles, must be done for poor platform like windows.
       disposeStyleOf();
       d_.curves_.clear();
       d_.curves_=null; 
    }
  }
  
  
  /** render graphic title.
   * @return the height of the title, 0 if title isn't displayed
   */
  protected int renderGraphicTitle( DChartStyle _g_style ) throws ChartLocation
  {
    int title_h = 0;

    boolean display_title = d_.G_.getProperties().get(DGraphic.P_SHOW_TITLE, true);
    String title = d_.G_.getTitle();  
    
    if( !display_title || DrawUtil.isEmpty( title ) ) return title_h;
    
    d_.gc_.setFont( _g_style.font_ );
    FontMetrics fm = d_.gc_.getFontMetrics();      
    title_h = fm.getHeight();
  
    String txt = DrawUtil.truncateTextH( d_.gc_, title, d_.R_.w() );
    int    aln = DrawUtil.HCENTER | DrawUtil.TOP;
    Rect r_ttl = new Rect( d_.R_.x(), d_.R_.y(), d_.R_.w(), title_h );

    if( d_.drawing() )
    {
      d_.gc_.setForeground( _g_style.fore_ );
      DrawUtil.drawText( d_.gc_, r_ttl, aln, txt);
    }
    else
    {
      //locating
      Rect br = DrawUtil.boundingRect( d_.gc_, r_ttl, aln, txt );
      br.union( r_ttl );

      if( br.contains( d_.lx_, d_.ly_ ) )
      {
        throw new ChartLocation( DLocated.InGraphicTitleText, d_.G_, br, txt );
      }
      if( r_ttl.contains( d_.lx_, d_.ly_ ) )
      {
        throw new ChartLocation( DLocated.InGraphicTitleArea, d_.G_, r_ttl );
      }
    }
 
    return title_h;
  }
  
  /** render legend and return the width, 0 if none */
  protected int renderLegend( int _title_h )  throws ChartLocation 
  {
    int legend_width = 0;

    //display legend ?
    boolean display_legend = d_.G_.getProperties().get(DGraphic.P_SHOW_LEGEND, true);
    if( !display_legend ) return 0;
    
    CurveStyle ccurve=null;

    int legend_height = 0;
    int legend_height_min =100000;
    for( int i=0; i<d_.curves_.size(); ++i )
    {
      ccurve = (CurveStyle)d_.curves_.get(i);
      d_.gc_.setFont( ccurve.style_.font_ );
      String name = ccurve.curve_.getName();
      if( !DrawUtil.isEmpty( name ) )
      {
        Point sz = d_.gc_.textExtent( name );
        if ( legend_width < sz.x ) legend_width=sz.x;
        if ( legend_height_min > sz.y ) legend_height_min=sz.y;
        legend_height += sz.y;
      }
    }

    int legend_rect=(legend_height_min-2);
    if(legend_rect>10) legend_rect=10; //limit to 10x10 pixel
    legend_width += legend_rect+margin_ ;
    //limit legend width to 1/3 of graphic
    if ( legend_width > (d_.R_.w()/3) ) legend_width=d_.R_.w()/3;
    int legend_x = d_.R_.w() - legend_width;
    int legend_H = d_.curves_.size() * 4 + legend_height;
    int legend_y = _title_h+(d_.R_.h()-legend_H)/2+2;

    Rect lgd_area= new Rect( d_.R_.right()-legend_width, d_.R_.top()+_title_h,
                             legend_width, d_.R_.h()-_title_h );
    //shortcut: on on y est pas, on passe la main ...
    //shortcut for locating: outside ?
    if( d_.locating() && !lgd_area.contains( d_.lx_, d_.ly_ ) )
    {
      return legend_width ;
    }
    
    Rect r_clr=new Rect();
    Rect r_lgd_area=new Rect();
    Rectangle r = new Rectangle(0,0,0,0);
    
    for( int i=0; i<d_.curves_.size(); ++i )
    {
      ccurve = (CurveStyle)d_.curves_.get(i);
      String name = ccurve.curve_.getName();
      if( DrawUtil.isEmpty( name )) continue;
      //text is choosen black, because fore color is curve color
      d_.gc_.setFont( ccurve.style_.font_ );
      int legend_h = d_.gc_.textExtent( name ).y;
      r_clr.setRect( legend_x,legend_y+(legend_h-legend_rect)/2, 
                     legend_rect, legend_rect );

      int available_width = legend_width-legend_rect-5;
      String txt = DrawUtil.truncateTextH( d_.gc_, name, available_width );

      if( d_.drawing() )
      {
        d_.gc_.setForeground( getBlack() );
        d_.gc_.setBackground( ccurve.style_.back_ );
        r_clr.toRectangle( r );
        d_.gc_.fillRectangle( r ); //fill it
        d_.gc_.drawRectangle( r ); //draw black square around it
      }
      else //locating
      {
        //inside color rectangle
        if( r_clr.contains( d_.lx_, d_.ly_ ) )
        {
          throw new ChartLocation( DLocated.LegendColor, ccurve.curve_, r_clr );
        }
      }
    
      r_lgd_area.setRect( legend_x+legend_rect+5, legend_y,
                          available_width,        legend_h );

      renderText( txt, r_lgd_area, DrawUtil.LEFT|DrawUtil.VCENTER,
                   ccurve.style_, DLocated.LegendText, ccurve.curve_ );

      legend_y += legend_h+4;
    }//for

    //dans la zone de la legende ??
    if( d_.locating() && lgd_area.contains( d_.lx_, d_.ly_ ) )
    {
      throw new ChartLocation( DLocated.LegendArea, null, lgd_area );
    }
  
    return legend_width;
  }    
  
  protected static class TruncateAxisTextData
  {
    public String sttl_ ; //string title
    public String sunit_; //string for unit
    public Rect   rttl_ ; //rectangle for title
    public Rect   runit_; //rectangle for unit
    public Rect   rtu_  ; //rectangle for both title and unit    
  }
  
  /** truncate unit and/or title text for an "horizontal" axis,
   * return also bounding rectangles.
   */
  protected TruncateAxisTextData truncateHAxisText(
      DAxis _axis, DChartStyle _axis_style,
      int _available_width, int _available_height
    )
  {
    Rect   rttl;
    Rect   runit;
    Rect   rtu;

    //reduction de la chaine de titre ou units..

    //title et unit sont geree dans des chaines/rect different cela va permettre
    //de les affichers avec des styles differents ... un jours prochains.
    String axis_title = _axis.getTitle();
    String axis_unit  = _axis.getUnit();
//TODO: mieux a faire, SWT ne supporte pas textExtent(null)    
    if( axis_title == null ) axis_title="";
    if( axis_unit  == null ) axis_unit="";

    //titre ( unit ) axe des Y
    d_.gc_.setFont( _axis_style.font_ );
    Point ext = d_.gc_.textExtent( axis_title );
    rttl = new Rect( 0,0, ext.x+2, _available_height );
    //unit a droite du titre
    ext = d_.gc_.textExtent( axis_unit );
    runit = new Rect( rttl.w()+5, 0, ext.x+2, _available_height );
    rtu = Rect.Union( rttl, runit ); //rectangle englobant titre et unite 

    //reduire d'abord le titre
    String n_axis_title = axis_title;
    while( rtu.w() > _available_width ) 
    {
      try {
        axis_title = axis_title.substring( axis_title.length()-2, axis_title.length()-1 );
      }
      catch( StringIndexOutOfBoundsException e )
      {
        n_axis_title="";
        break;
      }
      if( axis_title.length()==0 ) { n_axis_title=null; break; }
      n_axis_title = axis_title+"...";
      ext = d_.gc_.textExtent( n_axis_title );
      rttl.setWidth( ext.x+2 );
      runit.setX( rttl.w()+5 );
      // rtu = Rect.Union( rttl, runit ) but without allocation
      rtu.setRect( rttl );
      rtu.union( runit );
    }
    axis_title = n_axis_title ;
    
    //puis l'unite
    String n_axis_unit = axis_unit;
    while( rtu.w() > _available_width ) 
    {
      try {
        axis_unit = axis_unit.substring( axis_unit.length()-2, axis_unit.length()-1 );
      }
      catch( StringIndexOutOfBoundsException e )
      {
        n_axis_unit="";
        break;
      }
      if( axis_unit.length()==0 ) { n_axis_unit=null; break; }
      n_axis_unit = axis_unit+"...";
      ext = d_.gc_.textExtent( n_axis_unit );
      runit.setWidth( ext.x+2 );
      rtu.setRect( rttl );
      rtu.union( runit );
    }
    axis_unit = n_axis_unit;

    TruncateAxisTextData data = new TruncateAxisTextData();
    data.sttl_ = axis_title;
    data.sunit_= axis_unit;
    data.rttl_ = rttl;
    data.runit_= runit;
    data.rtu_  = rtu ;
    
    return data;
  }

    /** truncate unit and/or title text for an "vertical" axis,
     * return also bounding rectangles.
     */
    protected TruncateAxisTextData truncateVAxisText(
        DAxis _axis, DChartStyle _axis_style,
        int _available_width, int _available_height
      )
    {
      Rect   rttl;
      Rect   runit;
      Rect   rtu;
  
      //reduction de la chaine de titre ou units..
  
      //title et unit sont geree dans des chaines/rect different cela va permettre
      //de les affichers avec des styles differents ... un jours prochains.
      String axis_title = _axis.getTitle();
      String axis_unit  = _axis.getUnit();
  //TODO: mieux a faire, SWT ne supporte pas textExtent(null)    
      if( axis_title == null ) axis_title="";
      if( axis_unit  == null ) axis_unit="";
  
      //titre ( unit ) axe des Y
      d_.gc_.setFont( _axis_style.font_ );
      Point ext = DrawUtil.verticalTextExtent( d_.gc_, axis_title );
      rttl = new Rect( 0,0, _available_width, ext.y+2 );
      //unit a droite du titre
      ext = DrawUtil.verticalTextExtent( d_.gc_, axis_unit );
      runit = new Rect( 0, rttl.h()+5, _available_width, ext.y+2 );
      rtu = Rect.Union( rttl, runit ); //rectangle englobant titre et unite 
  
      //reduire d'abord le titre
      String n_axis_title = axis_title;
      while( rtu.w() > _available_height ) 
      {
        try {
          axis_title = axis_title.substring( axis_title.length()-2, axis_title.length()-1 );
        }
        catch( StringIndexOutOfBoundsException e )
        {
          n_axis_title="";
          break;
        }
        if( axis_title.length()==0 ) { n_axis_title=null; break; }
        //no sens for vertical text/ n_axis_title = axis_title+"...";
        ext = DrawUtil.verticalTextExtent( d_.gc_, n_axis_title );
        rttl.setHeight( ext.y+2 );
        runit.setY( rttl.h()+5 );
        // rtu = Rect.Union( rttl, runit ) but without allocation
        rtu.setRect( rttl );
        rtu.union( runit );
      }
      axis_title = n_axis_title ;
      
      //puis l'unite
      String n_axis_unit = axis_unit;
      while( rtu.w() > _available_height ) 
      {
        try {
          axis_unit = axis_unit.substring( axis_unit.length()-2, axis_unit.length()-1 );
        }
        catch( StringIndexOutOfBoundsException e )
        {
          n_axis_unit="";
          break;
        }
        if( axis_unit.length()==0 ) { n_axis_unit=null; break; }
        //no sense for vertical/ n_axis_unit = axis_unit+"...";
        ext = DrawUtil.verticalTextExtent( d_.gc_, n_axis_unit );
        runit.setHeight( ext.y+2 );
        rtu.setRect( rttl );
        rtu.union( runit );
      }
      axis_unit = n_axis_unit;
  
      TruncateAxisTextData data = new TruncateAxisTextData();
      data.sttl_ = axis_title;
      data.sunit_= axis_unit;
      data.rttl_ = rttl;
      data.runit_= runit;
      data.rtu_  = rtu ;
      
      return data;
    }
  
  /** Render property on Y axis, support for "3D-like".
   * @param _get_step_unit_width true to compute step unit width on axis, 
   *                             false to locate/draw.
   * @param _R rectangle to render axis.
   * @param _axis axis to render.
   * @param _property property identifier to render: STEP_LINE,STEP_UNIT,STEP_DOT.
   * @param _3d_line_x simulated 3d on x axis
   * @param _3d_line_y simulated 3d on y axis
   */
  protected int onYAxis( 
        boolean _get_step_unit_width, Rect _R,
        DAxis _axis, String _property,
        int _3d_line_x, int _3d_line_y
      )
    throws ChartLocation  
  {
    int max_step_text_width=0;

    if( !_axis.getProperties().hasKey( _property ) ) return 0;

    float value = _axis.getProperties().get( _property, 0.0f );
    if( value == 0.0f ) return 0;
     
    float min     = _axis.getProperties().get( DAxis.P_MIN, -1.0f );
    float max     = _axis.getProperties().get( DAxis.P_MAX, +1.0f );
    float max_min = max-min ;

    //prevent div/0
    if ( max_min == 0.0f ) return 0;
    
    float wsu = _R.h()*value/max_min;
    int xd = _R.x()-margin_;
    //la numerotation, les lignes ... se fait en modulo value
    float vdn = (float)Math.ceil(min/value);
    if ( Math.abs(vdn)<1e-6 ) vdn=0.0f; 
    vdn *= value;
    float yd = _R.bottom() - _R.h()*(vdn-min)/max_min;

DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);

    for( float vd = vdn; vd <= max; vd += value )
    {
       if( DAxis.P_STEP_UNIT.equals( _property ) )
       {
              if ( Math.abs(vd)<1e-6 ) vd=0.0f;
              String t=df.format(vd);
              org.eclipse.swt.graphics.Point ext = d_.gc_.textExtent( t );
              int wt = ext.x;
              if( wt>max_step_text_width) max_step_text_width=wt;
              Rect r_txt=new Rect( xd-wt, (int)yd-ext.y/2, wt, ext.y );
              if( d_.drawing() )
              {
                if( !_get_step_unit_width ) //mode calcul largeur ?
                {
                  DrawUtil.drawText( d_.gc_, r_txt, DrawUtil.CENTER, t );
                  d_.gc_.drawLine( xd+margin_-2, (int)yd, xd+margin_, (int)yd );
                }
              }
              else //locating
              {
                if ( r_txt.contains( d_.lx_, d_.ly_ ) )
                {
                  throw new ChartLocation( DLocated.InAxisYStepUnitText, _axis, r_txt, t );
                }
              }
       }
       else if ( DAxis.P_STEP_DOT.equals( _property ) )
       {
              if(d_.drawing())
               d_.gc_.drawLine( xd+margin_-2, (int)yd, xd+margin_, (int)yd );
       }
       else if ( DAxis.P_STEP_LINE.equals( _property ) )
       {
//TBD:locating: detection des lignes, mais apres les courbes, alors que le dessin se fait avant ...
            if(d_.drawing()) 
            {
              d_.gc_.drawLine( _R.left(), (int)yd, _R.right(), (int)yd );
              if( _3d_line_x!=0 && _3d_line_y!=0 )
              {
                d_.gc_.drawLine( _R.x(), (int)yd, _R.x()-_3d_line_x, (int)yd+_3d_line_y );
              }
            }
       }
       yd -= wsu;
    }

    return max_step_text_width;
  }
  
  /** Render property on X axis, no support for "3D-like".
   * @param _R rectangle to render axis.
   * @param _axis axis to render.
   * @param _property property identifier to render: STEP_LINE,STEP_UNIT,STEP_DOT.
   */
  protected void onXAxis( Rect _R, DAxis _axis, String _property )
   throws ChartLocation
  {
    if( !_axis.getProperties().hasKey( _property ) ) return ;
    float value = _axis.getProperties().get( _property, 0.0f );
    if( value==0.0f ) return ;
    float min     = _axis.getProperties().get( DAxis.P_MIN, -1.0f);
    float max     = _axis.getProperties().get( DAxis.P_MAX, +1.0f );
    float max_min = max - min ;

    if ( max_min == 0.0f ) return; //sinon div/0
    
    float wsu = (_R.w()*value/max_min);
    int yd = _R.y()+_R.h()+margin_;
    //la numerotation, les lignes ... se fait en modulo prop_todo
    float vdn = (float)Math.ceil(min/value) ;
    if ( Math.abs(vdn)<1e-6 ) vdn=0.0f; // x*0 = 0
    vdn = vdn * value ;
    float xd = _R.left()+_R.w()*(vdn-min)/max_min;
    float vd;
    Rect r_txt=new Rect();
DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);    

    for( vd  = vdn;//_axis->property(DAxis::MIN);
         vd <= max;
         vd += value
       )
    {
      if( DAxis.P_STEP_UNIT.equals( _property ) )
      {
          if ( Math.abs(vd)<1e-6 ) vd=0.0f;        
          String t=df.format( vd );
          Point ext = d_.gc_.textExtent( t );
          r_txt.setRect( (int)xd-ext.x/2, yd, ext.x, ext.y );
          if(d_.drawing())
          {
            DrawUtil.drawText( d_.gc_, r_txt, DrawUtil.CENTER, t );
            d_.gc_.drawLine( (int)xd, yd-margin_, (int)xd, yd-margin_+2 );
          }
          else //locating
          {
            if ( r_txt.contains( d_.lx_, d_.ly_ ) )
            {
              throw new ChartLocation( DLocated.InAxisXStepUnitText, _axis, r_txt, t );
            }
          }
      }
      else if ( DAxis.P_STEP_DOT.equals( _property ) )
      {
          d_.gc_.drawLine( (int)xd, yd-margin_, (int)xd, yd-margin_+2 );
      }
      else if ( DAxis.P_STEP_LINE.equals( _property ) )
      {
          d_.gc_.drawLine( (int)xd, _R.top(), (int)xd, _R.bottom() );
//TBD:locating: la ligne, mais apres la detection des courbes ...
      }
      xd += wsu;
    }//for
  }
  
  /** draw or locate for an horizontal/vertical text.
   * @param _text text to render.
   * @param _rect rectangle to draw text
   * @param _align LEFT/RIGHT/HCENTER/... flag
   * @param _style style to paint
   * @param _locid identifier for location
   * @param _item item for location
   */
  public void renderText( String _text, Rect _rect, int _align, DChartStyle _style,
                          String _locid, IDItem _item )
    throws ChartLocation                           
  {
    if( d_.drawing() )
    {
      //la valeur est affichee dans le style de la courbe
      d_.gc_.setForeground( _style.fore_ );
      d_.gc_.setFont( _style.font_ );
      DrawUtil.drawText( d_.gc_, _rect, _align, _text );
    }
    else //locating
    {
      Rect nrtxt = DrawUtil.boundingRect( d_.gc_, _rect, _align, _text );
      //parce que si le texte deborde de _rect, .. ne le clippe pas
      //nrtxt might be bigger than _rect, clip it
      nrtxt.intersection( _rect );

      if( (nrtxt.w()>0) && nrtxt.contains( d_.lx_, d_.ly_ ) )
      {
        throw new ChartLocation( _locid, _item, nrtxt, _text );
      }
    }
  }
  
  /** Render Histogram graphic family */
  protected void renderHistogram( Rect _RG, boolean _3d, boolean _bs )
    throws ChartLocation
  {
    //histogram must have only one axis.
    DAxis axis=null;
    for( IDItem i=d_.G_.getFirstChild(); i!=null; i=i.getNext() )
    {
      if( i instanceof DAxis )
      {
        axis=(DAxis)i;
        break;
      }
    }

    if ( axis==null )
    {
      error( "Error: Histogram without axis");
      return ;
    }

    DChartStyle axis_style = styleOf( axis );
    d_.gc_.setFont( axis_style.font_ );
    int axis_font_h = d_.gc_.getFontMetrics().getHeight();

    int max_h_title_unit = axis_font_h + 2 ;

    //title et unit sont geree dans des chaines/rect different cela va permettre
    //de les afficher avec des styles differents ... un jours prochains.
    TruncateAxisTextData tatd =null;
 /*   String axis_title ;
    String axis_unit  ;
    Rectangle rttly; //rect du titre de l'axe
    Rectangle runity;//rect de l'unit de l'axe
    Rectangle rtuy; //rect englobant les deux premiers.
*/
    //reduction de la chaine de titre ou units..
//XXXX    //available_width est un height a cause de la rotation de 90
    tatd = truncateVAxisText( axis, axis_style, _RG.h(), max_h_title_unit );

    int yg = _RG.y()+ axis_font_h/2; //numerotation de l'axe
    int hg = _RG.h()
           - axis_font_h/2  // 1/2 hauteur aff. mesure axe Y (au cas ou)
           - margin_
           ;

    //mesure sur l'axe des Y (avec marge):
    int wuy=0;
    hg -= axis_font_h+margin_ ;
    wuy = onYAxis( true, new Rect(), axis, DAxis.P_STEP_UNIT, 0, 0 );
  
    int wg = _RG.w() ;
    int xg = _RG.x() ;

    //display Y axis title
    {
      int axis_title_w = 0;
      //largeur du titre + unite => hauteur max. de police (tourne a 90 degre)
      if( !DrawUtil.isEmpty( tatd.sttl_ )
       || !DrawUtil.isEmpty( tatd.sunit_ ) ) axis_title_w = max_h_title_unit;

      //titre de l'axe + unite en vertical a droite de l'axe des Y
      int decoration_w = wuy //aff. mesure axe des Y (max.)
                       + margin_ //entre wuy et max_h_title 
                       + axis_title_w // titre et/ou unite
                       + margin_ //entre max_h_... et bord 
                       ;
      wg -= decoration_w ;
      xg += decoration_w ;
    }

//TBD: 20 a afficher comme valeur de 3D
    int  _3d_z    = 20 ;
    if ( _3d )
    {
      wg -= _3d_z;
      hg -= _3d_z;
      yg += _3d_z;
    }


    Rect rg = new Rect(xg,yg,wg,hg);

    //lost of Y axis enlightment
    {
      //rect global de placement titre/unit Y
      //a note: a cause de la rotation on utilise la translation du QPainter
      // pour le placement correct, les calculs sont fait a origine 0,0
      Rect grtuy= new Rect( 0, 0, _RG.h(), max_h_title_unit );
      //centrer le rect titre+unit Y
      tatd.rtu_.moveCenter( grtuy.centerX(), grtuy.centerY() );
      //replacer rect title et unite
      tatd.rttl_.moveBy(  tatd.rtu_.x(), tatd.rtu_.y() );
      tatd.runit_.moveBy(  tatd.rtu_.x(), tatd.rtu_.y() );

      if( d_.drawing() )
      {
//        d_.gc_.save();
//TODO: he les gars SWT on fait ca comment ????
//        d_.gc_.translate( _RG.x, _RG.y+_RG.height );
//        d_.gc_.rotate( -90.0f );
//ye! SWT tu fait comment la rotation ???        
        d_.gc_.setForeground( axis_style.fore_ );
        d_.gc_.setFont( axis_style.font_ );
      
        DrawUtil.drawText( d_.gc_, tatd.rttl_, DrawUtil.CENTER, tatd.sttl_ );
        DrawUtil.drawText( d_.gc_, tatd.runit_, DrawUtil.CENTER, tatd.sunit_ );
       
   //     d_.gc_.restore();
      }
      else //locating
      {
        //on deplace le pt en (0,0) sur grtuy et on invers x/y:
        //pour prendre en compte la rotation
        int lx=d_.lx_;
        int ly=d_.ly_;
        //rotation di -90: on inverse x/-y w/h
        grtuy.setRect(  -grtuy.y(), grtuy.x(), grtuy.h(), +grtuy.w() );
        tatd.rttl_.setRect(  -tatd.rttl_.y(), tatd.rttl_.x(), tatd.rttl_.h(), +tatd.rttl_.w());
        tatd.runit_.setRect(  -tatd.runit_.y(), tatd.runit_.x(), tatd.runit_.h(), +tatd.runit_.w() );
        //translation:
        grtuy.moveBy(   _RG.x(), _RG.y() /*+_RG.height() */);
        tatd.rttl_.moveBy(  _RG.x(), _RG.y() /*+_RG.height() */);
        tatd.runit_.moveBy(  _RG.x(), _RG.y() /*+_RG.height() */);
        if( grtuy.contains( lx, ly ) )
        {
          //dans la zone de l'axe
          if( !DrawUtil.isEmpty(tatd.sttl_) && tatd.rttl_.contains( lx, ly ) )
          {
            //dans la zone de titre
            throw new ChartLocation( DLocated.InAxisYTitle, axis, tatd.rttl_, tatd.sttl_ );                    
          }
          if( !DrawUtil.isEmpty(tatd.sunit_) && tatd.runit_.contains( lx, ly ) )
          {
            //dans la zone de titre
            throw new ChartLocation( DLocated.InAxisYUnit, axis, tatd.runit_, tatd.sunit_);
          }
          //donc dans la zone de l'axe Y
          throw new ChartLocation( DLocated.InAxisYArea, axis, grtuy );
        }
      }

      //lines, unit, 'n dot sur l'axe des Y
      Rect rg2 = new Rect(rg);
      if(_3d) rg2.moveBy( _3d_z,-_3d_z );

      //drawing et:ou locating:
      if(d_.drawing())
      {
        d_.gc_.setForeground( getLightGray() );
      }
      if(_3d)
      {
        onYAxis( false, rg2, axis, DAxis.P_STEP_LINE, _3d_z, _3d_z );
      } else {
        onYAxis( false, rg2, axis, DAxis.P_STEP_LINE, 0, 0 );
      }
      if(d_.drawing())
      {
        d_.gc_.setForeground( getBlack() ); //black
      }
      onYAxis( false, rg, axis, DAxis.P_STEP_DOT, 0,0  );
      if(d_.drawing())
      {
        d_.gc_.setForeground( axis_style.fore_ );
        d_.gc_.setFont( axis_style.font_ );
      }
      onYAxis( false, rg, axis, DAxis.P_STEP_UNIT, 0,0 );
    }
  
    if(d_.drawing())
    {
      d_.gc_.setForeground( axis_style.fore_ );
      d_.gc_.drawLine( xg, yg,    xg,    yg+hg ); //axe Y
      d_.gc_.drawLine( xg, yg+hg, xg+wg, yg+hg ); //axe X
    }

    //dessin des courbes dans l'ordre
    float ay_min=axis.getProperties().get( DAxis.P_MIN,  0.0f),
          ay_max=axis.getProperties().get( DAxis.P_MAX, +1.123412345f);
    float ky = ay_max - ay_min;
System.err.println("HISTO axis["+axis+"].MIN="+ay_min);    
System.err.println("HISTO axis["+axis+"].MAX="+ay_max);
System.out.println("  PROP=["+DAxis.P_MAX+"]");    
    

    //if min==max, force a delta of 10%
    if( Math.abs(ky)<1e-6 )
    {
      ay_max = (ay_max>0.0f ? 1.05f : 0.95f )* ay_max;
      ay_min = (ay_min>0.0f ? 0.95f : 1.05f )* ay_min;
      ky = ay_max - ay_min;     
      if( Math.abs(ky)<1e-6)
      {
        //smell like update document problem (MIN/MAX=0.0f)
        error("Error: can't draw document not updated");
        return ;
      }
    }

    ky = hg/ky;
    int nbar=d_.curves_.size();

    float wbar=rg.w()/(float)nbar;
    float wbar_bs = (wbar+1)/3.0f; // _BS

    float xbar=rg.x();
    int   y_bar_zero;
    float bar_zero;
    if ( ay_min > 0.0f ) {
      y_bar_zero=rg.bottom();
        bar_zero=ay_min;
    } else if ( ay_max < 0.0f ) {
      y_bar_zero=rg.y();
        bar_zero=ay_max;
    }else {
      y_bar_zero=rg.bottom()-(int)(ky*(-ay_min)); //pos de 0.0
        bar_zero=0.0f;
    }

    IDItem item=null;
    ChartLocation loc=null;

DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits( 2 );

    for( int icrv=0; icrv<d_.curves_.size(); ++icrv )
    {
      CurveStyle ccurve = (CurveStyle)d_.curves_.get(icrv);    
      for( item=ccurve.curve_.getFirstChild(); item!=null; item=item.getNext() )
      {
        //les courbes n'ont qu'un point
//TBD:on pourrait imaginer des histogram multi point en superposant les bar.???
        if( item instanceof DPoint )
        {
          DPoint point = (DPoint)item;
          float vy;
          DCoord coord=point.getCoord( axis );          
          if ( coord !=null )
          {
            vy = coord.getValue();
            //transformation.
            float bh = vy-bar_zero;
            int  ibh = (int)(ky*(bh));
            if (ibh>0)ibh+=1;else if (ibh<0) ibh-=1;        
            //curve==BAR forcement.
            Rect rbar;
            if( _bs )  // bandes separees
            {
              int dx = (int)(xbar+0.5f*(wbar-wbar_bs));
              rbar = new Rect( dx, y_bar_zero, (int)wbar_bs, -ibh );
            } else {
              rbar = new Rect( (int)xbar, y_bar_zero, (int)wbar+1, -ibh );
            }
            rbar.normalize();

            //barre principale
            if(d_.drawing())
            {
              d_.gc_.setForeground( getBlack() ); //black ://trait exterieur en noir
              d_.gc_.setBackground( ccurve.style_.back_ );
              d_.gc_.fillRectangle( rbar.x(), rbar.y(), rbar.w(), rbar.h() );
              d_.gc_.drawRectangle( rbar.x(), rbar.y(), rbar.w(), rbar.h() );
            } 
            else //locating
            {
              if( rbar.contains( d_.lx_, d_.ly_ ) )
              {
                throw new ChartLocation( DLocated.CurveBar, point,  rbar );
              }
            }

//TBD: a revoir parce que les barre value>0 sont mal affiches en 3d
            if ( _3d )
            {
//TBD:20: c'est le Z 3d
              int ar1[] = new int[8],
                  ar2[] = new int[8];
                    
              ar1[0] = rbar.x();                ar1[1] = rbar.y();
              ar1[2] = rbar.x()+20;             ar1[3] = rbar.y()-20;
              ar1[4] = (int)(rbar.x()+wbar+20); ar1[5] = rbar.y()-20;
              ar1[6] = (int)(rbar.x()+wbar+1);  ar1[7] = rbar.y();
    
              ar2[0] = (int)(rbar.x()+wbar);    ar2[1] = rbar.y();
              ar2[2] = ar1[4];                ar2[3] = ar1[5]; 
              ar2[4] = (int)(rbar.x()+wbar+20); ar2[5] = rbar.y()+rbar.h()-20;
              ar2[6] = (int)(rbar.x()+wbar);    ar2[7] = rbar.y()+rbar.h();

              if(d_.drawing())
              {
                Color clr   = ccurve.style_.back_;
                Color dark1 = DrawUtil.lighter( clr, 0.50f );
                Color dark2 = DrawUtil.lighter( clr, 0.75f );
                d_.gc_.setBackground( dark1 );
                d_.gc_.fillPolygon( ar1 );
                d_.gc_.drawPolygon( ar1 );
                d_.gc_.setBackground( dark2 );
                d_.gc_.fillPolygon( ar2 );
                d_.gc_.drawPolygon( ar2 );
              }
              else //locating
              //Il faudrait parcourir les barres a l'envers pour ne pas 
              //se tromper lorsque les barres 3D se recouvrent ...
              //MAIS, je preferre y aller comme le dessin et choisir
              //la derniere zone 3d designee par la souris.
              {
//evidement:20 c'est la simili projection 3d
                //l'ideal c'est que l'on est un ar1.contains( ... )
                boolean in_ar1 = Data.horizontal3DContains( rbar.x(), rbar.y(),
                                          rbar.w(), 20, d_.lx_,d_.ly_ );
                boolean in_ar2 = Data.vertical3DContains( rbar.right(), rbar.y(),
                                         rbar.h(), 20, d_.lx_,d_.ly_ );
                //pas encore vu ...
                if( in_ar1 )
                {
                  //memo...
                  Rect br = new Rect( rbar.x(), rbar.y()-20, rbar.w()+20, 20 );
                  loc = new ChartLocation( DLocated.CurveBar, point, br );
                }
                else if( in_ar2 )
                {
                  Rect br = new Rect( rbar.right(), rbar.y()-20, 20, rbar.h()+20 );
                  loc = new ChartLocation( DLocated.CurveBar, point, br );
                }
                //sinon si on en avait trouve, c'est la derniere...
                else if( loc!=null )
                {
                  throw loc;
                }
              }
            }

            //value is displayed below the bar.
            {
              String full_txt = df.format( vy );
              d_.gc_.setFont( ccurve.style_.font_ );
              String txt = DrawUtil.truncateTextH( d_.gc_, full_txt, (int)wbar-2 );
              if( !DrawUtil.isEmpty( txt ) )
              {
                int font_h = d_.gc_.getFontMetrics().getHeight();
                Rect rtxt = new Rect( (int)xbar, rg.bottom()+margin_,
                                                (int)wbar-2, font_h+2 );
                //drawn using curve style
                renderText( txt, rtxt, DrawUtil.CENTER, ccurve.style_,
                             DLocated.CurvePointText, point );
              }
            }
          }
          else
          {
            error( "Error: Bar without value." );
          }

          //si plusieurs valeur... xbar += wbar;
        }//if 
      }//for 
      xbar += wbar;
    }//for
 
    //pour la derniere zone 3d vue...
    if( d_.locating() && loc!=null )
    {
      throw loc;
    }
  }
  
  /** Render sector-ish graphic */
  protected   void renderSector( Rect _RG, boolean _3d ) throws ChartLocation
  {
    boolean display_values = true; //true to display value of sector
    int d = _RG.w();
    //DR2358: je fais ce que je peux...avec le temps que j'ai.
    // Comme ca bave un peu au dessus surl le titre du graphique,
    // je vais reserver une zone de 2*20 pixels (dessus/dessous)
    // pour le texte (si display_decoration).
    // Ok, 20 c'est sorti de la cuisse de Jupiter, car a ce stade
    // on ne sait pas si il y aura vraiment du texte dessus/dessous
    // (il peut etre un peu decale), ni si ce texte a un style dont
    // la police est <= 20 pixels...
    // Vous aurez compris que c'est totalement du pif !
    // En attendant ca calme M'sieu Telep, en attendant qu'il trouve 
    // que les % se chevauchent quand le graphique est petit...
    int h_for_text = display_values ? 40 : 0;
    if ( d > _RG.h()-h_for_text ) d=_RG.h()-h_for_text;

    //si on affiche les texte:
    if ( display_values )
    {
//TBD:il y a surement un calcul plus serieux que 3/4 !!!
      d = 3*d/4; //les secteurs occupent 3/4 de la place (prevoir pr texte)
    }

    float v_sum = 0.0f;

    CurveStyle ccurve=null;
    IDItem item=null;
    
    DAxis axis=null;
    for( item=d_.G_.getFirstChild(); item!=null; item=item.getNext() )
    {
      if( item instanceof DAxis ) axis=(DAxis)item;
    }
    if( axis==null )
    {
      error("Error: missing Axis in graphic");
      return;
    }
    String axis_unit = axis.getUnit();
    if( axis_unit==null ) axis_unit="";

    //cumul des valeurs des courbes (pour correspondance au 100% du cercle)
    int n_points=0;
    for( int i=0; i<d_.curves_.size(); ++i )
    {
      ccurve = (CurveStyle)d_.curves_.get(i);
      //courbes: plusieurs point a une seule coord possibles
      for( item=ccurve.curve_.getFirstChild(); item!=null; item=item.getNext())
      {
        if ( item instanceof DPoint )
        {
          DCoord coord= ((DPoint)item).getCoord( axis );
          if ( coord!=null )
          {
            v_sum += coord.getValue();
            n_points++;
          }
        } 
      }
    }

    //v_sum represente 100% du secteur.
    if ( v_sum == 0.0f )
    {
      error( "Error: can't print sectors (sum:0)" );
      return ;
    }

    int x3d = 0, y3d = 0, r3d=0, a3d=0;

    if ( _3d )
    {
      //parce que c'est a peut pres correct
      if( d >= 48 ) //ombre 4 fois moins importante que le pie chart
      {
        x3d = - 8;
        y3d = +12;
      } else {
        y3d = d / 4;
        x3d = -(int)(2.0f*y3d/3.0f);
        // on voit rien: circulez y'a rien a voir
        if( Math.abs(x3d) < 2 ) //au moins 2 pixels d'ombrage
        {
          _3d = false;
          x3d = y3d = 0;
        }
      }
      if( _3d )
      {
        r3d = (int)Math.sqrt( (double)(x3d*x3d + y3d*y3d) );
        float a = (float)Math.atan2( (float)y3d, (float)x3d ); //-PI..PI
        if( a < 0.0f ) a += 2*Math.PI ; 
        //a est dans 0..2PI 
        a3d = (int)(180*a/(Math.PI));
      }
    }

    LineAlg l3d = new LineAlg( 0, 0, x3d, y3d );

    //locating: les sector sont detectes dans l'ordre de dessin,
    //donc on les teste tous, et c'est le dernier qui gagne
    ChartLocation loc_sector=null;

DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);

    //because line separating two sector must be drawn after sector
    //otherwise fill next sector will overwrite line.
    //this array store x,y coordinates of point to draw joined to center or pie.
    int between_sectors[] = new int[n_points*2];
    int wbs=0;
    Rect r = new Rect();
    
    while( l3d.nextPoint() )
    {
      int dx3d = l3d.x();
      int dy3d = l3d.y();
      boolean top=(dx3d==x3d)&&(dy3d==y3d);

      int angle=0;
      int ray = d >> 1; 

      r.setRect( _RG.x()+(_RG.w() -d)/2 +dx3d,
                 _RG.y()+(_RG.h() -d)/2 -dy3d, d,d ); 
      if( r.w()==0 || r.h() == 0 ) continue;

      //locating: (ici pour optimiser le calcul)
      float loc_dist=-1.0f;
      int loc_angle=-1; //en 16eme de degre
      if(d_.locating())
      {
        Point rc=r.center();
        int dx=d_.lx_-rc.x, dy=d_.ly_-rc.y;
        loc_dist=(float)Math.sqrt( dx*dx + dy*dy );
        if(loc_dist<=ray)
        {
          float a = (float)Math.atan2( (float)-dy, (float)dx ); //-PI..PI
          if( a < 0.0f ) a += 2*Math.PI ; 
          //a est dans 0..2PI 
          loc_angle = (int)(360*a/(2*Math.PI));
        }
      }
      
      int ipt=0;
      for( int i=0; i<d_.curves_.size(); ++i )
      {
        ccurve = (CurveStyle)d_.curves_.get(i);
        //Courbes: plusieurs points possibles
        for( item=ccurve.curve_.getFirstChild(); item!=null; item=item.getNext() )
        {
          if ( item instanceof DPoint )
          {
            DPoint point = (DPoint)item;
            DCoord coord = point.getCoord( axis );
            if ( coord !=null )
            {
              ipt++;
              float v = coord.getValue()/v_sum; //% du disque
              int end_angle = angle+(int)(360*v); //degree;
              if( ipt >= n_points ) //last sector ?
              {
                end_angle=360;
              }

              if( d_.drawing() )
              {
                if(top)
                {
                  d_.gc_.setForeground( getBlack() ); //black
                  Color b = ccurve.style_.back_;
                  d_.gc_.setBackground( b );
                  d_.gc_.fillArc( r.x(),r.y(),r.w(),r.h(), angle, end_angle-angle );
                  d_.gc_.drawArc( r.x(),r.y(),r.w(),r.h(), angle, end_angle-angle );
                  int cx = r.centerX(), cy=r.centerY();
                  double a = (Math.PI*angle)/180.0f;
                  between_sectors[wbs++]= (int)( cx+r.w()*Math.cos( a )/2.0f );
                  between_sectors[wbs++]= (int)( cy-r.h()*Math.sin( a )/2.0f );                  
                } else {
                  Color dark = DrawUtil.lighter( ccurve.style_.back_, 0.75f );
                  d_.gc_.setForeground( dark );
                  d_.gc_.setBackground( dark );
                  d_.gc_.fillArc( r.x(),r.y(),r.w(),r.h(), angle, end_angle-angle );
                }
              }
              else //locating
              {
                if( (loc_dist <= ray)
                    &&(loc_angle>=angle)&&(loc_angle<=end_angle) )
                {
                  //memo pour connaitre le don secteur a la fin ..
                  loc_sector = new ChartLocation( DLocated.CurveSector, point, r );
                  ((DChartLocation)loc_sector.location_).setAngles( angle, end_angle );
                }
              }
              

              //affichage de la valeur + unite de l'axe
              if( top && display_values )
              {
                int atxt = angle+(end_angle-angle)/2;
                //calculer la compensation sur le rayon en mode _3d pour
                //que le texte ne recouvre pas l'ombre:
                int r3=0;
                if ( _3d )
                {
                  /* aller, j'en dis un peu plus:
                   * en 3d on va ajouter un rayon pour ne pas afficher le texte
                   * sur l'ombre, le + c'est que ce rayon depend de l'angle
                   * du texte ET de l'angle choisit pour l'ombrage, eh, eh, eh!
                   * C'est pas beau ca !
                   */
                  int da = (int)Math.abs( atxt-a3d-180 );
                  //compenser que dans la zone de l'ombre:
                  if( da < 90 ) //PI/2
                  {
                    //approx du 1er degre:
                    float k = 1.0f - da/90.f;
                    r3 = (int)(k * r3d);
                  }
                }
                float arad=(float)( 2*Math.PI-(2*Math.PI*atxt)/360 );
                int sector_ray = ray+margin_+r3 ;
                int tx = (int)(r.x()+(r.w() >>1)+sector_ray*Math.cos(arad));
                int ty = (int)(r.y()+(r.h()>>1)+sector_ray*Math.sin(arad));

                String txt = df.format( coord.getValue() ) + axis_unit;

                Point te = d_.gc_.textExtent( txt );
                int tw = te.x;
                int th = te.y;

                //on essaye de placer le texte au mieux
                     if ( atxt < 90  ) {}/*raf*/
                else if ( atxt < 180 ) { tx -= tw; }
                else if ( atxt < 270 ) { tx -= tw; ty+=th; }
                else                    {          ty+=th; };

                Rect rtxt = new Rect( tx, ty-th, tw, th );

                renderText( txt, rtxt,  DrawUtil.LEFT|DrawUtil.BOTTOM,
                             ccurve.style_, DLocated.CurvePointText,
                             point );
                if( loc_sector!=null ) throw loc_sector;
              }

              angle = end_angle;
            }
          }//if
        }//item
      }//curve
    }//3d
    
    if( d_.drawing() )
    {
      d_.gc_.setForeground( getBlack() ); //black
      int cx = r.centerX(), cy=r.centerY();
      for( int i=0; i<between_sectors.length; )
      {
        int x = between_sectors[i++];
        int y = between_sectors[i++];
        d_.gc_.drawLine( cx,cy, x,y );
      }
    }
    else //locating
    {
      if( loc_sector!=null ) throw loc_sector;
    }
  }

  /** Render sector-ish graphic */
  protected   void renderXY( Rect _RG ) throws ChartLocation
  {
    //decorations means axis titles/units/...
    boolean display_decorations = true;
    
    //search for axis "x" and "y", must exists
    DAxis x_axis=null, y_axis=null;
    for( IDItem item=d_.G_.getFirstChild(); item!=null; item=item.getNext() )
    {
      if( item instanceof DAxis )
      {
        DAxis axis = (DAxis)item;
        if( x_axis==null && "x".equals( axis.getName() ) )
        {
          x_axis = axis;
        }
        else if ( y_axis==null && "y".equals( axis.getName() ))
        {
          y_axis = axis;
        }
        if( x_axis!=null && y_axis!=null ) break;
      }
    }

    if ( x_axis==null || y_axis==null )
    {
      String txt;
      if( x_axis==null )
      {
        txt = "Error: missing 'x' axis";
      } else {
        txt = "Error: missing 'y' axis";
      }
      error( txt );
      return ;
    }

    DChartStyle x_axis_style = styleOf( x_axis );
    DChartStyle y_axis_style = styleOf( y_axis );

    d_.gc_.setFont( x_axis_style.font_ );
    int x_axis_style_h = d_.gc_.textExtent( "X" ).y+2;
    d_.gc_.setFont( y_axis_style.font_ );
    int y_axis_style_h = d_.gc_.textExtent( "X" ).y+2;
//evidement c'est sorti de la cuisse de jupiter...    

    //metrics titre ( unit ) axe des X pour les chaines/rect
    int xg,yg,hg,wg;

    xg = _RG.x();
    yg = _RG.y();
    hg = _RG.h();
    wg = _RG.w();

DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);
    if ( display_decorations )
    {  
      d_.gc_.setFont( y_axis_style.font_ );
      //compute steps width ont Y Axis...
      int wuy= onYAxis( true,  new Rect(), y_axis, DAxis.P_STEP_UNIT, 0,0 );

      //it's a worth measure.
      int wux=0;
      if ( x_axis.getProperties().hasValue(DAxis.P_STEP_UNIT))
      {
        String s = df.format( x_axis.getProperties().get(DAxis.P_MAX, +1.0f) );
        d_.gc_.setFont( x_axis_style.font_ );
        wux = d_.gc_.textExtent( s ).x;
      }

//Note: title and unit of Y axis are drawn verticaly
      int y_axis_title_unit_w = 0;
      if ( !DrawUtil.isEmpty( y_axis.getTitle())
        || !DrawUtil.isEmpty( y_axis.getUnit()) )
      {
//TODO: SWT sait pas faire le 90 !      
//TODO:pas de rotation:           y_axis_title_unit_w = y_axis_style_h ; //texte tourne a 90 degre.
        y_axis_title_unit_w = d_.gc_.textExtent( "X" ).x+2;           
      }

      wg -= y_axis_title_unit_w // titre ( unite ) axe des Y ecris en vertical
          + margin_         // marge en titre et mesure axe des Y
          + wuy             // aff. mesure axe des Y (maximum)
          + margin_         // entre aff mesure axe des Y et l'axe des Y
          + wux/2           // au pire on doit afficher la mesure maxi (axe X)
          ;
     
      xg += y_axis_title_unit_w
         + margin_
          + wuy
          + margin_
          ;

      int x_axis_title_unit_h = 0;
      if ( !DrawUtil.isEmpty( x_axis.getTitle() )
        || !DrawUtil.isEmpty( x_axis.getUnit() ) )
           x_axis_title_unit_h = y_axis_style_h;

      hg -= y_axis_style_h/2  // 1/2 hauteur aff. mesure axe Y (au cas ou)
          + margin_           // marge en axe X et affichage mesure des X
          + x_axis_style_h    // mesure des X  
          + margin_           // entre mesure des X et titre axe X
          + x_axis_title_unit_h // titre (unite) axe des X
          ;

      yg += y_axis_style_h/2 ; //1/2 nombre possible en haut axe des Y

      Rect rg = new Rect(xg,yg,wg,hg);

      //X Axis:
      TruncateAxisTextData tatdx
            = truncateHAxisText( x_axis, x_axis_style, wg, x_axis_style_h );

      //rect global de placement titre/unit X pour recenter les rect...
      Rect grtux=new Rect( xg, _RG.bottom()-margin_-x_axis_style_h, wg, x_axis_style_h );
      //centrer dans grtux
      tatdx.rtu_.moveCenter( grtux.center() );
      //donc:
      tatdx.rttl_.moveBy( tatdx.rtu_.x(), tatdx.rtu_.y() );
      tatdx.runit_.moveBy( tatdx.rtu_.x(), tatdx.rtu_.y() );

      d_.gc_.setFont( x_axis_style.font_ );
      if( d_.drawing() )
      {
        d_.gc_.setForeground( x_axis_style.fore_ );
     
        DrawUtil.drawText( d_.gc_, tatdx.rttl_, DrawUtil.CENTER, tatdx.sttl_ );
        DrawUtil.drawText( d_.gc_, tatdx.runit_, DrawUtil.CENTER, tatdx.sunit_ );

        //lines, unit, 'n dot sur l'axe des X
        d_.gc_.setForeground( getLightGray() ); 
//TBD:render
        onXAxis( rg, x_axis, DAxis.P_STEP_LINE );
        d_.gc_.setForeground( getBlack() );
        onXAxis( rg, x_axis, DAxis.P_STEP_DOT  );
        d_.gc_.setForeground ( x_axis_style.fore_  );
        d_.gc_.setFont( x_axis_style.font_ );
        onXAxis( rg, x_axis, DAxis.P_STEP_UNIT );
      }
      else //locating
      {
        if( grtux.contains( d_.lx_, d_.ly_ ) )
        {
          //inside axis area...
          if( !DrawUtil.isEmpty( tatdx.sttl_ ) && tatdx.rttl_.contains( d_.lx_, d_.ly_ ) )
          {
            //inside title area
            throw new ChartLocation( DLocated.InAxisXTitle, x_axis, tatdx.rttl_, tatdx.sttl_ );
          }
          if( !DrawUtil.isEmpty( tatdx.sunit_ ) && tatdx.runit_.contains( d_.lx_, d_.ly_ ) )
          {
            //dans la zone de titre
            throw new ChartLocation( DLocated.InAxisXUnit, x_axis,
                                      tatdx.runit_, tatdx.sunit_ );
          }
          //so, in axis area.
          throw new ChartLocation( DLocated.InAxisXArea, x_axis, 
                                    grtux, tatdx.sttl_+tatdx.sunit_ );
        }

        //locating unit step text ...
        onXAxis( rg, x_axis, DAxis.P_STEP_UNIT );
      }

      // Axe Y:
      TruncateAxisTextData tatdy
        = truncateVAxisText( y_axis, y_axis_style, xg-_RG.x(), hg );

      //rect global de placement titre/unit Y
//TODO: pas de text verticaux dans SWT.      
      //a note: a cause de la rotation on utilise la translation du QPainter
      // pour le placement correct, les calculs sont fait a origine 0,0
 //     Rect grtuy=new Rect( yg, 0, hg, y_axis_style_h );
 Rect grtuy = new Rect( 0, yg, xg-_RG.x(), hg ); //y_axis_style_h, hg );
      //centrer le rect titre+unit Y
      tatdy.rtu_.moveCenter( grtuy.center() );

      //replace title and unit rectangle (remember: drawn verticaly)
      tatdy.rttl_ .setLocation( tatdy.rtu_.x(), tatdy.rtu_.y() );
      tatdy.runit_.setLocation( tatdy.rtu_.x(), tatdy.rtu_.bottom()-tatdy.runit_.h() );

      d_.gc_.setFont( y_axis_style.font_ );
      if( d_.drawing() )
      {
//TODO: pas de texte verticaux...        
/*        d_->P_->save();
        d_->P_->translate( _RG.left(), _RG.top()+_RG.height() );
        d_->P_->rotate( -90.0f );
*/        
        d_.gc_.setForeground( y_axis_style.fore_ );

        DrawUtil.drawText( d_.gc_, tatdy.rttl_,  DrawUtil.CENTER|DrawUtil.VERTICAL, tatdy.sttl_ );
        DrawUtil.drawText( d_.gc_, tatdy.runit_, DrawUtil.CENTER|DrawUtil.VERTICAL, tatdy.sunit_  );
//        d_->P_->restore();

        //lines, unit, 'n dot sur l'axe des Y
        d_.gc_.setForeground( getLightGray() ); 
        onYAxis( false, rg, y_axis, DAxis.P_STEP_LINE, 0,0 );
        d_.gc_.setForeground( getBlack() );
        onYAxis( false, rg, y_axis, DAxis.P_STEP_DOT, 0,0  );
        d_.gc_.setForeground( y_axis_style.fore_  );
        d_.gc_.setFont( y_axis_style.font_  );
        onYAxis( false, rg, y_axis, DAxis.P_STEP_UNIT, 0,0 );
      } 
      else //locating
      {
//TODO: pas de texte verticaux, locating faux        
        //on deplace le pt en (0,0) sur grtuy et on invers x/y:
        //pour prendre en compte la rotation
        int lx=d_.lx_;
        int ly=d_.ly_;
        //rotation di -90: on inverse x/-y w/h
        grtuy.setRect( -grtuy.y(), grtuy.x(), grtuy.h(), +grtuy.w() );
        tatdy.rttl_.setRect( -tatdy.rttl_.y(), tatdy.rttl_.x(), tatdy.rttl_.h(), +tatdy.rttl_.w() );
        tatdy.runit_.setRect( -tatdy.runit_.y(), tatdy.runit_.x(), tatdy.runit_.h(), +tatdy.runit_.w() );
        //translation:
        grtuy.moveBy(  _RG.left(), _RG.top() /*+_RG.height() */);
        tatdy.rttl_.moveBy(  _RG.left(), _RG.top() /*+_RG.height() */);
        tatdy.runit_.moveBy(  _RG.left(), _RG.top() /*+_RG.height() */);
        if( grtuy.contains( lx, ly ) )
        {
          //in axis area... 
          if( !DrawUtil.isEmpty( tatdy.sttl_ ) && tatdy.rttl_.contains( lx, ly ) )
          {
            //in axis title area
            throw new ChartLocation( DLocated.InAxisYTitle, y_axis,
                                      tatdy.rttl_, tatdy.sttl_ );
          }
          if( !DrawUtil.isEmpty( tatdy.sunit_ ) && tatdy.runit_.contains( lx, ly ) )
          {
            //in unit text area
            throw new ChartLocation( DLocated.InAxisYUnit, y_axis,
                                      tatdy.runit_, tatdy.sunit_ );
          }
          //so in axis area...
          throw new ChartLocation( DLocated.InAxisYArea, y_axis, 
                                    grtuy, tatdy.sttl_+tatdy.sunit_ );
        }

        //locating unit step text ...
        onYAxis( false, rg, y_axis, DAxis.P_STEP_UNIT, 0,0 );
      }
    }

    //graph rect without axis
    Rect rg = new Rect(xg,yg,wg,hg);

    if( d_.drawing() )
    {
      d_.gc_.setForeground( getBlack() );
      d_.gc_.drawLine( xg, yg+hg, xg+wg, yg+hg ); // X axis
      d_.gc_.drawLine( xg, yg,    xg,    yg+hg ); // Y axis
    }

    //draw curves in ordre.
    float x_min = x_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
          x_max = x_axis.getProperties().get( DAxis.P_MAX, -1.0f ),
          y_min = y_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
          y_max = y_axis.getProperties().get( DAxis.P_MAX, +1.0f );
    float kx = x_max - x_min;
    float ky = y_max - y_min;
    if ( (kx==0.0f)||(ky==0.0f) )
    {
      error( "Error: can't draw curves (divison by 0)" );
      return;
    }
  
    kx = wg/kx;
    ky = hg/ky;
    //basis for histogram bar (xy curve might be bars)
    int y_bar_zero ;
    float bar_zero;

    //except if we haven't 0
    if ( y_min > 0.0f ) {
      y_bar_zero = rg.bottom(); //DAxis.MIN
        bar_zero = y_min;
    } else if ( y_max < 0.0f ) {
      y_bar_zero = rg.top(); //DAxis.MAX
        bar_zero = y_max;
    } else {
      //position of 0on Y axis:
      y_bar_zero = rg.bottom()-(int)(ky*(-y_min));
        bar_zero = 0.0f;
    }
     
    int last_x=-1, last_y=-1; 
    float  last_vy=-1;
    CurveStyle ccurve=null;
    Rect rp=new Rect();
    Rect r_bar=new Rect();

    //for locating mode, curve are checking in reverse order to handle covering.
    int ic_inc = d_.locating() ? -1 : +1;
    int ic_size = d_.curves_.size();
    for( int ic=d_.locating() ? ic_size-1 : 0;
         d_.locating()? (ic>=0) : (ic<ic_size);
         ic += ic_inc )
    {
      ccurve = (CurveStyle)d_.curves_.get(ic);      
      boolean first=true;
      IDItem last_item=null;
      DPoint point=null;
      for( IDItem item=ccurve.curve_.getFirstChild(); item!=null; item=item.getNext() )
      {
        if ( item instanceof DPoint )
        {
          point = (DPoint)item;
          DCoord cx = point.getCoord( x_axis );
          DCoord cy = point.getCoord( y_axis );
          if ( cx!=null && cy!=null )
          {
            float vx = cx.getValue();
            float vy = cy.getValue();
            //transformation.
            int px = xg+(int)(kx*(vx-x_min));
            int py = yg+hg-(int)(ky*(vy-y_min));
            if ((!first)|| ccurve.curve_.isType( DCurve.T_POINTS ) )
            {
              if( ccurve.curve_.isType( DCurve.T_LINE ) )
              {
                if( d_.drawing() )
                {
                  d_.gc_.setForeground( ccurve.style_.back_ );
                  //d_->P_->setBrush(Qt::NoBrush);
                  d_.gc_.drawLine( last_x, last_y, px,py );
                }
                else //locating
                {
                  if( DrawUtil.segmentContains( last_x, last_y, px, py, d_.lx_, d_.ly_ ) )
                  {
                    throw new ChartLocation( DLocated.CurveLine, point, new Rect(last_x,last_y,px-last_x,py-last_y));
                  }
                }
              }
              else if ( ccurve.curve_.isType( DCurve.T_POINTS ))
              {
//TBD:choix de 6x6 pour la dimension des points
                rp.setRect( px-3, py-3, 6, 6 );
                if( d_.drawing() )
                {
                  d_.gc_.setForeground( getBlack() );
                  d_.gc_.setBackground( ccurve.style_.back_ );
                  d_.gc_.fillRectangle( rp.x(), rp.y(), 6,6 );
                  d_.gc_.drawRectangle( rp.x(), rp.y(), 6,6 );
                }
                else //locating
                {
                  if( rp.contains( d_.lx_, d_.ly_ ) )
                  {
                     throw new ChartLocation( DLocated.CurvePoint, point, rp );
                  }
                }
              }
              else if ( ccurve.curve_.isType( DCurve.T_BAR ))
              {   
                float bh=last_vy-bar_zero;
                int  ibh=(int)(ky*(bh));
                if (ibh>0)ibh+=1;else if (ibh<0) ibh-=1;
                r_bar.setRect( last_x, y_bar_zero, px-last_x+1, -ibh );
                r_bar.normalize();
/* cf ci-dessus               if(r_bar.h()<0) 
                {
                     //redresser le rectangle sinon contains marche pas
                     r_bar.setRect( r_bar.x(), r_bar.y()+r_bar.height(),
                                    r_bar.width(), -r_bar.height() );
                   }*/
                if( d_.drawing() )
                {
                  d_.gc_.setForeground( getBlack() );
                  d_.gc_.setBackground( ccurve.style_.back_ );
                  d_.gc_.fillRectangle( r_bar.x(), r_bar.y(), r_bar.w(), r_bar.h() );
                  d_.gc_.drawRectangle( r_bar.x(), r_bar.y(), r_bar.w(), r_bar.h() );
                }
                else //locating
                {
                  //on s'arrete ici parce que l'on recouvre tout...
                  if( r_bar.contains( d_.lx_, d_.ly_ ) )
                  {
                    throw new ChartLocation( DLocated.CurveBar, last_item, r_bar );
                  }
                }
              }
            } 
            else
            {
              first=false;
            }
           last_x = px;
           last_y = py;
           last_vy= vy;
         }
         else
         {
error( "Warning:Point without (x,y) coord.");
          }
       }
 
       last_item=item;
     }//for 

      //last histogram bar
//TBD: c'est pas terrible, mais sinon ca veut dire que la derniere
// valeur d'une courbe histo d'a de sens que pour X (termine la valeur precedente).
     if ( !first && (ccurve.curve_.isType( DCurve.T_BAR ) ))
     {
       int px=rg.right(); //on force jusqu'a la fin du graphe
       r_bar.setRect( last_x-1, last_y, px-last_x+1, rg.bottom()-last_y+1 );
       r_bar.normalize();
       if( d_.drawing() )
       {
         d_.gc_.setForeground( getBlack() );
         d_.gc_.setBackground( ccurve.style_.back_);
         d_.gc_.fillRectangle( r_bar.x(), r_bar.y(), r_bar.w(), r_bar.h() );
         d_.gc_.drawRectangle( r_bar.x(), r_bar.y(), r_bar.w(), r_bar.h() );
       }
       else //locating
       {
         if( r_bar.contains( d_.lx_, d_.ly_ ) )
         {
           throw new ChartLocation( DLocated.CurveBar, point, r_bar );
         }
       }
     }

    }//for
  }
  
  private static Color c_black_=null;
  protected Color getBlack()
  {
    if( c_black_==null)
    {
      c_black_=new Color(null,0,0,0);
    }
    return c_black_;
  }
  
  private static Color c_light_gray_=null;
  protected Color getLightGray()
  {
    if( c_light_gray_==null )
    {
      c_light_gray_ = new Color( null, 160,160,160);
    }
    return c_light_gray_;
  }
  
  private static Color c_red_ = null;
  protected Color getRed()
  {
    if( c_red_==null )
    {
      c_red_ = new Color( null, 255,0,0);
    }
    return c_red_;
  }
  
  private static class ForceCoord implements ICoordTransformation
  {
    public float value_ = 0.0f;
    public float apply( float _f ) { return value_; }
  }
  private static ForceCoord force_coord = new ForceCoord();
  
  protected void drawBox3D( View3D _view, Box3D _bb )
  {
    Point2D p[] = new Point2D[8];
    float xn = _bb.getXMin(), xx = _bb.getXMax();
    float yn = _bb.getYMin(), yx = _bb.getYMax();
    float zn = _bb.getZMin(), zx = _bb.getZMax();
    p[0] = _view.projection( xn,yn,zn, new Point2D() );
    p[1] = _view.projection( xn,yn,zx, new Point2D() );
    p[2] = _view.projection( xx,yn,zx, new Point2D() );
    p[3] = _view.projection( xx,yn,zn, new Point2D() );
    p[4] = _view.projection( xn,yx,zx, new Point2D() );
    p[5] = _view.projection( xx,yx,zx, new Point2D() );
    p[6] = _view.projection( xx,yx,zn, new Point2D() );
    p[7] = _view.projection( xn,yx,zn, new Point2D() );

    d_.gc_.drawLine( (int)p[0].getX(),(int)p[0].getY(), (int)p[1].getX(),(int)p[1].getY());
    d_.gc_.drawLine( (int)p[0].getX(),(int)p[0].getY(), (int)p[3].getX(),(int)p[3].getY());
    d_.gc_.drawLine( (int)p[0].getX(),(int)p[0].getY(), (int)p[7].getX(),(int)p[7].getY());
    d_.gc_.drawLine( (int)p[2].getX(),(int)p[2].getY(), (int)p[1].getX(),(int)p[1].getY());
    d_.gc_.drawLine( (int)p[2].getX(),(int)p[2].getY(), (int)p[3].getX(),(int)p[3].getY());
    d_.gc_.drawLine( (int)p[2].getX(),(int)p[2].getY(), (int)p[5].getX(),(int)p[5].getY());
    d_.gc_.drawLine( (int)p[4].getX(),(int)p[4].getY(), (int)p[1].getX(),(int)p[1].getY());
    d_.gc_.drawLine( (int)p[4].getX(),(int)p[4].getY(), (int)p[5].getX(),(int)p[5].getY());
    d_.gc_.drawLine( (int)p[4].getX(),(int)p[4].getY(), (int)p[7].getX(),(int)p[7].getY());
    d_.gc_.drawLine( (int)p[6].getX(),(int)p[6].getY(), (int)p[3].getX(),(int)p[3].getY());
    d_.gc_.drawLine( (int)p[6].getX(),(int)p[6].getY(), (int)p[5].getX(),(int)p[5].getY());
    d_.gc_.drawLine( (int)p[6].getX(),(int)p[6].getY(), (int)p[7].getX(),(int)p[7].getY());
  }
  
  //identifiers for axis X,Y,Z...
  protected static final int AxisX = 1;
  protected static final int AxisY = 2;
  protected static final int AxisZ = 3;
  
  protected String axis3DName( DAxis _axis )
  {
    String title = _axis.getTitle();

    if( title != null )
    {
      title = title + _axis.getUnit();
    } else {
      title = _axis.getUnit();
    }
  
    return title ;
  }
  
  protected void renderAnAxis3D( 
      View3D _view, Box3D _bb, DAxis _axis, DChartStyle _axis_style, int _axis_type
    )
    throws ChartLocation
  {
    float x_proj = _bb.getXMin(), x_proj2 = _bb.getXMax(),
          y_proj = _bb.getYMin(), y_proj2 = _bb.getYMax(),
          z_proj = _bb.getZMin(), z_proj2 = _bb.getZMax();

    float coord_min, coord_max;
    switch( _axis_type )
    {
      case AxisX: coord_min = _bb.getXMin(); coord_max = _bb.getXMax(); break;
      case AxisY: coord_min = _bb.getYMin(); coord_max = _bb.getYMax(); break;
      case AxisZ: coord_min = _bb.getZMin(); coord_max = _bb.getZMax(); break;
      default: return ;
    }

    Point2D p0 = new Point2D(), 
            p1 = new Point2D(),
            p2 = new Point2D();

    /* 
     * Lignes des reperes
     */
    if ( d_.drawing() && _axis.getProperties().hasValue( DAxis.P_STEP_LINE ) )
    {
      float coord_step = _axis.getProperties().get( DAxis.P_STEP_LINE, 1.0f );
      if ( coord_step != 0.0f )
      {
        float coord;
        float coord_start = (float)( coord_step*Math.ceil(coord_min/coord_step) );

        d_.gc_.setForeground( getLightGray() );

                
        //ca a l'air plus complique, mais comme ca on fait le switch une
        //seule fois par for...
        switch( _axis_type )
        {
          case AxisX: for( coord = coord_start; coord < coord_max; coord += coord_step )
                      {
                        _view.projection( coord, y_proj2, z_proj, p0 );
                        _view.projection( coord, y_proj, z_proj, p1 );
                        _view.projection( coord, y_proj, z_proj2, p2 );
          
                        d_.gc_.drawLine( (int)p0.getX(), (int)p0.getY(), (int)p1.getX(), (int)p1.getY() );
                        d_.gc_.drawLine( (int)p2.getX(), (int)p2.getY(), (int)p1.getX(), (int)p1.getY() );
                      }
                      break;
          case AxisY: for( coord = coord_start; coord < coord_max; coord += coord_step )
                      {
                        _view.projection( x_proj2, coord, z_proj, p0 );
                        _view.projection( x_proj,  coord, z_proj, p1 );
                        _view.projection( x_proj,  coord, z_proj2, p2 );
          
                        d_.gc_.drawLine( (int)p0.getX(), (int)p0.getY(), (int)p1.getX(), (int)p1.getY() );
                        d_.gc_.drawLine( (int)p2.getX(), (int)p2.getY(), (int)p1.getX(), (int)p1.getY() );
                      }
                      break;
          case AxisZ: for( coord = coord_start; coord < coord_max; coord += coord_step )
                      {
                        _view.projection( x_proj2, y_proj,  coord, p0 );
                        _view.projection( x_proj,  y_proj,  coord, p1 );
                        _view.projection( x_proj,  y_proj2, coord, p2 );
          
                        d_.gc_.drawLine( (int)p0.getX(), (int)p0.getY(), (int)p1.getX(), (int)p1.getY() );
                        d_.gc_.drawLine( (int)p2.getX(), (int)p2.getY(), (int)p1.getX(), (int)p1.getY() );
                      }
                      break;
        }
      }
    }

    /*
     * Graduation (texte) de l'axe 
     */
    String axis_title = axis3DName( _axis );
    boolean have_step_unit = false;
    
    if ( _axis.getProperties().hasValue( DAxis.P_STEP_UNIT )
      || !DrawUtil.isEmpty( axis_title )
       )
    {
      have_step_unit = true;
      float d_angle = 0.0f;
      switch( _axis_type )
      {
        case AxisX:
            _view.projection( _bb.getXMin(),_bb.getYMin(),_bb.getZMin(), p0 );
            _view.projection( _bb.getXMax(),_bb.getYMin(),_bb.getZMin(), p1 );
            d_angle = 180;
          break;
        case AxisY:
             _view.projection( _bb.getXMin(),_bb.getYMin(),_bb.getZMin(), p0 );
             _view.projection( _bb.getXMin(),_bb.getYMax(),_bb.getZMin(), p1 );
             d_angle = 0;
           break;
        case AxisZ:
             _view.projection( _bb.getXMin(),_bb.getYMin(),_bb.getZMin(), p0 );
             _view.projection( _bb.getXMin(),_bb.getYMin(),_bb.getZMax(), p1 );
             d_angle = 180;
           break;
      }
       
      float angle = (float)Math.atan2( p1.getY()-p0.getY(), p1.getX()-p0.getX() ); //rad.
      angle = (float)(180*angle/Math.PI + d_angle); //degree

//SWT en est loin de ca...      QWMatrix m ;

      if( d_.drawing() )
      {
        d_.gc_.setForeground(  _axis_style.fore_ );
        d_.gc_.setFont( _axis_style.font_ );
      }
DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);      
      if ( _axis.getProperties().hasValue( DAxis.P_STEP_UNIT ) )
      {
        float coord_step = _axis.getProperties().get( DAxis.P_STEP_UNIT, +1.0f );
        if (  coord_step != 0.0f )
        {
          float coord;
//          float *cx, *cy, *cz;
          //optim: fait le switch une seule fois par for...
          /*
          switch( _axis_type )
          { 
            case AxisX:
          cx = &coord;  cy = &y_proj;  cz = &z_proj2 ; break;
            case AxisY:
          cx = &x_proj; cy = &coord;   cz = &z_proj2 ; break;
            case AxisZ:
          cx = &x_proj; cy = &y_proj2; cz = &coord ; break;
          }*/
 Rect rtxt = new Rect();
 int fh = d_.gc_.getFontMetrics().getHeight();
          for( coord = (float)(coord_step*Math.ceil(coord_min/coord_step));
               coord < coord_max;
               coord += coord_step )
          {
            switch( _axis_type )
            { 
              case AxisX: _view.projection(  coord,  y_proj,  z_proj2, p2 ); break;
              case AxisY: _view.projection(  x_proj, coord,   z_proj2, p2 ); break;
              case AxisZ: _view.projection(  x_proj, y_proj2, coord,   p2 ); break;
            }
            
//SWT?? tu fais quoi avec ca ???            
   /*         m.reset();
            m.translate( px[2], py[2] );
            m.rotate( angle );*/

            String txt = df.format( coord );
            int wt = d_.gc_.textExtent( txt ).x;
            rtxt.setRect( -wt/2, -3-fh, wt, fh );
rtxt.moveBy( (int)p2.getX(), (int)p2.getY() );            

            if( d_.drawing() )
            {
//SWT???
/*              
                d_->P_->save();
                d_->P_->setWorldMatrix( m );
                d_->P_->drawText( rtxt, Qt::AlignCenter, txt );
                d_->P_->restore();*/
                DrawUtil.drawText( d_.gc_, rtxt, DrawUtil.CENTER, txt );
              }
              else //locating
              {
/*SWT*declare forfait*                QRegion reg( rtxt );
                reg = m * reg;
                if ( reg.contains( _pt_to_locate ) )
                {
                  DChartLocation::Location location;
                  switch( _axis_type )
                  {
                    case AxisX:
                      location = DChartLocation::AxisXStepUnitText; break;
                    case AxisY:
                      location = DChartLocation::AxisYStepUnitText; break;
                    case AxisZ:
                      location = DChartLocation::AxisZStepUnitText; break;
                  }
                  d_->location_= DChartLocation( location, _axis, reg.boundingRect(), txt );
                  return ;
                }
                */
              }
            }
          }
        }
      }

      /*
       * Axis title.
       */
      if ( !DrawUtil.isEmpty( axis_title ) )
      {
        String location=null;
 Rect rtxt = new Rect();
 int fh = d_.gc_.getFontMetrics().getHeight();       
        switch( _axis_type )
        {
          case AxisX:
                _view.projection( _bb.getXMin(),y_proj,z_proj2, p0 );
                _view.projection( _bb.getXMax(),y_proj,z_proj2, p1 );
               location = DLocated.InAxisXTitle;
            break;
          case AxisY:
                _view.projection( x_proj,_bb.getYMin(),z_proj2, p0 );
                _view.projection( x_proj,_bb.getYMax(),z_proj2, p1 );
                location = DLocated.InAxisYTitle; 
            break;

          case AxisZ:
                 _view.projection( x_proj,y_proj2,_bb.getZMin(), p0 );
                 _view.projection( x_proj,y_proj2,_bb.getZMax(), p1 );
                 location = DLocated.InAxisZTitle; 
            break;
        }

        float center_x=(p0.getX()+p1.getX())/2,
              center_y=(p0.getY()+p1.getY())/2;
/*SWT*? toujours pas ?
          m.reset();
          m.translate( center_x, center_y );
          m.rotate( angle );
*/
          int wt = d_.gc_.textExtent( axis_title ).x;
          rtxt.setRect( -wt/2, -6-2*fh, wt, fh );

//TODO: pas de rotation ... donc "a plat"
//-X-8 au dessus des unites et marge...
rtxt.moveCenter( (int)center_x, (int)center_y - (have_step_unit?(fh+4):0) -4 );

          if( d_.drawing() )
          {
/*            
            d_->P_->save();
            d_->P_->setWorldMatrix( m );
            d_->P_->drawText( rtxt, Qt::AlignCenter, axis_title );
      d_->P_->restore();
*/
  DrawUtil.drawText( d_.gc_, rtxt, DrawUtil.CENTER, axis_title );      
          }
          else //locating
          {
  if( rtxt.contains( d_.lx_, d_.ly_ ))
  {
    throw new ChartLocation( location, _axis, rtxt ); 
  }
/*SWT...            
            QRegion reg( rtxt );
            reg = m * reg ;
            if( reg.contains( _pt_to_locate ) )
            {
              d_->location_= DChartLocation( location, _axis, reg.boundingRect(),
                                    axis_title );
              return;
            }
*/            
          }
        }
      }
    
   protected static final int XProj=1;
   protected static final int YProj=2;
   protected static final int ZProj=3;
   protected static final int Curve=4;
   protected static final int Finished=5;

  
  /** render XYZ curves */
  void renderXYZ( Rect _RG ) throws ChartLocation
  {
    //decorations means axis title/units...
    boolean display_decorations = true;
    boolean is_surface = d_.G_.isGraphicType( DGraphic.T_SURFACE );

    DAxis x_axis = null;
    DAxis y_axis = null;
    DAxis z_axis = null;
    int   axis_missed = 3;
    for( IDItem item = d_.G_.getFirstChild(); item!=null; item=item.getNext() )
    {
      if( item instanceof DAxis )
      {
        DAxis axis = (DAxis)item;
        if( x_axis==null && "x".equals( axis.getName() ))
        {
          x_axis = axis; axis_missed--;
        }
        else if ( y_axis==null & "y".equals( axis.getName() ))
        {
          y_axis = axis; axis_missed--;
        }
        else if ( z_axis==null & "z".equals( axis.getName() ))
        {
          z_axis = axis; axis_missed--;
        }
        if( axis_missed==0 ) break;
      }
    }

    if ( axis_missed > 0 )
    {
      String txt;
      if(x_axis==null) {
        txt = "Error: missing 'x' axis";
      } else 
      if(y_axis==null) {
        txt = "Error: missing 'y' axis";
      } else {
        txt = "Error: missing 'z' axis";
      }
      error( txt );
      return ;
    }

    DChartStyle x_axis_style = styleOf( x_axis );
    DChartStyle y_axis_style = styleOf( y_axis );
    DChartStyle z_axis_style = styleOf( z_axis );

    //garder un peu d'espace pour les texte sur les axes 
    //c'est completement choisi au hazard (ou a peu pres!).
    int max_font_h  = 0 ;
    if ( display_decorations )
    {
      d_.gc_.setFont( x_axis_style.font_ );
      max_font_h = d_.gc_.getFontMetrics().getHeight();
      d_.gc_.setFont( y_axis_style.font_ );
      int hf = d_.gc_.getFontMetrics().getHeight();
      if ( max_font_h < hf ) max_font_h = hf;
      d_.gc_.setFont( z_axis_style.font_ );
      hf = d_.gc_.getFontMetrics().getHeight();
      if ( max_font_h < hf ) max_font_h = hf;
      //prevoir le titre + les unites sur l'axe
      max_font_h = 2*max_font_h +6;
    }

    Rect gR = new Rect( _RG.left()+max_font_h, _RG.top()+max_font_h,
                        _RG.w()-2*max_font_h,  _RG.h()-max_font_h );

    //boite englobante des courbes a afficher.
    Box3D bb = new Box3D( x_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
                          x_axis.getProperties().get( DAxis.P_MAX, +1.0f ),
                          y_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
                          y_axis.getProperties().get( DAxis.P_MAX, +1.0f ),
                          z_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
                          z_axis.getProperties().get( DAxis.P_MAX, +1.0f ) );

    float phi   = d_.G_.getProperties().get( DGraphic.P_XYZ_PHI,   0.3f );
    float theta = d_.G_.getProperties().get( DGraphic.P_XYZ_THETA, 0.2f );

    View3D view = new View3D( phi, theta, 0, 1.0f, bb );

    //mise a l'echelle de la vue 3D en fonction du rectangle de dessin
    view.scaleTo( gR.x(), gR.y(), gR.w(), gR.h() );

    int min_w_h = gR.w() ;
    if (min_w_h > gR.h() ) min_w_h = gR.h();

    //bounding Box
    if ( display_decorations )
    {
      if( d_.drawing() )
      {
        d_.gc_.setForeground( getLightGray() );
        drawBox3D( view, bb );
      }

      // render des axes...
      renderAnAxis3D( view, bb, x_axis, x_axis_style, AxisX );
//      if( d_->located() ) return ;
      renderAnAxis3D( view, bb, y_axis, y_axis_style, AxisY );
//      if( d_->located() ) return ;
      renderAnAxis3D( view, bb, z_axis, z_axis_style, AxisZ );
//      if( d_->located() ) return ;
    }

    //pour projection des courbes.
    float x_proj = bb.getXMin(), y_proj = bb.getYMin(), z_proj = bb.getZMin(); 
    CurveStyle ccurve =null;
    CurveStyle p_ccurve=null; //courbe precedente (connection plan)
    IDItem         item=null;
    IDItem       p_item=null;
    float       last_x=0, last_y=0;

    //memo h,s,v du fond pour les couleurs de projection des courbes...
    DChartStyle bgsty = styleOf( d_.G_ );
    Color background_color = bgsty.back_;
    HSV bhsv = new HSV( background_color );

    // check property status for curves projections
    boolean do_xproj = d_.G_.getProperties().get( DGraphic.P_XYZ_XPROJ, true);
    boolean do_yproj = d_.G_.getProperties().get( DGraphic.P_XYZ_YPROJ, true);
    boolean do_zproj = d_.G_.getProperties().get( DGraphic.P_XYZ_ZPROJ, true);

    Point2D p = new Point2D();
    
    for( int ic=0; ic<d_.curves_.size(); ++ic )
    {
      ccurve = (CurveStyle)d_.curves_.get(ic);
      //la courbe est dessinee 4 fois: 3 projections + la courbe
      int what=XProj;

      if( display_decorations
//       && d_.G_.display3DCurveProjection() 
       && d_.drawing() /*en locating on ne fait pas les projections*/
        )
      {
        //courbes projetees: couleur de la courbe en plus clair ...
        Color proj_color = DrawUtil.lighter( ccurve.style_.back_, 1.75f );
        HSV hsv = new HSV( proj_color );
        //sauf si ecart < 20, me demandez pas pourquoi, c'est une heuristique
        //finement elaboree...
        int ecart=(int)(Math.abs(bhsv.getV()-hsv.getV())+Math.abs(bhsv.getS()-hsv.getS()))>>1;
        if(ecart < 20 )
        {
          //couleur de projection trop proche du fond, on bascule
          //en couleur plus fonce..
          proj_color = DrawUtil.lighter( ccurve.style_.back_, 0.75f );
        }
        d_.gc_.setForeground( proj_color );
        what = XProj ;
      } else {
        what = Curve ;
      }
  for( ;what<Finished; )
  {
 
      //reset des projecitons
      view.resetCoordTransformation();
      switch( what )
      {
        case XProj: { if(!do_xproj) { what=YProj; continue; }
                      force_coord.value_ = x_proj ;
                      view.setXCoordTransformation( force_coord );
                      what = YProj;
                      break; }
        case YProj: { if(!do_yproj) { what=ZProj; continue; }
                      force_coord.value_ = y_proj ;
                      view.setYCoordTransformation( force_coord );
                      what = ZProj;
                      break; }
        case ZProj: { if(!do_zproj) { what=Curve; continue; }
                      force_coord.value_ = z_proj ;
                      view.setZCoordTransformation( force_coord );
                      what = Curve;
                      break; }
        case Curve: { if(d_.drawing()) d_.gc_.setForeground( ccurve.style_.back_ ); 
          what = Finished;
                      break; }
        case Finished:break;
      }//switch

      if ( is_surface )
      {
        p_item = p_ccurve!=null ? p_ccurve.curve_.getFirstChild(): null;
      }

      boolean first=true; 
      for( item = ccurve.curve_.getFirstChild(); item!=null; item=item.getNext() )
      {
        DPoint point   = (DPoint)item;
        DPoint p_point = p_item!=null ? (DPoint)p_item: null; 
        if( point!=null )
        {
          float v_x=0, v_y=0, v_z=0 ;
          DCoord c=null;
          boolean ok=true;
          if( (c=point.getCoord( x_axis))!=null ) { v_x=c.getValue(); } else { ok=false; }
          if( (c=point.getCoord( y_axis))!=null ) { v_y=c.getValue(); } else { ok=false; }
          if( (c=point.getCoord( z_axis))!=null ) { v_z=c.getValue(); } else { ok=false; }
  
          if( ok )
          {
            float p_x,p_y;
            view.projection( v_x,v_y,v_z, p );
            {
              p_x = p.getX();
              p_y = p.getY();
              if ( first )
              {
                first=false;
              }
              else
              {
                if(d_.drawing())
                {
                  d_.gc_.drawLine( (int)last_x, (int)last_y, (int)p_x, (int)p_y );
                }
                else //locating
                {
                  if( what==Finished ) //pas les projections
                  {
                    if( DrawUtil.segmentContains( (int)last_x, (int)last_y, (int)p_x, (int)p_y, d_.lx_,d_.ly_))
                    {
                      throw new ChartLocation( DLocated.CurveLine, item,
                             new Rect((int)p_x,(int)p_y, (int)(last_x-p_x),(int)(last_y-p_y) ));
                    }
                  }
                }
              }
            }
            last_x=p_x;
            last_y=p_y;

            //lien avec la courbe precedente (en mode surface)
            if ( p_point!=null )
            {
              float p_v_x=0,p_v_y=0,p_v_z=0;
              ok = true;
              ok = ok&&((c=p_point.getCoord( x_axis ))!=null); if(ok) {p_v_x=c.getValue(); }
              ok = ok&&((c=p_point.getCoord( y_axis ))!=null); if(ok) {p_v_y=c.getValue(); }
              ok = ok&&((c=p_point.getCoord( z_axis ))!=null); if(ok) {p_v_z=c.getValue(); }
              if( ok )
              {
                view.projection( p_v_x,p_v_y,p_v_z, p);
                {
                  float p_p_x = p.getX(),  p_p_y=p.getY();
                  if( d_.drawing() )
                  {
                    d_.gc_.drawLine( (int)p_x,(int)p_y, (int)p_p_x,(int)p_p_y);
                  } 
                  else // locating
                  {
                    if( what==Finished )
                    {
 //TBD en 3D(2D) plusieurs courbes peuvent se superposer...
                      if( DrawUtil.segmentContains( (int)p_x,(int)p_y, (int)p_p_x,(int)p_p_y, d_.lx_,d_.ly_ ))
                      {
                        throw new ChartLocation( DLocated.CurveLine, item, 
                            new Rect( (int)p_x,(int)p_y,(int)(p_p_x-p_x),(int)(p_p_y-p_y)));
                      } 
                    }
                  }
                }
              }
            }
          }
          if ( p_item!=null ) p_item = p_item.getNext();
        }
      }//for

  }//for:what

      p_ccurve=ccurve;
    }//for
  }

  
/*
class TESTRT_EXPORT DChartPainter
{
 protected:
  void renderHistor ( const QRect & );
  void renderBargraph ( const QRect & );

  QString valueWithUnit( float _value, DAxis *_axis ) const ;

*/
}
//Source file: C:\\Program Files\\eclipse21\\workspace\\apidoc\\com\\rational\\apidoc\\drivers\\builder\\TBuilder.java

//Source file: C:\\Program Files\\eclipse21\\workspace\\apidoc\\com\\rational\\apidoc\\drivers\\ui\\TBuilder.java

package com.rational.apidoc.drivers.builder;

import com.rational.apidoc.core.*;
import com.rational.apidoc.drawutil.DrawUtil;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.Device;
import java.util.Enumeration;
import java.util.Hashtable;
import com.rational.apidoc.parsers.DParser;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.FontMetrics;
import org.eclipse.swt.graphics.Rectangle;


public class TBuilder extends DParser 
{
   private Image brokenImage = null;
   private int alignment = DParagraph.LEFT;
   private int bottomMargin = 0;
   private int cellSpacing = 0;
   private TCell currentCell = null;
   private int currentHLine = 0;
   private int currentNLine = 0;
   private int currentNPage = 0;
   private TCell currentPage = null;
   private int currentPosition = 1;
   private int currentWLine = 0;
   private int currentXPos = 0;
   private int currentYPos = 0;
   private Font defaultFont = null;
   private Font previousFont = null;
   private Device device = null;
   private boolean drawPopup = true;
   private TCell firstCellLine = null;
   private boolean flatPopup = false;
   private Hashtable fonts = new Hashtable();
   private TCell footerCell = null;
   private GC gc = null;
   private TCell headerCell = null;
   private Hashtable imagePool = new Hashtable();
   private TCell lastCellLine = null;
   private int leftMargin = 0;
   private int lineSpacing = 0;
   private int listOffsetSpacing = 0;
   private int maxWidth = 100;
   private int minWidth = 20;
   private int pageBegin = 0;
   private int pageEnd = 0;
   private int pageHeight = 0;
   private int prefixSpacing = 0;
   private int rightMargin = 0;
   private int titleSpacing = 0;
   private int topMargin = 0;
   private boolean wordWrap = true;
   private int xoffset = 0;
   private DFooter footer = null;
   private DHeader header = null;
   private TCell firstCell = null;
   static int level = 0;
   
   private int currentNColumn = 0;
   private int currentTotalColumn = 0;
   private int currentNRow = 0;
   private DRow currentRow = null;
   private DTable currentTable = null;
   private int maxCellHeight = 0;
   
   
   private class CellInfo {
		  TCell cell = null;
		  int rowspan = 0;
   }
   
   private CellInfo[] currentRowSpan = null;
   
   private class Range 
   {
		public int start = 0;
		public int end = 0;
      
		/**
		 * @roseuid 3F0168C202A7
		 */
		public Range() 
		{
       
		}
   }
   
   /**
    * @param gc
    * @param d
    * @roseuid 3EF82BBF01A5
    */
   public TBuilder(GC gc, Device d) 
   {
      this.gc = gc;
	  this.device = d; 
            
      if (gc!=null)    
         defaultFont = gc.getFont();
      else
	     defaultFont = new Font(d, "courier", 12, SWT.NORMAL);   
   }
   
   
   private TBuilder _newBuilder(int size)
	  {
		 TBuilder b = new TBuilder(getGc(), getDevice());
		 b.setFlatPopup(getFlatPopup());
		 b.setCurrentPosition(getCurrentPosition());
		 b.setWordWrap(getWordWrap());
		 b.setMaxWidth(size);
		 b.setImagePool(getImagePool());
		 b.setFontPool(getImagePool());
		 b.setImageSearchPath(getImageSearchPath());
		 b.setCurrentNPage(getCurrentNPage());
		 return b;
	  }
   
	  private void _disposeBuilder(TBuilder b)
	  {
		 b.setImagePool(null);
		 b.setImageSearchPath(null);
		 b.setFontPool(null);
	  }
   
	  /**
	   * @param i
	   * @roseuid 3F004C9B00F9
	   */
	  private void _addListItemPrefix(IDItem i) 
	  {  
		 if (!(i.getParent() instanceof DList)) return;
   	
		 addListEntry();
		 String rval = getListStringPrefix(i.getParent());
      
		 if (rval != null)
		 {
			Point p = getGc().textExtent(rval);
			addXSpacing(-p.x);
			TCell c = addCell(p.x, p.y, i, rval.length());
			c.setText(rval);
			addXSpacing(getTitleSpacing());
		 }    
	  }
   
	  /**
	   * @param i
	   * @param text
	   * @param r
	   * @roseuid 3F0168EB0089
	   */
	  private void _buildPourTextBuffer(IDItem i, String text, Range r) 
	  {
		 if (text.length() == 0) return;
      
		 int w, wprev;
		 int count = r.end - r.start;
		 TCell item = null;
      
		 FontMetrics fm = getGc().getFontMetrics();
		 int lineSpacing = fm.getAscent()+fm.getDescent();
      
		 // rien a traiter
		 if (r.start >= r.end)
		 {
		   if (text.getBytes()[r.start] == '\n')
		   {
			   item = addCell(0, lineSpacing, i, 0);
		   }
		   else
		   {	
			   addYSpacing(lineSpacing);	
		   }
		   return;
		 }
      
		 String qs = null;
		 if (r.end == text.length())
			qs = text;
		 else
			qs = text.substring(r.start, r.end);
      
		 GC gc = getGc();
		 Point p = gc.textExtent(qs);
      
		 // Si on est en non WORD WRAP ou que le buffer rentre
		 // dans la ligne alors on ajoute ce dernier
		 // sur la ligne courante
		 if (!getWordWrap() 
			 || (getCurrentWLine() + p.x <= getMaxWidth()))
		 {
		   item = addCell(p.x, p.y, i, count);
		   item.setText(text, r.start, count); 
		   return;
		 }
         
		 // sinon on va couper le buffer sur un separateur de mot afin
		 // de la faire rentrer sur plusieurs lignes
		 wprev = 0;
      
		 Range r2 = new Range();
		 r2.start = r.start; // position dernier caractere de la sous chaine
      
      
		 while (r2.start < r.end)
		 {
			r2.end = 0;   // taille en pixel de la sous chaine
      	 
			_buildGetNextWord(qs, 
							  getMaxWidth() - getCurrentWLine(),
							  r2);
              
			count = r2.start - r.start;
			TCell cell = addCell(r2.end, lineSpacing, i, count);
			cell.setText(text, r.start, count);
			newLine();
         
			r.start = r2.start;
		 }    
	  }
   
	  /**
	   * @param b
	   * @param i
	   * @param text
	   * @roseuid 3F003ACA012D
	   */
	  private void _buildTextBuffer(IDItem i, String text) 
	  {
		 if (text.length()==0) return;
       
		 Range r = new Range();
		 r.start = 0;
		 r.end = text.indexOf("\n");
      
		 // tant que l'on a pas atteint la fin de la ligne
		 while (r.end != -1)
		 {
		   _buildPourTextBuffer(i, text, r);
      	
		   // ligne non adaptative car terminee par \n
		   // on ajoute donc une nouvelle ligne
		   newLine();
      	
		   r.start = r.end + 1;
      	
		   // recherche du \n suivant 
		   r.end = text.indexOf("\n", r.start);
		 }
      
		 if (r.start < text.length())
		 {
			r.end = text.length();
		   _buildPourTextBuffer(i, text, r);
		 }    
	  }
   
	  /**
	   * @param c
	   * @return int
	   * @roseuid 3F003121028F
	   */
	  private int _getColumnSizeForCell(DCell c) 
	  {
		  int size = 0;
		  if (currentTable.getAutoResize())  // toute la largeur du document 
		  {
				 float ws = (float)getMaxWidth() * currentTable.getWidthScale();
				 for (int i=0; i<c.getColSpan()+1; i = i + 1)
				 {
				   float factor = 1.0f;
				   int s = 0;
				   if ((currentTable.getColumnWidth() != null)
						&& (currentTable.getColumnWidth().length > currentNColumn + i))
				   {
					   factor = currentTable.getColumnWidth()[currentNColumn + i];
					   s = (int)(ws * factor);
				   }
				   else
				   {
					   s = (int)(ws / currentTotalColumn);
				   }
			  	
				   if (s < getMinWidth())
				   {
					 s = getMinWidth();
				   }
				   size = size + s;
				 }
		 }
		 else 
		 {
				size = getMaxWidth();
		 }
		 return size;    
	  }
   
	  /**
	   * @param b
	   * @param c
	   * @roseuid 3F0030E403A0
	   */
	  private void _setBuilderMarginForCell(TBuilder b) 
	  {
		  int m = getCellSpacing();
	   
	      b.setLeftMargin(m + currentTable.getBorder());	  
	      b.setTopMargin( m + currentTable.getBorder());   	  
	      b.setBottomMargin(m + currentTable.getBorder());
	      b.setRightMargin(m + currentTable.getBorder());
	  }
   
   
   /**
    * Access method for the alignment property.
    * 
    * @return   the current value of the alignment property
    */
   public int getAlignment() 
   {
      return alignment;    
   }
   
   /**
    * Sets the value of the alignment property.
    * 
    * @param aAlignment the new value of the alignment property
    */
   public void setAlignment(int aAlignment) 
   {
      alignment = aAlignment;    
   }
   
   /**
    * Access method for the bottomMargin property.
    * 
    * @return   the current value of the bottomMargin property
    */
   public int getBottomMargin() 
   {
      return bottomMargin;    
   }
   
   /**
    * Sets the value of the bottomMargin property.
    * 
    * @param aBottomMargin the new value of the bottomMargin property
    */
   public void setBottomMargin(int aBottomMargin) 
   {
      bottomMargin = aBottomMargin;    
   }
   
   /**
    * Access method for the cellSpacing property.
    * 
    * @return   the current value of the cellSpacing property
    */
   public int getCellSpacing() 
   {
      return cellSpacing;    
   }
   
   /**
    * Sets the value of the cellSpacing property.
    * 
    * @param aCellSpacing the new value of the cellSpacing property
    */
   public void setCellSpacing(int aCellSpacing) 
   {
      cellSpacing = aCellSpacing;    
   }
   
   /**
    * Access method for the currentCell property.
    * 
    * @return   the current value of the currentCell property
    */
   public TCell getCurrentCell() 
   {
      return currentCell;    
   }
   
   /**
    * Access method for the currentHLine property.
    * 
    * @return   the current value of the currentHLine property
    */
   public int getCurrentHLine() 
   {
      return currentHLine;    
   }
   
   /**
    * Access method for the currentNPage property.
    * 
    * @return   the current value of the currentNPage property
    */
   public int getCurrentNPage() 
   {
      return currentNPage;    
   }
   
   /**
    * Access method for the currentNPage property.
    * 
	* @param i
	*/
   public void setCurrentNPage(int i) {
	   currentNPage = i;
   }
   
   /**
    * Access method for the currentPage property.
    * 
    * @return   the current value of the currentPage property
    */
   public TCell getCurrentPage() 
   {
      return currentPage;    
   }
   
   /**
    * Access method for the currentPosition property.
    * 
    * @return   the current value of the currentPosition property
    */
   public int getCurrentPosition() 
   {
      return currentPosition;    
   }
   
   /**
    * Sets the value of the currentPosition property.
    * 
    * @param aCurrentPosition the new value of the currentPosition property
    */
   public void setCurrentPosition(int aCurrentPosition) 
   {
      currentPosition = aCurrentPosition;    
   }
   
   /**
    * Access method for the currentWLine property.
    * 
    * @return   the current value of the currentWLine property
    */
   public int getCurrentWLine() 
   {
      return currentWLine;    
   }
   
   /**
    * Access method for the currentXPos property.
    * 
    * @return   the current value of the currentXPos property
    */
   public int getCurrentXPos() 
   {
      return currentXPos;    
   }
   
   /**
    * Access method for the currentYPos property.
    * 
    * @return   the current value of the currentYPos property
    */
   public int getCurrentYPos() 
   {
      return currentYPos;    
   }
   
   /**
    * Access method for the defaultFont property.
    * 
    * @return   the current value of the defaultFont property
    */
   public Font getDefaultFont() 
   {
      return defaultFont;    
   }
   
   /**
    * Sets the value of the defaultFont property.
    * 
    * @param aDefaultFont the new value of the defaultFont property
    */
   public void setDefaultFont(Font aDefaultFont) 
   {
      defaultFont = aDefaultFont;    
   }
   
   /**
    * Access method for the device property.
    * 
    * @return   the current value of the device property
    */
   public Device getDevice() 
   {
      return device;    
   }
   
   /**
    * Determines if the drawPopup property is true.
    * 
    * @return   <code>true<code> if the drawPopup property is true
    */
   public boolean getDrawPopup() 
   {
      return drawPopup;    
   }
   
   /**
    * Sets the value of the drawPopup property.
    * 
    * @param aDrawPopup the new value of the drawPopup property
    */
   public void setDrawPopup(boolean aDrawPopup) 
   {
      drawPopup = aDrawPopup;    
   }
   
   /**
    * Access method for the firstCellLine property.
    * 
    * @return   the current value of the firstCellLine property
    */
   public TCell getFirstCellLine() 
   {
      return firstCellLine;    
   }
   
   /**
    * Determines if the flatPopup property is true.
    * 
    * @return   <code>true<code> if the flatPopup property is true
    */
   public boolean getFlatPopup() 
   {
      return flatPopup;    
   }
   
   /**
    * Sets the value of the flatPopup property.
    * 
    * @param aFlatPopup the new value of the flatPopup property
    */
   public void setFlatPopup(boolean aFlatPopup) 
   {
      flatPopup = aFlatPopup;    
   }
   
   /**
    * Access method for the footerCell property.
    * 
    * @return   the current value of the footerCell property
    */
   public TCell getFooterCell() 
   {
      return footerCell;    
   }
   
   /**
    * Sets the value of the footerCell property.
    * 
    * @param aFooterCell the new value of the footerCell property
    */
   public void setFooterCell(TCell aFooterCell) 
   {
      footerCell = aFooterCell;    
   }
   
   /**
    * Access method for the gc property.
    * 
    * @return   the current value of the gc property
    */
   public GC getGc() 
   {
      return gc;    
   }
   
   /**
    * Access method for the headerCell property.
    * 
    * @return   the current value of the headerCell property
    */
   public TCell getHeaderCell() 
   {
      return headerCell;    
   }
   
   /**
    * Sets the value of the headerCell property.
    * 
    * @param aHeaderCell the new value of the headerCell property
    */
   public void setHeaderCell(TCell aHeaderCell) 
   {
      headerCell = aHeaderCell;    
   }
   
   /**
    * Access method for the imagePool property.
    * 
    * @return   the current value of the imagePool property
    */
   public Hashtable getImagePool() 
   {
      return imagePool;    
   }
   
   /**
    * Sets the value of the imagePool property.
    * 
    * @param aImagePool the new value of the imagePool property
    */
   public void setImagePool(Hashtable aImagePool) 
   {
      imagePool = aImagePool;    
   }
   
   /**
	* Access method for the imagePool property.
	* 
	* @return   the current value of the imagePool property
	*/
   public Hashtable getFontPool() 
   {
		 return fonts;    
   } 
   
   /**
	* Sets the value of the imagePool property.
	* 
	* @param aImagePool the new value of the imagePool property
	*/
   public void setFontPool(Hashtable aFontPool) 
   {
		 fonts = aFontPool;    
   }
   
   /**
    * Access method for the leftMargin property.
    * 
    * @return   the current value of the leftMargin property
    */
   public int getLeftMargin() 
   {
      return leftMargin;    
   }
   
   /**
    * Sets the value of the leftMargin property.
    * 
    * @param aLeftMargin the new value of the leftMargin property
    */
   public void setLeftMargin(int aLeftMargin) 
   {
      leftMargin = aLeftMargin;    
   }
   
   /**
    * Access method for the lineSpacing property.
    * 
    * @return   the current value of the lineSpacing property
    */
   public int getLineSpacing() 
   {
      return lineSpacing;    
   }
   
   /**
    * Sets the value of the lineSpacing property.
    * 
    * @param aLineSpacing the new value of the lineSpacing property
    */
   public void setLineSpacing(int aLineSpacing) 
   {
      lineSpacing = aLineSpacing;    
   }
   
   /**
    * Access method for the listOffsetSpacing property.
    * 
    * @return   the current value of the listOffsetSpacing property
    */
   public int getListOffsetSpacing() 
   {
      return listOffsetSpacing;    
   }
   
   /**
    * Sets the value of the listOffsetSpacing property.
    * 
    * @param aListOffsetSpacing the new value of the listOffsetSpacing property
    */
   public void setListOffsetSpacing(int aListOffsetSpacing) 
   {
      listOffsetSpacing = aListOffsetSpacing;    
   }
   
   /**
    * Access method for the maxWidth property.
    * 
    * @return   the current value of the maxWidth property
    */
   public int getMaxWidth() 
   {
      return maxWidth;    
   }
   
   /**
    * Sets the value of the maxWidth property.
    * 
    * @param aMaxWidth the new value of the maxWidth property
    */
   public void setMaxWidth(int aMaxWidth) 
   {
	  maxWidth = Math.max(aMaxWidth, getMinWidth());
   }
   
   /**
    * Access method for the minWidth property.
    * 
    * @return   the current value of the minWidth property
    */
   public int getMinWidth() 
   {
      return minWidth;    
   }
   
   /**
    * Sets the value of the minWidth property.
    * 
    * @param aMinWidth the new value of the minWidth property
    */
   public void setMinWidth(int aMinWidth) 
   {
      minWidth = aMinWidth;    
   }
   
   /**
    * Determines if the modePage property is true.
    * 
    * @return   <code>true<code> if the modePage property is true
    */
   public boolean getModePage() 
   {
      return (getPageHeight()>0);    
   }
   
   /**
    * Access method for the pageBegin property.
    * 
    * @return   the current value of the pageBegin property
    */
   public int getPageBegin() 
   {
      return pageBegin;    
   }
   
   /**
    * Sets the value of the pageBegin property.
    * 
    * @param aPageBegin the new value of the pageBegin property
    */
   public void setPageBegin(int aPageBegin) 
   {
      pageBegin = aPageBegin;    
   }
   
   /**
    * Access method for the pageEnd property.
    * 
    * @return   the current value of the pageEnd property
    */
   public int getPageEnd() 
   {
      return pageEnd;    
   }
   
   /**
    * Sets the value of the pageEnd property.
    * 
    * @param aPageEnd the new value of the pageEnd property
    */
   public void setPageEnd(int aPageEnd) 
   {
      pageEnd = aPageEnd;    
   }
   
   /**
    * Access method for the pageHeight property.
    * 
    * @return   the current value of the pageHeight property
    */
   public int getPageHeight() 
   {
      return pageHeight;    
   }
   
   /**
    * Sets the value of the pageHeight property.
    * 
    * @param aPageHeight the new value of the pageHeight property
    */
   public void setPageHeight(int aPageHeight) 
   {
      pageHeight = aPageHeight;    
   }
   
   /**
    * Access method for the rightMargin property.
    * 
    * @return   the current value of the rightMargin property
    */
   public int getRightMargin() 
   {
      return rightMargin;    
   }
   
   /**
    * Sets the value of the rightMargin property.
    * 
    * @param aRightMargin the new value of the rightMargin property
    */
   public void setRightMargin(int aRightMargin) 
   {
      rightMargin = aRightMargin;    
   }
   
   /**
    * Access method for the titleSpacing property.
    * 
    * @return   the current value of the titleSpacing property
    */
   public int getTitleSpacing() 
   {
      return titleSpacing;    
   }
   
   /**
    * Sets the value of the titleSpacing property.
    * 
    * @param aTitleSpacing the new value of the titleSpacing property
    */
   public void setTitleSpacing(int aTitleSpacing) 
   {
      titleSpacing = aTitleSpacing;    
   }
   
   /**
    * Access method for the topMargin property.
    * 
    * @return   the current value of the topMargin property
    */
   public int getTopMargin() 
   {
      return topMargin;    
   }
   
   /**
    * Sets the value of the topMargin property.
    * 
    * @param aTopMargin the new value of the topMargin property
    */
   public void setTopMargin(int aTopMargin) 
   {
      topMargin = aTopMargin;    
   }
   
   /**
    * Determines if the wordWrap property is true.
    * 
    * @return   <code>true<code> if the wordWrap property is true
    */
   public boolean getWordWrap() 
   {
      return wordWrap;    
   }
   
   /**
    * Sets the value of the wordWrap property.
    * 
    * @param aWordWrap the new value of the wordWrap property
    */
   public void setWordWrap(boolean aWordWrap) 
   {
      wordWrap = aWordWrap;    
   }
   
   /**
    * Access method for the xoffset property.
    * 
    * @return   the current value of the xoffset property
    */
   public int getXoffset() 
   {
      return xoffset;    
   }
   
   /**
    * Sets the value of the xoffset property.
    * 
    * @param aXoffset the new value of the xoffset property
    */
   public void setXoffset(int aXoffset) 
   {
      xoffset = aXoffset;    
   }
   
   /**
    * Access method for the footer property.
    * 
    * @return   the current value of the footer property
    */
   public DFooter getFooter() 
   {
      return footer;    
   }
   
   /**
    * Sets the value of the footer property.
    * 
    * @param aFooter the new value of the footer property
    */
   public void setFooter(DFooter aFooter) 
   {
      footer = aFooter;    
   }
   
   /**
    * Access method for the header property.
    * 
    * @return   the current value of the header property
    */
   public DHeader getHeader() 
   {
      return header;    
   }
   
   /**
    * Sets the value of the header property.
    * 
    * @param aHeader the new value of the header property
    */
   public void setHeader(DHeader aHeader) 
   {
      header = aHeader;    
   }
   
   /**
    * Access method for the firstCell property.
    * 
    * @return   the current value of the firstCell property
    */
   public TCell getFirstCell() 
   {
      return firstCell;    
   }
   
   /**
    * Sets the value of the firstCell property.
    * 
    * @param aFirstCell the new value of the firstCell property
    */
   public void setFirstCell(TCell aFirstCell) 
   {
      firstCell = aFirstCell;    
   }
   
   /**
    * @param w
    * @param h
    * @param item
    * @param count
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF9510A011C
    */
   public TCell addCell(int w, int h, IDItem item, int count) 
   {
   	  TCell c = new TCell(); //currentPage, currentCell);
   	  addCell(w, h, item, count, c);
   	  return c;    
   }
   
   /**
    * @param w
    * @param h
    * @param item
    * @param count
    * @param c
    * @roseuid 3F00143402E1
    */
   public void addCell(int w, int h, IDItem item, int count, TCell c) 
   {   
   	   if (currentPage==null)
   	   {
		   TCell pageCell = new TCell();
		   if (firstCell == null)
		   {
				firstCell  = pageCell;
		   }
		   currentPage  = pageCell;
		   currentCell  = null;
		   currentYPos  = 0;
		   currentXPos = getXoffset();
   	   }
   	
   	   // accrochage dans l'arbre
   	   if (currentPage.getChild() == null)
	       currentPage.setChild(c);
       c.setParent(currentPage);
       c.setPrevious(currentCell);
       
       if (currentCell != null)
          currentCell.setNext(c);
       
	   if (firstCellLine == null)
	   {
			// new line
			c.setRect(currentXPos + getLeftMargin(), currentYPos + getTopMargin(), w, h);
			firstCellLine = c;
			lastCellLine = c;
   	  	
			currentWLine = currentWLine + w;
			currentHLine = h;
			currentXPos = w;
			currentNLine = currentNLine + 1;
	   }
	   else
	   {
			c.setRect(currentXPos + getLeftMargin(), currentYPos + getTopMargin(), w, h);
			currentHLine = Math.max(currentHLine, h);
			currentWLine = currentWLine + w;
			currentXPos = currentXPos + w;
	   }
       
       // position des caracteres
	   c.setPosition(currentPosition);
	   c.setNumberLine(currentNLine);
	   c.setItem(item);
	   currentPosition = currentPosition + count;   
	   
	   // mise a jour de la taille du parent
	   if (currentPage != null)
	   {
		  currentPage.getRect().height = getTopMargin()+ currentYPos + currentHLine + getBottomMargin();
	      currentPage.getRect().width  = Math.max(currentPage.getRect().width, getLeftMargin() + currentWLine + getRightMargin());
	   }
       
       currentCell = c;    
   }
   
   /**
    * @param s
    * @roseuid 3EFAF6900056
    */
   public void addXSpacing(int s) 
   {
      currentWLine = currentWLine + s;    
   }
   
   /**
    * @param s
    * @roseuid 3EFAF6BA00BA
    */
   public void addYSpacing(int s) 
   {
      currentHLine = currentHLine + s;    
   }
   
   /**
    * @param item
    * @roseuid 3EFAA5FD0218
    */
   protected void applyFont(IDItem item) 
   {
   	   Font font = null;
   	   previousFont = getGc().getFont();
	   	
	   if ((item.getStyle() != null)
			&& (item.getStyle().getFont() != null))
	   {
			IDFont f = item.getStyle().getFont();
			font = (Font)fonts.get(f.getKey());
			
			if (font == null)
			{
				font = DrawUtil.getFont(f);
				fonts.put(f.getKey(), font);
			}
	   }
	   else
	   {
	   	   font = getDefaultFont();
	   }   
	   
	   getGc().setFont(font);    
   }
   
   protected void restoreFont()
   {
   	  if (previousFont!=null)
   	     getGc().setFont(previousFont);
   }
   
   
   
   /**
    * @roseuid 3EF849AE0212
    */
   public void dispose() 
   {
   	  clear();
   	  
      // dispose all fonts
      if (fonts != null)
      {
	     for (Enumeration e = fonts.elements() ; e.hasMoreElements() ;) 
	     {
		    ((Font)e.nextElement()).dispose();
	     }
	     fonts.clear();
      }
	   
      //	reset all imagePool
      if (imagePool != null)
      {
	     for (Enumeration e = imagePool.elements() ; e.hasMoreElements() ;) 
	     {
		    ((Image)e.nextElement()).dispose();
	     }
	     imagePool.clear();
      }
   }
   
   
   /**
    * @return int
    * @roseuid 3EF8475300C0
    */
   public int getHeight() 
   {
	  int h = 0;
	  for (TCell c = getFirstCell(); c != null; c = c.getNext() )
	  {
		h = h +  c.getRect().height;
	  }
	  return h;    
   }
   
   /**
    * @param i
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EFACC7B0159
    */
   public TCell getPage(int i) 
   {
	 int j = 0;
     for (TCell c = getFirstCell(); c != null; c = c.getNext() )
	 {
	    if (i == j) return c;
		j = j + 1; 
	 }
     return  null;    
   }
   
   /**
    * @return int
    * @roseuid 3EFACC4101BA
    */
   public int getPageCount() 
   {
   	  int i = 0;
	  for (TCell c = getFirstCell(); c != null; c = c.getNext() )
	  {
		 i = i + 1;
	  }
	  return i;    
   }
   
   /**
    * @return int
    * @roseuid 3EF847650152
    */
   public int getWidth() 
   {
   	  int w = 0;
	  for (TCell c = getFirstCell(); c != null; c = c.getNext() )
	  {
		 w = Math.max(w, c.getRect().width);
	  }
      return w;    
   }
   
   /**
    * @param p
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF95FE600FE
    */
   public TCell mapPointToCell(Point p) 
   {
	  return mapPointToCell(getFirstCell(), p);    
   }
   
   /**
    * @param c
    * @param p
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF96C530167
    */
   protected TCell mapPointToCell(TCell c, Point p) 
   {
	 TCell i;
	 for (i = c.getChild(); i != null; i = i.getNext())
	 {
	   if ( ((i.getRect().y <= p.y) && (p.y <= i.getRect().y + i.getRect().height))
			&& ((i.getRect().x <= p.x) && (p.x <= i.getRect().x + i.getRect().width)))
	   {
				break;
	   }
	 }
	   
	 if (i==null) return null;
	   
	 // la cellule est trouvee si elle en contient d'autre on y va en 
	 // profondeur
	 if (i.getChild()!=null)
	 {
	 	 p.x = p.x - i.getRect().x;
	 	 p.y = p.y - i.getRect().y;
		 TCell cf = mapPointToCell(i, p);
	   	 
		 if (cf!=null) return cf;
	 }
	   
	 p.x = p.x - i.getRect().x;
	 p.y = p.y - i.getRect().y;
	 return  i;    
   }
   
   /**
    * @param pos
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF9610F0282
    */
   public TCell mapPositionToCell(int pos) 
   {
      return mapPositionToCell(getFirstCell(), pos);    
   }
   
   /**
    * @param c
    * @param pos
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF96C810159
    */
   protected TCell mapPositionToCell(TCell c, int pos) 
   {
	 TCell i;
     for (i = c.getChild(); i != null; i = i.getNext())
	 {
	 	if (i.getChild()!=null)
	 	{
	 		TCell j = mapPositionToCell(i, pos);
	 		if (j!=null) return j;
	 	}
	 	if ((i.getPosition() <= pos)
	 	     && (pos <= i.getPosition() + i.getLen()))
	 	{
	 		return i;
	 	}
	 }
	 return null;    
   }
   
   /**
    * @roseuid 3EF9A6E0028C
    */
   public void newLine() 
   {
      if (firstCellLine != null)
      {
      	int xoffset = getLeftMargin();
      	
      	if (getAlignment() == DParagraph.RIGHT)
      	       xoffset = xoffset + getWidth() - currentWLine;
      	else
      	if (getAlignment() == DParagraph.CENTER)
      	{
      	       xoffset = xoffset + (getWidth() - currentWLine)/2; 
      	}
      	
      	int maxDsc = 0;
      	TCell c;
      	for (c = firstCellLine; c!=null; c = c.getNext())
      	{
      		if (c.getItem() != null)
      		{
      			applyFont(c.getItem());
      			maxDsc=Math.max(maxDsc, getGc().getFontMetrics().getDescent());
				restoreFont();
      		}
      	}
      	
      	for (c = firstCellLine; c!=null; c = c.getNext())
      	{
      		int locDsc;
      		if (c.getItem()!=null)
      		{
				applyFont(c.getItem());
      			locDsc = getGc().getFontMetrics().getDescent();
				restoreFont();
      			c.setRect(xoffset + c.getRect().x, 
      			          c.getRect().y - (maxDsc-locDsc)+(currentHLine-c.getRect().height), 
      			          c.getRect().width, c.getRect().height);
      			//xoffset = xoffset + c.getRect().width;
      		}
      	}
      }
      
      if (getModePage())
      {
      	 int fh = 0;
      	 if (getFooterCell()!=null)
      	 {
			fh = fh + getFooterCell().getRect().height;
      	 }
      	 
      	 if ((currentYPos + currentHLine + fh) > getPageHeight())
      	 { 
      	 	newPage();
      	 }
      	 else
      	 {
			currentYPos = currentYPos + currentHLine;
      	 }
      }
      else
      {   
         currentYPos = currentYPos + currentHLine;
      }
      
	  currentXPos = getXoffset(); 
      
      currentHLine  = 0;
      currentWLine  = 0;
      firstCellLine = null;    
   }
   
   /**
    * @roseuid 3EFAF84300A1
    */
   public void newPage() 
   {
	  // creation de la page
      TCell pageCell = new TCell();
      if (firstCell == null)
	     firstCell  = pageCell;
      else
      {
		 currentPage.setNext(pageCell);
      }
         
      currentPage  = pageCell;
      currentNPage = currentNPage + 1;
      currentCell  = null;
      int hLine = 0;
	  
      // ajout de l'entete s'il y en a une
      if (getHeader()!=null)
      {
	    doItem(getHeader());
	    
	    if (getHeaderCell()!=null)
	    {
	       // retour a la ligne
		   currentYPos  = 0;
		   currentXPos  = getXoffset();
		   currentWLine = 0;
		   currentHLine = 0;
	    
	       addCell(getHeaderCell().getRect().width, 
						 getHeaderCell().getRect().height, 
						 getHeader(), 0, getHeaderCell());
						 
		   hLine = getHeaderCell().getRect().height;
	    }
      }   
      else
      {
	    if (firstCellLine != null)
	    {
		   firstCellLine.setPrevious(null);
		   pageCell.setChild(firstCellLine);
	    }
      }
      
      // ajout du bas de page en fin de la page
      if (getFooter()!=null)
      {
		doItem(getFooter());
      	
		if (getFooterCell()!=null)
		{
            //	retour a la ligne plus positionnement en bas
			currentYPos  = getPageHeight()-getFooterCell().getRect().height;
			currentXPos  = getXoffset();
			currentWLine = 0;
			currentHLine = 0;	
			 
			addCell(getFooterCell().getRect().width, 
					getFooterCell().getRect().height, 
					getFooter(), 0, getFooterCell());
		}     	
      }
         
      TCell c;
      for (c = firstCellLine; c!=null; c = c.getNext())
      {
	     c.setParent(pageCell);
      }    
      
	  currentYPos = hLine;
	  currentXPos = getXoffset();
      currentHLine  = 0;
      currentWLine  = 0;
      firstCellLine = null;    
   }
   
   /**
    * @roseuid 3EFAF25A020E
    */
   public void clear() 
   {
   	   // clears all pages
	   for (TCell c = getFirstCell(); c != null; )
	   {
			 TCell p = c.getNext();
			 c.clear();
			 c.setRect(0,0,0,0);
			 c.setPosition(1);
		 
			 c.setNext(null);
			 c = p;
	   }
	   
	   super.clear();
	   
       currentPosition = 1;
	   currentNLine = 0; 
	   currentXPos = 0;
	   currentYPos = 0;
	   firstCellLine = null;
	   lastCellLine = null;
	   currentCell = null;
	   currentHLine = 0;
	   currentWLine = 0;    
	   alignment = DParagraph.LEFT;
	   currentPage = null;    
	   firstCell = null;
	   currentNPage = 0;    
   }
   
   /**
    * @param sp
    * @param ep
    * @roseuid 3EF9A56C002E
    */
   public void setPageRange(int sp, int ep) 
   {
      setPageBegin(sp);
      setPageEnd(ep);    
   }
   
   /**
    * @roseuid 3F7294070388
    */
   public void dump() 
   {
   	  int i = 0;
   	  for (TCell c = firstCell; c != null; c = c.getNext())
   	  {
   	  	 i = i +1;
   	  	 System.out.println("------------------ Page (" + i + ")-------------------");
   	  	 c.dump(0);
   	  }
   }


   /**
    * @param i
    */
   public void setCurrentYPos(int i) 
   {
	   currentYPos = i;
   }

    /**
     * @return
     */
    public Image getBrokenImage() {
	    if(brokenImage == null){
		   brokenImage = new Image(getDevice(),30,30);
	    }
	    return brokenImage;
   }

   /**
     * @param image
    */
   public void setBrokenImage(Image image) {
	   brokenImage = image;
   }

   /**
	* @param c
	* @roseuid 3EFC03260070
	*/
   public void doItem(DCell c) 
   {
   	  applyFont(c);
   	
	  int size = _getColumnSizeForCell(c);
   	  
	  TBuilder b = _newBuilder(size);
	 	
	  _setBuilderMarginForCell(b);
	
	  b.doChildrenItem(c);
		
	  TCell cl = b.getFirstCell();
	  if (cl != null)
	  {
		if (currentTable.getAutoResize())
			addCell(b.getMaxWidth(), cl.getRect().height, c, 0, cl);
		else
			addCell(cl.getRect().width, cl.getRect().height, c, 0, cl);
	  	
		setCurrentPosition(b.getCurrentPosition());
	    
		// mise en place des marges
		if (currentNColumn > 0)
		{
			cl.getRect().x -= currentTable.getBorder()-1;
		}
		if (currentNRow > 0)
		{
			cl.getRect().y -= currentTable.getBorder()-1;
		}
	  }

	  // on met fin au constructeur 
	  _disposeBuilder(b);
	  
	  CellInfo cf = new CellInfo();
	  cf.rowspan = c.getRowSpan();
	  cf.cell = cl;
	  
	  if (c.getRowSpan()==0)
	  {
		maxCellHeight = Math.max(maxCellHeight, cl.getRect().height);
	  }
	  
	  currentRowSpan[currentNColumn] = cf;
	  currentNColumn = currentNColumn + c.getColSpan() + 1;
	  
	  restoreFont();    
   }
   
   /**
	* @param c
	* @roseuid 3EFC0332009F
	*/
   public void doItem(DCellText c) 
   {
   	  applyFont(c);
   	
	  int size = _getColumnSizeForCell(c);
	  
	  TBuilder b = _newBuilder (size);
	  _setBuilderMarginForCell(b);
	 
	  _buildTextBuffer(c, c.getText());	 
	 	
	  TCell cl = b.getFirstCell();
	   
	  addCell(cl.getRect().width, cl.getRect().height, c, 0, b.getFirstCell());
	  setCurrentPosition(b.getCurrentPosition());
	  
	  // on met fin au constructeur 
	  _disposeBuilder(b);	
	 
	  CellInfo cf = new CellInfo();
	  cf.rowspan = c.getRowSpan();
	  cf.cell = cl;
	  
	  if (c.getRowSpan()==0)
	  {
		  maxCellHeight = Math.max(maxCellHeight, cl.getRect().height);
	  }
	  
	  currentRowSpan[currentNColumn] = cf;
	  currentNColumn = currentNColumn + c.getColSpan() + 1;
	  
	  restoreFont();    
   }
   
   /**
	* @param d
	* @roseuid 3EFC0213029B
	*/
   public void doItem(DDocument d) 
   {
   	  applyFont(d);
   	
	  setAlignment(DParagraph.LEFT);
	  doChildrenItem(d);
	  
	  restoreFont();    
   }
   
   /**
	* @param g
	* @roseuid 3EFC03750268
	*/
   public void doItem(DGraphic g) 
   {
   	  applyFont(g);
   	
	  int w = (int)((float)getMaxWidth() * g.getWidthScale());
	  int h = (int)((float)getMaxWidth() * g.getHeightScale());
	  float ratio = g.getWidthScale()/g.getHeightScale();
	  if (ratio <= 1.0f)
	  {
		if (w < 100)
		{
			w = 100;
			h = (int)(w/ratio);
		}
		else
		if (h < 100)
		{
			h = 100;
			w = (int)(h*ratio);
      		
		}
	  }
      
	  addCell(w, h, g, 0);
	  
	  restoreFont();     
   }
   
   /**
	* @param i
	* @roseuid 3EFC02FC00CA
	*/
   public void doItem(DImage i) 
   {
	 _addListItemPrefix(i);
	  
	 // search image 
	 Image img = (Image)getImagePool().get(i.getName());
	 if (img == null)
	 {
		String imgfilename = searchImage(i.getName());
		if (imgfilename != null)
		{
		   img = new Image(getDevice(), i.getName());
		   getImagePool().put(i.getName(), img);
		}
		else
		{
		   img = (Image)getBrokenImage();
		}
	 }
  
	 Rectangle r = img.getBounds();
	 int w = (int)(r.width * i.getWidthScale());
	 int h = (int)(r.height * i.getHeightScale());
	      
	 if (getWordWrap() 
		 && (w + getCurrentWLine()) > getMaxWidth())
	 {
		 newLine();	
	 }
     
	 TCell c = addCell(w, h, i, 0);
	 c.setImage(img);    
   }
   
   
   /**
	* @param l
	* @roseuid 3EFC02AF03CC
	*/
   public void doItem(DList l) 
   {
   	  applyFont(l);
   	
	  newLine();

	  IDStyle s = l.getStyle();
	
	  addListLevel();
	  setXoffset(getXoffset() * getListOffsetSpacing());
		
	  doChildrenItem(l);
	
	  removeListLevel();
	  setXoffset(getXoffset() * getListOffsetSpacing());
	
	
	  newLine();   
	
	  restoreFont(); 
   }
   
   /**
	* @param l
	* @roseuid 3EFC02AF03CC
	*/
   public void doItem(DLine l) 
   {
	   newLine();
	
	   addCell(getMaxWidth(),
				l.getSize()+getCellSpacing()*2,l,0);
	
	   newLine();    
   }
   
   /**
	* @param pb
	* @roseuid 3F004C3A029E
	*/
   public void doItem(DPageBreak pb) 
   {    
	  if (getModePage())
		  newPage();
	  else
		  newLine();
   }
   
   /**
	* @param pc
	* @roseuid 3F004C20039B
	*/
   public void doItem(DPageCounter pc) 
   { 
   	  applyFont(pc);
   	
	  String s = new Integer(getCurrentNPage()).toString();
      
	  Point p = getGc().textExtent(s);
	  TCell c = addCell(p.x, p.y, pc, s.length());
	  c.setText(s); 
	  
	  restoreFont();   
   }
   
   /**
	* @param p
	* @roseuid 3EFC025D0388
	*/
   public void doItem(DParagraph p) 
   {
   	  applyFont(p);
   	
	  int al = getAlignment();
	  newLine();
	  setAlignment(p.getAlign());
	  
	  _addListItemPrefix(p);
	  
	  doChildrenItem(p);
	  
	  newLine();
	  setAlignment(al);
	  
	  restoreFont();    
   }
   
   /**
	* @param p
	* @roseuid 3EFC02DE0329
	*/
   public void doItem(DPopup p) 
   {
   	  applyFont(p);
   	
	  if (getDrawPopup())
	  {
		setAlignment(DParagraph.LEFT);
		doChildrenItem(p);
	  }    
	  
	  restoreFont();
   }
   
   /**
	* @param p
	* @roseuid 3EFC02CA0280
	*/
   public void doItem(DPopupLink p) 
   {
   	 applyFont(p);
   	
	 doChildrenItem(p);
	 if (!getFlatPopup()) return;
	 
	 DLinkResolver r = new DLinkResolver(p);
	 DPopup popup = (DPopup)r.getTaggedItem();
	 
	 if (popup!=null)
	 {
		TBuilder b = _newBuilder(getMaxWidth());
	 	
		int m = getCellSpacing();
		b.setTopMargin(m);
		b.setBottomMargin(m);
		b.setLeftMargin(m);
		b.setRightMargin(m);
		
		b.doChildrenItem(popup);
		
		TCell c = b.getFirstCell();
		
		addCell(c.getRect().width, c.getRect().height, popup, 0);   
		setCurrentPosition(b.getCurrentPosition());

		// on met fin au constructeur 
		_disposeBuilder(b);	
	 }    
	 
	 restoreFont();
   }
   
   /**
	* @param r
	* @roseuid 3EFC03190216
	*/
   public void doItem(DRow r) 
   {
   	  applyFont(r);
   	
	  currentNColumn = 0;
	  currentRow = r;
	  maxCellHeight = 0;
	  
	  // calcul numero de colonne de la premiere cellule de la liste
	  // a partir de la deuxieme ligne
	  if (currentNRow>0)
	  {
		for (currentNColumn=0; 
			   (currentRowSpan[currentNColumn].rowspan>0) 
				  && (currentNColumn<currentRowSpan.length); currentNColumn++);
	  }
	  
	  
	  doChildrenItem(r);
	  
	  if (currentNRow>0)
	  {
		int w = 0;
		for (int i=0; i<currentRowSpan.length; i++)
		{
		  if (currentRowSpan[i]==null) continue;
		  TCell c = currentRowSpan[i].cell;
		  
		  if (currentRowSpan[i].rowspan>0)
		  { 
			 w += c.getRect().width;
			 currentRowSpan[i].rowspan -= 1;
			 if (currentRowSpan[i].rowspan==0)
			 {
				c.getRect().height = (getCurrentYPos() + maxCellHeight) - c.getRect().y;
				currentRowSpan[i] = null;
			 }
		  }
		  else
		  {
			c.getRect().x = c.getRect().x + w;			
			c.getRect().height = maxCellHeight;
			w += c.getRect().width;
		  }
		}
	  }
	  
	  int currentYPos = getCurrentYPos();
	  newLine();
	  setCurrentYPos(maxCellHeight + currentYPos);
	  currentNRow = currentNRow + 1;
	  
	  restoreFont();    
   }
   
   /**
	* @param s
	* @roseuid 3EFC02410247
	*/
   public void doItem(DSection s) 
   {
   	  applyFont(s);
   	
	  setHeader(s.getHeader());
	  setFooter(s.getFooter());
	  
	  if (getModePage())
		  newPage();
	  else
		  newLine();
          
	  addCell(0,0,s,0);
      
	  doChildrenItem(s);
	  
	  restoreFont();    
   }
   
   /**
	* @param t
	* @roseuid 3EFC030901EB
	*/
   public void doItem(DTable t) 
   {
   	  applyFont(t);
   	
	  newLine();
	  currentTable = t;
	  currentNRow = 0;
	  currentTotalColumn = t.getColumnCount();
	  currentRowSpan = new CellInfo[t.getColumnCount()];
      
      
	  doChildrenItem(t);    
	  
	  currentTable = null;
	  currentRow = null;
	  currentRowSpan = null;
	  
	  restoreFont(); 
   }
   
   /**
	* @param t
	* @roseuid 3EFC03A001FC
	*/
   public void doItem(DTag t) 
   {
	  addCell(0,0,t,0);    
   }
   
   /**
	* @param t
	* @roseuid 3EFC02EE0142
	*/
   public void doItem(DText t) 
   {
   	  applyFont(t);
   	  
	  _addListItemPrefix(t);
	  _buildTextBuffer(t, t.getText());
	  
	  restoreFont();    
   }
   
   /**
	* @param t
	* @roseuid 3EFC02A00244
	*/
   public void doItem(DTitle t) 
   {
   	  applyFont(t);
   	
	  addTitleLevel(t.getLevel());
	  String st = getTitleStringPrefix(t);
	  
	  Point r = getGc().textExtent(st);
	  
	  addYSpacing(getTitleSpacing());
	  newLine();
	  
	  TCell c = addCell(r.x, r.y + getTitleSpacing(), t, st.length());
	  c.setText(st);
	  
	  doChildrenItem(t);
	  
	  newLine();   
	  
	  restoreFont(); 
   }
   
   /**
	* @param f
	* @roseuid 3F02AEBE025D
	*/
   public void doItem(DFooter f) 
   {
   	   applyFont(f);
   	
	   //	fabrication de l'entete et pied de base
	   TBuilder b = _newBuilder (getMaxWidth());
		
	   b.doChildrenItem(f);
		
	   setFooterCell(b.getFirstCell());
	   b.setFirstCell(null);
			
	   // on met fin au constructeur 
	   _disposeBuilder(b);
	   
	   restoreFont();    
   }
   
   /**
	* @param h
	* @roseuid 3F02AEAC03AC
	*/
   public void doItem(DHeader h) 
   {
   	   applyFont(h);
   	
	   // fabrication de l'entete et pied de base
	   TBuilder b = _newBuilder(getMaxWidth());
	 	
	   b.doChildrenItem(h);
		
	   setHeaderCell(b.getFirstCell());
	   b.setFirstCell(null);
			
	   // on met fin au constructeur 
	   _disposeBuilder(b);
	   
	   restoreFont();    
   }
   
   /**
	* @param text
	* @param w
	* @param r
	* @roseuid 3F02AF060329
	*/
   private void _buildGetNextWord(String text, int w, Range r) 
   {
	  String qs = text.substring(r.start);
	  int wt = getGc().textExtent(qs).x;
	  if (wt <= w)
	  {
		r.end = wt;
		r.start = text.length(); 
	  }
	  else
	  {
		int pos = ((w * qs.length()) / wt);
		for (int i=pos; i > 0; i = i -1)
		{
		   char c = qs.charAt(i);
		   if (!Character.isSpaceChar(c)
				&& (c!='.') && (c!=';') && (c!=',') && (c!='?')
				&& (c!=':') && (c!='(') && (c!=')') && (c!='/')
				&& (c!='[') && (c!=']') && (c!='\\'))
		   {
			  continue;
		   }

		   wt = getGc().textExtent(qs.substring(0, i)).x;
		   if (wt < w)
		   {
			  r.end = wt;
			  r.start = r.start + i;
			  return;
		   }
		}
		// on est pas arrive a decouper la chaine sur des separateurs
		// on va decouper les mots par dicotomie
		int a = qs.length() / 2;
		int b = qs.length();
		while (b-a>1)
		{
		   String s = qs.substring(0, a);
		   wt = getGc().textExtent(s).x;
		   if (wt < w)
		   {
			  a = a + (b-a)/2;
		   }
		   else
		   {
			  b = a;
			  a = a / 2;
		   }
		}
		r.end = wt;
		r.start = a + r.start;
	  }
   }

}



/*
 * Created on Sep 19, 2003
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package com.rational.apidoc.drivers.html;

import java.util.LinkedList;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;

import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.ImageLoader;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.SWTError;

import com.rational.apidoc.parsers.DParser;
import com.rational.apidoc.drivers.IWriter;
import com.rational.apidoc.core.*;
import com.rational.apidoc.examples.*;
import com.rational.apidoc.chart.ChartSnapShot;

/**
 * @author sduguet
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
public class DHtmlDriver extends DParser implements IWriter {

	private ConfParser confparser;
	private PrintStream out;

	private String dirname, filename;
	private DDocument document;
	
	private static Item default_hitem = new Item();

	/**
	 * 
	 */
	public DHtmlDriver() {
		super();
		confparser = new ConfParser();
		if (!confparser.parse()) {
			System.err.println("Html Configuration parsing failed");
			confparser = null;
			return;
		}
	}

	public void setDestDirAndFileName(String dname, String fname) {
		dirname = dname != null ? dname : ".";
		filename =
			dirname + "/" + fname + confparser.getHtmlFile().getExtension();
	}

	public PrintStream getStream() {
		return out;
	}

	public void htmlConvertDump(String string) {
		int space_count = 0;
		for (int i = 0; i < string.length(); i++) {
			char c = string.charAt(i);
			if (c == ' ') {
				space_count++;
				if (space_count == 1) {
					out.print(confparser.getConvertText().getFirstSpace());
				} else {
					out.print(confparser.getConvertText().getExtraSpace());
				}
			} else {
				space_count = 0;
				Character ch = new Character(c);
				String repl = confparser.getConvertText().getReplace(ch);
				String cstr = repl != null ? repl : ch.toString();
				try {
					byte a[] = cstr.getBytes("UTF-8");
					out.write(a, 0, a.length);
				} catch (UnsupportedEncodingException e) {
					out.print(cstr);
				}
			}
		}
	}

	public void htmlDump(String string) {
		try {
			byte a[] = string.getBytes("UTF-8");
			out.write(a, 0, a.length);
		} catch (UnsupportedEncodingException e) {
			out.print(string);
		}
	}

	public void htmlDump(int value) {
		out.print(value);
	}

	public void htmlDump(LinkedList list) {
		for (int i = 0; i < list.size(); i++) {
			String elt = (String) list.get(i);
			String value = confparser.getVariableValue(elt);
			if (value == "#") {
				htmlDump("${" + elt + "}");
			} else {
				htmlDump(value != null ? value : elt);
			}
		}
	}

	private void setTagVariable(String name, String tag, IDItem item) {
		confparser.setVariableValue(
			name,
			tag != null ? tag : "_" + item.hashCode(),
			item);
	}

	private String to2DigitsHexa(int color) {
		return Integer.toHexString((color % 256) / 16)
			+ Integer.toHexString(color % 16);
	}

	private String getColorHexa(IDColor color, String default_value) {
		if (color == null) {
			return default_value;
		}
		String result = "#";
		result += to2DigitsHexa(color.getRed());
		result += to2DigitsHexa(color.getGreen());
		result += to2DigitsHexa(color.getBlue());
		return result;
	}

	private String getFontAttrsBegin(IDStyle style) {
		IDFont font = style.getFont();
		IDColor color = style.getForeColor();
		if (font == null && color == null) {
			return "";
		}
		String result = "";
		if (font != null) {
			if (font.getBold()) {
				result += "<B>";
			}
			if (font.getUnderline() || font.getDashed()) {
				result += "<U>";
			}
			if (font.getItalic()) {
				result += "<I>";
			}
		}
		result += "<FONT COLOR=" + getColorHexa(color, "#000000");
		if (font != null && font.getSize() != -1) {
			int s = font.getSize() / 3; // size 12 -> 4 (normal size)
			if (s < 1) {
				s = 1; // 1 is the minimum
			}
			if (s > 7) {
				s = 7; // 7 is the maximum
			}
			result += " SIZE=\"" + Integer.toString(s) + "\"";
		}
		result += ">";
		if (font != null && confparser.isTTFont(font.getName())) {
			result += "<TT>";
		}

		return result;
	}

	private String getFontAttrsEnd(IDStyle style) {
		IDFont font = style.getFont();
		IDColor color = style.getForeColor();
		if (font == null && color == null) {
			return "";
		}
		String result = "";
		if (font != null && confparser.isTTFont(font.getName())) {
			result += "</TT>";
		}
		result += "</FONT>";
		if (font != null) {
			if (font.getBold()) {
				result += "</B>";
			}
			if (font.getUnderline() || font.getDashed()) {
				result += "</U>";
			}
			if (font.getItalic()) {
				result += "</I>";
			}
		}
		return result;
	}

	private void updateStyleVariables(IDItem item) {
		// PRE: item.getStyle() not null (to be checked by the caller)
		confparser.setVariableValue(
			"HTML_FONT_ATTRS_BEGIN",
			getFontAttrsBegin(item.getStyle()),
			item);
		confparser.setVariableValue(
			"HTML_FONT_ATTRS_END",
			getFontAttrsEnd(item.getStyle()),
			item);
		String fore_color =
			getColorHexa(item.getStyle().getForeColor(), "#000000");
		confparser.setVariableValue(
			"HTML_COLOR",
			"=\"" + fore_color + "\"",
			item);
		confparser.setVariableValue("HTML_FORE_COLOR", fore_color, item);
		String back_color =
			getColorHexa(item.getStyle().getBackColor(), "#ffffff");
		confparser.setVariableValue(
			"HTML_BGCOLOR",
			"=\"" + back_color + "\"",
			item);
		confparser.setVariableValue("HTML_BACK_COLOR", back_color, item);
	}

	public Item getItem(IDItem item) {
		// Perform some checks
		Item hitem = confparser.getItem(item.getClass());
		if (hitem == null) {
			System.err.println("No html parser found for " + item.getClass());
			hitem = default_hitem;
		}
		return hitem;
	}

	public void doOneItemAfter(IDItem item) {
		if (item instanceof DText) {
		} else if (item instanceof DCellText) {
		} else if (item instanceof DGraphic) {
			graphicDump((DGraphic) item);
		}
	}

	public void doItem(DDocument item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("DOCUMENT_TAG", item.getTag(), item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DSection item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("SECTION_TAG", ((DSection) item).getTag(), item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DHeader item) {
		// ignored
	}

	public void doItem(DFooter item) {
		// ignored
	}

	public void doItem(DParagraph item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"PARAGRAPH_ALIGN",
			((DParagraph) item).alignToString(((DParagraph) item).getAlign()),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DTag item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("TAG_NAME", ((DTag) item).getTag(), item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DText item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		htmlDump(hitem.getBegin());
		htmlConvertDump(item.getText());
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DImage item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"IMAGE_NAME",
			((DImage) item).getName(),
			item);
		htmlDump(hitem.getBegin());
		// no childs
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DList item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		if (((DList) item).getFormat() == DStyle.BULLET) {
			confparser.setVariableValue("LIST_TYPE", "BULLETED", item);
			confparser.setVariableValue("HTML_LIST_TYPE", "UL", item);
		} else if (((DList) item).getFormat() == DStyle.NUMBER) {
			confparser.setVariableValue("LIST_TYPE", "NUMBERED", item);
			confparser.setVariableValue("HTML_LIST_TYPE", "OL", item);
		} else {
			confparser.setVariableValue("LIST_TYPE", "", item);
			confparser.setVariableValue("HTML_LIST_TYPE", "UL", item);
		}
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DTable item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		if (((DTable) item).getAutoResize()) {
			float w = 0;
			float wc[] = ((DTable) item).getColumnWidth();
			if (wc == null || wc.length < ((DTable) item).getColumnCount()) {
				w = ((DTable) item).getWidthScale();
			} else {
				for (int i = 0; i < wc.length; i++) {
					w += wc[i];
				}
			}
			confparser.setVariableValue("TABLE_WIDTH", Float.toString(w), item);
			confparser.setVariableValue(
				"HTML_TABLE_WIDTH",
				Integer.toString((int) (w * 100.0f)),
				item);
		} else {
			float w = 0.0f;
			confparser.setVariableValue("TABLE_WIDTH", Float.toString(w), item);
			confparser.setVariableValue(
				"HTML_TABLE_WIDTH",
				Integer.toString((int) (w * 100.0f)),
				item);
		}
		confparser.setVariableValue(
			"TABLE_BORDER",
			Integer.toString(((DTable) item).getBorder()),
			item);
		confparser.setVariableValue(
			"TABLE_COLUMNS",
			Integer.toString(((DTable) item).getColumnCount()),
			item);
		confparser.setVariableValue(
			"TABLE_ROWS",
			Integer.toString(((DTable) item).getRowCount()),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DRow item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"ROW_CELLS",
			Integer.toString(((DRow) item).getCellCount()),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	private void setCellVariables(DCell item) {
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"CELL_COLSPAN",
			Integer.toString(((DCell) item).getColSpan()),
			item);
		confparser.setVariableValue(
			"HTML_CELL_COLSPAN",
			Integer.toString(((DCell) item).getColSpan() + 1),
			item);
		confparser.setVariableValue(
			"CELL_ROWSPAN",
			Integer.toString(((DCell) item).getRowSpan()),
			item);
		confparser.setVariableValue(
			"HTML_CELL_ROWSPAN",
			Integer.toString(((DCell) item).getRowSpan() + 1),
			item);
		confparser.setVariableValue(
			"CELL_COL",
			Integer.toString(((DCell) item).getColumnNumber()),
			item);
		confparser.setVariableValue(
			"CELL_ROW",
			Integer.toString(((DCell) item).getLineNumber()),
			item);
		DRow row = (DRow) item.getParent();
		DTable table = (DTable) row.getParent();
		if (table.getColumnWidth() != null
			&& table.getColumnCount() > ((DCell) item).getColumnNumber()) {
			float w = table.getColumnWidth()[((DCell) item).getColumnNumber()];
			confparser.setVariableValue("CELL_WIDTH", Float.toString(w), item);
			confparser.setVariableValue(
				"HTML_CELL_WIDTH",
				Integer.toString((int) (w * 100.0f)),
				item);
		}
	}

	public void doItem(DCell item) {
		Item hitem = getItem(item);
		setCellVariables(item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DCellText item) {
		Item hitem = getItem(item);
		setCellVariables(item);
		htmlDump(hitem.getBegin());
		htmlConvertDump(((DCellText) item).getText());
		// no childs
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DLink item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		String url =
			(((DLink) item).isInternal() ? "" : ((DLink) item).getMethod())
				+ ":"
				+ ((DLink) item).getPath();
		confparser.setVariableValue("LINK_TARGET", url, item);
		confparser.setVariableValue("HTML_LINK", url, item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DLine item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"HTML_LINE_SIZE",
			Integer.toString(((DLine) item).getSize()),
			item);
		htmlDump(hitem.getBegin());
		// no childs
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DPageBreak item) {
		// ignored
	}

	public void doItem(DPageCounter item) {
		// ignored
	}

	public void doItem(DPopup item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("POPUP_TAG", ((DPopup) item).getTag(), item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DPopupLink item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"POPUP_LINK_METHOD",
			((DPopupLink) item).typeToString(((DPopupLink) item).getType()),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DSummary item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DTitle item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("TITLE_TAG", ((DTitle) item).getTag(), item);
		addTitleLevel(((DTitle) item).getLevel());
		confparser.setVariableValue(
			"TITLE_LEVEL",
			Integer.toString(getTitleLevel()),
			item);
		confparser.setVariableValue(
			"HTML_TITLE_LEVEL",
			Integer.toString(getTitleLevel()),
			item);
		confparser.setVariableValue(
			"TITLE_FULL_NUMBER",
			getTitleStringPrefix(item),
			item);
		confparser.setVariableValue(
			"TITLE_NUMBER",
			Integer.toString(titleTab[getTitleLevel()]),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DIndex item) {
	}

	public void doItem(DIndexEntry item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DTimerItemContainer item) {
		// ignored
	}

	private void graphicDump(DGraphic graphic) {
		String image_filename =
			dirname
				+ "/"
				+ confparser.getVariableValue("HTML_GRAPHIC_FILE_NAME")
				+ ".jpg";
		int w = 500, h = w;
		w *= graphic.getWidthScale();
		h *= graphic.getHeightScale();
		Rectangle r = new Rectangle(0, 0, w, h);
		Image img = ChartSnapShot.Process(r, graphic);
		ImageLoader loader = new ImageLoader();
		if (loader.data == null) {
			loader.data = new ImageData[1];
		}
		loader.data[0] = img.getImageData();
		try {
			loader.save(image_filename, SWT.IMAGE_JPEG);
		} catch (SWTException _e) {
			System.err.println(
				"Error saving image:" + image_filename + ": " + _e);
		} catch (SWTError _r) {
			System.err.println(
				"Error saving image:" + image_filename + ": " + _r);
		}
		img.dispose();
	}

	public void doItem(DGraphic item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"GRAPHIC_TITLE",
			((DGraphic) item).getTitle(),
			item);
		confparser.setVariableValue(
			"GRAPHIC_WIDTH",
			Float.toString(((DGraphic) item).getWidthScale()),
			item);
		confparser.setVariableValue(
			"HTML_GRAPHIC_WIDTH",
			Integer.toString(
				(int) (((DGraphic) item).getWidthScale() * 100.0f)),
			item);
		confparser.setVariableValue(
			"GRAPHIC_HEIGHT",
			Float.toString(((DGraphic) item).getHeightScale()),
			item);
		confparser.setVariableValue(
			"HTML_GRAPHIC_FILE_NAME",
			"img" + Integer.toHexString(item.hashCode()),
			item);
		htmlDump(hitem.getBegin());
		graphicDump(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DAxis item) {
		// already made from the enclosing DGraphic
	}

	public void doItem(DCurve item) {
		// already made from the enclosing DGraphic
	}

	public void doItem(DPoint item) {
		// already made from the enclosing DGraphic
	}

	public void write(IDItem doc) {
		document = (DDocument) doc;

		if (confparser == null) {
			System.err.println("Html Configuration parsing failed");
			return;
		}
		try {
			out = new PrintStream(new FileOutputStream(filename));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return;
		}

		// Some variables at document level
		confparser.setVariableValue("HTML_FILE_NAME", filename, document);
		confparser.setVariableValue(
			"DOCUMENT_TITLE",
			document.getTitle(),
			document);
		confparser.setVariableValue("HTML_BGCOLOR", "=\"#ffffff\"", document);

		setTitleMaxLevel(confparser.getHtmlFile().getMaxTitleLevel() - 1);
		HtmlFile hf = confparser.getHtmlFile();
		htmlDump(hf.getBegin());
		doItem(document);
		htmlDump(hf.getEnd());

		out.close();
	}

	public static void main(String[] args) {
		System.out.println("initialize...");
		DHtmlDriver hd = new DHtmlDriver();

		System.out.println("create...");

//		Example1 ex1 = new Example1();
//		hd.setDestDirAndFileName(".", "file1");
//		System.out.println("translate...");
//		hd.write(ex1.doc);
//		hd.clear();
//
//		Example2 ex2 = new Example2();
//		hd.setDestDirAndFileName(".", "file2");
//		System.out.println("translate...");
//		hd.write(ex2.doc);
//		hd.clear();
//
//		Example3 ex3 = new Example3();
//		hd.setDestDirAndFileName(".", "file3");
//		System.out.println("translate...");
//		hd.write(ex3.doc);
//		hd.clear();
//
//		Example4 ex4 = new Example4();
//		hd.setDestDirAndFileName(".", "file4");
//		System.out.println("translate...");
//		hd.write(ex4.doc);
//		hd.clear();
//
//		Example5 ex5 = new Example5();
//		hd.setDestDirAndFileName(".", "file5");
//		System.out.println("translate...");
//		hd.write(ex5.doc);
//		hd.clear();
//
//		Example6 ex6 = new Example6();
//		hd.setDestDirAndFileName(".", "file6");
//		System.out.println("translate...");
//		hd.write(ex6.doc);
//		hd.clear();
//
		Source so = new Source();
		hd.setDestDirAndFileName(".", "source");
		System.out.println("translate...");
		hd.write(so.doc);
		hd.clear();
		
		System.out.println("done!");
	}
}
/*
 * Created on 26 juin 2003
 *
 */
package com.rational.apidoc.chart;

import org.eclipse.swt.SWT;

import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Layout;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Button;

import com.rational.apidoc.drawutil.*;
import com.rational.apidoc.core.* ;

/**
 * Example to create and render a new graphic type for ApiDoc.
 * The new graphic will draw polar curves.
 * 
 * To extends graphics type, follow these steps:
 * - define a new id for graphic type: see T_POLAR field.
 * - define an IChartRender that able to draw your graphic
 *   if render have support to location, extends IChartLocation 
 *   (thrown by your ChartRender).
 * 
 * - don't forget to register the new IChartRender using DChartRenderRegistry
 *   (see it in main() method)
 * 
 * - now you can build your document containing new graphic type, see this
 *   in BuildPolarGraphic() method.
 */
public class TestUserGraphicType
{  
   /** Define a new graphic type identifier: */   
   public static final String T_POLAR = "T_POLAR.sample.chart.apidoc.rational.com";
   
   /** First axis a Polar graphic must provide */
   public static final String AXIS_RAY = "ray-axis";
   /** Second axis a Polar graphic must provide */
   public static final String AXIS_ANGLE ="angle-axis";
   
   /** Class used to identify where is located a point on a polar graphic */
   public static class PolarLocation implements IChartLocation
   {
      /** Located on graphic title area */
      public static final String GRAPHIC_TITLE="graphic-title";
      /** Located on a segment of a curve (item is first point of segment) */
      public static final String CURVE_SEGMENT="curve-segment";
      /** located on a point of a curve (only for curve type DCurve.POINT) */
      public static final String CURVE_POINT  ="curve-point";
      
      private String id_;
      private IDItem item_;
      private Rectangle rect_;
      
      public PolarLocation( String _id, IDItem _item, Rectangle _r )
      {
        id_=_id; item_=_item; rect_=_r;
      }
      
      public String getId() { return id_; }
      public IDItem getItem() { return item_; }
      public Rectangle getRectangle() { return rect_; }
   }
   
   /** 
    * Implements an IChartRender for this new graphic type
    */   
   private static class PolarChartRender implements IChartRender
   {
     public boolean canRender(String _graphic_type)
     {
       if( T_POLAR.equals( _graphic_type )) return true;
       return false;
     }

     public void draw(GC _gc, Rectangle _R, DGraphic _g)
     {   
       render( true, 0,0, _gc, _R, _g );
     }
     
     public IChartLocation locate(int _x, int _y, GC _gc, Rectangle _R, DGraphic _G)
     {
       return render( false, _x, _y, _gc, _R, _G );
     }
     
     /** support for both draw and location modes */
     public IChartLocation render( boolean _draw, int _lx, int _ly, GC _gc, Rectangle _R, DGraphic _g)
     {   
       if( _draw )
       {
         _gc.setBackground( new Color( null, 255,255,255 ));
         _gc.fillRectangle( _R );                
       }
       else
       {
         if( !_R.contains( _lx, _ly ) ) return null; //outside graphic
       }
       
       //search axis
       DAxis r_axis = null;
       DAxis a_axis = null;
       for( IDItem item=_g.getFirstChild(); item!=null; item=item.getNext() )
       {
         if( item instanceof DAxis )
         {
           DAxis axis = (DAxis)item;
           if( AXIS_RAY.equals( axis.getName() ) ) r_axis=axis;
           else if ( AXIS_ANGLE.equals( axis.getName() ) ) a_axis=axis;
         }
       }
       if( r_axis==null || a_axis==null )
       {
         if( _draw )
         {
           Color red = new Color( null, 255,0,0 );
           String missed = (r_axis==null?"'AXIS_RAY'":"")+" "+(a_axis==null?"'AXIS_ANGLE'":"");
           _gc.setForeground( red );
           _gc.drawText( "Error: missing axis "+missed+"' in graphic item", 10, 30 );
         }
         return null;
       }
       
       final int margin = 10;
       int lx = _R.x+margin;
       int ty = _R.y+margin;
       int rx = _R.x+_R.width-margin;
       int by = _R.y+_R.height-margin;
       int cx = (lx+rx)/2;
       int cy = (ty+by)/2; 
       
       //render title if any
       String title = _g.getTitle();
       if( !DrawUtil.isEmpty( title )
         && _g.getProperties().get( DGraphic.P_SHOW_TITLE, true ) )
       {
         IDStyle style = _g.getStyle();
         Font save = _gc.getFont();
         Font font = DrawUtil.getFont( style.getFont() );
         int aln = DrawUtil.HCENTER|DrawUtil.VCENTER|DrawUtil.VERTICAL;
         if( font!=null ) _gc.setFont( font );
         int tw = DrawUtil.textExtent( _gc, title, aln ).x;
         if (_draw)
         {
           //gradient background
           _gc.setForeground( new Color(null,205,220,255));
           _gc.fillGradientRectangle( lx-1, ty, tw+3, by-ty, true );
           //title color and text
           _gc.setForeground( DrawUtil.getColor(style.getForeColor()) );
           DrawUtil.drawText(_gc, lx, ty, tw, by - ty, aln, title);
         } else {
           Rect r = DrawUtil.boundingRect( _gc, lx, ty, tw, by-ty, aln, title );
           if( r.contains( _lx,_ly ) )
           {
             return new PolarLocation( PolarLocation.GRAPHIC_TITLE, _g, r.toRectangle() );
           }
           if( _lx <= lx+tw )
           {
             //located on graphic title area (outside text),
             //no need to search in curve, but no id is defined for this...
             return null;
           }
         }

         _gc.setFont( save );
         
         lx+=tw+margin;
         cx = (lx+rx)/2;
       }
       
       //first search curve and points to compute min/max
       //an array of DCurve is stored to be able to iterate in reverse order
       //as location mode need.
       DCurve curves[] = new DCurve[ _g.getChildCount() ];
       int    n_curves=0; //because curves.length take into account DAxis too... ;-(
       float xmin=0.0f, xmax=0.0f, ymin=0.0f, ymax=0.0f;
       float ray_max=0.0f;
       boolean init=false;       
       
       for( IDItem item=_g.getFirstChild(); item!=null; item=item.getNext() )
       {
         if( !(item instanceof DCurve) ) continue;
         DCurve curve = (DCurve)item;
         curves[ n_curves++ ] = curve;
         for( IDItem item2=curve.getFirstChild(); item2!=null; item2=item2.getNext())
         {
           if(!(item2 instanceof DPoint)) continue;
           DPoint point = (DPoint)item2;
           DCoord c_angle = point.getCoord( a_axis );
           DCoord c_ray   = point.getCoord( r_axis );
           if( c_angle==null || c_ray==null )
           {
             //no error message here, skip point only
             continue;
           }
           float r = c_ray.getValue();
           float a = c_angle.getValue();
           if( r > ray_max ) ray_max = r;
           //convert to pixels
           float ptx = (float)( r*Math.cos(a) );
           float pty = (float)( r*Math.sin(a) );
           if( init )
           {
             if( ptx < xmin ) xmin=ptx; else if ( ptx > xmax ) xmax=ptx;
             if( pty < ymin ) ymin=pty; else if ( pty > ymax ) ymax=pty;
           }
           else
           {
             xmin=xmax=ptx; 
             ymin=ymax=pty;
             init=true;
           }
         }
       }
       
       float K=1.0f;
       float c_max ;
       {
         c_max = xmax;
         float x = Math.abs( xmin ); if( x > c_max) c_max=x;
         if( ymax > c_max ) c_max=ymax;
         x = Math.abs( ymin ); if ( x > c_max ) c_max=x;
         int pix_min = cx-lx < cy-ty ? cx-lx : cy-ty;
         if( c_max > 0.0f ) K = pix_min / c_max;
       }
       
       //draw axis lines
       if( _draw )
       {
         _gc.setForeground( new Color( null, 220,220,220 ));
         Rectangle r = _gc.getClipping();
         _gc.setClipping( lx, ty, rx-lx, by-ty );
         //Line Step property is set on 'ray' axis ?
         float ray_step_line = r_axis.getProperties().get( DAxis.P_STEP_LINE, 0.0f );
         if( ray_step_line >=0.0f )
         {
            for( float ray=ray_step_line; ray <= c_max; ray += ray_step_line )
            {
              int p_ray = (int)( K*ray );
              int p_ray2 = 2*p_ray;
              _gc.drawOval( cx-p_ray, cy-p_ray, p_ray2, p_ray2 );
            }
         }
         
         //Line Step property is set on 'angle' axis ?
         float ang_step_line = a_axis.getProperties().get( DAxis.P_STEP_LINE, 0.0f );
         if( ang_step_line >=0.0f )
         {
           float dpi = (float)( 2*Math.PI );
           for( float a = ang_step_line; a < dpi; a+=ang_step_line )
           {
             int px = (int)( cx + K*c_max*Math.cos(a) );
             int py = (int)( cy + K*c_max*Math.sin(a) );
             _gc.drawLine( cx, cy, px, py );
           }
         }
         //Dot Step property is set on 'angle' axis ?
         float ang_step_dot = a_axis.getProperties().get( DAxis.P_STEP_DOT, 0.0f );
         if( ang_step_dot >=0.0f )
         {
           float dpi = (float)( 2*Math.PI );
           _gc.setLineStyle( SWT.LINE_DOT );
           for( float a = ang_step_dot; a < dpi; a+=ang_step_dot )
           {
             int px = (int)( cx + K*c_max*Math.cos(a) );
             int py = (int)( cy + K*c_max*Math.sin(a) );
             _gc.drawLine( cx, cy, px, py );
           }
           _gc.setLineStyle( SWT.LINE_SOLID );
         }
         
         _gc.setForeground( new Color( null, 0,0,0 ));
         _gc.drawLine( lx, cy, rx, cy );
         _gc.drawLine( cx, ty, cx, by );
         

         _gc.setClipping( r );
       }

       
       Color def_clr = new Color( null, 0,150,0 );
       //render curves...
       // - draw in forward order
       // - locate in reverse order as the painter algorythm request.
       int crv_shift= _draw ? 0: n_curves-1 ;
       for( int icrv=0; icrv<n_curves; ++icrv )
       {
         DCurve curve = curves[ Math.abs( icrv - crv_shift) ];
         //curve type is dots or lines ?
         boolean dots = DCurve.T_POINTS.equals( curve.getType() );
         if( _draw )
         {        
           Color clr = def_clr;
           IDStyle style = curve.getStyle();
           if( style!=null )
           {
             clr = DrawUtil.getColor( style.getBackColor() );
             if( dots )
             {
               _gc.setBackground( clr );
               clr = DrawUtil.getColor( style.getForeColor() );
             }
           }
           _gc.setForeground( clr );
         }
         
         boolean draw_line=false;
         int ppx=0, ppy=0;
         DPoint last_point=null;
         for( IDItem item2=curve.getFirstChild(); item2!=null; item2=item2.getNext())
         {
           if(!(item2 instanceof DPoint)) continue;
           DPoint point = (DPoint)item2;
           DCoord c_angle = point.getCoord( a_axis );
           DCoord c_ray   = point.getCoord( r_axis );
           if( c_angle==null || c_ray==null )
           {
             //no error message here, skip point only
             draw_line=false;
             last_point=null;
             continue;
           }
           float r = c_ray.getValue();
           float a = c_angle.getValue();
           //convert to pixels
           int px = (int)( cx + K*r*Math.cos(a) );
           int py = (int)( cy + K*r*Math.sin(a) );
           if( dots )
           {
             Rectangle rect = new Rectangle( px-2, py-2, 4, 4 );
             if( _draw )
             {
               _gc.fillRectangle( rect );
               _gc.drawRectangle( rect );
             }
             else
             {
               //point located on the curve Points
               if( rect.contains(_lx, _ly ) )
               {
                 return new PolarLocation( PolarLocation.CURVE_POINT, point, rect );
               }
             }
           }
           else //other curves draw using lines
           {
             if( draw_line )
             {
               if( _draw )
               {
                 _gc.drawLine( ppx, ppy, px, py );
               }
               else
               {
                 //point near this segment ?
                 if( DrawUtil.segmentContains( ppx, ppy, px, py, _lx,_ly ) )
                 {
                   return new PolarLocation( PolarLocation.CURVE_SEGMENT, last_point,
                           new Rectangle( _lx-1,_ly-1,2,2) ); 
                 }
               }
             } else {
               draw_line=true;
             } 
             ppx=px;
             ppy=py;
             last_point=point;
           }
         }
       }  
       
       return null;     
     }

   }

/*---------------------------------------------------------------------------------------
 * Code from here up to the end of file are only made to build the test.
 * You can see how to use your graphic type in an apidoc document in the
 * BuildPolarGraphic method.
 *---------------------------------------------------------------------------------------
 */
 
   /** used only for test shell */
   public static interface IBuildGraphic
   {
     DGraphic buildGraphic();
   }   

  private static DColor black = new DColor(0,0,0);
  private static DColor blue1 = new DColor(0,0,255);
  private static DColor blue2 = new DColor(80,80,255);
  private static DColor blue3 = new DColor(150,150,255);
  private static DColor blue4 = new DColor(180,180,255);
  private static DColor green = new DColor(0,255,0);
  private static DColor red   = new DColor(255,0,0);
  private static DColor magenta = new DColor(255,0,255);
  private static DColor white = new DColor(255,255,255);

  private static DFont font1 = new DFont( "f1",null,null, 14, DFont.BOLD );
  private static DFont df    = new DFont( "df", "Arial",null,10,DFont.NORMAL );
  
  private static IDStyle sd = new DStyle( "sd", df, black, white );        
  private static IDStyle s0 = new DStyle( "s0", font1, green, white );
  private static IDStyle s1 = new DStyle( "s1", df, black, red );
  private static IDStyle s2 = new DStyle( "s2", df, black, blue2 );
  private static IDStyle s3 = new DStyle( "s3", df, black, blue4 );
  private static IDStyle s4 = new DStyle( "s4", df, black, green );
  private static IDStyle s5 = new DStyle( "s5", df, black, magenta );
  private static IDStyle s6 = new DStyle( "s6", df, black, green );
     
  private static DPoint newDPoint( DCurve _p, DAxis _x, DAxis _y, float _cx, float _cy )
  {
    DPoint pt = new DPoint();
    _p.addChild( pt );
    pt.add( new DCoord( _x, _cx ));
    pt.add( new DCoord( _y, _cy ));
    return pt;
  }
  private static DPoint newDPoint3( DCurve _p, DAxis _x, DAxis _y, DAxis _z, float _cx, float _cy, float _cz )
  {
    DPoint pt = new DPoint();
    _p.addChild( pt );
    pt.add( new DCoord( _x, _cx ));
    pt.add( new DCoord( _y, _cy ));
    pt.add( new DCoord( _z, _cz ));
    return pt;
  }
    /**
   * Graphic document builder for XY graphic 
   */
  private static class BuildXYGraphic implements IBuildGraphic
  {
    public DGraphic buildGraphic()
   {
     DGraphic  gph = new DGraphic();  
     gph.setGraphicType( DGraphic.T_XY );
     gph.setTitle( "Graphic XY" );
     gph.setStyle( s0 );
//gph.setProperties().store(DGraphic.P_SHOW_TITLE, false);     
//gph.setProperties().store(DGraphic.P_SHOW_LEGEND, false); 

     DAxis  x_axs = new DAxis("x");    gph.addChild( x_axs );
     DAxis  y_axs = new DAxis("y");    gph.addChild( y_axs );
     x_axs.setStyle( sd );
     y_axs.setStyle( sd );
     x_axs.getProperties().store( DAxis.P_STEP_UNIT, 2.0f );
     x_axs.getProperties().store( DAxis.P_STEP_DOT, 0.5f );
     y_axs.getProperties().store( DAxis.P_STEP_UNIT, 2.0f );
     y_axs.getProperties().store( DAxis.P_STEP_DOT, 0.5f );
     x_axs.setTitle( "X Axis");
     x_axs.setUnit("(x)");
     y_axs.setTitle( "Y Axis");
     y_axs.setUnit("(unit)");
     
     
     DCurve    cv0 = new DCurve();            gph.addChild( cv0 );
     cv0.setName( "Bars" );
     cv0.setType( DCurve.T_BAR );
     cv0.setStyle( s4 );
     newDPoint( cv0, x_axs, y_axs, 0, 0.9f );
     newDPoint( cv0, x_axs, y_axs, 1, 1.2f );
     newDPoint( cv0, x_axs, y_axs, 2, 1.6f );
     newDPoint( cv0, x_axs, y_axs, 3, 3.4f );
     newDPoint( cv0, x_axs, y_axs, 4, 4.6f );

     DCurve    cv1 = new DCurve();            gph.addChild( cv1 );
     cv1.setName("Line");
     cv1.setStyle( s1 );
     newDPoint( cv1, x_axs, y_axs, 0,0 );
     newDPoint( cv1, x_axs, y_axs, 1,1.2f );
     newDPoint( cv1, x_axs, y_axs, 2,3.2f );
     newDPoint( cv1, x_axs, y_axs, 3,2.4f );
     newDPoint( cv1, x_axs, y_axs, 4,0.6f );

     DCurve    cv2 = new DCurve();            gph.addChild( cv2 );
     cv2.setName( "Points" );
     cv2.setType( DCurve.T_POINTS );
     cv2.setStyle( s2 );
     newDPoint( cv2, x_axs, y_axs, 0, 1.9f );
     newDPoint( cv2, x_axs, y_axs, 1, 1.2f );
     newDPoint( cv2, x_axs, y_axs, 2, 1.1f );
     newDPoint( cv2, x_axs, y_axs, 3, 0.4f );
     newDPoint( cv2, x_axs, y_axs, 4, 0.0f );
     
     x_axs.getProperties().store( DAxis.P_MIN, 0.0f );
     x_axs.getProperties().store( DAxis.P_MAX, 5.0f );
     x_axs.getProperties().store( DAxis.P_STEP_LINE, 1.0f );
     y_axs.getProperties().store( DAxis.P_MIN, 0.0f );
     y_axs.getProperties().store( DAxis.P_MAX, 5.0f );
     y_axs.getProperties().store( DAxis.P_STEP_LINE, 1.0f );
     
     return gph;
   }
  }
  
  /**
   * Graphic document builder for Polar graphic 
   */
  private static class BuildPolarGraphic implements IBuildGraphic
  {
    public DGraphic buildGraphic()
   {
     DGraphic  gph = new DGraphic();  
     // Set my graphic type 
     gph.setGraphicType( T_POLAR );
     gph.setTitle( "Polar Graphic" );
     gph.setStyle( s0 );

     DAxis  ray_axs = new DAxis( AXIS_RAY );    gph.addChild( ray_axs );
     DAxis  angle_axs = new DAxis( AXIS_ANGLE );    gph.addChild( angle_axs );
     ray_axs.setStyle( sd );
     angle_axs.setStyle( sd );
     ray_axs.getProperties().store( DAxis.P_STEP_UNIT, 2.0f );
     ray_axs.getProperties().store( DAxis.P_STEP_DOT, 0.5f );
     angle_axs.getProperties().store( DAxis.P_STEP_UNIT, Math.PI/8 );
     angle_axs.getProperties().store( DAxis.P_STEP_DOT, Math.PI/16 );
     
     DCurve    cv0 = new DCurve();            gph.addChild( cv0 );
     cv0.setName( "Petals" );
     cv0.setType( DCurve.T_LINE );
     cv0.setStyle( s4 );
     DCurve    cv1 = new DCurve();            gph.addChild( cv1 );
     cv1.setName("Spiral");
     cv1.setType( DCurve.T_POINTS );
     cv1.setStyle( s1 );
     DCurve    cv2 = new DCurve();            gph.addChild( cv2 );
     cv2.setName("Curve 3");
     cv2.setStyle( s2 );
     DCurve    cv3 = new DCurve();            gph.addChild( cv3 );
     cv3.setName("Curve 4");
     cv3.setStyle( s5 );
     
     //populate curve using various function ( ray = f( angle ) )
     float da = (float)( Math.PI/100 );
     float r;
     float rmax=0.0f;
     int cnt=10;
     for( float a=0.0f; a<6*Math.PI; a+=da )
     {
       float ca = (float)Math.cos(a);
       float sa = (float)Math.sin(a);
       
       r = (float)( (a/(10*da))*ca*sa );
       if( r > rmax ) rmax=r;
       newDPoint( cv0, ray_axs, angle_axs, r, a );
       r = (float)( a );
       if( --cnt == 0 )
       {
         if( r > rmax ) rmax=r;
         newDPoint( cv1, ray_axs, angle_axs, r, a );
         cnt=10;
       }
       r = (float)(8*ca - 10.0f*sa/(1.1f+ca));
       if( r > rmax ) rmax=r;
       newDPoint( cv2, ray_axs, angle_axs, r, a );
       r = (float)(20+3*Math.cos(10*a)-20*Math.sin(5*a)*Math.cos(2*a));
       if( r > rmax ) rmax=r;
       newDPoint( cv3, ray_axs, angle_axs, r, a );
     }
   
     ray_axs.getProperties().store( DAxis.P_MIN, 0.0f );
     ray_axs.getProperties().store( DAxis.P_MAX, 1.1f*rmax );
     ray_axs.getProperties().store( DAxis.P_STEP_LINE, rmax/5 );
     angle_axs.getProperties().store( DAxis.P_MIN, 0.0f );
     angle_axs.getProperties().store( DAxis.P_MAX, 5.0f );
     angle_axs.getProperties().store( DAxis.P_STEP_LINE, Math.PI/8 );
     
     return gph;
   }
  }
  
  private static class RenderGraphic implements MouseListener, MouseMoveListener, PaintListener
  {
    private Canvas canvas_;
    private IChartRender render_;
    private DGraphic gph_;
    private boolean mouse_down_=false;
    private int mx_=0, my_=0;
    
    public RenderGraphic( Canvas _cvs )
    {
      canvas_=_cvs; 
      canvas_.addMouseListener( this );
      canvas_.addMouseMoveListener( this );
      canvas_.addPaintListener( this );
    }
    
    public void setGraphic( DGraphic _g )
    {
      gph_ = _g;
      render_ = DChartRenderRegistry.GetRender( _g.getGraphicType() );
      canvas_.redraw();
    }
    
    public void mouseMove(MouseEvent e)
    {
      if( render_==null || gph_==null ) return ;
      if( mouse_down_ &&( gph_.isGraphicType( DGraphic.T_XYZ )
                      || gph_.isGraphicType( DGraphic.T_SURFACE)) )
      {
        move3DView(e);
        return;
      }
          GC gc = new GC(canvas_);   
          Rectangle r = canvas_.getClientArea();
          IChartLocation loc = render_.locate( e.x,e.y, gc, r, gph_ );
//System.out.println("GOT LOC="+loc);          
          if( loc!=null && loc.getItem() instanceof DPoint )
          {
            DPoint pt = (DPoint)loc.getItem();
            DCurve crv= (DCurve)pt.getParent();
            DGraphic gph=(DGraphic)crv.getParent();
            String name = crv.getName();
            if( name==null ) name=""; else name=name+": ";
            for( IDItem item=gph.getFirstChild(); item!=null; item=item.getNext() )
            {
              if( item instanceof DAxis )
              {
                DAxis axis=(DAxis)item;
                DCoord crd= pt.getCoord( axis );
                if( crd!=null )
                {
                  name = name + Float.toString( crd.getValue() )+"  ";
                }
              }
            }
            canvas_.setToolTipText( name );
          } else {
            canvas_.setToolTipText( null );
          }
          gc.dispose();
          gc = null;
        }
    public void mouseDoubleClick(MouseEvent e){}

    public void mouseDown(MouseEvent e)
    {
      mouse_down_=true;
      mx_=e.x;
      my_=e.y;      
    }
    public void mouseUp(MouseEvent e)
    {
      mouse_down_=false;
    }     
    protected void move3DView(MouseEvent e)
    {  
      if( render_==null || gph_==null ) return ;

      float phi   = gph_.getProperties().get( DGraphic.P_XYZ_PHI, 0.3f);
      float theta = gph_.getProperties().get( DGraphic.P_XYZ_THETA, 0.2f);
      int dx = e.x -mx_;
      int dy = e.y -my_;
      
      phi   += 0.02f*dy;
      theta -= 0.02f*dx;

      gph_.getProperties().store( DGraphic.P_XYZ_PHI,  phi    );
      gph_.getProperties().store( DGraphic.P_XYZ_THETA, theta );
      mx_ = e.x;
      my_ = e.y;    
      canvas_.redraw();  
    }
    
    private Image img_;
    public void paintControl(PaintEvent e)
    {
       if( render_==null || gph_==null ) return ;
       Rectangle r = canvas_.getClientArea();
       if( img_==null || img_.getBounds().width!=r.width || img_.getBounds().height!=r.height)
       {
         if( img_!=null ) img_.dispose();
         img_ = new Image(null,r.width,r.height);
       }
       GC gc = new GC( img_ );
       render_.draw( gc, r, gph_ );
       e.gc.drawImage( img_, 0, 0 );
       gc.dispose();
    }
   }
   
   /* Layout horizontally children control, all width of them (but not the last one)
    * is limited (as the FillLayout doesn't do it).
    */
   private static class HLimLayout extends Layout
   {
     private int lim_;
     public HLimLayout(int _lim){lim_=_lim;}
     protected Point computeSize(Composite composite, int wHint, int hHint, boolean flushCache)
     {
       return new Point( 100, hHint );
     }

     protected void layout(Composite composite, boolean flushCache)
     {
       Rectangle r = composite.getClientArea();
       Control c[] = composite.getChildren();
       if( c==null ) return ;
       int x = r.x;
       for( int i=0; i<c.length; ++i )
       {
         if( (i+1)==c.length)
         {
           c[i].setBounds( x, r.y, r.x+r.width-x, r.height );
         } else {
           c[i].setBounds( r.x, r.y, lim_, r.height );
           x += lim_;
         }
       }
      
     }
   }
   
   public static void main( String[] argv )
   {
     //This is very important to register the render for new graphic type.
     DChartRenderRegistry.AddRender( new PolarChartRender() );
     
     Display disp = Display.getDefault();
     final Shell shell = new Shell(disp, SWT.DIALOG_TRIM|SWT.RESIZE );
     shell.setText( "Test ApiDoc: User Graphic" );
     shell.setLayout(new HLimLayout( 150 ));

     Composite cb = new Composite( shell, SWT.NONE );
     cb.setLayout( new FillLayout( SWT.VERTICAL ));
     Group gb = new Group( cb, SWT.NONE);
     gb.setLayout( new FillLayout(SWT.VERTICAL));
     gb.setText("Graphic Type:");
     final Button btn[] = new Button[2];
     int b=-1;

     btn[++b] = new Button( gb, SWT.RADIO ); btn[b].setText("XY (standard)");
     btn[b].setData( new BuildXYGraphic());     

     btn[++b] = new Button( gb, SWT.RADIO ); btn[b].setText("Polar (user defined)");
     btn[b].setData( new BuildPolarGraphic());     
       
     final Canvas canvas = new Canvas( shell, SWT.NO_BACKGROUND );
     final RenderGraphic rg = new RenderGraphic( canvas );
    
     //just on first button is enough
     SelectionListener listen_button= new SelectionListener(){
       public void widgetSelected(SelectionEvent e)
       {   
         for(int i=0; i<btn.length; ++i)
         {
           if( btn[i].getSelection())
           {
             DGraphic g = ((IBuildGraphic)btn[i].getData()).buildGraphic();
             rg.setGraphic( g );
             break;            
           }
         }
       }
 
       public void widgetDefaultSelected(SelectionEvent e){} //never called
     };
     
     //set selection listeners to all buttons...
     for( int i=0; i<btn.length; ++i )
     {
       btn[i].addSelectionListener( listen_button );
     }
         
     shell.setSize( 600, 500 );
     shell.open();
 
     
      while (!shell.isDisposed())
        if (!disp.readAndDispatch())
          disp.sleep();
   }
}
/*
 * Created on 25 juin 2003
 *
 */
package com.rational.apidoc.chart.internal;

import java.text.DecimalFormat;
import java.util.Vector;

import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.graphics.FontMetrics;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.SWT;
import com.rational.apidoc.core.*;
import com.rational.apidoc.chart.IChartLocation;
import com.rational.apidoc.chart.IChartRender;
import com.rational.apidoc.drawutil.HSV;
import com.rational.apidoc.drawutil.DrawUtil;
import com.rational.apidoc.drawutil.Rect;
import com.rational.apidoc.drawutil.Point2D;

/**
 * ChartRender is made to 
 * - draw DGraphic item in an area
 * - locate a point on a DGraphic area.
 */
public class DefaultChartRender implements IChartRender
{
  /** margin for render */
  protected int margin_; 
  /** Default style to draw */
  protected DChartStyle default_;
 
  /** temporary data during draw/locate */
  protected Data d_ ; 
  /** used to dispose font/color from styles (MUST BE disposed for windows systems)
   *  (?huh? work'in java and take care of windows ... smell bad (but smell'SWT!) */
  protected Vector styleof_;

  /** Create a chart render */
  public DefaultChartRender()
  {
    margin_=10;
    default_ = new DChartStyle();
    d_=null;
  }
  
  //paire courbe / style
  private static class CurveStyle
  {
    public DCurve curve_;
    public DChartStyle style_;
    public CurveStyle( DCurve _c, DChartStyle _s )
    {
      curve_=_c; style_=_s;
    }
  }
  
  private static class Data
  {
     
    public boolean drawing() { return draw_; }
    public boolean locating() { return !draw_; }
    
    public boolean draw_;
    public GC gc_; //mode drawing
    public Font default_font_;
    public int lx_, ly_; //mode locating
    public DGraphic G_; 
    public Rect R_;
    public Vector curves_; //list of CurveStyle
    
    public Data( boolean _draw, DGraphic _g, Rect _R, GC _gc )
    {
      gc_=_gc;
      draw_=_draw;
      G_=_g;
      R_=_R;
      lx_=ly_=-1;
      curves_ = new Vector();
      default_font_ = _gc.getFont();
    }        

    static boolean horizontal3DContains( int _lx, int _ly, int _w, int _3d, int _px, int _py )
    {
      //? si (_px,_py) est contenus dans le rectangle cisaille en X:
      //    _lx+_3d
      //       +---------+ _ly-_3d
      //      /         /
      // _ly +---------+
      //    _lx     _lx+_w
      //!! les Y ecrans sont inverses
      if( _py < (_ly-_3d) ) return false;
      if( _py >  _ly      ) return false;
      if( _py==_ly )
      {
        return (_px>=_lx)&&(_px<=(_lx+_w));
      }
      if( _py==(_ly-_3d))
      {
        return (_px>=(_lx+_3d))&&(_px<=(_lx+_w+_3d));
      }
      //dans le rectangle cisaille en X
      int dx = _ly-_py; //OK: car proj. 3d en pente 45 deg.
      return (_px>=(_lx+dx))&&(_px<=(_lx+dx+_w));
    }
    
    static boolean vertical3DContains( int _lx, int _ly, int _h, int _3d, int _px, int _py )
    {
      //? si (_px,_py) est contenus dans le rectangle cisaille en Y:
      //       _lx+_3d
      //          + _ly-_3d
      //         /|
      //    _ly + |
      //        | |
      //        | +
      //        |/
      // _ly+_h +
      //      _lx
      if( _px <  _lx     ) return false;
      if( _px > (_lx+_3d)) return false;
      if( _px == _lx      ) return (_py>=_ly)&&(_py<=(_ly+_h));
      ////if( _px == (_lx+_3d)) return (_py>=(_ly-_3d))&&(_py<=(_ly-_3d+_h));
      //!! les Y ecrans sont inverses
      //dans le rectangle cisaille en Y
      int dy = _px-_lx; //OK: car proj. 3d en pente 45 deg.
      int npy = _py+dy;
     return (npy>=_ly)&&(npy<=(_ly+_h));
    }

  }
  
  protected static class ChartLocation extends Throwable
  {
    public IChartLocation location_;
    public ChartLocation( IChartLocation _l )
    {
      location_=_l;
    }
    public ChartLocation( String _id, IDItem _item, Rect _rect )
    {
      location_=new DChartLocation( _id, _item, _rect );
    }
    public ChartLocation( String _id, IDItem _item, Rect _rect, String _dtxt )
    {
      location_=new DChartLocation( _id, _item, _rect, _dtxt );
    }
  }
  
  /** @return true since default render can draw anything */
  public boolean canRender( String _graph_type )
  {
    return true;
  }
  
  /** Draw graphic _g on rectangle _R using GC _gc. */
  public void draw( GC _gc, Rectangle _R, DGraphic _g )
  {
    if( _g==null ) return ;
    d_ = new Data( true, _g, new Rect(_R), _gc );
    try {
      render();
    }
    catch( ChartLocation _l) {}//musn't appears
    d_ = null;
  }
  
  /** Return a location for the given point, GC must be provided to compute text extends.
   * can return null if no location is found.
   */
  public IChartLocation locate( int _x, int _y, GC _gc, Rectangle _R, DGraphic _G )
  {
    if(_G==null) return null;
    if(!_R.contains(_x,_y) ) return null;

    d_ = new Data( false, _G, new Rect(_R), _gc );
    d_.lx_ = _x;
    d_.ly_ = _y;
    
    try {
      render();
    }
    catch( ChartLocation _l )
    {
      return _l.location_;
    }
    finally
    {
      d_=null;
    }
    return null;
  }
  
  /** Return style for item, walk upstair in tree to try to complete style. */
  protected DChartStyle styleOf( IDItem _item )
  {
    IDStyle apidoc_style=null;
    IDFont  apidoc_font =null;
    IDColor apidoc_color=null;
    DChartStyle style=new DChartStyle();
    char missing = 0x07; //font|fore|back

    for( IDItem item=_item; item!=null; item=item.getParent() )
    {
      if( (apidoc_style=item.getStyle())!=null )
      {
        //font
        if ( (missing & 0x01)!=0 && (apidoc_font=apidoc_style.getFont())!=null ) 
        {          
          style.font_ = DrawUtil.getFont( apidoc_font ); 
          missing &= ~0x01 ;       
        }
        //fore color
        if ( (missing & 0x02)!=0 && (apidoc_color=apidoc_style.getForeColor())!=null )
        {
          style.fore_ = DrawUtil.getColor( apidoc_color );
          missing &= ~0x02 ;
        }
        //back color
        if ( (missing & 0x04)!=0 && (apidoc_color=apidoc_style.getBackColor())!=null )
        {
          style.back_ = DrawUtil.getColor( apidoc_color );
          missing &= ~0x04 ;
        }
        //style completed.
        if( missing == 0x00 ) break ;
      }
    }//for

    if( styleof_==null ) styleof_ = new Vector();
    styleof_.add( style );
    return style;
  }
  
  protected void disposeStyleOf()
  {
    if( styleof_==null ) return ;
    for( int i=0; i<styleof_.size(); ++i )
    {
      DChartStyle s = (DChartStyle)styleof_.get(i);
      s.dispose();
    }
    styleof_.clear();
  }
  
  /** print error occured while rendering chart. */
  void error( String _message )
  {
//TBD:juste pour mise au point: affichage de ce qui a merde
    if( d_.drawing() )
    {
      Font font = d_.gc_.getFont();
      FontData fd[] = font.getFontData();
      FontData nfd = new FontData( fd[0].getName(), fd[0].getHeight(), fd[0].getStyle()|SWT.BOLD );
      Font nf = new Font( null, nfd );
      
      d_.gc_.setFont( nf );
      d_.gc_.setForeground( getRed() );
      d_.gc_.drawText( _message, d_.R_.x()+5, d_.R_.y()+5+12 );

      nf.dispose();
    }
  }
  
  /** draw or locate point using d_ Data */
  protected void render() throws ChartLocation
  {
    try
    {    
      DChartStyle g_style = styleOf( d_.G_ );

      //prepare curve list, and their styles.
      d_.curves_.clear();
      IDItem  item=null;
      DCurve curve=null;
      for( item=d_.G_.getFirstChild(); item!=null; item=item.getNext() )
      {
        if( item instanceof DCurve )
        {
          curve = (DCurve)item;
          DChartStyle cs = styleOf( curve );                
          d_.curves_.add( new CurveStyle( curve, cs ) );
        }
      }
    
      //graphics background
      if( d_.drawing() )
      {
        d_.gc_.setBackground( g_style.back_ );
        d_.gc_.fillRectangle( d_.R_.x(), d_.R_.y(), d_.R_.w(), d_.R_.h() );
      }

      //graphic title
      int title_h = renderGraphicTitle( g_style );

      //legend on the right side
      int legend_width = renderLegend( title_h );

      //the rest is available to render graphic
      Rect rg = new Rect( margin_,
                          title_h+margin_,
                          d_.R_.w()-legend_width-2*margin_,
                          d_.R_.h()-title_h-2*margin_ );
  
      //graphic is rendered depending on it type:
           if( d_.G_.isGraphicType( DGraphic.T_SECTORS      )) renderSector( rg, false );
      else if ( d_.G_.isGraphicType( DGraphic.T_SECTORS3D    )) renderSector( rg, true  );
      else if ( d_.G_.isGraphicType( DGraphic.T_HISTOGRAM    )) renderHistogram( rg, false, false );
      else if ( d_.G_.isGraphicType( DGraphic.T_HISTOGRAM3D  )) renderHistogram( rg, true, false );
      else if ( d_.G_.isGraphicType( DGraphic.T_HISTOGRAM_BS )) renderHistogram( rg, false, true );
      else if ( d_.G_.isGraphicType( DGraphic.T_XY           )) renderXY( rg );
      else if ( d_.G_.isGraphicType( DGraphic.T_XYZ          )) renderXYZ( rg );
      else if ( d_.G_.isGraphicType( DGraphic.T_SURFACE      )) renderXYZ( rg );
/*
      else if ( d_.G_.isGraphicType( DGraphic.T_HISTORS )) renderHistors( rg )
*/
      else
      {
        error( "Unknwon Graphic type '"+d_.G_.getGraphicType()+"'" );
        return ;
      }
    
      g_style.dispose();
    
      if( d_.locating() && rg.contains( d_.lx_, d_.ly_ ) )
      {
        throw new ChartLocation( DLocated.InGraphicArea, d_.G_, rg );
      }
    }
    finally
    {
       //dispose styles, must be done for poor platform like windows.
       disposeStyleOf();
       d_.curves_.clear();
       d_.curves_=null; 
    }
  }
  
  
  /** render graphic title.
   * @return the height of the title, 0 if title isn't displayed
   */
  protected int renderGraphicTitle( DChartStyle _g_style ) throws ChartLocation
  {
    int title_h = 0;

    boolean display_title = d_.G_.getProperties().get(DGraphic.P_SHOW_TITLE, true);
    String title = d_.G_.getTitle();  
    
    if( !display_title || DrawUtil.isEmpty( title ) ) return title_h;
    
    d_.gc_.setFont( _g_style.font_ );
    FontMetrics fm = d_.gc_.getFontMetrics();      
    title_h = fm.getHeight();
  
    String txt = DrawUtil.truncateTextH( d_.gc_, title, d_.R_.w() );
    int    aln = DrawUtil.HCENTER | DrawUtil.TOP;
    Rect r_ttl = new Rect( d_.R_.x(), d_.R_.y(), d_.R_.w(), title_h );

    if( d_.drawing() )
    {
      d_.gc_.setForeground( _g_style.fore_ );
      DrawUtil.drawText( d_.gc_, r_ttl, aln, txt);
    }
    else
    {
      //locating
      Rect br = DrawUtil.boundingRect( d_.gc_, r_ttl, aln, txt );
      br.union( r_ttl );

      if( br.contains( d_.lx_, d_.ly_ ) )
      {
        throw new ChartLocation( DLocated.InGraphicTitleText, d_.G_, br, txt );
      }
      if( r_ttl.contains( d_.lx_, d_.ly_ ) )
      {
        throw new ChartLocation( DLocated.InGraphicTitleArea, d_.G_, r_ttl );
      }
    }
 
    return title_h;
  }
  
  /** render legend and return the width, 0 if none */
  protected int renderLegend( int _title_h )  throws ChartLocation 
  {
    int legend_width = 0;

    //display legend ?
    boolean display_legend = d_.G_.getProperties().get(DGraphic.P_SHOW_LEGEND, true);
    if( !display_legend ) return 0;
    
    CurveStyle ccurve=null;

    int legend_height = 0;
    int legend_height_min =100000;
    for( int i=0; i<d_.curves_.size(); ++i )
    {
      ccurve = (CurveStyle)d_.curves_.get(i);
      d_.gc_.setFont( ccurve.style_.font_ );
      String name = ccurve.curve_.getName();
      if( !DrawUtil.isEmpty( name ) )
      {
        Point sz = d_.gc_.textExtent( name );
        if ( legend_width < sz.x ) legend_width=sz.x;
        if ( legend_height_min > sz.y ) legend_height_min=sz.y;
        legend_height += sz.y;
      }
    }

    int legend_rect=(legend_height_min-2);
    if(legend_rect>10) legend_rect=10; //limit to 10x10 pixel
    legend_width += legend_rect+margin_ ;
    //limit legend width to 1/3 of graphic
    if ( legend_width > (d_.R_.w()/3) ) legend_width=d_.R_.w()/3;
    int legend_x = d_.R_.w() - legend_width;
    int legend_H = d_.curves_.size() * 4 + legend_height;
    int legend_y = _title_h+(d_.R_.h()-legend_H)/2+2;

    Rect lgd_area= new Rect( d_.R_.right()-legend_width, d_.R_.top()+_title_h,
                             legend_width, d_.R_.h()-_title_h );
    //shortcut: on on y est pas, on passe la main ...
    //shortcut for locating: outside ?
    if( d_.locating() && !lgd_area.contains( d_.lx_, d_.ly_ ) )
    {
      return legend_width ;
    }
    
    Rect r_clr=new Rect();
    Rect r_lgd_area=new Rect();
    Rectangle r = new Rectangle(0,0,0,0);
    
    for( int i=0; i<d_.curves_.size(); ++i )
    {
      ccurve = (CurveStyle)d_.curves_.get(i);
      String name = ccurve.curve_.getName();
      if( DrawUtil.isEmpty( name )) continue;
      //text is choosen black, because fore color is curve color
      d_.gc_.setFont( ccurve.style_.font_ );
      int legend_h = d_.gc_.textExtent( name ).y;
      r_clr.setRect( legend_x,legend_y+(legend_h-legend_rect)/2, 
                     legend_rect, legend_rect );

      int available_width = legend_width-legend_rect-5;
      String txt = DrawUtil.truncateTextH( d_.gc_, name, available_width );

      if( d_.drawing() )
      {
        d_.gc_.setForeground( getBlack() );
        d_.gc_.setBackground( ccurve.style_.back_ );
        r_clr.toRectangle( r );
        d_.gc_.fillRectangle( r ); //fill it
        d_.gc_.drawRectangle( r ); //draw black square around it
      }
      else //locating
      {
        //inside color rectangle
        if( r_clr.contains( d_.lx_, d_.ly_ ) )
        {
          throw new ChartLocation( DLocated.LegendColor, ccurve.curve_, r_clr );
        }
      }
    
      r_lgd_area.setRect( legend_x+legend_rect+5, legend_y,
                          available_width,        legend_h );

      renderText( txt, r_lgd_area, DrawUtil.LEFT|DrawUtil.VCENTER,
                   ccurve.style_, DLocated.LegendText, ccurve.curve_ );

      legend_y += legend_h+4;
    }//for

    //dans la zone de la legende ??
    if( d_.locating() && lgd_area.contains( d_.lx_, d_.ly_ ) )
    {
      throw new ChartLocation( DLocated.LegendArea, null, lgd_area );
    }
  
    return legend_width;
  }    
  
  protected static class TruncateAxisTextData
  {
    public String sttl_ ; //string title
    public String sunit_; //string for unit
    public Rect   rttl_ ; //rectangle for title
    public Rect   runit_; //rectangle for unit
    public Rect   rtu_  ; //rectangle for both title and unit    
  }
  
  /** truncate unit and/or title text for an "horizontal" axis,
   * return also bounding rectangles.
   */
  protected TruncateAxisTextData truncateHAxisText(
      DAxis _axis, DChartStyle _axis_style,
      int _available_width, int _available_height
    )
  {
    Rect   rttl;
    Rect   runit;
    Rect   rtu;

    //reduction de la chaine de titre ou units..

    //title et unit sont geree dans des chaines/rect different cela va permettre
    //de les affichers avec des styles differents ... un jours prochains.
    String axis_title = _axis.getTitle();
    String axis_unit  = _axis.getUnit();
//TODO: mieux a faire, SWT ne supporte pas textExtent(null)    
    if( axis_title == null ) axis_title="";
    if( axis_unit  == null ) axis_unit="";

    //titre ( unit ) axe des Y
    d_.gc_.setFont( _axis_style.font_ );
    Point ext = d_.gc_.textExtent( axis_title );
    rttl = new Rect( 0,0, ext.x+2, _available_height );
    //unit a droite du titre
    ext = d_.gc_.textExtent( axis_unit );
    runit = new Rect( rttl.w()+5, 0, ext.x+2, _available_height );
    rtu = Rect.Union( rttl, runit ); //rectangle englobant titre et unite 

    //reduire d'abord le titre
    String n_axis_title = axis_title;
    while( rtu.w() > _available_width ) 
    {
      try {
        axis_title = axis_title.substring( axis_title.length()-2, axis_title.length()-1 );
      }
      catch( StringIndexOutOfBoundsException e )
      {
        n_axis_title="";
        break;
      }
      if( axis_title.length()==0 ) { n_axis_title=null; break; }
      n_axis_title = axis_title+"...";
      ext = d_.gc_.textExtent( n_axis_title );
      rttl.setWidth( ext.x+2 );
      runit.setX( rttl.w()+5 );
      // rtu = Rect.Union( rttl, runit ) but without allocation
      rtu.setRect( rttl );
      rtu.union( runit );
    }
    axis_title = n_axis_title ;
    
    //puis l'unite
    String n_axis_unit = axis_unit;
    while( rtu.w() > _available_width ) 
    {
      try {
        axis_unit = axis_unit.substring( axis_unit.length()-2, axis_unit.length()-1 );
      }
      catch( StringIndexOutOfBoundsException e )
      {
        n_axis_unit="";
        break;
      }
      if( axis_unit.length()==0 ) { n_axis_unit=null; break; }
      n_axis_unit = axis_unit+"...";
      ext = d_.gc_.textExtent( n_axis_unit );
      runit.setWidth( ext.x+2 );
      rtu.setRect( rttl );
      rtu.union( runit );
    }
    axis_unit = n_axis_unit;

    TruncateAxisTextData data = new TruncateAxisTextData();
    data.sttl_ = axis_title;
    data.sunit_= axis_unit;
    data.rttl_ = rttl;
    data.runit_= runit;
    data.rtu_  = rtu ;
    
    return data;
  }

    /** truncate unit and/or title text for an "vertical" axis,
     * return also bounding rectangles.
     */
    protected TruncateAxisTextData truncateVAxisText(
        DAxis _axis, DChartStyle _axis_style,
        int _available_width, int _available_height
      )
    {
      Rect   rttl;
      Rect   runit;
      Rect   rtu;
  
      //reduction de la chaine de titre ou units..
  
      //title et unit sont geree dans des chaines/rect different cela va permettre
      //de les affichers avec des styles differents ... un jours prochains.
      String axis_title = _axis.getTitle();
      String axis_unit  = _axis.getUnit();
  //TODO: mieux a faire, SWT ne supporte pas textExtent(null)    
      if( axis_title == null ) axis_title="";
      if( axis_unit  == null ) axis_unit="";
  
      //titre ( unit ) axe des Y
      d_.gc_.setFont( _axis_style.font_ );
      Point ext = DrawUtil.verticalTextExtent( d_.gc_, axis_title );
      rttl = new Rect( 0,0, _available_width, ext.y+2 );
      //unit a droite du titre
      ext = DrawUtil.verticalTextExtent( d_.gc_, axis_unit );
      runit = new Rect( 0, rttl.h()+5, _available_width, ext.y+2 );
      rtu = Rect.Union( rttl, runit ); //rectangle englobant titre et unite 
  
      //reduire d'abord le titre
      String n_axis_title = axis_title;
      while( rtu.w() > _available_height ) 
      {
        try {
          axis_title = axis_title.substring( axis_title.length()-2, axis_title.length()-1 );
        }
        catch( StringIndexOutOfBoundsException e )
        {
          n_axis_title="";
          break;
        }
        if( axis_title.length()==0 ) { n_axis_title=null; break; }
        //no sens for vertical text/ n_axis_title = axis_title+"...";
        ext = DrawUtil.verticalTextExtent( d_.gc_, n_axis_title );
        rttl.setHeight( ext.y+2 );
        runit.setY( rttl.h()+5 );
        // rtu = Rect.Union( rttl, runit ) but without allocation
        rtu.setRect( rttl );
        rtu.union( runit );
      }
      axis_title = n_axis_title ;
      
      //puis l'unite
      String n_axis_unit = axis_unit;
      while( rtu.w() > _available_height ) 
      {
        try {
          axis_unit = axis_unit.substring( axis_unit.length()-2, axis_unit.length()-1 );
        }
        catch( StringIndexOutOfBoundsException e )
        {
          n_axis_unit="";
          break;
        }
        if( axis_unit.length()==0 ) { n_axis_unit=null; break; }
        //no sense for vertical/ n_axis_unit = axis_unit+"...";
        ext = DrawUtil.verticalTextExtent( d_.gc_, n_axis_unit );
        runit.setHeight( ext.y+2 );
        rtu.setRect( rttl );
        rtu.union( runit );
      }
      axis_unit = n_axis_unit;
  
      TruncateAxisTextData data = new TruncateAxisTextData();
      data.sttl_ = axis_title;
      data.sunit_= axis_unit;
      data.rttl_ = rttl;
      data.runit_= runit;
      data.rtu_  = rtu ;
      
      return data;
    }
  
  /** Render property on Y axis, support for "3D-like".
   * @param _get_step_unit_width true to compute step unit width on axis, 
   *                             false to locate/draw.
   * @param _R rectangle to render axis.
   * @param _axis axis to render.
   * @param _property property identifier to render: STEP_LINE,STEP_UNIT,STEP_DOT.
   * @param _3d_line_x simulated 3d on x axis
   * @param _3d_line_y simulated 3d on y axis
   */
  protected int onYAxis( 
        boolean _get_step_unit_width, Rect _R,
        DAxis _axis, String _property,
        int _3d_line_x, int _3d_line_y
      )
    throws ChartLocation  
  {
    int max_step_text_width=0;

    if( !_axis.getProperties().hasKey( _property ) ) return 0;

    float value = _axis.getProperties().get( _property, 0.0f );
    if( value == 0.0f ) return 0;
     
    float min     = _axis.getProperties().get( DAxis.P_MIN, -1.0f );
    float max     = _axis.getProperties().get( DAxis.P_MAX, +1.0f );
    float max_min = max-min ;

    //prevent div/0
    if ( max_min == 0.0f ) return 0;
    
    float wsu = _R.h()*value/max_min;
    int xd = _R.x()-margin_;
    //la numerotation, les lignes ... se fait en modulo value
    float vdn = (float)Math.ceil(min/value);
    if ( Math.abs(vdn)<1e-6 ) vdn=0.0f; 
    vdn *= value;
    float yd = _R.bottom() - _R.h()*(vdn-min)/max_min;

DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);

    for( float vd = vdn; vd <= max; vd += value )
    {
       if( DAxis.P_STEP_UNIT.equals( _property ) )
       {
              if ( Math.abs(vd)<1e-6 ) vd=0.0f;
              String t=df.format(vd);
              org.eclipse.swt.graphics.Point ext = d_.gc_.textExtent( t );
              int wt = ext.x;
              if( wt>max_step_text_width) max_step_text_width=wt;
              Rect r_txt=new Rect( xd-wt, (int)yd-ext.y/2, wt, ext.y );
              if( d_.drawing() )
              {
                if( !_get_step_unit_width ) //mode calcul largeur ?
                {
                  DrawUtil.drawText( d_.gc_, r_txt, DrawUtil.CENTER, t );
                  d_.gc_.drawLine( xd+margin_-2, (int)yd, xd+margin_, (int)yd );
                }
              }
              else //locating
              {
                if ( r_txt.contains( d_.lx_, d_.ly_ ) )
                {
                  throw new ChartLocation( DLocated.InAxisYStepUnitText, _axis, r_txt, t );
                }
              }
       }
       else if ( DAxis.P_STEP_DOT.equals( _property ) )
       {
              if(d_.drawing())
               d_.gc_.drawLine( xd+margin_-2, (int)yd, xd+margin_, (int)yd );
       }
       else if ( DAxis.P_STEP_LINE.equals( _property ) )
       {
//TBD:locating: detection des lignes, mais apres les courbes, alors que le dessin se fait avant ...
            if(d_.drawing()) 
            {
              d_.gc_.drawLine( _R.left(), (int)yd, _R.right(), (int)yd );
              if( _3d_line_x!=0 && _3d_line_y!=0 )
              {
                d_.gc_.drawLine( _R.x(), (int)yd, _R.x()-_3d_line_x, (int)yd+_3d_line_y );
              }
            }
       }
       yd -= wsu;
    }

    return max_step_text_width;
  }
  
  /** Render property on X axis, no support for "3D-like".
   * @param _R rectangle to render axis.
   * @param _axis axis to render.
   * @param _property property identifier to render: STEP_LINE,STEP_UNIT,STEP_DOT.
   */
  protected void onXAxis( Rect _R, DAxis _axis, String _property )
   throws ChartLocation
  {
    if( !_axis.getProperties().hasKey( _property ) ) return ;
    float value = _axis.getProperties().get( _property, 0.0f );
    if( value==0.0f ) return ;
    float min     = _axis.getProperties().get( DAxis.P_MIN, -1.0f);
    float max     = _axis.getProperties().get( DAxis.P_MAX, +1.0f );
    float max_min = max - min ;

    if ( max_min == 0.0f ) return; //sinon div/0
    
    float wsu = (_R.w()*value/max_min);
    int yd = _R.y()+_R.h()+margin_;
    //la numerotation, les lignes ... se fait en modulo prop_todo
    float vdn = (float)Math.ceil(min/value) ;
    if ( Math.abs(vdn)<1e-6 ) vdn=0.0f; // x*0 = 0
    vdn = vdn * value ;
    float xd = _R.left()+_R.w()*(vdn-min)/max_min;
    float vd;
    Rect r_txt=new Rect();
DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);    

    for( vd  = vdn;//_axis->property(DAxis::MIN);
         vd <= max;
         vd += value
       )
    {
      if( DAxis.P_STEP_UNIT.equals( _property ) )
      {
          if ( Math.abs(vd)<1e-6 ) vd=0.0f;        
          String t=df.format( vd );
          Point ext = d_.gc_.textExtent( t );
          r_txt.setRect( (int)xd-ext.x/2, yd, ext.x, ext.y );
          if(d_.drawing())
          {
            DrawUtil.drawText( d_.gc_, r_txt, DrawUtil.CENTER, t );
            d_.gc_.drawLine( (int)xd, yd-margin_, (int)xd, yd-margin_+2 );
          }
          else //locating
          {
            if ( r_txt.contains( d_.lx_, d_.ly_ ) )
            {
              throw new ChartLocation( DLocated.InAxisXStepUnitText, _axis, r_txt, t );
            }
          }
      }
      else if ( DAxis.P_STEP_DOT.equals( _property ) )
      {
          d_.gc_.drawLine( (int)xd, yd-margin_, (int)xd, yd-margin_+2 );
      }
      else if ( DAxis.P_STEP_LINE.equals( _property ) )
      {
          d_.gc_.drawLine( (int)xd, _R.top(), (int)xd, _R.bottom() );
//TBD:locating: la ligne, mais apres la detection des courbes ...
      }
      xd += wsu;
    }//for
  }
  
  /** draw or locate for an horizontal/vertical text.
   * @param _text text to render.
   * @param _rect rectangle to draw text
   * @param _align LEFT/RIGHT/HCENTER/... flag
   * @param _style style to paint
   * @param _locid identifier for location
   * @param _item item for location
   */
  public void renderText( String _text, Rect _rect, int _align, DChartStyle _style,
                          String _locid, IDItem _item )
    throws ChartLocation                           
  {
    if( d_.drawing() )
    {
      //la valeur est affichee dans le style de la courbe
      d_.gc_.setForeground( _style.fore_ );
      d_.gc_.setFont( _style.font_ );
      DrawUtil.drawText( d_.gc_, _rect, _align, _text );
    }
    else //locating
    {
      Rect nrtxt = DrawUtil.boundingRect( d_.gc_, _rect, _align, _text );
      //parce que si le texte deborde de _rect, .. ne le clippe pas
      //nrtxt might be bigger than _rect, clip it
      nrtxt.intersection( _rect );

      if( (nrtxt.w()>0) && nrtxt.contains( d_.lx_, d_.ly_ ) )
      {
        throw new ChartLocation( _locid, _item, nrtxt, _text );
      }
    }
  }
  
  /** Render Histogram graphic family */
  protected void renderHistogram( Rect _RG, boolean _3d, boolean _bs )
    throws ChartLocation
  {
    //histogram must have only one axis.
    DAxis axis=null;
    for( IDItem i=d_.G_.getFirstChild(); i!=null; i=i.getNext() )
    {
      if( i instanceof DAxis )
      {
        axis=(DAxis)i;
        break;
      }
    }

    if ( axis==null )
    {
      error( "Error: Histogram without axis");
      return ;
    }

    DChartStyle axis_style = styleOf( axis );
    d_.gc_.setFont( axis_style.font_ );
    int axis_font_h = d_.gc_.getFontMetrics().getHeight();

    int max_h_title_unit = axis_font_h + 2 ;

    //title et unit sont geree dans des chaines/rect different cela va permettre
    //de les afficher avec des styles differents ... un jours prochains.
    TruncateAxisTextData tatd =null;
 /*   String axis_title ;
    String axis_unit  ;
    Rectangle rttly; //rect du titre de l'axe
    Rectangle runity;//rect de l'unit de l'axe
    Rectangle rtuy; //rect englobant les deux premiers.
*/
    //reduction de la chaine de titre ou units..
//XXXX    //available_width est un height a cause de la rotation de 90
    tatd = truncateVAxisText( axis, axis_style, _RG.h(), max_h_title_unit );

    int yg = _RG.y()+ axis_font_h/2; //numerotation de l'axe
    int hg = _RG.h()
           - axis_font_h/2  // 1/2 hauteur aff. mesure axe Y (au cas ou)
           - margin_
           ;

    //mesure sur l'axe des Y (avec marge):
    int wuy=0;
    hg -= axis_font_h+margin_ ;
    wuy = onYAxis( true, new Rect(), axis, DAxis.P_STEP_UNIT, 0, 0 );
  
    int wg = _RG.w() ;
    int xg = _RG.x() ;

    //display Y axis title
    {
      int axis_title_w = 0;
      //largeur du titre + unite => hauteur max. de police (tourne a 90 degre)
      if( !DrawUtil.isEmpty( tatd.sttl_ )
       || !DrawUtil.isEmpty( tatd.sunit_ ) ) axis_title_w = max_h_title_unit;

      //titre de l'axe + unite en vertical a droite de l'axe des Y
      int decoration_w = wuy //aff. mesure axe des Y (max.)
                       + margin_ //entre wuy et max_h_title 
                       + axis_title_w // titre et/ou unite
                       + margin_ //entre max_h_... et bord 
                       ;
      wg -= decoration_w ;
      xg += decoration_w ;
    }

//TBD: 20 a afficher comme valeur de 3D
    int  _3d_z    = 20 ;
    if ( _3d )
    {
      wg -= _3d_z;
      hg -= _3d_z;
      yg += _3d_z;
    }


    Rect rg = new Rect(xg,yg,wg,hg);

    //lost of Y axis enlightment
    {
      //rect global de placement titre/unit Y
      //a note: a cause de la rotation on utilise la translation du QPainter
      // pour le placement correct, les calculs sont fait a origine 0,0
      Rect grtuy= new Rect( 0, 0, _RG.h(), max_h_title_unit );
      //centrer le rect titre+unit Y
      tatd.rtu_.moveCenter( grtuy.centerX(), grtuy.centerY() );
      //replacer rect title et unite
      tatd.rttl_.moveBy(  tatd.rtu_.x(), tatd.rtu_.y() );
      tatd.runit_.moveBy(  tatd.rtu_.x(), tatd.rtu_.y() );

      if( d_.drawing() )
      {
//        d_.gc_.save();
//TODO: he les gars SWT on fait ca comment ????
//        d_.gc_.translate( _RG.x, _RG.y+_RG.height );
//        d_.gc_.rotate( -90.0f );
//ye! SWT tu fait comment la rotation ???        
        d_.gc_.setForeground( axis_style.fore_ );
        d_.gc_.setFont( axis_style.font_ );
      
        DrawUtil.drawText( d_.gc_, tatd.rttl_, DrawUtil.CENTER, tatd.sttl_ );
        DrawUtil.drawText( d_.gc_, tatd.runit_, DrawUtil.CENTER, tatd.sunit_ );
       
   //     d_.gc_.restore();
      }
      else //locating
      {
        //on deplace le pt en (0,0) sur grtuy et on invers x/y:
        //pour prendre en compte la rotation
        int lx=d_.lx_;
        int ly=d_.ly_;
        //rotation di -90: on inverse x/-y w/h
        grtuy.setRect(  -grtuy.y(), grtuy.x(), grtuy.h(), +grtuy.w() );
        tatd.rttl_.setRect(  -tatd.rttl_.y(), tatd.rttl_.x(), tatd.rttl_.h(), +tatd.rttl_.w());
        tatd.runit_.setRect(  -tatd.runit_.y(), tatd.runit_.x(), tatd.runit_.h(), +tatd.runit_.w() );
        //translation:
        grtuy.moveBy(   _RG.x(), _RG.y() /*+_RG.height() */);
        tatd.rttl_.moveBy(  _RG.x(), _RG.y() /*+_RG.height() */);
        tatd.runit_.moveBy(  _RG.x(), _RG.y() /*+_RG.height() */);
        if( grtuy.contains( lx, ly ) )
        {
          //dans la zone de l'axe
          if( !DrawUtil.isEmpty(tatd.sttl_) && tatd.rttl_.contains( lx, ly ) )
          {
            //dans la zone de titre
            throw new ChartLocation( DLocated.InAxisYTitle, axis, tatd.rttl_, tatd.sttl_ );                    
          }
          if( !DrawUtil.isEmpty(tatd.sunit_) && tatd.runit_.contains( lx, ly ) )
          {
            //dans la zone de titre
            throw new ChartLocation( DLocated.InAxisYUnit, axis, tatd.runit_, tatd.sunit_);
          }
          //donc dans la zone de l'axe Y
          throw new ChartLocation( DLocated.InAxisYArea, axis, grtuy );
        }
      }

      //lines, unit, 'n dot sur l'axe des Y
      Rect rg2 = new Rect(rg);
      if(_3d) rg2.moveBy( _3d_z,-_3d_z );

      //drawing et:ou locating:
      if(d_.drawing())
      {
        d_.gc_.setForeground( getLightGray() );
      }
      if(_3d)
      {
        onYAxis( false, rg2, axis, DAxis.P_STEP_LINE, _3d_z, _3d_z );
      } else {
        onYAxis( false, rg2, axis, DAxis.P_STEP_LINE, 0, 0 );
      }
      if(d_.drawing())
      {
        d_.gc_.setForeground( getBlack() ); //black
      }
      onYAxis( false, rg, axis, DAxis.P_STEP_DOT, 0,0  );
      if(d_.drawing())
      {
        d_.gc_.setForeground( axis_style.fore_ );
        d_.gc_.setFont( axis_style.font_ );
      }
      onYAxis( false, rg, axis, DAxis.P_STEP_UNIT, 0,0 );
    }
  
    if(d_.drawing())
    {
      d_.gc_.setForeground( axis_style.fore_ );
      d_.gc_.drawLine( xg, yg,    xg,    yg+hg ); //axe Y
      d_.gc_.drawLine( xg, yg+hg, xg+wg, yg+hg ); //axe X
    }

    //dessin des courbes dans l'ordre
    float ay_min=axis.getProperties().get( DAxis.P_MIN,  0.0f),
          ay_max=axis.getProperties().get( DAxis.P_MAX, +1.123412345f);
    float ky = ay_max - ay_min;
System.err.println("HISTO axis["+axis+"].MIN="+ay_min);    
System.err.println("HISTO axis["+axis+"].MAX="+ay_max);
System.out.println("  PROP=["+DAxis.P_MAX+"]");    
    

    //if min==max, force a delta of 10%
    if( Math.abs(ky)<1e-6 )
    {
      ay_max = (ay_max>0.0f ? 1.05f : 0.95f )* ay_max;
      ay_min = (ay_min>0.0f ? 0.95f : 1.05f )* ay_min;
      ky = ay_max - ay_min;     
      if( Math.abs(ky)<1e-6)
      {
        //smell like update document problem (MIN/MAX=0.0f)
        error("Error: can't draw document not updated");
        return ;
      }
    }

    ky = hg/ky;
    int nbar=d_.curves_.size();

    float wbar=rg.w()/(float)nbar;
    float wbar_bs = (wbar+1)/3.0f; // _BS

    float xbar=rg.x();
    int   y_bar_zero;
    float bar_zero;
    if ( ay_min > 0.0f ) {
      y_bar_zero=rg.bottom();
        bar_zero=ay_min;
    } else if ( ay_max < 0.0f ) {
      y_bar_zero=rg.y();
        bar_zero=ay_max;
    }else {
      y_bar_zero=rg.bottom()-(int)(ky*(-ay_min)); //pos de 0.0
        bar_zero=0.0f;
    }

    IDItem item=null;
    ChartLocation loc=null;

DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits( 2 );

    for( int icrv=0; icrv<d_.curves_.size(); ++icrv )
    {
      CurveStyle ccurve = (CurveStyle)d_.curves_.get(icrv);    
      for( item=ccurve.curve_.getFirstChild(); item!=null; item=item.getNext() )
      {
        //les courbes n'ont qu'un point
//TBD:on pourrait imaginer des histogram multi point en superposant les bar.???
        if( item instanceof DPoint )
        {
          DPoint point = (DPoint)item;
          float vy;
          DCoord coord=point.getCoord( axis );          
          if ( coord !=null )
          {
            vy = coord.getValue();
            //transformation.
            float bh = vy-bar_zero;
            int  ibh = (int)(ky*(bh));
            if (ibh>0)ibh+=1;else if (ibh<0) ibh-=1;        
            //curve==BAR forcement.
            Rect rbar;
            if( _bs )  // bandes separees
            {
              int dx = (int)(xbar+0.5f*(wbar-wbar_bs));
              rbar = new Rect( dx, y_bar_zero, (int)wbar_bs, -ibh );
            } else {
              rbar = new Rect( (int)xbar, y_bar_zero, (int)wbar+1, -ibh );
            }
            rbar.normalize();

            //barre principale
            if(d_.drawing())
            {
              d_.gc_.setForeground( getBlack() ); //black ://trait exterieur en noir
              d_.gc_.setBackground( ccurve.style_.back_ );
              d_.gc_.fillRectangle( rbar.x(), rbar.y(), rbar.w(), rbar.h() );
              d_.gc_.drawRectangle( rbar.x(), rbar.y(), rbar.w(), rbar.h() );
            } 
            else //locating
            {
              if( rbar.contains( d_.lx_, d_.ly_ ) )
              {
                throw new ChartLocation( DLocated.CurveBar, point,  rbar );
              }
            }

//TBD: a revoir parce que les barre value>0 sont mal affiches en 3d
            if ( _3d )
            {
//TBD:20: c'est le Z 3d
              int ar1[] = new int[8],
                  ar2[] = new int[8];
                    
              ar1[0] = rbar.x();                ar1[1] = rbar.y();
              ar1[2] = rbar.x()+20;             ar1[3] = rbar.y()-20;
              ar1[4] = (int)(rbar.x()+wbar+20); ar1[5] = rbar.y()-20;
              ar1[6] = (int)(rbar.x()+wbar+1);  ar1[7] = rbar.y();
    
              ar2[0] = (int)(rbar.x()+wbar);    ar2[1] = rbar.y();
              ar2[2] = ar1[4];                ar2[3] = ar1[5]; 
              ar2[4] = (int)(rbar.x()+wbar+20); ar2[5] = rbar.y()+rbar.h()-20;
              ar2[6] = (int)(rbar.x()+wbar);    ar2[7] = rbar.y()+rbar.h();

              if(d_.drawing())
              {
                Color clr   = ccurve.style_.back_;
                Color dark1 = DrawUtil.lighter( clr, 0.50f );
                Color dark2 = DrawUtil.lighter( clr, 0.75f );
                d_.gc_.setBackground( dark1 );
                d_.gc_.fillPolygon( ar1 );
                d_.gc_.drawPolygon( ar1 );
                d_.gc_.setBackground( dark2 );
                d_.gc_.fillPolygon( ar2 );
                d_.gc_.drawPolygon( ar2 );
              }
              else //locating
              //Il faudrait parcourir les barres a l'envers pour ne pas 
              //se tromper lorsque les barres 3D se recouvrent ...
              //MAIS, je preferre y aller comme le dessin et choisir
              //la derniere zone 3d designee par la souris.
              {
//evidement:20 c'est la simili projection 3d
                //l'ideal c'est que l'on est un ar1.contains( ... )
                boolean in_ar1 = Data.horizontal3DContains( rbar.x(), rbar.y(),
                                          rbar.w(), 20, d_.lx_,d_.ly_ );
                boolean in_ar2 = Data.vertical3DContains( rbar.right(), rbar.y(),
                                         rbar.h(), 20, d_.lx_,d_.ly_ );
                //pas encore vu ...
                if( in_ar1 )
                {
                  //memo...
                  Rect br = new Rect( rbar.x(), rbar.y()-20, rbar.w()+20, 20 );
                  loc = new ChartLocation( DLocated.CurveBar, point, br );
                }
                else if( in_ar2 )
                {
                  Rect br = new Rect( rbar.right(), rbar.y()-20, 20, rbar.h()+20 );
                  loc = new ChartLocation( DLocated.CurveBar, point, br );
                }
                //sinon si on en avait trouve, c'est la derniere...
                else if( loc!=null )
                {
                  throw loc;
                }
              }
            }

            //value is displayed below the bar.
            {
              String full_txt = df.format( vy );
              d_.gc_.setFont( ccurve.style_.font_ );
              String txt = DrawUtil.truncateTextH( d_.gc_, full_txt, (int)wbar-2 );
              if( !DrawUtil.isEmpty( txt ) )
              {
                int font_h = d_.gc_.getFontMetrics().getHeight();
                Rect rtxt = new Rect( (int)xbar, rg.bottom()+margin_,
                                                (int)wbar-2, font_h+2 );
                //drawn using curve style
                renderText( txt, rtxt, DrawUtil.CENTER, ccurve.style_,
                             DLocated.CurvePointText, point );
              }
            }
          }
          else
          {
            error( "Error: Bar without value." );
          }

          //si plusieurs valeur... xbar += wbar;
        }//if 
      }//for 
      xbar += wbar;
    }//for
 
    //pour la derniere zone 3d vue...
    if( d_.locating() && loc!=null )
    {
      throw loc;
    }
  }
  
  /** Render sector-ish graphic */
  protected   void renderSector( Rect _RG, boolean _3d ) throws ChartLocation
  {
    boolean display_values = true; //true to display value of sector
    int d = _RG.w();
    //DR2358: je fais ce que je peux...avec le temps que j'ai.
    // Comme ca bave un peu au dessus surl le titre du graphique,
    // je vais reserver une zone de 2*20 pixels (dessus/dessous)
    // pour le texte (si display_decoration).
    // Ok, 20 c'est sorti de la cuisse de Jupiter, car a ce stade
    // on ne sait pas si il y aura vraiment du texte dessus/dessous
    // (il peut etre un peu decale), ni si ce texte a un style dont
    // la police est <= 20 pixels...
    // Vous aurez compris que c'est totalement du pif !
    // En attendant ca calme M'sieu Telep, en attendant qu'il trouve 
    // que les % se chevauchent quand le graphique est petit...
    int h_for_text = display_values ? 40 : 0;
    if ( d > _RG.h()-h_for_text ) d=_RG.h()-h_for_text;

    //si on affiche les texte:
    if ( display_values )
    {
//TBD:il y a surement un calcul plus serieux que 3/4 !!!
      d = 3*d/4; //les secteurs occupent 3/4 de la place (prevoir pr texte)
    }

    float v_sum = 0.0f;

    CurveStyle ccurve=null;
    IDItem item=null;
    
    DAxis axis=null;
    for( item=d_.G_.getFirstChild(); item!=null; item=item.getNext() )
    {
      if( item instanceof DAxis ) axis=(DAxis)item;
    }
    if( axis==null )
    {
      error("Error: missing Axis in graphic");
      return;
    }
    String axis_unit = axis.getUnit();
    if( axis_unit==null ) axis_unit="";

    //cumul des valeurs des courbes (pour correspondance au 100% du cercle)
    int n_points=0;
    for( int i=0; i<d_.curves_.size(); ++i )
    {
      ccurve = (CurveStyle)d_.curves_.get(i);
      //courbes: plusieurs point a une seule coord possibles
      for( item=ccurve.curve_.getFirstChild(); item!=null; item=item.getNext())
      {
        if ( item instanceof DPoint )
        {
          DCoord coord= ((DPoint)item).getCoord( axis );
          if ( coord!=null )
          {
            v_sum += coord.getValue();
            n_points++;
          }
        } 
      }
    }

    //v_sum represente 100% du secteur.
    if ( v_sum == 0.0f )
    {
      error( "Error: can't print sectors (sum:0)" );
      return ;
    }

    int x3d = 0, y3d = 0, r3d=0, a3d=0;

    if ( _3d )
    {
      //parce que c'est a peut pres correct
      if( d >= 48 ) //ombre 4 fois moins importante que le pie chart
      {
        x3d = - 8;
        y3d = +12;
      } else {
        y3d = d / 4;
        x3d = -(int)(2.0f*y3d/3.0f);
        // on voit rien: circulez y'a rien a voir
        if( Math.abs(x3d) < 2 ) //au moins 2 pixels d'ombrage
        {
          _3d = false;
          x3d = y3d = 0;
        }
      }
      if( _3d )
      {
        r3d = (int)Math.sqrt( (double)(x3d*x3d + y3d*y3d) );
        float a = (float)Math.atan2( (float)y3d, (float)x3d ); //-PI..PI
        if( a < 0.0f ) a += 2*Math.PI ; 
        //a est dans 0..2PI 
        a3d = (int)(180*a/(Math.PI));
      }
    }

    LineAlg l3d = new LineAlg( 0, 0, x3d, y3d );

    //locating: les sector sont detectes dans l'ordre de dessin,
    //donc on les teste tous, et c'est le dernier qui gagne
    ChartLocation loc_sector=null;

DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);

    //because line separating two sector must be drawn after sector
    //otherwise fill next sector will overwrite line.
    //this array store x,y coordinates of point to draw joined to center or pie.
    int between_sectors[] = new int[n_points*2];
    int wbs=0;
    Rect r = new Rect();
    
    while( l3d.nextPoint() )
    {
      int dx3d = l3d.x();
      int dy3d = l3d.y();
      boolean top=(dx3d==x3d)&&(dy3d==y3d);

      int angle=0;
      int ray = d >> 1; 

      r.setRect( _RG.x()+(_RG.w() -d)/2 +dx3d,
                 _RG.y()+(_RG.h() -d)/2 -dy3d, d,d ); 
      if( r.w()==0 || r.h() == 0 ) continue;

      //locating: (ici pour optimiser le calcul)
      float loc_dist=-1.0f;
      int loc_angle=-1; //en 16eme de degre
      if(d_.locating())
      {
        Point rc=r.center();
        int dx=d_.lx_-rc.x, dy=d_.ly_-rc.y;
        loc_dist=(float)Math.sqrt( dx*dx + dy*dy );
        if(loc_dist<=ray)
        {
          float a = (float)Math.atan2( (float)-dy, (float)dx ); //-PI..PI
          if( a < 0.0f ) a += 2*Math.PI ; 
          //a est dans 0..2PI 
          loc_angle = (int)(360*a/(2*Math.PI));
        }
      }
      
      int ipt=0;
      for( int i=0; i<d_.curves_.size(); ++i )
      {
        ccurve = (CurveStyle)d_.curves_.get(i);
        //Courbes: plusieurs points possibles
        for( item=ccurve.curve_.getFirstChild(); item!=null; item=item.getNext() )
        {
          if ( item instanceof DPoint )
          {
            DPoint point = (DPoint)item;
            DCoord coord = point.getCoord( axis );
            if ( coord !=null )
            {
              ipt++;
              float v = coord.getValue()/v_sum; //% du disque
              int end_angle = angle+(int)(360*v); //degree;
              if( ipt >= n_points ) //last sector ?
              {
                end_angle=360;
              }

              if( d_.drawing() )
              {
                if(top)
                {
                  d_.gc_.setForeground( getBlack() ); //black
                  Color b = ccurve.style_.back_;
                  d_.gc_.setBackground( b );
                  d_.gc_.fillArc( r.x(),r.y(),r.w(),r.h(), angle, end_angle-angle );
                  d_.gc_.drawArc( r.x(),r.y(),r.w(),r.h(), angle, end_angle-angle );
                  int cx = r.centerX(), cy=r.centerY();
                  double a = (Math.PI*angle)/180.0f;
                  between_sectors[wbs++]= (int)( cx+r.w()*Math.cos( a )/2.0f );
                  between_sectors[wbs++]= (int)( cy-r.h()*Math.sin( a )/2.0f );                  
                } else {
                  Color dark = DrawUtil.lighter( ccurve.style_.back_, 0.75f );
                  d_.gc_.setForeground( dark );
                  d_.gc_.setBackground( dark );
                  d_.gc_.fillArc( r.x(),r.y(),r.w(),r.h(), angle, end_angle-angle );
                }
              }
              else //locating
              {
                if( (loc_dist <= ray)
                    &&(loc_angle>=angle)&&(loc_angle<=end_angle) )
                {
                  //memo pour connaitre le don secteur a la fin ..
                  loc_sector = new ChartLocation( DLocated.CurveSector, point, r );
                  ((DChartLocation)loc_sector.location_).setAngles( angle, end_angle );
                }
              }
              

              //affichage de la valeur + unite de l'axe
              if( top && display_values )
              {
                int atxt = angle+(end_angle-angle)/2;
                //calculer la compensation sur le rayon en mode _3d pour
                //que le texte ne recouvre pas l'ombre:
                int r3=0;
                if ( _3d )
                {
                  /* aller, j'en dis un peu plus:
                   * en 3d on va ajouter un rayon pour ne pas afficher le texte
                   * sur l'ombre, le + c'est que ce rayon depend de l'angle
                   * du texte ET de l'angle choisit pour l'ombrage, eh, eh, eh!
                   * C'est pas beau ca !
                   */
                  int da = (int)Math.abs( atxt-a3d-180 );
                  //compenser que dans la zone de l'ombre:
                  if( da < 90 ) //PI/2
                  {
                    //approx du 1er degre:
                    float k = 1.0f - da/90.f;
                    r3 = (int)(k * r3d);
                  }
                }
                float arad=(float)( 2*Math.PI-(2*Math.PI*atxt)/360 );
                int sector_ray = ray+margin_+r3 ;
                int tx = (int)(r.x()+(r.w() >>1)+sector_ray*Math.cos(arad));
                int ty = (int)(r.y()+(r.h()>>1)+sector_ray*Math.sin(arad));

                String txt = df.format( coord.getValue() ) + axis_unit;

                Point te = d_.gc_.textExtent( txt );
                int tw = te.x;
                int th = te.y;

                //on essaye de placer le texte au mieux
                     if ( atxt < 90  ) {}/*raf*/
                else if ( atxt < 180 ) { tx -= tw; }
                else if ( atxt < 270 ) { tx -= tw; ty+=th; }
                else                    {          ty+=th; };

                Rect rtxt = new Rect( tx, ty-th, tw, th );

                renderText( txt, rtxt,  DrawUtil.LEFT|DrawUtil.BOTTOM,
                             ccurve.style_, DLocated.CurvePointText,
                             point );
                if( loc_sector!=null ) throw loc_sector;
              }

              angle = end_angle;
            }
          }//if
        }//item
      }//curve
    }//3d
    
    if( d_.drawing() )
    {
      d_.gc_.setForeground( getBlack() ); //black
      int cx = r.centerX(), cy=r.centerY();
      for( int i=0; i<between_sectors.length; )
      {
        int x = between_sectors[i++];
        int y = between_sectors[i++];
        d_.gc_.drawLine( cx,cy, x,y );
      }
    }
    else //locating
    {
      if( loc_sector!=null ) throw loc_sector;
    }
  }

  /** Render sector-ish graphic */
  protected   void renderXY( Rect _RG ) throws ChartLocation
  {
    //decorations means axis titles/units/...
    boolean display_decorations = true;
    
    //search for axis "x" and "y", must exists
    DAxis x_axis=null, y_axis=null;
    for( IDItem item=d_.G_.getFirstChild(); item!=null; item=item.getNext() )
    {
      if( item instanceof DAxis )
      {
        DAxis axis = (DAxis)item;
        if( x_axis==null && "x".equals( axis.getName() ) )
        {
          x_axis = axis;
        }
        else if ( y_axis==null && "y".equals( axis.getName() ))
        {
          y_axis = axis;
        }
        if( x_axis!=null && y_axis!=null ) break;
      }
    }

    if ( x_axis==null || y_axis==null )
    {
      String txt;
      if( x_axis==null )
      {
        txt = "Error: missing 'x' axis";
      } else {
        txt = "Error: missing 'y' axis";
      }
      error( txt );
      return ;
    }

    DChartStyle x_axis_style = styleOf( x_axis );
    DChartStyle y_axis_style = styleOf( y_axis );

    d_.gc_.setFont( x_axis_style.font_ );
    int x_axis_style_h = d_.gc_.textExtent( "X" ).y+2;
    d_.gc_.setFont( y_axis_style.font_ );
    int y_axis_style_h = d_.gc_.textExtent( "X" ).y+2;
//evidement c'est sorti de la cuisse de jupiter...    

    //metrics titre ( unit ) axe des X pour les chaines/rect
    int xg,yg,hg,wg;

    xg = _RG.x();
    yg = _RG.y();
    hg = _RG.h();
    wg = _RG.w();

DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);
    if ( display_decorations )
    {  
      d_.gc_.setFont( y_axis_style.font_ );
      //compute steps width ont Y Axis...
      int wuy= onYAxis( true,  new Rect(), y_axis, DAxis.P_STEP_UNIT, 0,0 );

      //it's a worth measure.
      int wux=0;
      if ( x_axis.getProperties().hasValue(DAxis.P_STEP_UNIT))
      {
        String s = df.format( x_axis.getProperties().get(DAxis.P_MAX, +1.0f) );
        d_.gc_.setFont( x_axis_style.font_ );
        wux = d_.gc_.textExtent( s ).x;
      }

//Note: title and unit of Y axis are drawn verticaly
      int y_axis_title_unit_w = 0;
      if ( !DrawUtil.isEmpty( y_axis.getTitle())
        || !DrawUtil.isEmpty( y_axis.getUnit()) )
      {
//TODO: SWT sait pas faire le 90 !      
//TODO:pas de rotation:           y_axis_title_unit_w = y_axis_style_h ; //texte tourne a 90 degre.
        y_axis_title_unit_w = d_.gc_.textExtent( "X" ).x+2;           
      }

      wg -= y_axis_title_unit_w // titre ( unite ) axe des Y ecris en vertical
          + margin_         // marge en titre et mesure axe des Y
          + wuy             // aff. mesure axe des Y (maximum)
          + margin_         // entre aff mesure axe des Y et l'axe des Y
          + wux/2           // au pire on doit afficher la mesure maxi (axe X)
          ;
     
      xg += y_axis_title_unit_w
         + margin_
          + wuy
          + margin_
          ;

      int x_axis_title_unit_h = 0;
      if ( !DrawUtil.isEmpty( x_axis.getTitle() )
        || !DrawUtil.isEmpty( x_axis.getUnit() ) )
           x_axis_title_unit_h = y_axis_style_h;

      hg -= y_axis_style_h/2  // 1/2 hauteur aff. mesure axe Y (au cas ou)
          + margin_           // marge en axe X et affichage mesure des X
          + x_axis_style_h    // mesure des X  
          + margin_           // entre mesure des X et titre axe X
          + x_axis_title_unit_h // titre (unite) axe des X
          ;

      yg += y_axis_style_h/2 ; //1/2 nombre possible en haut axe des Y

      Rect rg = new Rect(xg,yg,wg,hg);

      //X Axis:
      TruncateAxisTextData tatdx
            = truncateHAxisText( x_axis, x_axis_style, wg, x_axis_style_h );

      //rect global de placement titre/unit X pour recenter les rect...
      Rect grtux=new Rect( xg, _RG.bottom()-margin_-x_axis_style_h, wg, x_axis_style_h );
      //centrer dans grtux
      tatdx.rtu_.moveCenter( grtux.center() );
      //donc:
      tatdx.rttl_.moveBy( tatdx.rtu_.x(), tatdx.rtu_.y() );
      tatdx.runit_.moveBy( tatdx.rtu_.x(), tatdx.rtu_.y() );

      d_.gc_.setFont( x_axis_style.font_ );
      if( d_.drawing() )
      {
        d_.gc_.setForeground( x_axis_style.fore_ );
     
        DrawUtil.drawText( d_.gc_, tatdx.rttl_, DrawUtil.CENTER, tatdx.sttl_ );
        DrawUtil.drawText( d_.gc_, tatdx.runit_, DrawUtil.CENTER, tatdx.sunit_ );

        //lines, unit, 'n dot sur l'axe des X
        d_.gc_.setForeground( getLightGray() ); 
//TBD:render
        onXAxis( rg, x_axis, DAxis.P_STEP_LINE );
        d_.gc_.setForeground( getBlack() );
        onXAxis( rg, x_axis, DAxis.P_STEP_DOT  );
        d_.gc_.setForeground ( x_axis_style.fore_  );
        d_.gc_.setFont( x_axis_style.font_ );
        onXAxis( rg, x_axis, DAxis.P_STEP_UNIT );
      }
      else //locating
      {
        if( grtux.contains( d_.lx_, d_.ly_ ) )
        {
          //inside axis area...
          if( !DrawUtil.isEmpty( tatdx.sttl_ ) && tatdx.rttl_.contains( d_.lx_, d_.ly_ ) )
          {
            //inside title area
            throw new ChartLocation( DLocated.InAxisXTitle, x_axis, tatdx.rttl_, tatdx.sttl_ );
          }
          if( !DrawUtil.isEmpty( tatdx.sunit_ ) && tatdx.runit_.contains( d_.lx_, d_.ly_ ) )
          {
            //dans la zone de titre
            throw new ChartLocation( DLocated.InAxisXUnit, x_axis,
                                      tatdx.runit_, tatdx.sunit_ );
          }
          //so, in axis area.
          throw new ChartLocation( DLocated.InAxisXArea, x_axis, 
                                    grtux, tatdx.sttl_+tatdx.sunit_ );
        }

        //locating unit step text ...
        onXAxis( rg, x_axis, DAxis.P_STEP_UNIT );
      }

      // Axe Y:
      TruncateAxisTextData tatdy
        = truncateVAxisText( y_axis, y_axis_style, xg-_RG.x(), hg );

      //rect global de placement titre/unit Y
//TODO: pas de text verticaux dans SWT.      
      //a note: a cause de la rotation on utilise la translation du QPainter
      // pour le placement correct, les calculs sont fait a origine 0,0
 //     Rect grtuy=new Rect( yg, 0, hg, y_axis_style_h );
 Rect grtuy = new Rect( 0, yg, xg-_RG.x(), hg ); //y_axis_style_h, hg );
      //centrer le rect titre+unit Y
      tatdy.rtu_.moveCenter( grtuy.center() );

      //replace title and unit rectangle (remember: drawn verticaly)
      tatdy.rttl_ .setLocation( tatdy.rtu_.x(), tatdy.rtu_.y() );
      tatdy.runit_.setLocation( tatdy.rtu_.x(), tatdy.rtu_.bottom()-tatdy.runit_.h() );

      d_.gc_.setFont( y_axis_style.font_ );
      if( d_.drawing() )
      {
//TODO: pas de texte verticaux...        
/*        d_->P_->save();
        d_->P_->translate( _RG.left(), _RG.top()+_RG.height() );
        d_->P_->rotate( -90.0f );
*/        
        d_.gc_.setForeground( y_axis_style.fore_ );

        DrawUtil.drawText( d_.gc_, tatdy.rttl_,  DrawUtil.CENTER|DrawUtil.VERTICAL, tatdy.sttl_ );
        DrawUtil.drawText( d_.gc_, tatdy.runit_, DrawUtil.CENTER|DrawUtil.VERTICAL, tatdy.sunit_  );
//        d_->P_->restore();

        //lines, unit, 'n dot sur l'axe des Y
        d_.gc_.setForeground( getLightGray() ); 
        onYAxis( false, rg, y_axis, DAxis.P_STEP_LINE, 0,0 );
        d_.gc_.setForeground( getBlack() );
        onYAxis( false, rg, y_axis, DAxis.P_STEP_DOT, 0,0  );
        d_.gc_.setForeground( y_axis_style.fore_  );
        d_.gc_.setFont( y_axis_style.font_  );
        onYAxis( false, rg, y_axis, DAxis.P_STEP_UNIT, 0,0 );
      } 
      else //locating
      {
//TODO: pas de texte verticaux, locating faux        
        //on deplace le pt en (0,0) sur grtuy et on invers x/y:
        //pour prendre en compte la rotation
        int lx=d_.lx_;
        int ly=d_.ly_;
        //rotation di -90: on inverse x/-y w/h
        grtuy.setRect( -grtuy.y(), grtuy.x(), grtuy.h(), +grtuy.w() );
        tatdy.rttl_.setRect( -tatdy.rttl_.y(), tatdy.rttl_.x(), tatdy.rttl_.h(), +tatdy.rttl_.w() );
        tatdy.runit_.setRect( -tatdy.runit_.y(), tatdy.runit_.x(), tatdy.runit_.h(), +tatdy.runit_.w() );
        //translation:
        grtuy.moveBy(  _RG.left(), _RG.top() /*+_RG.height() */);
        tatdy.rttl_.moveBy(  _RG.left(), _RG.top() /*+_RG.height() */);
        tatdy.runit_.moveBy(  _RG.left(), _RG.top() /*+_RG.height() */);
        if( grtuy.contains( lx, ly ) )
        {
          //in axis area... 
          if( !DrawUtil.isEmpty( tatdy.sttl_ ) && tatdy.rttl_.contains( lx, ly ) )
          {
            //in axis title area
            throw new ChartLocation( DLocated.InAxisYTitle, y_axis,
                                      tatdy.rttl_, tatdy.sttl_ );
          }
          if( !DrawUtil.isEmpty( tatdy.sunit_ ) && tatdy.runit_.contains( lx, ly ) )
          {
            //in unit text area
            throw new ChartLocation( DLocated.InAxisYUnit, y_axis,
                                      tatdy.runit_, tatdy.sunit_ );
          }
          //so in axis area...
          throw new ChartLocation( DLocated.InAxisYArea, y_axis, 
                                    grtuy, tatdy.sttl_+tatdy.sunit_ );
        }

        //locating unit step text ...
        onYAxis( false, rg, y_axis, DAxis.P_STEP_UNIT, 0,0 );
      }
    }

    //graph rect without axis
    Rect rg = new Rect(xg,yg,wg,hg);

    if( d_.drawing() )
    {
      d_.gc_.setForeground( getBlack() );
      d_.gc_.drawLine( xg, yg+hg, xg+wg, yg+hg ); // X axis
      d_.gc_.drawLine( xg, yg,    xg,    yg+hg ); // Y axis
    }

    //draw curves in ordre.
    float x_min = x_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
          x_max = x_axis.getProperties().get( DAxis.P_MAX, -1.0f ),
          y_min = y_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
          y_max = y_axis.getProperties().get( DAxis.P_MAX, +1.0f );
    float kx = x_max - x_min;
    float ky = y_max - y_min;
    if ( (kx==0.0f)||(ky==0.0f) )
    {
      error( "Error: can't draw curves (divison by 0)" );
      return;
    }
  
    kx = wg/kx;
    ky = hg/ky;
    //basis for histogram bar (xy curve might be bars)
    int y_bar_zero ;
    float bar_zero;

    //except if we haven't 0
    if ( y_min > 0.0f ) {
      y_bar_zero = rg.bottom(); //DAxis.MIN
        bar_zero = y_min;
    } else if ( y_max < 0.0f ) {
      y_bar_zero = rg.top(); //DAxis.MAX
        bar_zero = y_max;
    } else {
      //position of 0on Y axis:
      y_bar_zero = rg.bottom()-(int)(ky*(-y_min));
        bar_zero = 0.0f;
    }
     
    int last_x=-1, last_y=-1; 
    float  last_vy=-1;
    CurveStyle ccurve=null;
    Rect rp=new Rect();
    Rect r_bar=new Rect();

    //for locating mode, curve are checking in reverse order to handle covering.
    int ic_inc = d_.locating() ? -1 : +1;
    int ic_size = d_.curves_.size();
    for( int ic=d_.locating() ? ic_size-1 : 0;
         d_.locating()? (ic>=0) : (ic<ic_size);
         ic += ic_inc )
    {
      ccurve = (CurveStyle)d_.curves_.get(ic);      
      boolean first=true;
      IDItem last_item=null;
      DPoint point=null;
      for( IDItem item=ccurve.curve_.getFirstChild(); item!=null; item=item.getNext() )
      {
        if ( item instanceof DPoint )
        {
          point = (DPoint)item;
          DCoord cx = point.getCoord( x_axis );
          DCoord cy = point.getCoord( y_axis );
          if ( cx!=null && cy!=null )
          {
            float vx = cx.getValue();
            float vy = cy.getValue();
            //transformation.
            int px = xg+(int)(kx*(vx-x_min));
            int py = yg+hg-(int)(ky*(vy-y_min));
            if ((!first)|| ccurve.curve_.isType( DCurve.T_POINTS ) )
            {
              if( ccurve.curve_.isType( DCurve.T_LINE ) )
              {
                if( d_.drawing() )
                {
                  d_.gc_.setForeground( ccurve.style_.back_ );
                  //d_->P_->setBrush(Qt::NoBrush);
                  d_.gc_.drawLine( last_x, last_y, px,py );
                }
                else //locating
                {
                  if( DrawUtil.segmentContains( last_x, last_y, px, py, d_.lx_, d_.ly_ ) )
                  {
                    throw new ChartLocation( DLocated.CurveLine, point, new Rect(last_x,last_y,px-last_x,py-last_y));
                  }
                }
              }
              else if ( ccurve.curve_.isType( DCurve.T_POINTS ))
              {
//TBD:choix de 6x6 pour la dimension des points
                rp.setRect( px-3, py-3, 6, 6 );
                if( d_.drawing() )
                {
                  d_.gc_.setForeground( getBlack() );
                  d_.gc_.setBackground( ccurve.style_.back_ );
                  d_.gc_.fillRectangle( rp.x(), rp.y(), 6,6 );
                  d_.gc_.drawRectangle( rp.x(), rp.y(), 6,6 );
                }
                else //locating
                {
                  if( rp.contains( d_.lx_, d_.ly_ ) )
                  {
                     throw new ChartLocation( DLocated.CurvePoint, point, rp );
                  }
                }
              }
              else if ( ccurve.curve_.isType( DCurve.T_BAR ))
              {   
                float bh=last_vy-bar_zero;
                int  ibh=(int)(ky*(bh));
                if (ibh>0)ibh+=1;else if (ibh<0) ibh-=1;
                r_bar.setRect( last_x, y_bar_zero, px-last_x+1, -ibh );
                r_bar.normalize();
/* cf ci-dessus               if(r_bar.h()<0) 
                {
                     //redresser le rectangle sinon contains marche pas
                     r_bar.setRect( r_bar.x(), r_bar.y()+r_bar.height(),
                                    r_bar.width(), -r_bar.height() );
                   }*/
                if( d_.drawing() )
                {
                  d_.gc_.setForeground( getBlack() );
                  d_.gc_.setBackground( ccurve.style_.back_ );
                  d_.gc_.fillRectangle( r_bar.x(), r_bar.y(), r_bar.w(), r_bar.h() );
                  d_.gc_.drawRectangle( r_bar.x(), r_bar.y(), r_bar.w(), r_bar.h() );
                }
                else //locating
                {
                  //on s'arrete ici parce que l'on recouvre tout...
                  if( r_bar.contains( d_.lx_, d_.ly_ ) )
                  {
                    throw new ChartLocation( DLocated.CurveBar, last_item, r_bar );
                  }
                }
              }
            } 
            else
            {
              first=false;
            }
           last_x = px;
           last_y = py;
           last_vy= vy;
         }
         else
         {
error( "Warning:Point without (x,y) coord.");
          }
       }
 
       last_item=item;
     }//for 

      //last histogram bar
//TBD: c'est pas terrible, mais sinon ca veut dire que la derniere
// valeur d'une courbe histo d'a de sens que pour X (termine la valeur precedente).
     if ( !first && (ccurve.curve_.isType( DCurve.T_BAR ) ))
     {
       int px=rg.right(); //on force jusqu'a la fin du graphe
       r_bar.setRect( last_x-1, last_y, px-last_x+1, rg.bottom()-last_y+1 );
       r_bar.normalize();
       if( d_.drawing() )
       {
         d_.gc_.setForeground( getBlack() );
         d_.gc_.setBackground( ccurve.style_.back_);
         d_.gc_.fillRectangle( r_bar.x(), r_bar.y(), r_bar.w(), r_bar.h() );
         d_.gc_.drawRectangle( r_bar.x(), r_bar.y(), r_bar.w(), r_bar.h() );
       }
       else //locating
       {
         if( r_bar.contains( d_.lx_, d_.ly_ ) )
         {
           throw new ChartLocation( DLocated.CurveBar, point, r_bar );
         }
       }
     }

    }//for
  }
  
  private static Color c_black_=null;
  protected Color getBlack()
  {
    if( c_black_==null)
    {
      c_black_=new Color(null,0,0,0);
    }
    return c_black_;
  }
  
  private static Color c_light_gray_=null;
  protected Color getLightGray()
  {
    if( c_light_gray_==null )
    {
      c_light_gray_ = new Color( null, 160,160,160);
    }
    return c_light_gray_;
  }
  
  private static Color c_red_ = null;
  protected Color getRed()
  {
    if( c_red_==null )
    {
      c_red_ = new Color( null, 255,0,0);
    }
    return c_red_;
  }
  
  private static class ForceCoord implements ICoordTransformation
  {
    public float value_ = 0.0f;
    public float apply( float _f ) { return value_; }
  }
  private static ForceCoord force_coord = new ForceCoord();
  
  protected void drawBox3D( View3D _view, Box3D _bb )
  {
    Point2D p[] = new Point2D[8];
    float xn = _bb.getXMin(), xx = _bb.getXMax();
    float yn = _bb.getYMin(), yx = _bb.getYMax();
    float zn = _bb.getZMin(), zx = _bb.getZMax();
    p[0] = _view.projection( xn,yn,zn, new Point2D() );
    p[1] = _view.projection( xn,yn,zx, new Point2D() );
    p[2] = _view.projection( xx,yn,zx, new Point2D() );
    p[3] = _view.projection( xx,yn,zn, new Point2D() );
    p[4] = _view.projection( xn,yx,zx, new Point2D() );
    p[5] = _view.projection( xx,yx,zx, new Point2D() );
    p[6] = _view.projection( xx,yx,zn, new Point2D() );
    p[7] = _view.projection( xn,yx,zn, new Point2D() );

    d_.gc_.drawLine( (int)p[0].getX(),(int)p[0].getY(), (int)p[1].getX(),(int)p[1].getY());
    d_.gc_.drawLine( (int)p[0].getX(),(int)p[0].getY(), (int)p[3].getX(),(int)p[3].getY());
    d_.gc_.drawLine( (int)p[0].getX(),(int)p[0].getY(), (int)p[7].getX(),(int)p[7].getY());
    d_.gc_.drawLine( (int)p[2].getX(),(int)p[2].getY(), (int)p[1].getX(),(int)p[1].getY());
    d_.gc_.drawLine( (int)p[2].getX(),(int)p[2].getY(), (int)p[3].getX(),(int)p[3].getY());
    d_.gc_.drawLine( (int)p[2].getX(),(int)p[2].getY(), (int)p[5].getX(),(int)p[5].getY());
    d_.gc_.drawLine( (int)p[4].getX(),(int)p[4].getY(), (int)p[1].getX(),(int)p[1].getY());
    d_.gc_.drawLine( (int)p[4].getX(),(int)p[4].getY(), (int)p[5].getX(),(int)p[5].getY());
    d_.gc_.drawLine( (int)p[4].getX(),(int)p[4].getY(), (int)p[7].getX(),(int)p[7].getY());
    d_.gc_.drawLine( (int)p[6].getX(),(int)p[6].getY(), (int)p[3].getX(),(int)p[3].getY());
    d_.gc_.drawLine( (int)p[6].getX(),(int)p[6].getY(), (int)p[5].getX(),(int)p[5].getY());
    d_.gc_.drawLine( (int)p[6].getX(),(int)p[6].getY(), (int)p[7].getX(),(int)p[7].getY());
  }
  
  //identifiers for axis X,Y,Z...
  protected static final int AxisX = 1;
  protected static final int AxisY = 2;
  protected static final int AxisZ = 3;
  
  protected String axis3DName( DAxis _axis )
  {
    String title = _axis.getTitle();

    if( title != null )
    {
      title = title + _axis.getUnit();
    } else {
      title = _axis.getUnit();
    }
  
    return title ;
  }
  
  protected void renderAnAxis3D( 
      View3D _view, Box3D _bb, DAxis _axis, DChartStyle _axis_style, int _axis_type
    )
    throws ChartLocation
  {
    float x_proj = _bb.getXMin(), x_proj2 = _bb.getXMax(),
          y_proj = _bb.getYMin(), y_proj2 = _bb.getYMax(),
          z_proj = _bb.getZMin(), z_proj2 = _bb.getZMax();

    float coord_min, coord_max;
    switch( _axis_type )
    {
      case AxisX: coord_min = _bb.getXMin(); coord_max = _bb.getXMax(); break;
      case AxisY: coord_min = _bb.getYMin(); coord_max = _bb.getYMax(); break;
      case AxisZ: coord_min = _bb.getZMin(); coord_max = _bb.getZMax(); break;
      default: return ;
    }

    Point2D p0 = new Point2D(), 
            p1 = new Point2D(),
            p2 = new Point2D();

    /* 
     * Lignes des reperes
     */
    if ( d_.drawing() && _axis.getProperties().hasValue( DAxis.P_STEP_LINE ) )
    {
      float coord_step = _axis.getProperties().get( DAxis.P_STEP_LINE, 1.0f );
      if ( coord_step != 0.0f )
      {
        float coord;
        float coord_start = (float)( coord_step*Math.ceil(coord_min/coord_step) );

        d_.gc_.setForeground( getLightGray() );

                
        //ca a l'air plus complique, mais comme ca on fait le switch une
        //seule fois par for...
        switch( _axis_type )
        {
          case AxisX: for( coord = coord_start; coord < coord_max; coord += coord_step )
                      {
                        _view.projection( coord, y_proj2, z_proj, p0 );
                        _view.projection( coord, y_proj, z_proj, p1 );
                        _view.projection( coord, y_proj, z_proj2, p2 );
          
                        d_.gc_.drawLine( (int)p0.getX(), (int)p0.getY(), (int)p1.getX(), (int)p1.getY() );
                        d_.gc_.drawLine( (int)p2.getX(), (int)p2.getY(), (int)p1.getX(), (int)p1.getY() );
                      }
                      break;
          case AxisY: for( coord = coord_start; coord < coord_max; coord += coord_step )
                      {
                        _view.projection( x_proj2, coord, z_proj, p0 );
                        _view.projection( x_proj,  coord, z_proj, p1 );
                        _view.projection( x_proj,  coord, z_proj2, p2 );
          
                        d_.gc_.drawLine( (int)p0.getX(), (int)p0.getY(), (int)p1.getX(), (int)p1.getY() );
                        d_.gc_.drawLine( (int)p2.getX(), (int)p2.getY(), (int)p1.getX(), (int)p1.getY() );
                      }
                      break;
          case AxisZ: for( coord = coord_start; coord < coord_max; coord += coord_step )
                      {
                        _view.projection( x_proj2, y_proj,  coord, p0 );
                        _view.projection( x_proj,  y_proj,  coord, p1 );
                        _view.projection( x_proj,  y_proj2, coord, p2 );
          
                        d_.gc_.drawLine( (int)p0.getX(), (int)p0.getY(), (int)p1.getX(), (int)p1.getY() );
                        d_.gc_.drawLine( (int)p2.getX(), (int)p2.getY(), (int)p1.getX(), (int)p1.getY() );
                      }
                      break;
        }
      }
    }

    /*
     * Graduation (texte) de l'axe 
     */
    String axis_title = axis3DName( _axis );
    boolean have_step_unit = false;
    
    if ( _axis.getProperties().hasValue( DAxis.P_STEP_UNIT )
      || !DrawUtil.isEmpty( axis_title )
       )
    {
      have_step_unit = true;
      float d_angle = 0.0f;
      switch( _axis_type )
      {
        case AxisX:
            _view.projection( _bb.getXMin(),_bb.getYMin(),_bb.getZMin(), p0 );
            _view.projection( _bb.getXMax(),_bb.getYMin(),_bb.getZMin(), p1 );
            d_angle = 180;
          break;
        case AxisY:
             _view.projection( _bb.getXMin(),_bb.getYMin(),_bb.getZMin(), p0 );
             _view.projection( _bb.getXMin(),_bb.getYMax(),_bb.getZMin(), p1 );
             d_angle = 0;
           break;
        case AxisZ:
             _view.projection( _bb.getXMin(),_bb.getYMin(),_bb.getZMin(), p0 );
             _view.projection( _bb.getXMin(),_bb.getYMin(),_bb.getZMax(), p1 );
             d_angle = 180;
           break;
      }
       
      float angle = (float)Math.atan2( p1.getY()-p0.getY(), p1.getX()-p0.getX() ); //rad.
      angle = (float)(180*angle/Math.PI + d_angle); //degree

//SWT en est loin de ca...      QWMatrix m ;

      if( d_.drawing() )
      {
        d_.gc_.setForeground(  _axis_style.fore_ );
        d_.gc_.setFont( _axis_style.font_ );
      }
DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);      
      if ( _axis.getProperties().hasValue( DAxis.P_STEP_UNIT ) )
      {
        float coord_step = _axis.getProperties().get( DAxis.P_STEP_UNIT, +1.0f );
        if (  coord_step != 0.0f )
        {
          float coord;
//          float *cx, *cy, *cz;
          //optim: fait le switch une seule fois par for...
          /*
          switch( _axis_type )
          { 
            case AxisX:
          cx = &coord;  cy = &y_proj;  cz = &z_proj2 ; break;
            case AxisY:
          cx = &x_proj; cy = &coord;   cz = &z_proj2 ; break;
            case AxisZ:
          cx = &x_proj; cy = &y_proj2; cz = &coord ; break;
          }*/
 Rect rtxt = new Rect();
 int fh = d_.gc_.getFontMetrics().getHeight();
          for( coord = (float)(coord_step*Math.ceil(coord_min/coord_step));
               coord < coord_max;
               coord += coord_step )
          {
            switch( _axis_type )
            { 
              case AxisX: _view.projection(  coord,  y_proj,  z_proj2, p2 ); break;
              case AxisY: _view.projection(  x_proj, coord,   z_proj2, p2 ); break;
              case AxisZ: _view.projection(  x_proj, y_proj2, coord,   p2 ); break;
            }
            
//SWT?? tu fais quoi avec ca ???            
   /*         m.reset();
            m.translate( px[2], py[2] );
            m.rotate( angle );*/

            String txt = df.format( coord );
            int wt = d_.gc_.textExtent( txt ).x;
            rtxt.setRect( -wt/2, -3-fh, wt, fh );
rtxt.moveBy( (int)p2.getX(), (int)p2.getY() );            

            if( d_.drawing() )
            {
//SWT???
/*              
                d_->P_->save();
                d_->P_->setWorldMatrix( m );
                d_->P_->drawText( rtxt, Qt::AlignCenter, txt );
                d_->P_->restore();*/
                DrawUtil.drawText( d_.gc_, rtxt, DrawUtil.CENTER, txt );
              }
              else //locating
              {
/*SWT*declare forfait*                QRegion reg( rtxt );
                reg = m * reg;
                if ( reg.contains( _pt_to_locate ) )
                {
                  DChartLocation::Location location;
                  switch( _axis_type )
                  {
                    case AxisX:
                      location = DChartLocation::AxisXStepUnitText; break;
                    case AxisY:
                      location = DChartLocation::AxisYStepUnitText; break;
                    case AxisZ:
                      location = DChartLocation::AxisZStepUnitText; break;
                  }
                  d_->location_= DChartLocation( location, _axis, reg.boundingRect(), txt );
                  return ;
                }
                */
              }
            }
          }
        }
      }

      /*
       * Axis title.
       */
      if ( !DrawUtil.isEmpty( axis_title ) )
      {
        String location=null;
 Rect rtxt = new Rect();
 int fh = d_.gc_.getFontMetrics().getHeight();       
        switch( _axis_type )
        {
          case AxisX:
                _view.projection( _bb.getXMin(),y_proj,z_proj2, p0 );
                _view.projection( _bb.getXMax(),y_proj,z_proj2, p1 );
               location = DLocated.InAxisXTitle;
            break;
          case AxisY:
                _view.projection( x_proj,_bb.getYMin(),z_proj2, p0 );
                _view.projection( x_proj,_bb.getYMax(),z_proj2, p1 );
                location = DLocated.InAxisYTitle; 
            break;

          case AxisZ:
                 _view.projection( x_proj,y_proj2,_bb.getZMin(), p0 );
                 _view.projection( x_proj,y_proj2,_bb.getZMax(), p1 );
                 location = DLocated.InAxisZTitle; 
            break;
        }

        float center_x=(p0.getX()+p1.getX())/2,
              center_y=(p0.getY()+p1.getY())/2;
/*SWT*? toujours pas ?
          m.reset();
          m.translate( center_x, center_y );
          m.rotate( angle );
*/
          int wt = d_.gc_.textExtent( axis_title ).x;
          rtxt.setRect( -wt/2, -6-2*fh, wt, fh );

//TODO: pas de rotation ... donc "a plat"
//-X-8 au dessus des unites et marge...
rtxt.moveCenter( (int)center_x, (int)center_y - (have_step_unit?(fh+4):0) -4 );

          if( d_.drawing() )
          {
/*            
            d_->P_->save();
            d_->P_->setWorldMatrix( m );
            d_->P_->drawText( rtxt, Qt::AlignCenter, axis_title );
      d_->P_->restore();
*/
  DrawUtil.drawText( d_.gc_, rtxt, DrawUtil.CENTER, axis_title );      
          }
          else //locating
          {
  if( rtxt.contains( d_.lx_, d_.ly_ ))
  {
    throw new ChartLocation( location, _axis, rtxt ); 
  }
/*SWT...            
            QRegion reg( rtxt );
            reg = m * reg ;
            if( reg.contains( _pt_to_locate ) )
            {
              d_->location_= DChartLocation( location, _axis, reg.boundingRect(),
                                    axis_title );
              return;
            }
*/            
          }
        }
      }
    
   protected static final int XProj=1;
   protected static final int YProj=2;
   protected static final int ZProj=3;
   protected static final int Curve=4;
   protected static final int Finished=5;

  
  /** render XYZ curves */
  void renderXYZ( Rect _RG ) throws ChartLocation
  {
    //decorations means axis title/units...
    boolean display_decorations = true;
    boolean is_surface = d_.G_.isGraphicType( DGraphic.T_SURFACE );

    DAxis x_axis = null;
    DAxis y_axis = null;
    DAxis z_axis = null;
    int   axis_missed = 3;
    for( IDItem item = d_.G_.getFirstChild(); item!=null; item=item.getNext() )
    {
      if( item instanceof DAxis )
      {
        DAxis axis = (DAxis)item;
        if( x_axis==null && "x".equals( axis.getName() ))
        {
          x_axis = axis; axis_missed--;
        }
        else if ( y_axis==null & "y".equals( axis.getName() ))
        {
          y_axis = axis; axis_missed--;
        }
        else if ( z_axis==null & "z".equals( axis.getName() ))
        {
          z_axis = axis; axis_missed--;
        }
        if( axis_missed==0 ) break;
      }
    }

    if ( axis_missed > 0 )
    {
      String txt;
      if(x_axis==null) {
        txt = "Error: missing 'x' axis";
      } else 
      if(y_axis==null) {
        txt = "Error: missing 'y' axis";
      } else {
        txt = "Error: missing 'z' axis";
      }
      error( txt );
      return ;
    }

    DChartStyle x_axis_style = styleOf( x_axis );
    DChartStyle y_axis_style = styleOf( y_axis );
    DChartStyle z_axis_style = styleOf( z_axis );

    //garder un peu d'espace pour les texte sur les axes 
    //c'est completement choisi au hazard (ou a peu pres!).
    int max_font_h  = 0 ;
    if ( display_decorations )
    {
      d_.gc_.setFont( x_axis_style.font_ );
      max_font_h = d_.gc_.getFontMetrics().getHeight();
      d_.gc_.setFont( y_axis_style.font_ );
      int hf = d_.gc_.getFontMetrics().getHeight();
      if ( max_font_h < hf ) max_font_h = hf;
      d_.gc_.setFont( z_axis_style.font_ );
      hf = d_.gc_.getFontMetrics().getHeight();
      if ( max_font_h < hf ) max_font_h = hf;
      //prevoir le titre + les unites sur l'axe
      max_font_h = 2*max_font_h +6;
    }

    Rect gR = new Rect( _RG.left()+max_font_h, _RG.top()+max_font_h,
                        _RG.w()-2*max_font_h,  _RG.h()-max_font_h );

    //boite englobante des courbes a afficher.
    Box3D bb = new Box3D( x_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
                          x_axis.getProperties().get( DAxis.P_MAX, +1.0f ),
                          y_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
                          y_axis.getProperties().get( DAxis.P_MAX, +1.0f ),
                          z_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
                          z_axis.getProperties().get( DAxis.P_MAX, +1.0f ) );

    float phi   = d_.G_.getProperties().get( DGraphic.P_XYZ_PHI,   0.3f );
    float theta = d_.G_.getProperties().get( DGraphic.P_XYZ_THETA, 0.2f );

    View3D view = new View3D( phi, theta, 0, 1.0f, bb );

    //mise a l'echelle de la vue 3D en fonction du rectangle de dessin
    view.scaleTo( gR.x(), gR.y(), gR.w(), gR.h() );

    int min_w_h = gR.w() ;
    if (min_w_h > gR.h() ) min_w_h = gR.h();

    //bounding Box
    if ( display_decorations )
    {
      if( d_.drawing() )
      {
        d_.gc_.setForeground( getLightGray() );
        drawBox3D( view, bb );
      }

      // render des axes...
      renderAnAxis3D( view, bb, x_axis, x_axis_style, AxisX );
//      if( d_->located() ) return ;
      renderAnAxis3D( view, bb, y_axis, y_axis_style, AxisY );
//      if( d_->located() ) return ;
      renderAnAxis3D( view, bb, z_axis, z_axis_style, AxisZ );
//      if( d_->located() ) return ;
    }

    //pour projection des courbes.
    float x_proj = bb.getXMin(), y_proj = bb.getYMin(), z_proj = bb.getZMin(); 
    CurveStyle ccurve =null;
    CurveStyle p_ccurve=null; //courbe precedente (connection plan)
    IDItem         item=null;
    IDItem       p_item=null;
    float       last_x=0, last_y=0;

    //memo h,s,v du fond pour les couleurs de projection des courbes...
    DChartStyle bgsty = styleOf( d_.G_ );
    Color background_color = bgsty.back_;
    HSV bhsv = new HSV( background_color );

    // check property status for curves projections
    boolean do_xproj = d_.G_.getProperties().get( DGraphic.P_XYZ_XPROJ, true);
    boolean do_yproj = d_.G_.getProperties().get( DGraphic.P_XYZ_YPROJ, true);
    boolean do_zproj = d_.G_.getProperties().get( DGraphic.P_XYZ_ZPROJ, true);

    Point2D p = new Point2D();
    
    for( int ic=0; ic<d_.curves_.size(); ++ic )
    {
      ccurve = (CurveStyle)d_.curves_.get(ic);
      //la courbe est dessinee 4 fois: 3 projections + la courbe
      int what=XProj;

      if( display_decorations
//       && d_.G_.display3DCurveProjection() 
       && d_.drawing() /*en locating on ne fait pas les projections*/
        )
      {
        //courbes projetees: couleur de la courbe en plus clair ...
        Color proj_color = DrawUtil.lighter( ccurve.style_.back_, 1.75f );
        HSV hsv = new HSV( proj_color );
        //sauf si ecart < 20, me demandez pas pourquoi, c'est une heuristique
        //finement elaboree...
        int ecart=(int)(Math.abs(bhsv.getV()-hsv.getV())+Math.abs(bhsv.getS()-hsv.getS()))>>1;
        if(ecart < 20 )
        {
          //couleur de projection trop proche du fond, on bascule
          //en couleur plus fonce..
          proj_color = DrawUtil.lighter( ccurve.style_.back_, 0.75f );
        }
        d_.gc_.setForeground( proj_color );
        what = XProj ;
      } else {
        what = Curve ;
      }
  for( ;what<Finished; )
  {
 
      //reset des projecitons
      view.resetCoordTransformation();
      switch( what )
      {
        case XProj: { if(!do_xproj) { what=YProj; continue; }
                      force_coord.value_ = x_proj ;
                      view.setXCoordTransformation( force_coord );
                      what = YProj;
                      break; }
        case YProj: { if(!do_yproj) { what=ZProj; continue; }
                      force_coord.value_ = y_proj ;
                      view.setYCoordTransformation( force_coord );
                      what = ZProj;
                      break; }
        case ZProj: { if(!do_zproj) { what=Curve; continue; }
                      force_coord.value_ = z_proj ;
                      view.setZCoordTransformation( force_coord );
                      what = Curve;
                      break; }
        case Curve: { if(d_.drawing()) d_.gc_.setForeground( ccurve.style_.back_ ); 
          what = Finished;
                      break; }
        case Finished:break;
      }//switch

      if ( is_surface )
      {
        p_item = p_ccurve!=null ? p_ccurve.curve_.getFirstChild(): null;
      }

      boolean first=true; 
      for( item = ccurve.curve_.getFirstChild(); item!=null; item=item.getNext() )
      {
        DPoint point   = (DPoint)item;
        DPoint p_point = p_item!=null ? (DPoint)p_item: null; 
        if( point!=null )
        {
          float v_x=0, v_y=0, v_z=0 ;
          DCoord c=null;
          boolean ok=true;
          if( (c=point.getCoord( x_axis))!=null ) { v_x=c.getValue(); } else { ok=false; }
          if( (c=point.getCoord( y_axis))!=null ) { v_y=c.getValue(); } else { ok=false; }
          if( (c=point.getCoord( z_axis))!=null ) { v_z=c.getValue(); } else { ok=false; }
  
          if( ok )
          {
            float p_x,p_y;
            view.projection( v_x,v_y,v_z, p );
            {
              p_x = p.getX();
              p_y = p.getY();
              if ( first )
              {
                first=false;
              }
              else
              {
                if(d_.drawing())
                {
                  d_.gc_.drawLine( (int)last_x, (int)last_y, (int)p_x, (int)p_y );
                }
                else //locating
                {
                  if( what==Finished ) //pas les projections
                  {
                    if( DrawUtil.segmentContains( (int)last_x, (int)last_y, (int)p_x, (int)p_y, d_.lx_,d_.ly_))
                    {
                      throw new ChartLocation( DLocated.CurveLine, item,
                             new Rect((int)p_x,(int)p_y, (int)(last_x-p_x),(int)(last_y-p_y) ));
                    }
                  }
                }
              }
            }
            last_x=p_x;
            last_y=p_y;

            //lien avec la courbe precedente (en mode surface)
            if ( p_point!=null )
            {
              float p_v_x=0,p_v_y=0,p_v_z=0;
              ok = true;
              ok = ok&&((c=p_point.getCoord( x_axis ))!=null); if(ok) {p_v_x=c.getValue(); }
              ok = ok&&((c=p_point.getCoord( y_axis ))!=null); if(ok) {p_v_y=c.getValue(); }
              ok = ok&&((c=p_point.getCoord( z_axis ))!=null); if(ok) {p_v_z=c.getValue(); }
              if( ok )
              {
                view.projection( p_v_x,p_v_y,p_v_z, p);
                {
                  float p_p_x = p.getX(),  p_p_y=p.getY();
                  if( d_.drawing() )
                  {
                    d_.gc_.drawLine( (int)p_x,(int)p_y, (int)p_p_x,(int)p_p_y);
                  } 
                  else // locating
                  {
                    if( what==Finished )
                    {
 //TBD en 3D(2D) plusieurs courbes peuvent se superposer...
                      if( DrawUtil.segmentContains( (int)p_x,(int)p_y, (int)p_p_x,(int)p_p_y, d_.lx_,d_.ly_ ))
                      {
                        throw new ChartLocation( DLocated.CurveLine, item, 
                            new Rect( (int)p_x,(int)p_y,(int)(p_p_x-p_x),(int)(p_p_y-p_y)));
                      } 
                    }
                  }
                }
              }
            }
          }
          if ( p_item!=null ) p_item = p_item.getNext();
        }
      }//for

  }//for:what

      p_ccurve=ccurve;
    }//for
  }

  
/*
class TESTRT_EXPORT DChartPainter
{
 protected:
  void renderHistor ( const QRect & );
  void renderBargraph ( const QRect & );

  QString valueWithUnit( float _value, DAxis *_axis ) const ;

*/
}
//Source file: C:\\Program Files\\eclipse21\\workspace\\apidoc\\com\\rational\\apidoc\\drivers\\builder\\TBuilder.java

//Source file: C:\\Program Files\\eclipse21\\workspace\\apidoc\\com\\rational\\apidoc\\drivers\\ui\\TBuilder.java

package com.rational.apidoc.drivers.builder;

import com.rational.apidoc.core.*;
import com.rational.apidoc.drawutil.DrawUtil;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.Device;
import java.util.Enumeration;
import java.util.Hashtable;
import com.rational.apidoc.parsers.DParser;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.FontMetrics;
import org.eclipse.swt.graphics.Rectangle;


public class TBuilder extends DParser 
{
   private Image brokenImage = null;
   private int alignment = DParagraph.LEFT;
   private int bottomMargin = 0;
   private int cellSpacing = 0;
   private TCell currentCell = null;
   private int currentHLine = 0;
   private int currentNLine = 0;
   private int currentNPage = 0;
   private TCell currentPage = null;
   private int currentPosition = 1;
   private int currentWLine = 0;
   private int currentXPos = 0;
   private int currentYPos = 0;
   private Font defaultFont = null;
   private Font previousFont = null;
   private Device device = null;
   private boolean drawPopup = true;
   private TCell firstCellLine = null;
   private boolean flatPopup = false;
   private Hashtable fonts = new Hashtable();
   private TCell footerCell = null;
   private GC gc = null;
   private TCell headerCell = null;
   private Hashtable imagePool = new Hashtable();
   private TCell lastCellLine = null;
   private int leftMargin = 0;
   private int lineSpacing = 0;
   private int listOffsetSpacing = 0;
   private int maxWidth = 100;
   private int minWidth = 20;
   private int pageBegin = 0;
   private int pageEnd = 0;
   private int pageHeight = 0;
   private int prefixSpacing = 0;
   private int rightMargin = 0;
   private int titleSpacing = 0;
   private int topMargin = 0;
   private boolean wordWrap = true;
   private int xoffset = 0;
   private DFooter footer = null;
   private DHeader header = null;
   private TCell firstCell = null;
   static int level = 0;
   
   private int currentNColumn = 0;
   private int currentTotalColumn = 0;
   private int currentNRow = 0;
   private DRow currentRow = null;
   private DTable currentTable = null;
   private int maxCellHeight = 0;
   
   
   private class CellInfo {
		  TCell cell = null;
		  int rowspan = 0;
   }
   
   private CellInfo[] currentRowSpan = null;
   
   private class Range 
   {
		public int start = 0;
		public int end = 0;
      
		/**
		 * @roseuid 3F0168C202A7
		 */
		public Range() 
		{
       
		}
   }
   
   /**
    * @param gc
    * @param d
    * @roseuid 3EF82BBF01A5
    */
   public TBuilder(GC gc, Device d) 
   {
      this.gc = gc;
	  this.device = d; 
            
      if (gc!=null)    
         defaultFont = gc.getFont();
      else
	     defaultFont = new Font(d, "courier", 12, SWT.NORMAL);   
   }
   
   
   private TBuilder _newBuilder(int size)
	  {
		 TBuilder b = new TBuilder(getGc(), getDevice());
		 b.setFlatPopup(getFlatPopup());
		 b.setCurrentPosition(getCurrentPosition());
		 b.setWordWrap(getWordWrap());
		 b.setMaxWidth(size);
		 b.setImagePool(getImagePool());
		 b.setFontPool(getImagePool());
		 b.setImageSearchPath(getImageSearchPath());
		 b.setCurrentNPage(getCurrentNPage());
		 return b;
	  }
   
	  private void _disposeBuilder(TBuilder b)
	  {
		 b.setImagePool(null);
		 b.setImageSearchPath(null);
		 b.setFontPool(null);
	  }
   
	  /**
	   * @param i
	   * @roseuid 3F004C9B00F9
	   */
	  private void _addListItemPrefix(IDItem i) 
	  {  
		 if (!(i.getParent() instanceof DList)) return;
   	
		 addListEntry();
		 String rval = getListStringPrefix(i.getParent());
      
		 if (rval != null)
		 {
			Point p = getGc().textExtent(rval);
			addXSpacing(-p.x);
			TCell c = addCell(p.x, p.y, i, rval.length());
			c.setText(rval);
			addXSpacing(getTitleSpacing());
		 }    
	  }
   
	  /**
	   * @param i
	   * @param text
	   * @param r
	   * @roseuid 3F0168EB0089
	   */
	  private void _buildPourTextBuffer(IDItem i, String text, Range r) 
	  {
		 if (text.length() == 0) return;
      
		 int w, wprev;
		 int count = r.end - r.start;
		 TCell item = null;
      
		 FontMetrics fm = getGc().getFontMetrics();
		 int lineSpacing = fm.getAscent()+fm.getDescent();
      
		 // rien a traiter
		 if (r.start >= r.end)
		 {
		   if (text.getBytes()[r.start] == '\n')
		   {
			   item = addCell(0, lineSpacing, i, 0);
		   }
		   else
		   {	
			   addYSpacing(lineSpacing);	
		   }
		   return;
		 }
      
		 String qs = null;
		 if (r.end == text.length())
			qs = text;
		 else
			qs = text.substring(r.start, r.end);
      
		 GC gc = getGc();
		 Point p = gc.textExtent(qs);
      
		 // Si on est en non WORD WRAP ou que le buffer rentre
		 // dans la ligne alors on ajoute ce dernier
		 // sur la ligne courante
		 if (!getWordWrap() 
			 || (getCurrentWLine() + p.x <= getMaxWidth()))
		 {
		   item = addCell(p.x, p.y, i, count);
		   item.setText(text, r.start, count); 
		   return;
		 }
         
		 // sinon on va couper le buffer sur un separateur de mot afin
		 // de la faire rentrer sur plusieurs lignes
		 wprev = 0;
      
		 Range r2 = new Range();
		 r2.start = r.start; // position dernier caractere de la sous chaine
      
      
		 while (r2.start < r.end)
		 {
			r2.end = 0;   // taille en pixel de la sous chaine
      	 
			_buildGetNextWord(qs, 
							  getMaxWidth() - getCurrentWLine(),
							  r2);
              
			count = r2.start - r.start;
			TCell cell = addCell(r2.end, lineSpacing, i, count);
			cell.setText(text, r.start, count);
			newLine();
         
			r.start = r2.start;
		 }    
	  }
   
	  /**
	   * @param b
	   * @param i
	   * @param text
	   * @roseuid 3F003ACA012D
	   */
	  private void _buildTextBuffer(IDItem i, String text) 
	  {
		 if (text.length()==0) return;
       
		 Range r = new Range();
		 r.start = 0;
		 r.end = text.indexOf("\n");
      
		 // tant que l'on a pas atteint la fin de la ligne
		 while (r.end != -1)
		 {
		   _buildPourTextBuffer(i, text, r);
      	
		   // ligne non adaptative car terminee par \n
		   // on ajoute donc une nouvelle ligne
		   newLine();
      	
		   r.start = r.end + 1;
      	
		   // recherche du \n suivant 
		   r.end = text.indexOf("\n", r.start);
		 }
      
		 if (r.start < text.length())
		 {
			r.end = text.length();
		   _buildPourTextBuffer(i, text, r);
		 }    
	  }
   
	  /**
	   * @param c
	   * @return int
	   * @roseuid 3F003121028F
	   */
	  private int _getColumnSizeForCell(DCell c) 
	  {
		  int size = 0;
		  if (currentTable.getAutoResize())  // toute la largeur du document 
		  {
				 float ws = (float)getMaxWidth() * currentTable.getWidthScale();
				 for (int i=0; i<c.getColSpan()+1; i = i + 1)
				 {
				   float factor = 1.0f;
				   int s = 0;
				   if ((currentTable.getColumnWidth() != null)
						&& (currentTable.getColumnWidth().length > currentNColumn + i))
				   {
					   factor = currentTable.getColumnWidth()[currentNColumn + i];
					   s = (int)(ws * factor);
				   }
				   else
				   {
					   s = (int)(ws / currentTotalColumn);
				   }
			  	
				   if (s < getMinWidth())
				   {
					 s = getMinWidth();
				   }
				   size = size + s;
				 }
		 }
		 else 
		 {
				size = getMaxWidth();
		 }
		 return size;    
	  }
   
	  /**
	   * @param b
	   * @param c
	   * @roseuid 3F0030E403A0
	   */
	  private void _setBuilderMarginForCell(TBuilder b) 
	  {
		  int m = getCellSpacing();
	   
	      b.setLeftMargin(m + currentTable.getBorder());	  
	      b.setTopMargin( m + currentTable.getBorder());   	  
	      b.setBottomMargin(m + currentTable.getBorder());
	      b.setRightMargin(m + currentTable.getBorder());
	  }
   
   
   /**
    * Access method for the alignment property.
    * 
    * @return   the current value of the alignment property
    */
   public int getAlignment() 
   {
      return alignment;    
   }
   
   /**
    * Sets the value of the alignment property.
    * 
    * @param aAlignment the new value of the alignment property
    */
   public void setAlignment(int aAlignment) 
   {
      alignment = aAlignment;    
   }
   
   /**
    * Access method for the bottomMargin property.
    * 
    * @return   the current value of the bottomMargin property
    */
   public int getBottomMargin() 
   {
      return bottomMargin;    
   }
   
   /**
    * Sets the value of the bottomMargin property.
    * 
    * @param aBottomMargin the new value of the bottomMargin property
    */
   public void setBottomMargin(int aBottomMargin) 
   {
      bottomMargin = aBottomMargin;    
   }
   
   /**
    * Access method for the cellSpacing property.
    * 
    * @return   the current value of the cellSpacing property
    */
   public int getCellSpacing() 
   {
      return cellSpacing;    
   }
   
   /**
    * Sets the value of the cellSpacing property.
    * 
    * @param aCellSpacing the new value of the cellSpacing property
    */
   public void setCellSpacing(int aCellSpacing) 
   {
      cellSpacing = aCellSpacing;    
   }
   
   /**
    * Access method for the currentCell property.
    * 
    * @return   the current value of the currentCell property
    */
   public TCell getCurrentCell() 
   {
      return currentCell;    
   }
   
   /**
    * Access method for the currentHLine property.
    * 
    * @return   the current value of the currentHLine property
    */
   public int getCurrentHLine() 
   {
      return currentHLine;    
   }
   
   /**
    * Access method for the currentNPage property.
    * 
    * @return   the current value of the currentNPage property
    */
   public int getCurrentNPage() 
   {
      return currentNPage;    
   }
   
   /**
    * Access method for the currentNPage property.
    * 
	* @param i
	*/
   public void setCurrentNPage(int i) {
	   currentNPage = i;
   }
   
   /**
    * Access method for the currentPage property.
    * 
    * @return   the current value of the currentPage property
    */
   public TCell getCurrentPage() 
   {
      return currentPage;    
   }
   
   /**
    * Access method for the currentPosition property.
    * 
    * @return   the current value of the currentPosition property
    */
   public int getCurrentPosition() 
   {
      return currentPosition;    
   }
   
   /**
    * Sets the value of the currentPosition property.
    * 
    * @param aCurrentPosition the new value of the currentPosition property
    */
   public void setCurrentPosition(int aCurrentPosition) 
   {
      currentPosition = aCurrentPosition;    
   }
   
   /**
    * Access method for the currentWLine property.
    * 
    * @return   the current value of the currentWLine property
    */
   public int getCurrentWLine() 
   {
      return currentWLine;    
   }
   
   /**
    * Access method for the currentXPos property.
    * 
    * @return   the current value of the currentXPos property
    */
   public int getCurrentXPos() 
   {
      return currentXPos;    
   }
   
   /**
    * Access method for the currentYPos property.
    * 
    * @return   the current value of the currentYPos property
    */
   public int getCurrentYPos() 
   {
      return currentYPos;    
   }
   
   /**
    * Access method for the defaultFont property.
    * 
    * @return   the current value of the defaultFont property
    */
   public Font getDefaultFont() 
   {
      return defaultFont;    
   }
   
   /**
    * Sets the value of the defaultFont property.
    * 
    * @param aDefaultFont the new value of the defaultFont property
    */
   public void setDefaultFont(Font aDefaultFont) 
   {
      defaultFont = aDefaultFont;    
   }
   
   /**
    * Access method for the device property.
    * 
    * @return   the current value of the device property
    */
   public Device getDevice() 
   {
      return device;    
   }
   
   /**
    * Determines if the drawPopup property is true.
    * 
    * @return   <code>true<code> if the drawPopup property is true
    */
   public boolean getDrawPopup() 
   {
      return drawPopup;    
   }
   
   /**
    * Sets the value of the drawPopup property.
    * 
    * @param aDrawPopup the new value of the drawPopup property
    */
   public void setDrawPopup(boolean aDrawPopup) 
   {
      drawPopup = aDrawPopup;    
   }
   
   /**
    * Access method for the firstCellLine property.
    * 
    * @return   the current value of the firstCellLine property
    */
   public TCell getFirstCellLine() 
   {
      return firstCellLine;    
   }
   
   /**
    * Determines if the flatPopup property is true.
    * 
    * @return   <code>true<code> if the flatPopup property is true
    */
   public boolean getFlatPopup() 
   {
      return flatPopup;    
   }
   
   /**
    * Sets the value of the flatPopup property.
    * 
    * @param aFlatPopup the new value of the flatPopup property
    */
   public void setFlatPopup(boolean aFlatPopup) 
   {
      flatPopup = aFlatPopup;    
   }
   
   /**
    * Access method for the footerCell property.
    * 
    * @return   the current value of the footerCell property
    */
   public TCell getFooterCell() 
   {
      return footerCell;    
   }
   
   /**
    * Sets the value of the footerCell property.
    * 
    * @param aFooterCell the new value of the footerCell property
    */
   public void setFooterCell(TCell aFooterCell) 
   {
      footerCell = aFooterCell;    
   }
   
   /**
    * Access method for the gc property.
    * 
    * @return   the current value of the gc property
    */
   public GC getGc() 
   {
      return gc;    
   }
   
   /**
    * Access method for the headerCell property.
    * 
    * @return   the current value of the headerCell property
    */
   public TCell getHeaderCell() 
   {
      return headerCell;    
   }
   
   /**
    * Sets the value of the headerCell property.
    * 
    * @param aHeaderCell the new value of the headerCell property
    */
   public void setHeaderCell(TCell aHeaderCell) 
   {
      headerCell = aHeaderCell;    
   }
   
   /**
    * Access method for the imagePool property.
    * 
    * @return   the current value of the imagePool property
    */
   public Hashtable getImagePool() 
   {
      return imagePool;    
   }
   
   /**
    * Sets the value of the imagePool property.
    * 
    * @param aImagePool the new value of the imagePool property
    */
   public void setImagePool(Hashtable aImagePool) 
   {
      imagePool = aImagePool;    
   }
   
   /**
	* Access method for the imagePool property.
	* 
	* @return   the current value of the imagePool property
	*/
   public Hashtable getFontPool() 
   {
		 return fonts;    
   } 
   
   /**
	* Sets the value of the imagePool property.
	* 
	* @param aImagePool the new value of the imagePool property
	*/
   public void setFontPool(Hashtable aFontPool) 
   {
		 fonts = aFontPool;    
   }
   
   /**
    * Access method for the leftMargin property.
    * 
    * @return   the current value of the leftMargin property
    */
   public int getLeftMargin() 
   {
      return leftMargin;    
   }
   
   /**
    * Sets the value of the leftMargin property.
    * 
    * @param aLeftMargin the new value of the leftMargin property
    */
   public void setLeftMargin(int aLeftMargin) 
   {
      leftMargin = aLeftMargin;    
   }
   
   /**
    * Access method for the lineSpacing property.
    * 
    * @return   the current value of the lineSpacing property
    */
   public int getLineSpacing() 
   {
      return lineSpacing;    
   }
   
   /**
    * Sets the value of the lineSpacing property.
    * 
    * @param aLineSpacing the new value of the lineSpacing property
    */
   public void setLineSpacing(int aLineSpacing) 
   {
      lineSpacing = aLineSpacing;    
   }
   
   /**
    * Access method for the listOffsetSpacing property.
    * 
    * @return   the current value of the listOffsetSpacing property
    */
   public int getListOffsetSpacing() 
   {
      return listOffsetSpacing;    
   }
   
   /**
    * Sets the value of the listOffsetSpacing property.
    * 
    * @param aListOffsetSpacing the new value of the listOffsetSpacing property
    */
   public void setListOffsetSpacing(int aListOffsetSpacing) 
   {
      listOffsetSpacing = aListOffsetSpacing;    
   }
   
   /**
    * Access method for the maxWidth property.
    * 
    * @return   the current value of the maxWidth property
    */
   public int getMaxWidth() 
   {
      return maxWidth;    
   }
   
   /**
    * Sets the value of the maxWidth property.
    * 
    * @param aMaxWidth the new value of the maxWidth property
    */
   public void setMaxWidth(int aMaxWidth) 
   {
	  maxWidth = Math.max(aMaxWidth, getMinWidth());
   }
   
   /**
    * Access method for the minWidth property.
    * 
    * @return   the current value of the minWidth property
    */
   public int getMinWidth() 
   {
      return minWidth;    
   }
   
   /**
    * Sets the value of the minWidth property.
    * 
    * @param aMinWidth the new value of the minWidth property
    */
   public void setMinWidth(int aMinWidth) 
   {
      minWidth = aMinWidth;    
   }
   
   /**
    * Determines if the modePage property is true.
    * 
    * @return   <code>true<code> if the modePage property is true
    */
   public boolean getModePage() 
   {
      return (getPageHeight()>0);    
   }
   
   /**
    * Access method for the pageBegin property.
    * 
    * @return   the current value of the pageBegin property
    */
   public int getPageBegin() 
   {
      return pageBegin;    
   }
   
   /**
    * Sets the value of the pageBegin property.
    * 
    * @param aPageBegin the new value of the pageBegin property
    */
   public void setPageBegin(int aPageBegin) 
   {
      pageBegin = aPageBegin;    
   }
   
   /**
    * Access method for the pageEnd property.
    * 
    * @return   the current value of the pageEnd property
    */
   public int getPageEnd() 
   {
      return pageEnd;    
   }
   
   /**
    * Sets the value of the pageEnd property.
    * 
    * @param aPageEnd the new value of the pageEnd property
    */
   public void setPageEnd(int aPageEnd) 
   {
      pageEnd = aPageEnd;    
   }
   
   /**
    * Access method for the pageHeight property.
    * 
    * @return   the current value of the pageHeight property
    */
   public int getPageHeight() 
   {
      return pageHeight;    
   }
   
   /**
    * Sets the value of the pageHeight property.
    * 
    * @param aPageHeight the new value of the pageHeight property
    */
   public void setPageHeight(int aPageHeight) 
   {
      pageHeight = aPageHeight;    
   }
   
   /**
    * Access method for the rightMargin property.
    * 
    * @return   the current value of the rightMargin property
    */
   public int getRightMargin() 
   {
      return rightMargin;    
   }
   
   /**
    * Sets the value of the rightMargin property.
    * 
    * @param aRightMargin the new value of the rightMargin property
    */
   public void setRightMargin(int aRightMargin) 
   {
      rightMargin = aRightMargin;    
   }
   
   /**
    * Access method for the titleSpacing property.
    * 
    * @return   the current value of the titleSpacing property
    */
   public int getTitleSpacing() 
   {
      return titleSpacing;    
   }
   
   /**
    * Sets the value of the titleSpacing property.
    * 
    * @param aTitleSpacing the new value of the titleSpacing property
    */
   public void setTitleSpacing(int aTitleSpacing) 
   {
      titleSpacing = aTitleSpacing;    
   }
   
   /**
    * Access method for the topMargin property.
    * 
    * @return   the current value of the topMargin property
    */
   public int getTopMargin() 
   {
      return topMargin;    
   }
   
   /**
    * Sets the value of the topMargin property.
    * 
    * @param aTopMargin the new value of the topMargin property
    */
   public void setTopMargin(int aTopMargin) 
   {
      topMargin = aTopMargin;    
   }
   
   /**
    * Determines if the wordWrap property is true.
    * 
    * @return   <code>true<code> if the wordWrap property is true
    */
   public boolean getWordWrap() 
   {
      return wordWrap;    
   }
   
   /**
    * Sets the value of the wordWrap property.
    * 
    * @param aWordWrap the new value of the wordWrap property
    */
   public void setWordWrap(boolean aWordWrap) 
   {
      wordWrap = aWordWrap;    
   }
   
   /**
    * Access method for the xoffset property.
    * 
    * @return   the current value of the xoffset property
    */
   public int getXoffset() 
   {
      return xoffset;    
   }
   
   /**
    * Sets the value of the xoffset property.
    * 
    * @param aXoffset the new value of the xoffset property
    */
   public void setXoffset(int aXoffset) 
   {
      xoffset = aXoffset;    
   }
   
   /**
    * Access method for the footer property.
    * 
    * @return   the current value of the footer property
    */
   public DFooter getFooter() 
   {
      return footer;    
   }
   
   /**
    * Sets the value of the footer property.
    * 
    * @param aFooter the new value of the footer property
    */
   public void setFooter(DFooter aFooter) 
   {
      footer = aFooter;    
   }
   
   /**
    * Access method for the header property.
    * 
    * @return   the current value of the header property
    */
   public DHeader getHeader() 
   {
      return header;    
   }
   
   /**
    * Sets the value of the header property.
    * 
    * @param aHeader the new value of the header property
    */
   public void setHeader(DHeader aHeader) 
   {
      header = aHeader;    
   }
   
   /**
    * Access method for the firstCell property.
    * 
    * @return   the current value of the firstCell property
    */
   public TCell getFirstCell() 
   {
      return firstCell;    
   }
   
   /**
    * Sets the value of the firstCell property.
    * 
    * @param aFirstCell the new value of the firstCell property
    */
   public void setFirstCell(TCell aFirstCell) 
   {
      firstCell = aFirstCell;    
   }
   
   /**
    * @param w
    * @param h
    * @param item
    * @param count
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF9510A011C
    */
   public TCell addCell(int w, int h, IDItem item, int count) 
   {
   	  TCell c = new TCell(); //currentPage, currentCell);
   	  addCell(w, h, item, count, c);
   	  return c;    
   }
   
   /**
    * @param w
    * @param h
    * @param item
    * @param count
    * @param c
    * @roseuid 3F00143402E1
    */
   public void addCell(int w, int h, IDItem item, int count, TCell c) 
   {   
   	   if (currentPage==null)
   	   {
		   TCell pageCell = new TCell();
		   if (firstCell == null)
		   {
				firstCell  = pageCell;
		   }
		   currentPage  = pageCell;
		   currentCell  = null;
		   currentYPos  = 0;
		   currentXPos = getXoffset();
   	   }
   	
   	   // accrochage dans l'arbre
   	   if (currentPage.getChild() == null)
	       currentPage.setChild(c);
       c.setParent(currentPage);
       c.setPrevious(currentCell);
       
       if (currentCell != null)
          currentCell.setNext(c);
       
	   if (firstCellLine == null)
	   {
			// new line
			c.setRect(currentXPos + getLeftMargin(), currentYPos + getTopMargin(), w, h);
			firstCellLine = c;
			lastCellLine = c;
   	  	
			currentWLine = currentWLine + w;
			currentHLine = h;
			currentXPos = w;
			currentNLine = currentNLine + 1;
	   }
	   else
	   {
			c.setRect(currentXPos + getLeftMargin(), currentYPos + getTopMargin(), w, h);
			currentHLine = Math.max(currentHLine, h);
			currentWLine = currentWLine + w;
			currentXPos = currentXPos + w;
	   }
       
       // position des caracteres
	   c.setPosition(currentPosition);
	   c.setNumberLine(currentNLine);
	   c.setItem(item);
	   currentPosition = currentPosition + count;   
	   
	   // mise a jour de la taille du parent
	   if (currentPage != null)
	   {
		  currentPage.getRect().height = getTopMargin()+ currentYPos + currentHLine + getBottomMargin();
	      currentPage.getRect().width  = Math.max(currentPage.getRect().width, getLeftMargin() + currentWLine + getRightMargin());
	   }
       
       currentCell = c;    
   }
   
   /**
    * @param s
    * @roseuid 3EFAF6900056
    */
   public void addXSpacing(int s) 
   {
      currentWLine = currentWLine + s;    
   }
   
   /**
    * @param s
    * @roseuid 3EFAF6BA00BA
    */
   public void addYSpacing(int s) 
   {
      currentHLine = currentHLine + s;    
   }
   
   /**
    * @param item
    * @roseuid 3EFAA5FD0218
    */
   protected void applyFont(IDItem item) 
   {
   	   Font font = null;
   	   previousFont = getGc().getFont();
	   	
	   if ((item.getStyle() != null)
			&& (item.getStyle().getFont() != null))
	   {
			IDFont f = item.getStyle().getFont();
			font = (Font)fonts.get(f.getKey());
			
			if (font == null)
			{
				font = DrawUtil.getFont(f);
				fonts.put(f.getKey(), font);
			}
	   }
	   else
	   {
	   	   font = getDefaultFont();
	   }   
	   
	   getGc().setFont(font);    
   }
   
   protected void restoreFont()
   {
   	  if (previousFont!=null)
   	     getGc().setFont(previousFont);
   }
   
   
   
   /**
    * @roseuid 3EF849AE0212
    */
   public void dispose() 
   {
   	  clear();
   	  
      // dispose all fonts
      if (fonts != null)
      {
	     for (Enumeration e = fonts.elements() ; e.hasMoreElements() ;) 
	     {
		    ((Font)e.nextElement()).dispose();
	     }
	     fonts.clear();
      }
	   
      //	reset all imagePool
      if (imagePool != null)
      {
	     for (Enumeration e = imagePool.elements() ; e.hasMoreElements() ;) 
	     {
		    ((Image)e.nextElement()).dispose();
	     }
	     imagePool.clear();
      }
   }
   
   
   /**
    * @return int
    * @roseuid 3EF8475300C0
    */
   public int getHeight() 
   {
	  int h = 0;
	  for (TCell c = getFirstCell(); c != null; c = c.getNext() )
	  {
		h = h +  c.getRect().height;
	  }
	  return h;    
   }
   
   /**
    * @param i
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EFACC7B0159
    */
   public TCell getPage(int i) 
   {
	 int j = 0;
     for (TCell c = getFirstCell(); c != null; c = c.getNext() )
	 {
	    if (i == j) return c;
		j = j + 1; 
	 }
     return  null;    
   }
   
   /**
    * @return int
    * @roseuid 3EFACC4101BA
    */
   public int getPageCount() 
   {
   	  int i = 0;
	  for (TCell c = getFirstCell(); c != null; c = c.getNext() )
	  {
		 i = i + 1;
	  }
	  return i;    
   }
   
   /**
    * @return int
    * @roseuid 3EF847650152
    */
   public int getWidth() 
   {
   	  int w = 0;
	  for (TCell c = getFirstCell(); c != null; c = c.getNext() )
	  {
		 w = Math.max(w, c.getRect().width);
	  }
      return w;    
   }
   
   /**
    * @param p
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF95FE600FE
    */
   public TCell mapPointToCell(Point p) 
   {
	  return mapPointToCell(getFirstCell(), p);    
   }
   
   /**
    * @param c
    * @param p
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF96C530167
    */
   protected TCell mapPointToCell(TCell c, Point p) 
   {
	 TCell i;
	 for (i = c.getChild(); i != null; i = i.getNext())
	 {
	   if ( ((i.getRect().y <= p.y) && (p.y <= i.getRect().y + i.getRect().height))
			&& ((i.getRect().x <= p.x) && (p.x <= i.getRect().x + i.getRect().width)))
	   {
				break;
	   }
	 }
	   
	 if (i==null) return null;
	   
	 // la cellule est trouvee si elle en contient d'autre on y va en 
	 // profondeur
	 if (i.getChild()!=null)
	 {
	 	 p.x = p.x - i.getRect().x;
	 	 p.y = p.y - i.getRect().y;
		 TCell cf = mapPointToCell(i, p);
	   	 
		 if (cf!=null) return cf;
	 }
	   
	 p.x = p.x - i.getRect().x;
	 p.y = p.y - i.getRect().y;
	 return  i;    
   }
   
   /**
    * @param pos
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF9610F0282
    */
   public TCell mapPositionToCell(int pos) 
   {
      return mapPositionToCell(getFirstCell(), pos);    
   }
   
   /**
    * @param c
    * @param pos
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF96C810159
    */
   protected TCell mapPositionToCell(TCell c, int pos) 
   {
	 TCell i;
     for (i = c.getChild(); i != null; i = i.getNext())
	 {
	 	if (i.getChild()!=null)
	 	{
	 		TCell j = mapPositionToCell(i, pos);
	 		if (j!=null) return j;
	 	}
	 	if ((i.getPosition() <= pos)
	 	     && (pos <= i.getPosition() + i.getLen()))
	 	{
	 		return i;
	 	}
	 }
	 return null;    
   }
   
   /**
    * @roseuid 3EF9A6E0028C
    */
   public void newLine() 
   {
      if (firstCellLine != null)
      {
      	int xoffset = getLeftMargin();
      	
      	if (getAlignment() == DParagraph.RIGHT)
      	       xoffset = xoffset + getWidth() - currentWLine;
      	else
      	if (getAlignment() == DParagraph.CENTER)
      	{
      	       xoffset = xoffset + (getWidth() - currentWLine)/2; 
      	}
      	
      	int maxDsc = 0;
      	TCell c;
      	for (c = firstCellLine; c!=null; c = c.getNext())
      	{
      		if (c.getItem() != null)
      		{
      			applyFont(c.getItem());
      			maxDsc=Math.max(maxDsc, getGc().getFontMetrics().getDescent());
				restoreFont();
      		}
      	}
      	
      	for (c = firstCellLine; c!=null; c = c.getNext())
      	{
      		int locDsc;
      		if (c.getItem()!=null)
      		{
				applyFont(c.getItem());
      			locDsc = getGc().getFontMetrics().getDescent();
				restoreFont();
      			c.setRect(xoffset + c.getRect().x, 
      			          c.getRect().y - (maxDsc-locDsc)+(currentHLine-c.getRect().height), 
      			          c.getRect().width, c.getRect().height);
      			//xoffset = xoffset + c.getRect().width;
      		}
      	}
      }
      
      if (getModePage())
      {
      	 int fh = 0;
      	 if (getFooterCell()!=null)
      	 {
			fh = fh + getFooterCell().getRect().height;
      	 }
      	 
      	 if ((currentYPos + currentHLine + fh) > getPageHeight())
      	 { 
      	 	newPage();
      	 }
      	 else
      	 {
			currentYPos = currentYPos + currentHLine;
      	 }
      }
      else
      {   
         currentYPos = currentYPos + currentHLine;
      }
      
	  currentXPos = getXoffset(); 
      
      currentHLine  = 0;
      currentWLine  = 0;
      firstCellLine = null;    
   }
   
   /**
    * @roseuid 3EFAF84300A1
    */
   public void newPage() 
   {
	  // creation de la page
      TCell pageCell = new TCell();
      if (firstCell == null)
	     firstCell  = pageCell;
      else
      {
		 currentPage.setNext(pageCell);
      }
         
      currentPage  = pageCell;
      currentNPage = currentNPage + 1;
      currentCell  = null;
      int hLine = 0;
	  
      // ajout de l'entete s'il y en a une
      if (getHeader()!=null)
      {
	    doItem(getHeader());
	    
	    if (getHeaderCell()!=null)
	    {
	       // retour a la ligne
		   currentYPos  = 0;
		   currentXPos  = getXoffset();
		   currentWLine = 0;
		   currentHLine = 0;
	    
	       addCell(getHeaderCell().getRect().width, 
						 getHeaderCell().getRect().height, 
						 getHeader(), 0, getHeaderCell());
						 
		   hLine = getHeaderCell().getRect().height;
	    }
      }   
      else
      {
	    if (firstCellLine != null)
	    {
		   firstCellLine.setPrevious(null);
		   pageCell.setChild(firstCellLine);
	    }
      }
      
      // ajout du bas de page en fin de la page
      if (getFooter()!=null)
      {
		doItem(getFooter());
      	
		if (getFooterCell()!=null)
		{
            //	retour a la ligne plus positionnement en bas
			currentYPos  = getPageHeight()-getFooterCell().getRect().height;
			currentXPos  = getXoffset();
			currentWLine = 0;
			currentHLine = 0;	
			 
			addCell(getFooterCell().getRect().width, 
					getFooterCell().getRect().height, 
					getFooter(), 0, getFooterCell());
		}     	
      }
         
      TCell c;
      for (c = firstCellLine; c!=null; c = c.getNext())
      {
	     c.setParent(pageCell);
      }    
      
	  currentYPos = hLine;
	  currentXPos = getXoffset();
      currentHLine  = 0;
      currentWLine  = 0;
      firstCellLine = null;    
   }
   
   /**
    * @roseuid 3EFAF25A020E
    */
   public void clear() 
   {
   	   // clears all pages
	   for (TCell c = getFirstCell(); c != null; )
	   {
			 TCell p = c.getNext();
			 c.clear();
			 c.setRect(0,0,0,0);
			 c.setPosition(1);
		 
			 c.setNext(null);
			 c = p;
	   }
	   
	   super.clear();
	   
       currentPosition = 1;
	   currentNLine = 0; 
	   currentXPos = 0;
	   currentYPos = 0;
	   firstCellLine = null;
	   lastCellLine = null;
	   currentCell = null;
	   currentHLine = 0;
	   currentWLine = 0;    
	   alignment = DParagraph.LEFT;
	   currentPage = null;    
	   firstCell = null;
	   currentNPage = 0;    
   }
   
   /**
    * @param sp
    * @param ep
    * @roseuid 3EF9A56C002E
    */
   public void setPageRange(int sp, int ep) 
   {
      setPageBegin(sp);
      setPageEnd(ep);    
   }
   
   /**
    * @roseuid 3F7294070388
    */
   public void dump() 
   {
   	  int i = 0;
   	  for (TCell c = firstCell; c != null; c = c.getNext())
   	  {
   	  	 i = i +1;
   	  	 System.out.println("------------------ Page (" + i + ")-------------------");
   	  	 c.dump(0);
   	  }
   }


   /**
    * @param i
    */
   public void setCurrentYPos(int i) 
   {
	   currentYPos = i;
   }

    /**
     * @return
     */
    public Image getBrokenImage() {
	    if(brokenImage == null){
		   brokenImage = new Image(getDevice(),30,30);
	    }
	    return brokenImage;
   }

   /**
     * @param image
    */
   public void setBrokenImage(Image image) {
	   brokenImage = image;
   }

   /**
	* @param c
	* @roseuid 3EFC03260070
	*/
   public void doItem(DCell c) 
   {
   	  applyFont(c);
   	
	  int size = _getColumnSizeForCell(c);
   	  
	  TBuilder b = _newBuilder(size);
	 	
	  _setBuilderMarginForCell(b);
	
	  b.doChildrenItem(c);
		
	  TCell cl = b.getFirstCell();
	  if (cl != null)
	  {
		if (currentTable.getAutoResize())
			addCell(b.getMaxWidth(), cl.getRect().height, c, 0, cl);
		else
			addCell(cl.getRect().width, cl.getRect().height, c, 0, cl);
	  	
		setCurrentPosition(b.getCurrentPosition());
	    
		// mise en place des marges
		if (currentNColumn > 0)
		{
			cl.getRect().x -= currentTable.getBorder()-1;
		}
		if (currentNRow > 0)
		{
			cl.getRect().y -= currentTable.getBorder()-1;
		}
	  }

	  // on met fin au constructeur 
	  _disposeBuilder(b);
	  
	  CellInfo cf = new CellInfo();
	  cf.rowspan = c.getRowSpan();
	  cf.cell = cl;
	  
	  if (c.getRowSpan()==0)
	  {
		maxCellHeight = Math.max(maxCellHeight, cl.getRect().height);
	  }
	  
	  currentRowSpan[currentNColumn] = cf;
	  currentNColumn = currentNColumn + c.getColSpan() + 1;
	  
	  restoreFont();    
   }
   
   /**
	* @param c
	* @roseuid 3EFC0332009F
	*/
   public void doItem(DCellText c) 
   {
   	  applyFont(c);
   	
	  int size = _getColumnSizeForCell(c);
	  
	  TBuilder b = _newBuilder (size);
	  _setBuilderMarginForCell(b);
	 
	  _buildTextBuffer(c, c.getText());	 
	 	
	  TCell cl = b.getFirstCell();
	   
	  addCell(cl.getRect().width, cl.getRect().height, c, 0, b.getFirstCell());
	  setCurrentPosition(b.getCurrentPosition());
	  
	  // on met fin au constructeur 
	  _disposeBuilder(b);	
	 
	  CellInfo cf = new CellInfo();
	  cf.rowspan = c.getRowSpan();
	  cf.cell = cl;
	  
	  if (c.getRowSpan()==0)
	  {
		  maxCellHeight = Math.max(maxCellHeight, cl.getRect().height);
	  }
	  
	  currentRowSpan[currentNColumn] = cf;
	  currentNColumn = currentNColumn + c.getColSpan() + 1;
	  
	  restoreFont();    
   }
   
   /**
	* @param d
	* @roseuid 3EFC0213029B
	*/
   public void doItem(DDocument d) 
   {
   	  applyFont(d);
   	
	  setAlignment(DParagraph.LEFT);
	  doChildrenItem(d);
	  
	  restoreFont();    
   }
   
   /**
	* @param g
	* @roseuid 3EFC03750268
	*/
   public void doItem(DGraphic g) 
   {
   	  applyFont(g);
   	
	  int w = (int)((float)getMaxWidth() * g.getWidthScale());
	  int h = (int)((float)getMaxWidth() * g.getHeightScale());
	  float ratio = g.getWidthScale()/g.getHeightScale();
	  if (ratio <= 1.0f)
	  {
		if (w < 100)
		{
			w = 100;
			h = (int)(w/ratio);
		}
		else
		if (h < 100)
		{
			h = 100;
			w = (int)(h*ratio);
      		
		}
	  }
      
	  addCell(w, h, g, 0);
	  
	  restoreFont();     
   }
   
   /**
	* @param i
	* @roseuid 3EFC02FC00CA
	*/
   public void doItem(DImage i) 
   {
	 _addListItemPrefix(i);
	  
	 // search image 
	 Image img = (Image)getImagePool().get(i.getName());
	 if (img == null)
	 {
		String imgfilename = searchImage(i.getName());
		if (imgfilename != null)
		{
		   img = new Image(getDevice(), i.getName());
		   getImagePool().put(i.getName(), img);
		}
		else
		{
		   img = (Image)getBrokenImage();
		}
	 }
  
	 Rectangle r = img.getBounds();
	 int w = (int)(r.width * i.getWidthScale());
	 int h = (int)(r.height * i.getHeightScale());
	      
	 if (getWordWrap() 
		 && (w + getCurrentWLine()) > getMaxWidth())
	 {
		 newLine();	
	 }
     
	 TCell c = addCell(w, h, i, 0);
	 c.setImage(img);    
   }
   
   
   /**
	* @param l
	* @roseuid 3EFC02AF03CC
	*/
   public void doItem(DList l) 
   {
   	  applyFont(l);
   	
	  newLine();

	  IDStyle s = l.getStyle();
	
	  addListLevel();
	  setXoffset(getXoffset() * getListOffsetSpacing());
		
	  doChildrenItem(l);
	
	  removeListLevel();
	  setXoffset(getXoffset() * getListOffsetSpacing());
	
	
	  newLine();   
	
	  restoreFont(); 
   }
   
   /**
	* @param l
	* @roseuid 3EFC02AF03CC
	*/
   public void doItem(DLine l) 
   {
	   newLine();
	
	   addCell(getMaxWidth(),
				l.getSize()+getCellSpacing()*2,l,0);
	
	   newLine();    
   }
   
   /**
	* @param pb
	* @roseuid 3F004C3A029E
	*/
   public void doItem(DPageBreak pb) 
   {    
	  if (getModePage())
		  newPage();
	  else
		  newLine();
   }
   
   /**
	* @param pc
	* @roseuid 3F004C20039B
	*/
   public void doItem(DPageCounter pc) 
   { 
   	  applyFont(pc);
   	
	  String s = new Integer(getCurrentNPage()).toString();
      
	  Point p = getGc().textExtent(s);
	  TCell c = addCell(p.x, p.y, pc, s.length());
	  c.setText(s); 
	  
	  restoreFont();   
   }
   
   /**
	* @param p
	* @roseuid 3EFC025D0388
	*/
   public void doItem(DParagraph p) 
   {
   	  applyFont(p);
   	
	  int al = getAlignment();
	  newLine();
	  setAlignment(p.getAlign());
	  
	  _addListItemPrefix(p);
	  
	  doChildrenItem(p);
	  
	  newLine();
	  setAlignment(al);
	  
	  restoreFont();    
   }
   
   /**
	* @param p
	* @roseuid 3EFC02DE0329
	*/
   public void doItem(DPopup p) 
   {
   	  applyFont(p);
   	
	  if (getDrawPopup())
	  {
		setAlignment(DParagraph.LEFT);
		doChildrenItem(p);
	  }    
	  
	  restoreFont();
   }
   
   /**
	* @param p
	* @roseuid 3EFC02CA0280
	*/
   public void doItem(DPopupLink p) 
   {
   	 applyFont(p);
   	
	 doChildrenItem(p);
	 if (!getFlatPopup()) return;
	 
	 DLinkResolver r = new DLinkResolver(p);
	 DPopup popup = (DPopup)r.getTaggedItem();
	 
	 if (popup!=null)
	 {
		TBuilder b = _newBuilder(getMaxWidth());
	 	
		int m = getCellSpacing();
		b.setTopMargin(m);
		b.setBottomMargin(m);
		b.setLeftMargin(m);
		b.setRightMargin(m);
		
		b.doChildrenItem(popup);
		
		TCell c = b.getFirstCell();
		
		addCell(c.getRect().width, c.getRect().height, popup, 0);   
		setCurrentPosition(b.getCurrentPosition());

		// on met fin au constructeur 
		_disposeBuilder(b);	
	 }    
	 
	 restoreFont();
   }
   
   /**
	* @param r
	* @roseuid 3EFC03190216
	*/
   public void doItem(DRow r) 
   {
   	  applyFont(r);
   	
	  currentNColumn = 0;
	  currentRow = r;
	  maxCellHeight = 0;
	  
	  // calcul numero de colonne de la premiere cellule de la liste
	  // a partir de la deuxieme ligne
	  if (currentNRow>0)
	  {
		for (currentNColumn=0; 
			   (currentRowSpan[currentNColumn].rowspan>0) 
				  && (currentNColumn<currentRowSpan.length); currentNColumn++);
	  }
	  
	  
	  doChildrenItem(r);
	  
	  if (currentNRow>0)
	  {
		int w = 0;
		for (int i=0; i<currentRowSpan.length; i++)
		{
		  if (currentRowSpan[i]==null) continue;
		  TCell c = currentRowSpan[i].cell;
		  
		  if (currentRowSpan[i].rowspan>0)
		  { 
			 w += c.getRect().width;
			 currentRowSpan[i].rowspan -= 1;
			 if (currentRowSpan[i].rowspan==0)
			 {
				c.getRect().height = (getCurrentYPos() + maxCellHeight) - c.getRect().y;
				currentRowSpan[i] = null;
			 }
		  }
		  else
		  {
			c.getRect().x = c.getRect().x + w;			
			c.getRect().height = maxCellHeight;
			w += c.getRect().width;
		  }
		}
	  }
	  
	  int currentYPos = getCurrentYPos();
	  newLine();
	  setCurrentYPos(maxCellHeight + currentYPos);
	  currentNRow = currentNRow + 1;
	  
	  restoreFont();    
   }
   
   /**
	* @param s
	* @roseuid 3EFC02410247
	*/
   public void doItem(DSection s) 
   {
   	  applyFont(s);
   	
	  setHeader(s.getHeader());
	  setFooter(s.getFooter());
	  
	  if (getModePage())
		  newPage();
	  else
		  newLine();
          
	  addCell(0,0,s,0);
      
	  doChildrenItem(s);
	  
	  restoreFont();    
   }
   
   /**
	* @param t
	* @roseuid 3EFC030901EB
	*/
   public void doItem(DTable t) 
   {
   	  applyFont(t);
   	
	  newLine();
	  currentTable = t;
	  currentNRow = 0;
	  currentTotalColumn = t.getColumnCount();
	  currentRowSpan = new CellInfo[t.getColumnCount()];
      
      
	  doChildrenItem(t);    
	  
	  currentTable = null;
	  currentRow = null;
	  currentRowSpan = null;
	  
	  restoreFont(); 
   }
   
   /**
	* @param t
	* @roseuid 3EFC03A001FC
	*/
   public void doItem(DTag t) 
   {
	  addCell(0,0,t,0);    
   }
   
   /**
	* @param t
	* @roseuid 3EFC02EE0142
	*/
   public void doItem(DText t) 
   {
   	  applyFont(t);
   	  
	  _addListItemPrefix(t);
	  _buildTextBuffer(t, t.getText());
	  
	  restoreFont();    
   }
   
   /**
	* @param t
	* @roseuid 3EFC02A00244
	*/
   public void doItem(DTitle t) 
   {
   	  applyFont(t);
   	
	  addTitleLevel(t.getLevel());
	  String st = getTitleStringPrefix(t);
	  
	  Point r = getGc().textExtent(st);
	  
	  addYSpacing(getTitleSpacing());
	  newLine();
	  
	  TCell c = addCell(r.x, r.y + getTitleSpacing(), t, st.length());
	  c.setText(st);
	  
	  doChildrenItem(t);
	  
	  newLine();   
	  
	  restoreFont(); 
   }
   
   /**
	* @param f
	* @roseuid 3F02AEBE025D
	*/
   public void doItem(DFooter f) 
   {
   	   applyFont(f);
   	
	   //	fabrication de l'entete et pied de base
	   TBuilder b = _newBuilder (getMaxWidth());
		
	   b.doChildrenItem(f);
		
	   setFooterCell(b.getFirstCell());
	   b.setFirstCell(null);
			
	   // on met fin au constructeur 
	   _disposeBuilder(b);
	   
	   restoreFont();    
   }
   
   /**
	* @param h
	* @roseuid 3F02AEAC03AC
	*/
   public void doItem(DHeader h) 
   {
   	   applyFont(h);
   	
	   // fabrication de l'entete et pied de base
	   TBuilder b = _newBuilder(getMaxWidth());
	 	
	   b.doChildrenItem(h);
		
	   setHeaderCell(b.getFirstCell());
	   b.setFirstCell(null);
			
	   // on met fin au constructeur 
	   _disposeBuilder(b);
	   
	   restoreFont();    
   }
   
   /**
	* @param text
	* @param w
	* @param r
	* @roseuid 3F02AF060329
	*/
   private void _buildGetNextWord(String text, int w, Range r) 
   {
	  String qs = text.substring(r.start);
	  int wt = getGc().textExtent(qs).x;
	  if (wt <= w)
	  {
		r.end = wt;
		r.start = text.length(); 
	  }
	  else
	  {
		int pos = ((w * qs.length()) / wt);
		for (int i=pos; i > 0; i = i -1)
		{
		   char c = qs.charAt(i);
		   if (!Character.isSpaceChar(c)
				&& (c!='.') && (c!=';') && (c!=',') && (c!='?')
				&& (c!=':') && (c!='(') && (c!=')') && (c!='/')
				&& (c!='[') && (c!=']') && (c!='\\'))
		   {
			  continue;
		   }

		   wt = getGc().textExtent(qs.substring(0, i)).x;
		   if (wt < w)
		   {
			  r.end = wt;
			  r.start = r.start + i;
			  return;
		   }
		}
		// on est pas arrive a decouper la chaine sur des separateurs
		// on va decouper les mots par dicotomie
		int a = qs.length() / 2;
		int b = qs.length();
		while (b-a>1)
		{
		   String s = qs.substring(0, a);
		   wt = getGc().textExtent(s).x;
		   if (wt < w)
		   {
			  a = a + (b-a)/2;
		   }
		   else
		   {
			  b = a;
			  a = a / 2;
		   }
		}
		r.end = wt;
		r.start = a + r.start;
	  }
   }

}



/*
 * Created on Sep 19, 2003
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package com.rational.apidoc.drivers.html;

import java.util.LinkedList;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;

import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.ImageLoader;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.SWTError;

import com.rational.apidoc.parsers.DParser;
import com.rational.apidoc.drivers.IWriter;
import com.rational.apidoc.core.*;
import com.rational.apidoc.examples.*;
import com.rational.apidoc.chart.ChartSnapShot;

/**
 * @author sduguet
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
public class DHtmlDriver extends DParser implements IWriter {

	private ConfParser confparser;
	private PrintStream out;

	private String dirname, filename;
	private DDocument document;
	
	private static Item default_hitem = new Item();

	/**
	 * 
	 */
	public DHtmlDriver() {
		super();
		confparser = new ConfParser();
		if (!confparser.parse()) {
			System.err.println("Html Configuration parsing failed");
			confparser = null;
			return;
		}
	}

	public void setDestDirAndFileName(String dname, String fname) {
		dirname = dname != null ? dname : ".";
		filename =
			dirname + "/" + fname + confparser.getHtmlFile().getExtension();
	}

	public PrintStream getStream() {
		return out;
	}

	public void htmlConvertDump(String string) {
		int space_count = 0;
		for (int i = 0; i < string.length(); i++) {
			char c = string.charAt(i);
			if (c == ' ') {
				space_count++;
				if (space_count == 1) {
					out.print(confparser.getConvertText().getFirstSpace());
				} else {
					out.print(confparser.getConvertText().getExtraSpace());
				}
			} else {
				space_count = 0;
				Character ch = new Character(c);
				String repl = confparser.getConvertText().getReplace(ch);
				String cstr = repl != null ? repl : ch.toString();
				try {
					byte a[] = cstr.getBytes("UTF-8");
					out.write(a, 0, a.length);
				} catch (UnsupportedEncodingException e) {
					out.print(cstr);
				}
			}
		}
	}

	public void htmlDump(String string) {
		try {
			byte a[] = string.getBytes("UTF-8");
			out.write(a, 0, a.length);
		} catch (UnsupportedEncodingException e) {
			out.print(string);
		}
	}

	public void htmlDump(int value) {
		out.print(value);
	}

	public void htmlDump(LinkedList list) {
		for (int i = 0; i < list.size(); i++) {
			String elt = (String) list.get(i);
			String value = confparser.getVariableValue(elt);
			if (value == "#") {
				htmlDump("${" + elt + "}");
			} else {
				htmlDump(value != null ? value : elt);
			}
		}
	}

	private void setTagVariable(String name, String tag, IDItem item) {
		confparser.setVariableValue(
			name,
			tag != null ? tag : "_" + item.hashCode(),
			item);
	}

	private String to2DigitsHexa(int color) {
		return Integer.toHexString((color % 256) / 16)
			+ Integer.toHexString(color % 16);
	}

	private String getColorHexa(IDColor color, String default_value) {
		if (color == null) {
			return default_value;
		}
		String result = "#";
		result += to2DigitsHexa(color.getRed());
		result += to2DigitsHexa(color.getGreen());
		result += to2DigitsHexa(color.getBlue());
		return result;
	}

	private String getFontAttrsBegin(IDStyle style) {
		IDFont font = style.getFont();
		IDColor color = style.getForeColor();
		if (font == null && color == null) {
			return "";
		}
		String result = "";
		if (font != null) {
			if (font.getBold()) {
				result += "<B>";
			}
			if (font.getUnderline() || font.getDashed()) {
				result += "<U>";
			}
			if (font.getItalic()) {
				result += "<I>";
			}
		}
		result += "<FONT COLOR=" + getColorHexa(color, "#000000");
		if (font != null && font.getSize() != -1) {
			int s = font.getSize() / 3; // size 12 -> 4 (normal size)
			if (s < 1) {
				s = 1; // 1 is the minimum
			}
			if (s > 7) {
				s = 7; // 7 is the maximum
			}
			result += " SIZE=\"" + Integer.toString(s) + "\"";
		}
		result += ">";
		if (font != null && confparser.isTTFont(font.getName())) {
			result += "<TT>";
		}

		return result;
	}

	private String getFontAttrsEnd(IDStyle style) {
		IDFont font = style.getFont();
		IDColor color = style.getForeColor();
		if (font == null && color == null) {
			return "";
		}
		String result = "";
		if (font != null && confparser.isTTFont(font.getName())) {
			result += "</TT>";
		}
		result += "</FONT>";
		if (font != null) {
			if (font.getBold()) {
				result += "</B>";
			}
			if (font.getUnderline() || font.getDashed()) {
				result += "</U>";
			}
			if (font.getItalic()) {
				result += "</I>";
			}
		}
		return result;
	}

	private void updateStyleVariables(IDItem item) {
		// PRE: item.getStyle() not null (to be checked by the caller)
		confparser.setVariableValue(
			"HTML_FONT_ATTRS_BEGIN",
			getFontAttrsBegin(item.getStyle()),
			item);
		confparser.setVariableValue(
			"HTML_FONT_ATTRS_END",
			getFontAttrsEnd(item.getStyle()),
			item);
		String fore_color =
			getColorHexa(item.getStyle().getForeColor(), "#000000");
		confparser.setVariableValue(
			"HTML_COLOR",
			"=\"" + fore_color + "\"",
			item);
		confparser.setVariableValue("HTML_FORE_COLOR", fore_color, item);
		String back_color =
			getColorHexa(item.getStyle().getBackColor(), "#ffffff");
		confparser.setVariableValue(
			"HTML_BGCOLOR",
			"=\"" + back_color + "\"",
			item);
		confparser.setVariableValue("HTML_BACK_COLOR", back_color, item);
	}

	public Item getItem(IDItem item) {
		// Perform some checks
		Item hitem = confparser.getItem(item.getClass());
		if (hitem == null) {
			System.err.println("No html parser found for " + item.getClass());
			hitem = default_hitem;
		}
		return hitem;
	}

	public void doOneItemAfter(IDItem item) {
		if (item instanceof DText) {
		} else if (item instanceof DCellText) {
		} else if (item instanceof DGraphic) {
			graphicDump((DGraphic) item);
		}
	}

	public void doItem(DDocument item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("DOCUMENT_TAG", item.getTag(), item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DSection item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("SECTION_TAG", ((DSection) item).getTag(), item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DHeader item) {
		// ignored
	}

	public void doItem(DFooter item) {
		// ignored
	}

	public void doItem(DParagraph item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"PARAGRAPH_ALIGN",
			((DParagraph) item).alignToString(((DParagraph) item).getAlign()),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DTag item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("TAG_NAME", ((DTag) item).getTag(), item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DText item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		htmlDump(hitem.getBegin());
		htmlConvertDump(item.getText());
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DImage item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"IMAGE_NAME",
			((DImage) item).getName(),
			item);
		htmlDump(hitem.getBegin());
		// no childs
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DList item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		if (((DList) item).getFormat() == DStyle.BULLET) {
			confparser.setVariableValue("LIST_TYPE", "BULLETED", item);
			confparser.setVariableValue("HTML_LIST_TYPE", "UL", item);
		} else if (((DList) item).getFormat() == DStyle.NUMBER) {
			confparser.setVariableValue("LIST_TYPE", "NUMBERED", item);
			confparser.setVariableValue("HTML_LIST_TYPE", "OL", item);
		} else {
			confparser.setVariableValue("LIST_TYPE", "", item);
			confparser.setVariableValue("HTML_LIST_TYPE", "UL", item);
		}
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DTable item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		if (((DTable) item).getAutoResize()) {
			float w = 0;
			float wc[] = ((DTable) item).getColumnWidth();
			if (wc == null || wc.length < ((DTable) item).getColumnCount()) {
				w = ((DTable) item).getWidthScale();
			} else {
				for (int i = 0; i < wc.length; i++) {
					w += wc[i];
				}
			}
			confparser.setVariableValue("TABLE_WIDTH", Float.toString(w), item);
			confparser.setVariableValue(
				"HTML_TABLE_WIDTH",
				Integer.toString((int) (w * 100.0f)),
				item);
		} else {
			float w = 0.0f;
			confparser.setVariableValue("TABLE_WIDTH", Float.toString(w), item);
			confparser.setVariableValue(
				"HTML_TABLE_WIDTH",
				Integer.toString((int) (w * 100.0f)),
				item);
		}
		confparser.setVariableValue(
			"TABLE_BORDER",
			Integer.toString(((DTable) item).getBorder()),
			item);
		confparser.setVariableValue(
			"TABLE_COLUMNS",
			Integer.toString(((DTable) item).getColumnCount()),
			item);
		confparser.setVariableValue(
			"TABLE_ROWS",
			Integer.toString(((DTable) item).getRowCount()),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DRow item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"ROW_CELLS",
			Integer.toString(((DRow) item).getCellCount()),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	private void setCellVariables(DCell item) {
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"CELL_COLSPAN",
			Integer.toString(((DCell) item).getColSpan()),
			item);
		confparser.setVariableValue(
			"HTML_CELL_COLSPAN",
			Integer.toString(((DCell) item).getColSpan() + 1),
			item);
		confparser.setVariableValue(
			"CELL_ROWSPAN",
			Integer.toString(((DCell) item).getRowSpan()),
			item);
		confparser.setVariableValue(
			"HTML_CELL_ROWSPAN",
			Integer.toString(((DCell) item).getRowSpan() + 1),
			item);
		confparser.setVariableValue(
			"CELL_COL",
			Integer.toString(((DCell) item).getColumnNumber()),
			item);
		confparser.setVariableValue(
			"CELL_ROW",
			Integer.toString(((DCell) item).getLineNumber()),
			item);
		DRow row = (DRow) item.getParent();
		DTable table = (DTable) row.getParent();
		if (table.getColumnWidth() != null
			&& table.getColumnCount() > ((DCell) item).getColumnNumber()) {
			float w = table.getColumnWidth()[((DCell) item).getColumnNumber()];
			confparser.setVariableValue("CELL_WIDTH", Float.toString(w), item);
			confparser.setVariableValue(
				"HTML_CELL_WIDTH",
				Integer.toString((int) (w * 100.0f)),
				item);
		}
	}

	public void doItem(DCell item) {
		Item hitem = getItem(item);
		setCellVariables(item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DCellText item) {
		Item hitem = getItem(item);
		setCellVariables(item);
		htmlDump(hitem.getBegin());
		htmlConvertDump(((DCellText) item).getText());
		// no childs
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DLink item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		String url =
			(((DLink) item).isInternal() ? "" : ((DLink) item).getMethod())
				+ ":"
				+ ((DLink) item).getPath();
		confparser.setVariableValue("LINK_TARGET", url, item);
		confparser.setVariableValue("HTML_LINK", url, item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DLine item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"HTML_LINE_SIZE",
			Integer.toString(((DLine) item).getSize()),
			item);
		htmlDump(hitem.getBegin());
		// no childs
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DPageBreak item) {
		// ignored
	}

	public void doItem(DPageCounter item) {
		// ignored
	}

	public void doItem(DPopup item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("POPUP_TAG", ((DPopup) item).getTag(), item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DPopupLink item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"POPUP_LINK_METHOD",
			((DPopupLink) item).typeToString(((DPopupLink) item).getType()),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DSummary item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DTitle item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("TITLE_TAG", ((DTitle) item).getTag(), item);
		addTitleLevel(((DTitle) item).getLevel());
		confparser.setVariableValue(
			"TITLE_LEVEL",
			Integer.toString(getTitleLevel()),
			item);
		confparser.setVariableValue(
			"HTML_TITLE_LEVEL",
			Integer.toString(getTitleLevel()),
			item);
		confparser.setVariableValue(
			"TITLE_FULL_NUMBER",
			getTitleStringPrefix(item),
			item);
		confparser.setVariableValue(
			"TITLE_NUMBER",
			Integer.toString(titleTab[getTitleLevel()]),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DIndex item) {
	}

	public void doItem(DIndexEntry item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DTimerItemContainer item) {
		// ignored
	}

	private void graphicDump(DGraphic graphic) {
		String image_filename =
			dirname
				+ "/"
				+ confparser.getVariableValue("HTML_GRAPHIC_FILE_NAME")
				+ ".jpg";
		int w = 500, h = w;
		w *= graphic.getWidthScale();
		h *= graphic.getHeightScale();
		Rectangle r = new Rectangle(0, 0, w, h);
		Image img = ChartSnapShot.Process(r, graphic);
		ImageLoader loader = new ImageLoader();
		if (loader.data == null) {
			loader.data = new ImageData[1];
		}
		loader.data[0] = img.getImageData();
		try {
			loader.save(image_filename, SWT.IMAGE_JPEG);
		} catch (SWTException _e) {
			System.err.println(
				"Error saving image:" + image_filename + ": " + _e);
		} catch (SWTError _r) {
			System.err.println(
				"Error saving image:" + image_filename + ": " + _r);
		}
		img.dispose();
	}

	public void doItem(DGraphic item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"GRAPHIC_TITLE",
			((DGraphic) item).getTitle(),
			item);
		confparser.setVariableValue(
			"GRAPHIC_WIDTH",
			Float.toString(((DGraphic) item).getWidthScale()),
			item);
		confparser.setVariableValue(
			"HTML_GRAPHIC_WIDTH",
			Integer.toString(
				(int) (((DGraphic) item).getWidthScale() * 100.0f)),
			item);
		confparser.setVariableValue(
			"GRAPHIC_HEIGHT",
			Float.toString(((DGraphic) item).getHeightScale()),
			item);
		confparser.setVariableValue(
			"HTML_GRAPHIC_FILE_NAME",
			"img" + Integer.toHexString(item.hashCode()),
			item);
		htmlDump(hitem.getBegin());
		graphicDump(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DAxis item) {
		// already made from the enclosing DGraphic
	}

	public void doItem(DCurve item) {
		// already made from the enclosing DGraphic
	}

	public void doItem(DPoint item) {
		// already made from the enclosing DGraphic
	}

	public void write(IDItem doc) {
		document = (DDocument) doc;

		if (confparser == null) {
			System.err.println("Html Configuration parsing failed");
			return;
		}
		try {
			out = new PrintStream(new FileOutputStream(filename));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return;
		}

		// Some variables at document level
		confparser.setVariableValue("HTML_FILE_NAME", filename, document);
		confparser.setVariableValue(
			"DOCUMENT_TITLE",
			document.getTitle(),
			document);
		confparser.setVariableValue("HTML_BGCOLOR", "=\"#ffffff\"", document);

		setTitleMaxLevel(confparser.getHtmlFile().getMaxTitleLevel() - 1);
		HtmlFile hf = confparser.getHtmlFile();
		htmlDump(hf.getBegin());
		doItem(document);
		htmlDump(hf.getEnd());

		out.close();
	}

	public static void main(String[] args) {
		System.out.println("initialize...");
		DHtmlDriver hd = new DHtmlDriver();

		System.out.println("create...");

//		Example1 ex1 = new Example1();
//		hd.setDestDirAndFileName(".", "file1");
//		System.out.println("translate...");
//		hd.write(ex1.doc);
//		hd.clear();
//
//		Example2 ex2 = new Example2();
//		hd.setDestDirAndFileName(".", "file2");
//		System.out.println("translate...");
//		hd.write(ex2.doc);
//		hd.clear();
//
//		Example3 ex3 = new Example3();
//		hd.setDestDirAndFileName(".", "file3");
//		System.out.println("translate...");
//		hd.write(ex3.doc);
//		hd.clear();
//
//		Example4 ex4 = new Example4();
//		hd.setDestDirAndFileName(".", "file4");
//		System.out.println("translate...");
//		hd.write(ex4.doc);
//		hd.clear();
//
//		Example5 ex5 = new Example5();
//		hd.setDestDirAndFileName(".", "file5");
//		System.out.println("translate...");
//		hd.write(ex5.doc);
//		hd.clear();
//
//		Example6 ex6 = new Example6();
//		hd.setDestDirAndFileName(".", "file6");
//		System.out.println("translate...");
//		hd.write(ex6.doc);
//		hd.clear();
//
		Source so = new Source();
		hd.setDestDirAndFileName(".", "source");
		System.out.println("translate...");
		hd.write(so.doc);
		hd.clear();
		
		System.out.println("done!");
	}
}
/*
 * Created on 26 juin 2003
 *
 */
package com.rational.apidoc.chart;

import org.eclipse.swt.SWT;

import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Layout;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Button;

import com.rational.apidoc.drawutil.*;
import com.rational.apidoc.core.* ;

/**
 * Example to create and render a new graphic type for ApiDoc.
 * The new graphic will draw polar curves.
 * 
 * To extends graphics type, follow these steps:
 * - define a new id for graphic type: see T_POLAR field.
 * - define an IChartRender that able to draw your graphic
 *   if render have support to location, extends IChartLocation 
 *   (thrown by your ChartRender).
 * 
 * - don't forget to register the new IChartRender using DChartRenderRegistry
 *   (see it in main() method)
 * 
 * - now you can build your document containing new graphic type, see this
 *   in BuildPolarGraphic() method.
 */
public class TestUserGraphicType
{  
   /** Define a new graphic type identifier: */   
   public static final String T_POLAR = "T_POLAR.sample.chart.apidoc.rational.com";
   
   /** First axis a Polar graphic must provide */
   public static final String AXIS_RAY = "ray-axis";
   /** Second axis a Polar graphic must provide */
   public static final String AXIS_ANGLE ="angle-axis";
   
   /** Class used to identify where is located a point on a polar graphic */
   public static class PolarLocation implements IChartLocation
   {
      /** Located on graphic title area */
      public static final String GRAPHIC_TITLE="graphic-title";
      /** Located on a segment of a curve (item is first point of segment) */
      public static final String CURVE_SEGMENT="curve-segment";
      /** located on a point of a curve (only for curve type DCurve.POINT) */
      public static final String CURVE_POINT  ="curve-point";
      
      private String id_;
      private IDItem item_;
      private Rectangle rect_;
      
      public PolarLocation( String _id, IDItem _item, Rectangle _r )
      {
        id_=_id; item_=_item; rect_=_r;
      }
      
      public String getId() { return id_; }
      public IDItem getItem() { return item_; }
      public Rectangle getRectangle() { return rect_; }
   }
   
   /** 
    * Implements an IChartRender for this new graphic type
    */   
   private static class PolarChartRender implements IChartRender
   {
     public boolean canRender(String _graphic_type)
     {
       if( T_POLAR.equals( _graphic_type )) return true;
       return false;
     }

     public void draw(GC _gc, Rectangle _R, DGraphic _g)
     {   
       render( true, 0,0, _gc, _R, _g );
     }
     
     public IChartLocation locate(int _x, int _y, GC _gc, Rectangle _R, DGraphic _G)
     {
       return render( false, _x, _y, _gc, _R, _G );
     }
     
     /** support for both draw and location modes */
     public IChartLocation render( boolean _draw, int _lx, int _ly, GC _gc, Rectangle _R, DGraphic _g)
     {   
       if( _draw )
       {
         _gc.setBackground( new Color( null, 255,255,255 ));
         _gc.fillRectangle( _R );                
       }
       else
       {
         if( !_R.contains( _lx, _ly ) ) return null; //outside graphic
       }
       
       //search axis
       DAxis r_axis = null;
       DAxis a_axis = null;
       for( IDItem item=_g.getFirstChild(); item!=null; item=item.getNext() )
       {
         if( item instanceof DAxis )
         {
           DAxis axis = (DAxis)item;
           if( AXIS_RAY.equals( axis.getName() ) ) r_axis=axis;
           else if ( AXIS_ANGLE.equals( axis.getName() ) ) a_axis=axis;
         }
       }
       if( r_axis==null || a_axis==null )
       {
         if( _draw )
         {
           Color red = new Color( null, 255,0,0 );
           String missed = (r_axis==null?"'AXIS_RAY'":"")+" "+(a_axis==null?"'AXIS_ANGLE'":"");
           _gc.setForeground( red );
           _gc.drawText( "Error: missing axis "+missed+"' in graphic item", 10, 30 );
         }
         return null;
       }
       
       final int margin = 10;
       int lx = _R.x+margin;
       int ty = _R.y+margin;
       int rx = _R.x+_R.width-margin;
       int by = _R.y+_R.height-margin;
       int cx = (lx+rx)/2;
       int cy = (ty+by)/2; 
       
       //render title if any
       String title = _g.getTitle();
       if( !DrawUtil.isEmpty( title )
         && _g.getProperties().get( DGraphic.P_SHOW_TITLE, true ) )
       {
         IDStyle style = _g.getStyle();
         Font save = _gc.getFont();
         Font font = DrawUtil.getFont( style.getFont() );
         int aln = DrawUtil.HCENTER|DrawUtil.VCENTER|DrawUtil.VERTICAL;
         if( font!=null ) _gc.setFont( font );
         int tw = DrawUtil.textExtent( _gc, title, aln ).x;
         if (_draw)
         {
           //gradient background
           _gc.setForeground( new Color(null,205,220,255));
           _gc.fillGradientRectangle( lx-1, ty, tw+3, by-ty, true );
           //title color and text
           _gc.setForeground( DrawUtil.getColor(style.getForeColor()) );
           DrawUtil.drawText(_gc, lx, ty, tw, by - ty, aln, title);
         } else {
           Rect r = DrawUtil.boundingRect( _gc, lx, ty, tw, by-ty, aln, title );
           if( r.contains( _lx,_ly ) )
           {
             return new PolarLocation( PolarLocation.GRAPHIC_TITLE, _g, r.toRectangle() );
           }
           if( _lx <= lx+tw )
           {
             //located on graphic title area (outside text),
             //no need to search in curve, but no id is defined for this...
             return null;
           }
         }

         _gc.setFont( save );
         
         lx+=tw+margin;
         cx = (lx+rx)/2;
       }
       
       //first search curve and points to compute min/max
       //an array of DCurve is stored to be able to iterate in reverse order
       //as location mode need.
       DCurve curves[] = new DCurve[ _g.getChildCount() ];
       int    n_curves=0; //because curves.length take into account DAxis too... ;-(
       float xmin=0.0f, xmax=0.0f, ymin=0.0f, ymax=0.0f;
       float ray_max=0.0f;
       boolean init=false;       
       
       for( IDItem item=_g.getFirstChild(); item!=null; item=item.getNext() )
       {
         if( !(item instanceof DCurve) ) continue;
         DCurve curve = (DCurve)item;
         curves[ n_curves++ ] = curve;
         for( IDItem item2=curve.getFirstChild(); item2!=null; item2=item2.getNext())
         {
           if(!(item2 instanceof DPoint)) continue;
           DPoint point = (DPoint)item2;
           DCoord c_angle = point.getCoord( a_axis );
           DCoord c_ray   = point.getCoord( r_axis );
           if( c_angle==null || c_ray==null )
           {
             //no error message here, skip point only
             continue;
           }
           float r = c_ray.getValue();
           float a = c_angle.getValue();
           if( r > ray_max ) ray_max = r;
           //convert to pixels
           float ptx = (float)( r*Math.cos(a) );
           float pty = (float)( r*Math.sin(a) );
           if( init )
           {
             if( ptx < xmin ) xmin=ptx; else if ( ptx > xmax ) xmax=ptx;
             if( pty < ymin ) ymin=pty; else if ( pty > ymax ) ymax=pty;
           }
           else
           {
             xmin=xmax=ptx; 
             ymin=ymax=pty;
             init=true;
           }
         }
       }
       
       float K=1.0f;
       float c_max ;
       {
         c_max = xmax;
         float x = Math.abs( xmin ); if( x > c_max) c_max=x;
         if( ymax > c_max ) c_max=ymax;
         x = Math.abs( ymin ); if ( x > c_max ) c_max=x;
         int pix_min = cx-lx < cy-ty ? cx-lx : cy-ty;
         if( c_max > 0.0f ) K = pix_min / c_max;
       }
       
       //draw axis lines
       if( _draw )
       {
         _gc.setForeground( new Color( null, 220,220,220 ));
         Rectangle r = _gc.getClipping();
         _gc.setClipping( lx, ty, rx-lx, by-ty );
         //Line Step property is set on 'ray' axis ?
         float ray_step_line = r_axis.getProperties().get( DAxis.P_STEP_LINE, 0.0f );
         if( ray_step_line >=0.0f )
         {
            for( float ray=ray_step_line; ray <= c_max; ray += ray_step_line )
            {
              int p_ray = (int)( K*ray );
              int p_ray2 = 2*p_ray;
              _gc.drawOval( cx-p_ray, cy-p_ray, p_ray2, p_ray2 );
            }
         }
         
         //Line Step property is set on 'angle' axis ?
         float ang_step_line = a_axis.getProperties().get( DAxis.P_STEP_LINE, 0.0f );
         if( ang_step_line >=0.0f )
         {
           float dpi = (float)( 2*Math.PI );
           for( float a = ang_step_line; a < dpi; a+=ang_step_line )
           {
             int px = (int)( cx + K*c_max*Math.cos(a) );
             int py = (int)( cy + K*c_max*Math.sin(a) );
             _gc.drawLine( cx, cy, px, py );
           }
         }
         //Dot Step property is set on 'angle' axis ?
         float ang_step_dot = a_axis.getProperties().get( DAxis.P_STEP_DOT, 0.0f );
         if( ang_step_dot >=0.0f )
         {
           float dpi = (float)( 2*Math.PI );
           _gc.setLineStyle( SWT.LINE_DOT );
           for( float a = ang_step_dot; a < dpi; a+=ang_step_dot )
           {
             int px = (int)( cx + K*c_max*Math.cos(a) );
             int py = (int)( cy + K*c_max*Math.sin(a) );
             _gc.drawLine( cx, cy, px, py );
           }
           _gc.setLineStyle( SWT.LINE_SOLID );
         }
         
         _gc.setForeground( new Color( null, 0,0,0 ));
         _gc.drawLine( lx, cy, rx, cy );
         _gc.drawLine( cx, ty, cx, by );
         

         _gc.setClipping( r );
       }

       
       Color def_clr = new Color( null, 0,150,0 );
       //render curves...
       // - draw in forward order
       // - locate in reverse order as the painter algorythm request.
       int crv_shift= _draw ? 0: n_curves-1 ;
       for( int icrv=0; icrv<n_curves; ++icrv )
       {
         DCurve curve = curves[ Math.abs( icrv - crv_shift) ];
         //curve type is dots or lines ?
         boolean dots = DCurve.T_POINTS.equals( curve.getType() );
         if( _draw )
         {        
           Color clr = def_clr;
           IDStyle style = curve.getStyle();
           if( style!=null )
           {
             clr = DrawUtil.getColor( style.getBackColor() );
             if( dots )
             {
               _gc.setBackground( clr );
               clr = DrawUtil.getColor( style.getForeColor() );
             }
           }
           _gc.setForeground( clr );
         }
         
         boolean draw_line=false;
         int ppx=0, ppy=0;
         DPoint last_point=null;
         for( IDItem item2=curve.getFirstChild(); item2!=null; item2=item2.getNext())
         {
           if(!(item2 instanceof DPoint)) continue;
           DPoint point = (DPoint)item2;
           DCoord c_angle = point.getCoord( a_axis );
           DCoord c_ray   = point.getCoord( r_axis );
           if( c_angle==null || c_ray==null )
           {
             //no error message here, skip point only
             draw_line=false;
             last_point=null;
             continue;
           }
           float r = c_ray.getValue();
           float a = c_angle.getValue();
           //convert to pixels
           int px = (int)( cx + K*r*Math.cos(a) );
           int py = (int)( cy + K*r*Math.sin(a) );
           if( dots )
           {
             Rectangle rect = new Rectangle( px-2, py-2, 4, 4 );
             if( _draw )
             {
               _gc.fillRectangle( rect );
               _gc.drawRectangle( rect );
             }
             else
             {
               //point located on the curve Points
               if( rect.contains(_lx, _ly ) )
               {
                 return new PolarLocation( PolarLocation.CURVE_POINT, point, rect );
               }
             }
           }
           else //other curves draw using lines
           {
             if( draw_line )
             {
               if( _draw )
               {
                 _gc.drawLine( ppx, ppy, px, py );
               }
               else
               {
                 //point near this segment ?
                 if( DrawUtil.segmentContains( ppx, ppy, px, py, _lx,_ly ) )
                 {
                   return new PolarLocation( PolarLocation.CURVE_SEGMENT, last_point,
                           new Rectangle( _lx-1,_ly-1,2,2) ); 
                 }
               }
             } else {
               draw_line=true;
             } 
             ppx=px;
             ppy=py;
             last_point=point;
           }
         }
       }  
       
       return null;     
     }

   }

/*---------------------------------------------------------------------------------------
 * Code from here up to the end of file are only made to build the test.
 * You can see how to use your graphic type in an apidoc document in the
 * BuildPolarGraphic method.
 *---------------------------------------------------------------------------------------
 */
 
   /** used only for test shell */
   public static interface IBuildGraphic
   {
     DGraphic buildGraphic();
   }   

  private static DColor black = new DColor(0,0,0);
  private static DColor blue1 = new DColor(0,0,255);
  private static DColor blue2 = new DColor(80,80,255);
  private static DColor blue3 = new DColor(150,150,255);
  private static DColor blue4 = new DColor(180,180,255);
  private static DColor green = new DColor(0,255,0);
  private static DColor red   = new DColor(255,0,0);
  private static DColor magenta = new DColor(255,0,255);
  private static DColor white = new DColor(255,255,255);

  private static DFont font1 = new DFont( "f1",null,null, 14, DFont.BOLD );
  private static DFont df    = new DFont( "df", "Arial",null,10,DFont.NORMAL );
  
  private static IDStyle sd = new DStyle( "sd", df, black, white );        
  private static IDStyle s0 = new DStyle( "s0", font1, green, white );
  private static IDStyle s1 = new DStyle( "s1", df, black, red );
  private static IDStyle s2 = new DStyle( "s2", df, black, blue2 );
  private static IDStyle s3 = new DStyle( "s3", df, black, blue4 );
  private static IDStyle s4 = new DStyle( "s4", df, black, green );
  private static IDStyle s5 = new DStyle( "s5", df, black, magenta );
  private static IDStyle s6 = new DStyle( "s6", df, black, green );
     
  private static DPoint newDPoint( DCurve _p, DAxis _x, DAxis _y, float _cx, float _cy )
  {
    DPoint pt = new DPoint();
    _p.addChild( pt );
    pt.add( new DCoord( _x, _cx ));
    pt.add( new DCoord( _y, _cy ));
    return pt;
  }
  private static DPoint newDPoint3( DCurve _p, DAxis _x, DAxis _y, DAxis _z, float _cx, float _cy, float _cz )
  {
    DPoint pt = new DPoint();
    _p.addChild( pt );
    pt.add( new DCoord( _x, _cx ));
    pt.add( new DCoord( _y, _cy ));
    pt.add( new DCoord( _z, _cz ));
    return pt;
  }
    /**
   * Graphic document builder for XY graphic 
   */
  private static class BuildXYGraphic implements IBuildGraphic
  {
    public DGraphic buildGraphic()
   {
     DGraphic  gph = new DGraphic();  
     gph.setGraphicType( DGraphic.T_XY );
     gph.setTitle( "Graphic XY" );
     gph.setStyle( s0 );
//gph.setProperties().store(DGraphic.P_SHOW_TITLE, false);     
//gph.setProperties().store(DGraphic.P_SHOW_LEGEND, false); 

     DAxis  x_axs = new DAxis("x");    gph.addChild( x_axs );
     DAxis  y_axs = new DAxis("y");    gph.addChild( y_axs );
     x_axs.setStyle( sd );
     y_axs.setStyle( sd );
     x_axs.getProperties().store( DAxis.P_STEP_UNIT, 2.0f );
     x_axs.getProperties().store( DAxis.P_STEP_DOT, 0.5f );
     y_axs.getProperties().store( DAxis.P_STEP_UNIT, 2.0f );
     y_axs.getProperties().store( DAxis.P_STEP_DOT, 0.5f );
     x_axs.setTitle( "X Axis");
     x_axs.setUnit("(x)");
     y_axs.setTitle( "Y Axis");
     y_axs.setUnit("(unit)");
     
     
     DCurve    cv0 = new DCurve();            gph.addChild( cv0 );
     cv0.setName( "Bars" );
     cv0.setType( DCurve.T_BAR );
     cv0.setStyle( s4 );
     newDPoint( cv0, x_axs, y_axs, 0, 0.9f );
     newDPoint( cv0, x_axs, y_axs, 1, 1.2f );
     newDPoint( cv0, x_axs, y_axs, 2, 1.6f );
     newDPoint( cv0, x_axs, y_axs, 3, 3.4f );
     newDPoint( cv0, x_axs, y_axs, 4, 4.6f );

     DCurve    cv1 = new DCurve();            gph.addChild( cv1 );
     cv1.setName("Line");
     cv1.setStyle( s1 );
     newDPoint( cv1, x_axs, y_axs, 0,0 );
     newDPoint( cv1, x_axs, y_axs, 1,1.2f );
     newDPoint( cv1, x_axs, y_axs, 2,3.2f );
     newDPoint( cv1, x_axs, y_axs, 3,2.4f );
     newDPoint( cv1, x_axs, y_axs, 4,0.6f );

     DCurve    cv2 = new DCurve();            gph.addChild( cv2 );
     cv2.setName( "Points" );
     cv2.setType( DCurve.T_POINTS );
     cv2.setStyle( s2 );
     newDPoint( cv2, x_axs, y_axs, 0, 1.9f );
     newDPoint( cv2, x_axs, y_axs, 1, 1.2f );
     newDPoint( cv2, x_axs, y_axs, 2, 1.1f );
     newDPoint( cv2, x_axs, y_axs, 3, 0.4f );
     newDPoint( cv2, x_axs, y_axs, 4, 0.0f );
     
     x_axs.getProperties().store( DAxis.P_MIN, 0.0f );
     x_axs.getProperties().store( DAxis.P_MAX, 5.0f );
     x_axs.getProperties().store( DAxis.P_STEP_LINE, 1.0f );
     y_axs.getProperties().store( DAxis.P_MIN, 0.0f );
     y_axs.getProperties().store( DAxis.P_MAX, 5.0f );
     y_axs.getProperties().store( DAxis.P_STEP_LINE, 1.0f );
     
     return gph;
   }
  }
  
  /**
   * Graphic document builder for Polar graphic 
   */
  private static class BuildPolarGraphic implements IBuildGraphic
  {
    public DGraphic buildGraphic()
   {
     DGraphic  gph = new DGraphic();  
     // Set my graphic type 
     gph.setGraphicType( T_POLAR );
     gph.setTitle( "Polar Graphic" );
     gph.setStyle( s0 );

     DAxis  ray_axs = new DAxis( AXIS_RAY );    gph.addChild( ray_axs );
     DAxis  angle_axs = new DAxis( AXIS_ANGLE );    gph.addChild( angle_axs );
     ray_axs.setStyle( sd );
     angle_axs.setStyle( sd );
     ray_axs.getProperties().store( DAxis.P_STEP_UNIT, 2.0f );
     ray_axs.getProperties().store( DAxis.P_STEP_DOT, 0.5f );
     angle_axs.getProperties().store( DAxis.P_STEP_UNIT, Math.PI/8 );
     angle_axs.getProperties().store( DAxis.P_STEP_DOT, Math.PI/16 );
     
     DCurve    cv0 = new DCurve();            gph.addChild( cv0 );
     cv0.setName( "Petals" );
     cv0.setType( DCurve.T_LINE );
     cv0.setStyle( s4 );
     DCurve    cv1 = new DCurve();            gph.addChild( cv1 );
     cv1.setName("Spiral");
     cv1.setType( DCurve.T_POINTS );
     cv1.setStyle( s1 );
     DCurve    cv2 = new DCurve();            gph.addChild( cv2 );
     cv2.setName("Curve 3");
     cv2.setStyle( s2 );
     DCurve    cv3 = new DCurve();            gph.addChild( cv3 );
     cv3.setName("Curve 4");
     cv3.setStyle( s5 );
     
     //populate curve using various function ( ray = f( angle ) )
     float da = (float)( Math.PI/100 );
     float r;
     float rmax=0.0f;
     int cnt=10;
     for( float a=0.0f; a<6*Math.PI; a+=da )
     {
       float ca = (float)Math.cos(a);
       float sa = (float)Math.sin(a);
       
       r = (float)( (a/(10*da))*ca*sa );
       if( r > rmax ) rmax=r;
       newDPoint( cv0, ray_axs, angle_axs, r, a );
       r = (float)( a );
       if( --cnt == 0 )
       {
         if( r > rmax ) rmax=r;
         newDPoint( cv1, ray_axs, angle_axs, r, a );
         cnt=10;
       }
       r = (float)(8*ca - 10.0f*sa/(1.1f+ca));
       if( r > rmax ) rmax=r;
       newDPoint( cv2, ray_axs, angle_axs, r, a );
       r = (float)(20+3*Math.cos(10*a)-20*Math.sin(5*a)*Math.cos(2*a));
       if( r > rmax ) rmax=r;
       newDPoint( cv3, ray_axs, angle_axs, r, a );
     }
   
     ray_axs.getProperties().store( DAxis.P_MIN, 0.0f );
     ray_axs.getProperties().store( DAxis.P_MAX, 1.1f*rmax );
     ray_axs.getProperties().store( DAxis.P_STEP_LINE, rmax/5 );
     angle_axs.getProperties().store( DAxis.P_MIN, 0.0f );
     angle_axs.getProperties().store( DAxis.P_MAX, 5.0f );
     angle_axs.getProperties().store( DAxis.P_STEP_LINE, Math.PI/8 );
     
     return gph;
   }
  }
  
  private static class RenderGraphic implements MouseListener, MouseMoveListener, PaintListener
  {
    private Canvas canvas_;
    private IChartRender render_;
    private DGraphic gph_;
    private boolean mouse_down_=false;
    private int mx_=0, my_=0;
    
    public RenderGraphic( Canvas _cvs )
    {
      canvas_=_cvs; 
      canvas_.addMouseListener( this );
      canvas_.addMouseMoveListener( this );
      canvas_.addPaintListener( this );
    }
    
    public void setGraphic( DGraphic _g )
    {
      gph_ = _g;
      render_ = DChartRenderRegistry.GetRender( _g.getGraphicType() );
      canvas_.redraw();
    }
    
    public void mouseMove(MouseEvent e)
    {
      if( render_==null || gph_==null ) return ;
      if( mouse_down_ &&( gph_.isGraphicType( DGraphic.T_XYZ )
                      || gph_.isGraphicType( DGraphic.T_SURFACE)) )
      {
        move3DView(e);
        return;
      }
          GC gc = new GC(canvas_);   
          Rectangle r = canvas_.getClientArea();
          IChartLocation loc = render_.locate( e.x,e.y, gc, r, gph_ );
//System.out.println("GOT LOC="+loc);          
          if( loc!=null && loc.getItem() instanceof DPoint )
          {
            DPoint pt = (DPoint)loc.getItem();
            DCurve crv= (DCurve)pt.getParent();
            DGraphic gph=(DGraphic)crv.getParent();
            String name = crv.getName();
            if( name==null ) name=""; else name=name+": ";
            for( IDItem item=gph.getFirstChild(); item!=null; item=item.getNext() )
            {
              if( item instanceof DAxis )
              {
                DAxis axis=(DAxis)item;
                DCoord crd= pt.getCoord( axis );
                if( crd!=null )
                {
                  name = name + Float.toString( crd.getValue() )+"  ";
                }
              }
            }
            canvas_.setToolTipText( name );
          } else {
            canvas_.setToolTipText( null );
          }
          gc.dispose();
          gc = null;
        }
    public void mouseDoubleClick(MouseEvent e){}

    public void mouseDown(MouseEvent e)
    {
      mouse_down_=true;
      mx_=e.x;
      my_=e.y;      
    }
    public void mouseUp(MouseEvent e)
    {
      mouse_down_=false;
    }     
    protected void move3DView(MouseEvent e)
    {  
      if( render_==null || gph_==null ) return ;

      float phi   = gph_.getProperties().get( DGraphic.P_XYZ_PHI, 0.3f);
      float theta = gph_.getProperties().get( DGraphic.P_XYZ_THETA, 0.2f);
      int dx = e.x -mx_;
      int dy = e.y -my_;
      
      phi   += 0.02f*dy;
      theta -= 0.02f*dx;

      gph_.getProperties().store( DGraphic.P_XYZ_PHI,  phi    );
      gph_.getProperties().store( DGraphic.P_XYZ_THETA, theta );
      mx_ = e.x;
      my_ = e.y;    
      canvas_.redraw();  
    }
    
    private Image img_;
    public void paintControl(PaintEvent e)
    {
       if( render_==null || gph_==null ) return ;
       Rectangle r = canvas_.getClientArea();
       if( img_==null || img_.getBounds().width!=r.width || img_.getBounds().height!=r.height)
       {
         if( img_!=null ) img_.dispose();
         img_ = new Image(null,r.width,r.height);
       }
       GC gc = new GC( img_ );
       render_.draw( gc, r, gph_ );
       e.gc.drawImage( img_, 0, 0 );
       gc.dispose();
    }
   }
   
   /* Layout horizontally children control, all width of them (but not the last one)
    * is limited (as the FillLayout doesn't do it).
    */
   private static class HLimLayout extends Layout
   {
     private int lim_;
     public HLimLayout(int _lim){lim_=_lim;}
     protected Point computeSize(Composite composite, int wHint, int hHint, boolean flushCache)
     {
       return new Point( 100, hHint );
     }

     protected void layout(Composite composite, boolean flushCache)
     {
       Rectangle r = composite.getClientArea();
       Control c[] = composite.getChildren();
       if( c==null ) return ;
       int x = r.x;
       for( int i=0; i<c.length; ++i )
       {
         if( (i+1)==c.length)
         {
           c[i].setBounds( x, r.y, r.x+r.width-x, r.height );
         } else {
           c[i].setBounds( r.x, r.y, lim_, r.height );
           x += lim_;
         }
       }
      
     }
   }
   
   public static void main( String[] argv )
   {
     //This is very important to register the render for new graphic type.
     DChartRenderRegistry.AddRender( new PolarChartRender() );
     
     Display disp = Display.getDefault();
     final Shell shell = new Shell(disp, SWT.DIALOG_TRIM|SWT.RESIZE );
     shell.setText( "Test ApiDoc: User Graphic" );
     shell.setLayout(new HLimLayout( 150 ));

     Composite cb = new Composite( shell, SWT.NONE );
     cb.setLayout( new FillLayout( SWT.VERTICAL ));
     Group gb = new Group( cb, SWT.NONE);
     gb.setLayout( new FillLayout(SWT.VERTICAL));
     gb.setText("Graphic Type:");
     final Button btn[] = new Button[2];
     int b=-1;

     btn[++b] = new Button( gb, SWT.RADIO ); btn[b].setText("XY (standard)");
     btn[b].setData( new BuildXYGraphic());     

     btn[++b] = new Button( gb, SWT.RADIO ); btn[b].setText("Polar (user defined)");
     btn[b].setData( new BuildPolarGraphic());     
       
     final Canvas canvas = new Canvas( shell, SWT.NO_BACKGROUND );
     final RenderGraphic rg = new RenderGraphic( canvas );
    
     //just on first button is enough
     SelectionListener listen_button= new SelectionListener(){
       public void widgetSelected(SelectionEvent e)
       {   
         for(int i=0; i<btn.length; ++i)
         {
           if( btn[i].getSelection())
           {
             DGraphic g = ((IBuildGraphic)btn[i].getData()).buildGraphic();
             rg.setGraphic( g );
             break;            
           }
         }
       }
 
       public void widgetDefaultSelected(SelectionEvent e){} //never called
     };
     
     //set selection listeners to all buttons...
     for( int i=0; i<btn.length; ++i )
     {
       btn[i].addSelectionListener( listen_button );
     }
         
     shell.setSize( 600, 500 );
     shell.open();
 
     
      while (!shell.isDisposed())
        if (!disp.readAndDispatch())
          disp.sleep();
   }
}
/*
 * Created on 25 juin 2003
 *
 */
package com.rational.apidoc.chart.internal;

import java.text.DecimalFormat;
import java.util.Vector;

import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.graphics.FontMetrics;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.SWT;
import com.rational.apidoc.core.*;
import com.rational.apidoc.chart.IChartLocation;
import com.rational.apidoc.chart.IChartRender;
import com.rational.apidoc.drawutil.HSV;
import com.rational.apidoc.drawutil.DrawUtil;
import com.rational.apidoc.drawutil.Rect;
import com.rational.apidoc.drawutil.Point2D;

/**
 * ChartRender is made to 
 * - draw DGraphic item in an area
 * - locate a point on a DGraphic area.
 */
public class DefaultChartRender implements IChartRender
{
  /** margin for render */
  protected int margin_; 
  /** Default style to draw */
  protected DChartStyle default_;
 
  /** temporary data during draw/locate */
  protected Data d_ ; 
  /** used to dispose font/color from styles (MUST BE disposed for windows systems)
   *  (?huh? work'in java and take care of windows ... smell bad (but smell'SWT!) */
  protected Vector styleof_;

  /** Create a chart render */
  public DefaultChartRender()
  {
    margin_=10;
    default_ = new DChartStyle();
    d_=null;
  }
  
  //paire courbe / style
  private static class CurveStyle
  {
    public DCurve curve_;
    public DChartStyle style_;
    public CurveStyle( DCurve _c, DChartStyle _s )
    {
      curve_=_c; style_=_s;
    }
  }
  
  private static class Data
  {
     
    public boolean drawing() { return draw_; }
    public boolean locating() { return !draw_; }
    
    public boolean draw_;
    public GC gc_; //mode drawing
    public Font default_font_;
    public int lx_, ly_; //mode locating
    public DGraphic G_; 
    public Rect R_;
    public Vector curves_; //list of CurveStyle
    
    public Data( boolean _draw, DGraphic _g, Rect _R, GC _gc )
    {
      gc_=_gc;
      draw_=_draw;
      G_=_g;
      R_=_R;
      lx_=ly_=-1;
      curves_ = new Vector();
      default_font_ = _gc.getFont();
    }        

    static boolean horizontal3DContains( int _lx, int _ly, int _w, int _3d, int _px, int _py )
    {
      //? si (_px,_py) est contenus dans le rectangle cisaille en X:
      //    _lx+_3d
      //       +---------+ _ly-_3d
      //      /         /
      // _ly +---------+
      //    _lx     _lx+_w
      //!! les Y ecrans sont inverses
      if( _py < (_ly-_3d) ) return false;
      if( _py >  _ly      ) return false;
      if( _py==_ly )
      {
        return (_px>=_lx)&&(_px<=(_lx+_w));
      }
      if( _py==(_ly-_3d))
      {
        return (_px>=(_lx+_3d))&&(_px<=(_lx+_w+_3d));
      }
      //dans le rectangle cisaille en X
      int dx = _ly-_py; //OK: car proj. 3d en pente 45 deg.
      return (_px>=(_lx+dx))&&(_px<=(_lx+dx+_w));
    }
    
    static boolean vertical3DContains( int _lx, int _ly, int _h, int _3d, int _px, int _py )
    {
      //? si (_px,_py) est contenus dans le rectangle cisaille en Y:
      //       _lx+_3d
      //          + _ly-_3d
      //         /|
      //    _ly + |
      //        | |
      //        | +
      //        |/
      // _ly+_h +
      //      _lx
      if( _px <  _lx     ) return false;
      if( _px > (_lx+_3d)) return false;
      if( _px == _lx      ) return (_py>=_ly)&&(_py<=(_ly+_h));
      ////if( _px == (_lx+_3d)) return (_py>=(_ly-_3d))&&(_py<=(_ly-_3d+_h));
      //!! les Y ecrans sont inverses
      //dans le rectangle cisaille en Y
      int dy = _px-_lx; //OK: car proj. 3d en pente 45 deg.
      int npy = _py+dy;
     return (npy>=_ly)&&(npy<=(_ly+_h));
    }

  }
  
  protected static class ChartLocation extends Throwable
  {
    public IChartLocation location_;
    public ChartLocation( IChartLocation _l )
    {
      location_=_l;
    }
    public ChartLocation( String _id, IDItem _item, Rect _rect )
    {
      location_=new DChartLocation( _id, _item, _rect );
    }
    public ChartLocation( String _id, IDItem _item, Rect _rect, String _dtxt )
    {
      location_=new DChartLocation( _id, _item, _rect, _dtxt );
    }
  }
  
  /** @return true since default render can draw anything */
  public boolean canRender( String _graph_type )
  {
    return true;
  }
  
  /** Draw graphic _g on rectangle _R using GC _gc. */
  public void draw( GC _gc, Rectangle _R, DGraphic _g )
  {
    if( _g==null ) return ;
    d_ = new Data( true, _g, new Rect(_R), _gc );
    try {
      render();
    }
    catch( ChartLocation _l) {}//musn't appears
    d_ = null;
  }
  
  /** Return a location for the given point, GC must be provided to compute text extends.
   * can return null if no location is found.
   */
  public IChartLocation locate( int _x, int _y, GC _gc, Rectangle _R, DGraphic _G )
  {
    if(_G==null) return null;
    if(!_R.contains(_x,_y) ) return null;

    d_ = new Data( false, _G, new Rect(_R), _gc );
    d_.lx_ = _x;
    d_.ly_ = _y;
    
    try {
      render();
    }
    catch( ChartLocation _l )
    {
      return _l.location_;
    }
    finally
    {
      d_=null;
    }
    return null;
  }
  
  /** Return style for item, walk upstair in tree to try to complete style. */
  protected DChartStyle styleOf( IDItem _item )
  {
    IDStyle apidoc_style=null;
    IDFont  apidoc_font =null;
    IDColor apidoc_color=null;
    DChartStyle style=new DChartStyle();
    char missing = 0x07; //font|fore|back

    for( IDItem item=_item; item!=null; item=item.getParent() )
    {
      if( (apidoc_style=item.getStyle())!=null )
      {
        //font
        if ( (missing & 0x01)!=0 && (apidoc_font=apidoc_style.getFont())!=null ) 
        {          
          style.font_ = DrawUtil.getFont( apidoc_font ); 
          missing &= ~0x01 ;       
        }
        //fore color
        if ( (missing & 0x02)!=0 && (apidoc_color=apidoc_style.getForeColor())!=null )
        {
          style.fore_ = DrawUtil.getColor( apidoc_color );
          missing &= ~0x02 ;
        }
        //back color
        if ( (missing & 0x04)!=0 && (apidoc_color=apidoc_style.getBackColor())!=null )
        {
          style.back_ = DrawUtil.getColor( apidoc_color );
          missing &= ~0x04 ;
        }
        //style completed.
        if( missing == 0x00 ) break ;
      }
    }//for

    if( styleof_==null ) styleof_ = new Vector();
    styleof_.add( style );
    return style;
  }
  
  protected void disposeStyleOf()
  {
    if( styleof_==null ) return ;
    for( int i=0; i<styleof_.size(); ++i )
    {
      DChartStyle s = (DChartStyle)styleof_.get(i);
      s.dispose();
    }
    styleof_.clear();
  }
  
  /** print error occured while rendering chart. */
  void error( String _message )
  {
//TBD:juste pour mise au point: affichage de ce qui a merde
    if( d_.drawing() )
    {
      Font font = d_.gc_.getFont();
      FontData fd[] = font.getFontData();
      FontData nfd = new FontData( fd[0].getName(), fd[0].getHeight(), fd[0].getStyle()|SWT.BOLD );
      Font nf = new Font( null, nfd );
      
      d_.gc_.setFont( nf );
      d_.gc_.setForeground( getRed() );
      d_.gc_.drawText( _message, d_.R_.x()+5, d_.R_.y()+5+12 );

      nf.dispose();
    }
  }
  
  /** draw or locate point using d_ Data */
  protected void render() throws ChartLocation
  {
    try
    {    
      DChartStyle g_style = styleOf( d_.G_ );

      //prepare curve list, and their styles.
      d_.curves_.clear();
      IDItem  item=null;
      DCurve curve=null;
      for( item=d_.G_.getFirstChild(); item!=null; item=item.getNext() )
      {
        if( item instanceof DCurve )
        {
          curve = (DCurve)item;
          DChartStyle cs = styleOf( curve );                
          d_.curves_.add( new CurveStyle( curve, cs ) );
        }
      }
    
      //graphics background
      if( d_.drawing() )
      {
        d_.gc_.setBackground( g_style.back_ );
        d_.gc_.fillRectangle( d_.R_.x(), d_.R_.y(), d_.R_.w(), d_.R_.h() );
      }

      //graphic title
      int title_h = renderGraphicTitle( g_style );

      //legend on the right side
      int legend_width = renderLegend( title_h );

      //the rest is available to render graphic
      Rect rg = new Rect( margin_,
                          title_h+margin_,
                          d_.R_.w()-legend_width-2*margin_,
                          d_.R_.h()-title_h-2*margin_ );
  
      //graphic is rendered depending on it type:
           if( d_.G_.isGraphicType( DGraphic.T_SECTORS      )) renderSector( rg, false );
      else if ( d_.G_.isGraphicType( DGraphic.T_SECTORS3D    )) renderSector( rg, true  );
      else if ( d_.G_.isGraphicType( DGraphic.T_HISTOGRAM    )) renderHistogram( rg, false, false );
      else if ( d_.G_.isGraphicType( DGraphic.T_HISTOGRAM3D  )) renderHistogram( rg, true, false );
      else if ( d_.G_.isGraphicType( DGraphic.T_HISTOGRAM_BS )) renderHistogram( rg, false, true );
      else if ( d_.G_.isGraphicType( DGraphic.T_XY           )) renderXY( rg );
      else if ( d_.G_.isGraphicType( DGraphic.T_XYZ          )) renderXYZ( rg );
      else if ( d_.G_.isGraphicType( DGraphic.T_SURFACE      )) renderXYZ( rg );
/*
      else if ( d_.G_.isGraphicType( DGraphic.T_HISTORS )) renderHistors( rg )
*/
      else
      {
        error( "Unknwon Graphic type '"+d_.G_.getGraphicType()+"'" );
        return ;
      }
    
      g_style.dispose();
    
      if( d_.locating() && rg.contains( d_.lx_, d_.ly_ ) )
      {
        throw new ChartLocation( DLocated.InGraphicArea, d_.G_, rg );
      }
    }
    finally
    {
       //dispose styles, must be done for poor platform like windows.
       disposeStyleOf();
       d_.curves_.clear();
       d_.curves_=null; 
    }
  }
  
  
  /** render graphic title.
   * @return the height of the title, 0 if title isn't displayed
   */
  protected int renderGraphicTitle( DChartStyle _g_style ) throws ChartLocation
  {
    int title_h = 0;

    boolean display_title = d_.G_.getProperties().get(DGraphic.P_SHOW_TITLE, true);
    String title = d_.G_.getTitle();  
    
    if( !display_title || DrawUtil.isEmpty( title ) ) return title_h;
    
    d_.gc_.setFont( _g_style.font_ );
    FontMetrics fm = d_.gc_.getFontMetrics();      
    title_h = fm.getHeight();
  
    String txt = DrawUtil.truncateTextH( d_.gc_, title, d_.R_.w() );
    int    aln = DrawUtil.HCENTER | DrawUtil.TOP;
    Rect r_ttl = new Rect( d_.R_.x(), d_.R_.y(), d_.R_.w(), title_h );

    if( d_.drawing() )
    {
      d_.gc_.setForeground( _g_style.fore_ );
      DrawUtil.drawText( d_.gc_, r_ttl, aln, txt);
    }
    else
    {
      //locating
      Rect br = DrawUtil.boundingRect( d_.gc_, r_ttl, aln, txt );
      br.union( r_ttl );

      if( br.contains( d_.lx_, d_.ly_ ) )
      {
        throw new ChartLocation( DLocated.InGraphicTitleText, d_.G_, br, txt );
      }
      if( r_ttl.contains( d_.lx_, d_.ly_ ) )
      {
        throw new ChartLocation( DLocated.InGraphicTitleArea, d_.G_, r_ttl );
      }
    }
 
    return title_h;
  }
  
  /** render legend and return the width, 0 if none */
  protected int renderLegend( int _title_h )  throws ChartLocation 
  {
    int legend_width = 0;

    //display legend ?
    boolean display_legend = d_.G_.getProperties().get(DGraphic.P_SHOW_LEGEND, true);
    if( !display_legend ) return 0;
    
    CurveStyle ccurve=null;

    int legend_height = 0;
    int legend_height_min =100000;
    for( int i=0; i<d_.curves_.size(); ++i )
    {
      ccurve = (CurveStyle)d_.curves_.get(i);
      d_.gc_.setFont( ccurve.style_.font_ );
      String name = ccurve.curve_.getName();
      if( !DrawUtil.isEmpty( name ) )
      {
        Point sz = d_.gc_.textExtent( name );
        if ( legend_width < sz.x ) legend_width=sz.x;
        if ( legend_height_min > sz.y ) legend_height_min=sz.y;
        legend_height += sz.y;
      }
    }

    int legend_rect=(legend_height_min-2);
    if(legend_rect>10) legend_rect=10; //limit to 10x10 pixel
    legend_width += legend_rect+margin_ ;
    //limit legend width to 1/3 of graphic
    if ( legend_width > (d_.R_.w()/3) ) legend_width=d_.R_.w()/3;
    int legend_x = d_.R_.w() - legend_width;
    int legend_H = d_.curves_.size() * 4 + legend_height;
    int legend_y = _title_h+(d_.R_.h()-legend_H)/2+2;

    Rect lgd_area= new Rect( d_.R_.right()-legend_width, d_.R_.top()+_title_h,
                             legend_width, d_.R_.h()-_title_h );
    //shortcut: on on y est pas, on passe la main ...
    //shortcut for locating: outside ?
    if( d_.locating() && !lgd_area.contains( d_.lx_, d_.ly_ ) )
    {
      return legend_width ;
    }
    
    Rect r_clr=new Rect();
    Rect r_lgd_area=new Rect();
    Rectangle r = new Rectangle(0,0,0,0);
    
    for( int i=0; i<d_.curves_.size(); ++i )
    {
      ccurve = (CurveStyle)d_.curves_.get(i);
      String name = ccurve.curve_.getName();
      if( DrawUtil.isEmpty( name )) continue;
      //text is choosen black, because fore color is curve color
      d_.gc_.setFont( ccurve.style_.font_ );
      int legend_h = d_.gc_.textExtent( name ).y;
      r_clr.setRect( legend_x,legend_y+(legend_h-legend_rect)/2, 
                     legend_rect, legend_rect );

      int available_width = legend_width-legend_rect-5;
      String txt = DrawUtil.truncateTextH( d_.gc_, name, available_width );

      if( d_.drawing() )
      {
        d_.gc_.setForeground( getBlack() );
        d_.gc_.setBackground( ccurve.style_.back_ );
        r_clr.toRectangle( r );
        d_.gc_.fillRectangle( r ); //fill it
        d_.gc_.drawRectangle( r ); //draw black square around it
      }
      else //locating
      {
        //inside color rectangle
        if( r_clr.contains( d_.lx_, d_.ly_ ) )
        {
          throw new ChartLocation( DLocated.LegendColor, ccurve.curve_, r_clr );
        }
      }
    
      r_lgd_area.setRect( legend_x+legend_rect+5, legend_y,
                          available_width,        legend_h );

      renderText( txt, r_lgd_area, DrawUtil.LEFT|DrawUtil.VCENTER,
                   ccurve.style_, DLocated.LegendText, ccurve.curve_ );

      legend_y += legend_h+4;
    }//for

    //dans la zone de la legende ??
    if( d_.locating() && lgd_area.contains( d_.lx_, d_.ly_ ) )
    {
      throw new ChartLocation( DLocated.LegendArea, null, lgd_area );
    }
  
    return legend_width;
  }    
  
  protected static class TruncateAxisTextData
  {
    public String sttl_ ; //string title
    public String sunit_; //string for unit
    public Rect   rttl_ ; //rectangle for title
    public Rect   runit_; //rectangle for unit
    public Rect   rtu_  ; //rectangle for both title and unit    
  }
  
  /** truncate unit and/or title text for an "horizontal" axis,
   * return also bounding rectangles.
   */
  protected TruncateAxisTextData truncateHAxisText(
      DAxis _axis, DChartStyle _axis_style,
      int _available_width, int _available_height
    )
  {
    Rect   rttl;
    Rect   runit;
    Rect   rtu;

    //reduction de la chaine de titre ou units..

    //title et unit sont geree dans des chaines/rect different cela va permettre
    //de les affichers avec des styles differents ... un jours prochains.
    String axis_title = _axis.getTitle();
    String axis_unit  = _axis.getUnit();
//TODO: mieux a faire, SWT ne supporte pas textExtent(null)    
    if( axis_title == null ) axis_title="";
    if( axis_unit  == null ) axis_unit="";

    //titre ( unit ) axe des Y
    d_.gc_.setFont( _axis_style.font_ );
    Point ext = d_.gc_.textExtent( axis_title );
    rttl = new Rect( 0,0, ext.x+2, _available_height );
    //unit a droite du titre
    ext = d_.gc_.textExtent( axis_unit );
    runit = new Rect( rttl.w()+5, 0, ext.x+2, _available_height );
    rtu = Rect.Union( rttl, runit ); //rectangle englobant titre et unite 

    //reduire d'abord le titre
    String n_axis_title = axis_title;
    while( rtu.w() > _available_width ) 
    {
      try {
        axis_title = axis_title.substring( axis_title.length()-2, axis_title.length()-1 );
      }
      catch( StringIndexOutOfBoundsException e )
      {
        n_axis_title="";
        break;
      }
      if( axis_title.length()==0 ) { n_axis_title=null; break; }
      n_axis_title = axis_title+"...";
      ext = d_.gc_.textExtent( n_axis_title );
      rttl.setWidth( ext.x+2 );
      runit.setX( rttl.w()+5 );
      // rtu = Rect.Union( rttl, runit ) but without allocation
      rtu.setRect( rttl );
      rtu.union( runit );
    }
    axis_title = n_axis_title ;
    
    //puis l'unite
    String n_axis_unit = axis_unit;
    while( rtu.w() > _available_width ) 
    {
      try {
        axis_unit = axis_unit.substring( axis_unit.length()-2, axis_unit.length()-1 );
      }
      catch( StringIndexOutOfBoundsException e )
      {
        n_axis_unit="";
        break;
      }
      if( axis_unit.length()==0 ) { n_axis_unit=null; break; }
      n_axis_unit = axis_unit+"...";
      ext = d_.gc_.textExtent( n_axis_unit );
      runit.setWidth( ext.x+2 );
      rtu.setRect( rttl );
      rtu.union( runit );
    }
    axis_unit = n_axis_unit;

    TruncateAxisTextData data = new TruncateAxisTextData();
    data.sttl_ = axis_title;
    data.sunit_= axis_unit;
    data.rttl_ = rttl;
    data.runit_= runit;
    data.rtu_  = rtu ;
    
    return data;
  }

    /** truncate unit and/or title text for an "vertical" axis,
     * return also bounding rectangles.
     */
    protected TruncateAxisTextData truncateVAxisText(
        DAxis _axis, DChartStyle _axis_style,
        int _available_width, int _available_height
      )
    {
      Rect   rttl;
      Rect   runit;
      Rect   rtu;
  
      //reduction de la chaine de titre ou units..
  
      //title et unit sont geree dans des chaines/rect different cela va permettre
      //de les affichers avec des styles differents ... un jours prochains.
      String axis_title = _axis.getTitle();
      String axis_unit  = _axis.getUnit();
  //TODO: mieux a faire, SWT ne supporte pas textExtent(null)    
      if( axis_title == null ) axis_title="";
      if( axis_unit  == null ) axis_unit="";
  
      //titre ( unit ) axe des Y
      d_.gc_.setFont( _axis_style.font_ );
      Point ext = DrawUtil.verticalTextExtent( d_.gc_, axis_title );
      rttl = new Rect( 0,0, _available_width, ext.y+2 );
      //unit a droite du titre
      ext = DrawUtil.verticalTextExtent( d_.gc_, axis_unit );
      runit = new Rect( 0, rttl.h()+5, _available_width, ext.y+2 );
      rtu = Rect.Union( rttl, runit ); //rectangle englobant titre et unite 
  
      //reduire d'abord le titre
      String n_axis_title = axis_title;
      while( rtu.w() > _available_height ) 
      {
        try {
          axis_title = axis_title.substring( axis_title.length()-2, axis_title.length()-1 );
        }
        catch( StringIndexOutOfBoundsException e )
        {
          n_axis_title="";
          break;
        }
        if( axis_title.length()==0 ) { n_axis_title=null; break; }
        //no sens for vertical text/ n_axis_title = axis_title+"...";
        ext = DrawUtil.verticalTextExtent( d_.gc_, n_axis_title );
        rttl.setHeight( ext.y+2 );
        runit.setY( rttl.h()+5 );
        // rtu = Rect.Union( rttl, runit ) but without allocation
        rtu.setRect( rttl );
        rtu.union( runit );
      }
      axis_title = n_axis_title ;
      
      //puis l'unite
      String n_axis_unit = axis_unit;
      while( rtu.w() > _available_height ) 
      {
        try {
          axis_unit = axis_unit.substring( axis_unit.length()-2, axis_unit.length()-1 );
        }
        catch( StringIndexOutOfBoundsException e )
        {
          n_axis_unit="";
          break;
        }
        if( axis_unit.length()==0 ) { n_axis_unit=null; break; }
        //no sense for vertical/ n_axis_unit = axis_unit+"...";
        ext = DrawUtil.verticalTextExtent( d_.gc_, n_axis_unit );
        runit.setHeight( ext.y+2 );
        rtu.setRect( rttl );
        rtu.union( runit );
      }
      axis_unit = n_axis_unit;
  
      TruncateAxisTextData data = new TruncateAxisTextData();
      data.sttl_ = axis_title;
      data.sunit_= axis_unit;
      data.rttl_ = rttl;
      data.runit_= runit;
      data.rtu_  = rtu ;
      
      return data;
    }
  
  /** Render property on Y axis, support for "3D-like".
   * @param _get_step_unit_width true to compute step unit width on axis, 
   *                             false to locate/draw.
   * @param _R rectangle to render axis.
   * @param _axis axis to render.
   * @param _property property identifier to render: STEP_LINE,STEP_UNIT,STEP_DOT.
   * @param _3d_line_x simulated 3d on x axis
   * @param _3d_line_y simulated 3d on y axis
   */
  protected int onYAxis( 
        boolean _get_step_unit_width, Rect _R,
        DAxis _axis, String _property,
        int _3d_line_x, int _3d_line_y
      )
    throws ChartLocation  
  {
    int max_step_text_width=0;

    if( !_axis.getProperties().hasKey( _property ) ) return 0;

    float value = _axis.getProperties().get( _property, 0.0f );
    if( value == 0.0f ) return 0;
     
    float min     = _axis.getProperties().get( DAxis.P_MIN, -1.0f );
    float max     = _axis.getProperties().get( DAxis.P_MAX, +1.0f );
    float max_min = max-min ;

    //prevent div/0
    if ( max_min == 0.0f ) return 0;
    
    float wsu = _R.h()*value/max_min;
    int xd = _R.x()-margin_;
    //la numerotation, les lignes ... se fait en modulo value
    float vdn = (float)Math.ceil(min/value);
    if ( Math.abs(vdn)<1e-6 ) vdn=0.0f; 
    vdn *= value;
    float yd = _R.bottom() - _R.h()*(vdn-min)/max_min;

DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);

    for( float vd = vdn; vd <= max; vd += value )
    {
       if( DAxis.P_STEP_UNIT.equals( _property ) )
       {
              if ( Math.abs(vd)<1e-6 ) vd=0.0f;
              String t=df.format(vd);
              org.eclipse.swt.graphics.Point ext = d_.gc_.textExtent( t );
              int wt = ext.x;
              if( wt>max_step_text_width) max_step_text_width=wt;
              Rect r_txt=new Rect( xd-wt, (int)yd-ext.y/2, wt, ext.y );
              if( d_.drawing() )
              {
                if( !_get_step_unit_width ) //mode calcul largeur ?
                {
                  DrawUtil.drawText( d_.gc_, r_txt, DrawUtil.CENTER, t );
                  d_.gc_.drawLine( xd+margin_-2, (int)yd, xd+margin_, (int)yd );
                }
              }
              else //locating
              {
                if ( r_txt.contains( d_.lx_, d_.ly_ ) )
                {
                  throw new ChartLocation( DLocated.InAxisYStepUnitText, _axis, r_txt, t );
                }
              }
       }
       else if ( DAxis.P_STEP_DOT.equals( _property ) )
       {
              if(d_.drawing())
               d_.gc_.drawLine( xd+margin_-2, (int)yd, xd+margin_, (int)yd );
       }
       else if ( DAxis.P_STEP_LINE.equals( _property ) )
       {
//TBD:locating: detection des lignes, mais apres les courbes, alors que le dessin se fait avant ...
            if(d_.drawing()) 
            {
              d_.gc_.drawLine( _R.left(), (int)yd, _R.right(), (int)yd );
              if( _3d_line_x!=0 && _3d_line_y!=0 )
              {
                d_.gc_.drawLine( _R.x(), (int)yd, _R.x()-_3d_line_x, (int)yd+_3d_line_y );
              }
            }
       }
       yd -= wsu;
    }

    return max_step_text_width;
  }
  
  /** Render property on X axis, no support for "3D-like".
   * @param _R rectangle to render axis.
   * @param _axis axis to render.
   * @param _property property identifier to render: STEP_LINE,STEP_UNIT,STEP_DOT.
   */
  protected void onXAxis( Rect _R, DAxis _axis, String _property )
   throws ChartLocation
  {
    if( !_axis.getProperties().hasKey( _property ) ) return ;
    float value = _axis.getProperties().get( _property, 0.0f );
    if( value==0.0f ) return ;
    float min     = _axis.getProperties().get( DAxis.P_MIN, -1.0f);
    float max     = _axis.getProperties().get( DAxis.P_MAX, +1.0f );
    float max_min = max - min ;

    if ( max_min == 0.0f ) return; //sinon div/0
    
    float wsu = (_R.w()*value/max_min);
    int yd = _R.y()+_R.h()+margin_;
    //la numerotation, les lignes ... se fait en modulo prop_todo
    float vdn = (float)Math.ceil(min/value) ;
    if ( Math.abs(vdn)<1e-6 ) vdn=0.0f; // x*0 = 0
    vdn = vdn * value ;
    float xd = _R.left()+_R.w()*(vdn-min)/max_min;
    float vd;
    Rect r_txt=new Rect();
DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);    

    for( vd  = vdn;//_axis->property(DAxis::MIN);
         vd <= max;
         vd += value
       )
    {
      if( DAxis.P_STEP_UNIT.equals( _property ) )
      {
          if ( Math.abs(vd)<1e-6 ) vd=0.0f;        
          String t=df.format( vd );
          Point ext = d_.gc_.textExtent( t );
          r_txt.setRect( (int)xd-ext.x/2, yd, ext.x, ext.y );
          if(d_.drawing())
          {
            DrawUtil.drawText( d_.gc_, r_txt, DrawUtil.CENTER, t );
            d_.gc_.drawLine( (int)xd, yd-margin_, (int)xd, yd-margin_+2 );
          }
          else //locating
          {
            if ( r_txt.contains( d_.lx_, d_.ly_ ) )
            {
              throw new ChartLocation( DLocated.InAxisXStepUnitText, _axis, r_txt, t );
            }
          }
      }
      else if ( DAxis.P_STEP_DOT.equals( _property ) )
      {
          d_.gc_.drawLine( (int)xd, yd-margin_, (int)xd, yd-margin_+2 );
      }
      else if ( DAxis.P_STEP_LINE.equals( _property ) )
      {
          d_.gc_.drawLine( (int)xd, _R.top(), (int)xd, _R.bottom() );
//TBD:locating: la ligne, mais apres la detection des courbes ...
      }
      xd += wsu;
    }//for
  }
  
  /** draw or locate for an horizontal/vertical text.
   * @param _text text to render.
   * @param _rect rectangle to draw text
   * @param _align LEFT/RIGHT/HCENTER/... flag
   * @param _style style to paint
   * @param _locid identifier for location
   * @param _item item for location
   */
  public void renderText( String _text, Rect _rect, int _align, DChartStyle _style,
                          String _locid, IDItem _item )
    throws ChartLocation                           
  {
    if( d_.drawing() )
    {
      //la valeur est affichee dans le style de la courbe
      d_.gc_.setForeground( _style.fore_ );
      d_.gc_.setFont( _style.font_ );
      DrawUtil.drawText( d_.gc_, _rect, _align, _text );
    }
    else //locating
    {
      Rect nrtxt = DrawUtil.boundingRect( d_.gc_, _rect, _align, _text );
      //parce que si le texte deborde de _rect, .. ne le clippe pas
      //nrtxt might be bigger than _rect, clip it
      nrtxt.intersection( _rect );

      if( (nrtxt.w()>0) && nrtxt.contains( d_.lx_, d_.ly_ ) )
      {
        throw new ChartLocation( _locid, _item, nrtxt, _text );
      }
    }
  }
  
  /** Render Histogram graphic family */
  protected void renderHistogram( Rect _RG, boolean _3d, boolean _bs )
    throws ChartLocation
  {
    //histogram must have only one axis.
    DAxis axis=null;
    for( IDItem i=d_.G_.getFirstChild(); i!=null; i=i.getNext() )
    {
      if( i instanceof DAxis )
      {
        axis=(DAxis)i;
        break;
      }
    }

    if ( axis==null )
    {
      error( "Error: Histogram without axis");
      return ;
    }

    DChartStyle axis_style = styleOf( axis );
    d_.gc_.setFont( axis_style.font_ );
    int axis_font_h = d_.gc_.getFontMetrics().getHeight();

    int max_h_title_unit = axis_font_h + 2 ;

    //title et unit sont geree dans des chaines/rect different cela va permettre
    //de les afficher avec des styles differents ... un jours prochains.
    TruncateAxisTextData tatd =null;
 /*   String axis_title ;
    String axis_unit  ;
    Rectangle rttly; //rect du titre de l'axe
    Rectangle runity;//rect de l'unit de l'axe
    Rectangle rtuy; //rect englobant les deux premiers.
*/
    //reduction de la chaine de titre ou units..
//XXXX    //available_width est un height a cause de la rotation de 90
    tatd = truncateVAxisText( axis, axis_style, _RG.h(), max_h_title_unit );

    int yg = _RG.y()+ axis_font_h/2; //numerotation de l'axe
    int hg = _RG.h()
           - axis_font_h/2  // 1/2 hauteur aff. mesure axe Y (au cas ou)
           - margin_
           ;

    //mesure sur l'axe des Y (avec marge):
    int wuy=0;
    hg -= axis_font_h+margin_ ;
    wuy = onYAxis( true, new Rect(), axis, DAxis.P_STEP_UNIT, 0, 0 );
  
    int wg = _RG.w() ;
    int xg = _RG.x() ;

    //display Y axis title
    {
      int axis_title_w = 0;
      //largeur du titre + unite => hauteur max. de police (tourne a 90 degre)
      if( !DrawUtil.isEmpty( tatd.sttl_ )
       || !DrawUtil.isEmpty( tatd.sunit_ ) ) axis_title_w = max_h_title_unit;

      //titre de l'axe + unite en vertical a droite de l'axe des Y
      int decoration_w = wuy //aff. mesure axe des Y (max.)
                       + margin_ //entre wuy et max_h_title 
                       + axis_title_w // titre et/ou unite
                       + margin_ //entre max_h_... et bord 
                       ;
      wg -= decoration_w ;
      xg += decoration_w ;
    }

//TBD: 20 a afficher comme valeur de 3D
    int  _3d_z    = 20 ;
    if ( _3d )
    {
      wg -= _3d_z;
      hg -= _3d_z;
      yg += _3d_z;
    }


    Rect rg = new Rect(xg,yg,wg,hg);

    //lost of Y axis enlightment
    {
      //rect global de placement titre/unit Y
      //a note: a cause de la rotation on utilise la translation du QPainter
      // pour le placement correct, les calculs sont fait a origine 0,0
      Rect grtuy= new Rect( 0, 0, _RG.h(), max_h_title_unit );
      //centrer le rect titre+unit Y
      tatd.rtu_.moveCenter( grtuy.centerX(), grtuy.centerY() );
      //replacer rect title et unite
      tatd.rttl_.moveBy(  tatd.rtu_.x(), tatd.rtu_.y() );
      tatd.runit_.moveBy(  tatd.rtu_.x(), tatd.rtu_.y() );

      if( d_.drawing() )
      {
//        d_.gc_.save();
//TODO: he les gars SWT on fait ca comment ????
//        d_.gc_.translate( _RG.x, _RG.y+_RG.height );
//        d_.gc_.rotate( -90.0f );
//ye! SWT tu fait comment la rotation ???        
        d_.gc_.setForeground( axis_style.fore_ );
        d_.gc_.setFont( axis_style.font_ );
      
        DrawUtil.drawText( d_.gc_, tatd.rttl_, DrawUtil.CENTER, tatd.sttl_ );
        DrawUtil.drawText( d_.gc_, tatd.runit_, DrawUtil.CENTER, tatd.sunit_ );
       
   //     d_.gc_.restore();
      }
      else //locating
      {
        //on deplace le pt en (0,0) sur grtuy et on invers x/y:
        //pour prendre en compte la rotation
        int lx=d_.lx_;
        int ly=d_.ly_;
        //rotation di -90: on inverse x/-y w/h
        grtuy.setRect(  -grtuy.y(), grtuy.x(), grtuy.h(), +grtuy.w() );
        tatd.rttl_.setRect(  -tatd.rttl_.y(), tatd.rttl_.x(), tatd.rttl_.h(), +tatd.rttl_.w());
        tatd.runit_.setRect(  -tatd.runit_.y(), tatd.runit_.x(), tatd.runit_.h(), +tatd.runit_.w() );
        //translation:
        grtuy.moveBy(   _RG.x(), _RG.y() /*+_RG.height() */);
        tatd.rttl_.moveBy(  _RG.x(), _RG.y() /*+_RG.height() */);
        tatd.runit_.moveBy(  _RG.x(), _RG.y() /*+_RG.height() */);
        if( grtuy.contains( lx, ly ) )
        {
          //dans la zone de l'axe
          if( !DrawUtil.isEmpty(tatd.sttl_) && tatd.rttl_.contains( lx, ly ) )
          {
            //dans la zone de titre
            throw new ChartLocation( DLocated.InAxisYTitle, axis, tatd.rttl_, tatd.sttl_ );                    
          }
          if( !DrawUtil.isEmpty(tatd.sunit_) && tatd.runit_.contains( lx, ly ) )
          {
            //dans la zone de titre
            throw new ChartLocation( DLocated.InAxisYUnit, axis, tatd.runit_, tatd.sunit_);
          }
          //donc dans la zone de l'axe Y
          throw new ChartLocation( DLocated.InAxisYArea, axis, grtuy );
        }
      }

      //lines, unit, 'n dot sur l'axe des Y
      Rect rg2 = new Rect(rg);
      if(_3d) rg2.moveBy( _3d_z,-_3d_z );

      //drawing et:ou locating:
      if(d_.drawing())
      {
        d_.gc_.setForeground( getLightGray() );
      }
      if(_3d)
      {
        onYAxis( false, rg2, axis, DAxis.P_STEP_LINE, _3d_z, _3d_z );
      } else {
        onYAxis( false, rg2, axis, DAxis.P_STEP_LINE, 0, 0 );
      }
      if(d_.drawing())
      {
        d_.gc_.setForeground( getBlack() ); //black
      }
      onYAxis( false, rg, axis, DAxis.P_STEP_DOT, 0,0  );
      if(d_.drawing())
      {
        d_.gc_.setForeground( axis_style.fore_ );
        d_.gc_.setFont( axis_style.font_ );
      }
      onYAxis( false, rg, axis, DAxis.P_STEP_UNIT, 0,0 );
    }
  
    if(d_.drawing())
    {
      d_.gc_.setForeground( axis_style.fore_ );
      d_.gc_.drawLine( xg, yg,    xg,    yg+hg ); //axe Y
      d_.gc_.drawLine( xg, yg+hg, xg+wg, yg+hg ); //axe X
    }

    //dessin des courbes dans l'ordre
    float ay_min=axis.getProperties().get( DAxis.P_MIN,  0.0f),
          ay_max=axis.getProperties().get( DAxis.P_MAX, +1.123412345f);
    float ky = ay_max - ay_min;
System.err.println("HISTO axis["+axis+"].MIN="+ay_min);    
System.err.println("HISTO axis["+axis+"].MAX="+ay_max);
System.out.println("  PROP=["+DAxis.P_MAX+"]");    
    

    //if min==max, force a delta of 10%
    if( Math.abs(ky)<1e-6 )
    {
      ay_max = (ay_max>0.0f ? 1.05f : 0.95f )* ay_max;
      ay_min = (ay_min>0.0f ? 0.95f : 1.05f )* ay_min;
      ky = ay_max - ay_min;     
      if( Math.abs(ky)<1e-6)
      {
        //smell like update document problem (MIN/MAX=0.0f)
        error("Error: can't draw document not updated");
        return ;
      }
    }

    ky = hg/ky;
    int nbar=d_.curves_.size();

    float wbar=rg.w()/(float)nbar;
    float wbar_bs = (wbar+1)/3.0f; // _BS

    float xbar=rg.x();
    int   y_bar_zero;
    float bar_zero;
    if ( ay_min > 0.0f ) {
      y_bar_zero=rg.bottom();
        bar_zero=ay_min;
    } else if ( ay_max < 0.0f ) {
      y_bar_zero=rg.y();
        bar_zero=ay_max;
    }else {
      y_bar_zero=rg.bottom()-(int)(ky*(-ay_min)); //pos de 0.0
        bar_zero=0.0f;
    }

    IDItem item=null;
    ChartLocation loc=null;

DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits( 2 );

    for( int icrv=0; icrv<d_.curves_.size(); ++icrv )
    {
      CurveStyle ccurve = (CurveStyle)d_.curves_.get(icrv);    
      for( item=ccurve.curve_.getFirstChild(); item!=null; item=item.getNext() )
      {
        //les courbes n'ont qu'un point
//TBD:on pourrait imaginer des histogram multi point en superposant les bar.???
        if( item instanceof DPoint )
        {
          DPoint point = (DPoint)item;
          float vy;
          DCoord coord=point.getCoord( axis );          
          if ( coord !=null )
          {
            vy = coord.getValue();
            //transformation.
            float bh = vy-bar_zero;
            int  ibh = (int)(ky*(bh));
            if (ibh>0)ibh+=1;else if (ibh<0) ibh-=1;        
            //curve==BAR forcement.
            Rect rbar;
            if( _bs )  // bandes separees
            {
              int dx = (int)(xbar+0.5f*(wbar-wbar_bs));
              rbar = new Rect( dx, y_bar_zero, (int)wbar_bs, -ibh );
            } else {
              rbar = new Rect( (int)xbar, y_bar_zero, (int)wbar+1, -ibh );
            }
            rbar.normalize();

            //barre principale
            if(d_.drawing())
            {
              d_.gc_.setForeground( getBlack() ); //black ://trait exterieur en noir
              d_.gc_.setBackground( ccurve.style_.back_ );
              d_.gc_.fillRectangle( rbar.x(), rbar.y(), rbar.w(), rbar.h() );
              d_.gc_.drawRectangle( rbar.x(), rbar.y(), rbar.w(), rbar.h() );
            } 
            else //locating
            {
              if( rbar.contains( d_.lx_, d_.ly_ ) )
              {
                throw new ChartLocation( DLocated.CurveBar, point,  rbar );
              }
            }

//TBD: a revoir parce que les barre value>0 sont mal affiches en 3d
            if ( _3d )
            {
//TBD:20: c'est le Z 3d
              int ar1[] = new int[8],
                  ar2[] = new int[8];
                    
              ar1[0] = rbar.x();                ar1[1] = rbar.y();
              ar1[2] = rbar.x()+20;             ar1[3] = rbar.y()-20;
              ar1[4] = (int)(rbar.x()+wbar+20); ar1[5] = rbar.y()-20;
              ar1[6] = (int)(rbar.x()+wbar+1);  ar1[7] = rbar.y();
    
              ar2[0] = (int)(rbar.x()+wbar);    ar2[1] = rbar.y();
              ar2[2] = ar1[4];                ar2[3] = ar1[5]; 
              ar2[4] = (int)(rbar.x()+wbar+20); ar2[5] = rbar.y()+rbar.h()-20;
              ar2[6] = (int)(rbar.x()+wbar);    ar2[7] = rbar.y()+rbar.h();

              if(d_.drawing())
              {
                Color clr   = ccurve.style_.back_;
                Color dark1 = DrawUtil.lighter( clr, 0.50f );
                Color dark2 = DrawUtil.lighter( clr, 0.75f );
                d_.gc_.setBackground( dark1 );
                d_.gc_.fillPolygon( ar1 );
                d_.gc_.drawPolygon( ar1 );
                d_.gc_.setBackground( dark2 );
                d_.gc_.fillPolygon( ar2 );
                d_.gc_.drawPolygon( ar2 );
              }
              else //locating
              //Il faudrait parcourir les barres a l'envers pour ne pas 
              //se tromper lorsque les barres 3D se recouvrent ...
              //MAIS, je preferre y aller comme le dessin et choisir
              //la derniere zone 3d designee par la souris.
              {
//evidement:20 c'est la simili projection 3d
                //l'ideal c'est que l'on est un ar1.contains( ... )
                boolean in_ar1 = Data.horizontal3DContains( rbar.x(), rbar.y(),
                                          rbar.w(), 20, d_.lx_,d_.ly_ );
                boolean in_ar2 = Data.vertical3DContains( rbar.right(), rbar.y(),
                                         rbar.h(), 20, d_.lx_,d_.ly_ );
                //pas encore vu ...
                if( in_ar1 )
                {
                  //memo...
                  Rect br = new Rect( rbar.x(), rbar.y()-20, rbar.w()+20, 20 );
                  loc = new ChartLocation( DLocated.CurveBar, point, br );
                }
                else if( in_ar2 )
                {
                  Rect br = new Rect( rbar.right(), rbar.y()-20, 20, rbar.h()+20 );
                  loc = new ChartLocation( DLocated.CurveBar, point, br );
                }
                //sinon si on en avait trouve, c'est la derniere...
                else if( loc!=null )
                {
                  throw loc;
                }
              }
            }

            //value is displayed below the bar.
            {
              String full_txt = df.format( vy );
              d_.gc_.setFont( ccurve.style_.font_ );
              String txt = DrawUtil.truncateTextH( d_.gc_, full_txt, (int)wbar-2 );
              if( !DrawUtil.isEmpty( txt ) )
              {
                int font_h = d_.gc_.getFontMetrics().getHeight();
                Rect rtxt = new Rect( (int)xbar, rg.bottom()+margin_,
                                                (int)wbar-2, font_h+2 );
                //drawn using curve style
                renderText( txt, rtxt, DrawUtil.CENTER, ccurve.style_,
                             DLocated.CurvePointText, point );
              }
            }
          }
          else
          {
            error( "Error: Bar without value." );
          }

          //si plusieurs valeur... xbar += wbar;
        }//if 
      }//for 
      xbar += wbar;
    }//for
 
    //pour la derniere zone 3d vue...
    if( d_.locating() && loc!=null )
    {
      throw loc;
    }
  }
  
  /** Render sector-ish graphic */
  protected   void renderSector( Rect _RG, boolean _3d ) throws ChartLocation
  {
    boolean display_values = true; //true to display value of sector
    int d = _RG.w();
    //DR2358: je fais ce que je peux...avec le temps que j'ai.
    // Comme ca bave un peu au dessus surl le titre du graphique,
    // je vais reserver une zone de 2*20 pixels (dessus/dessous)
    // pour le texte (si display_decoration).
    // Ok, 20 c'est sorti de la cuisse de Jupiter, car a ce stade
    // on ne sait pas si il y aura vraiment du texte dessus/dessous
    // (il peut etre un peu decale), ni si ce texte a un style dont
    // la police est <= 20 pixels...
    // Vous aurez compris que c'est totalement du pif !
    // En attendant ca calme M'sieu Telep, en attendant qu'il trouve 
    // que les % se chevauchent quand le graphique est petit...
    int h_for_text = display_values ? 40 : 0;
    if ( d > _RG.h()-h_for_text ) d=_RG.h()-h_for_text;

    //si on affiche les texte:
    if ( display_values )
    {
//TBD:il y a surement un calcul plus serieux que 3/4 !!!
      d = 3*d/4; //les secteurs occupent 3/4 de la place (prevoir pr texte)
    }

    float v_sum = 0.0f;

    CurveStyle ccurve=null;
    IDItem item=null;
    
    DAxis axis=null;
    for( item=d_.G_.getFirstChild(); item!=null; item=item.getNext() )
    {
      if( item instanceof DAxis ) axis=(DAxis)item;
    }
    if( axis==null )
    {
      error("Error: missing Axis in graphic");
      return;
    }
    String axis_unit = axis.getUnit();
    if( axis_unit==null ) axis_unit="";

    //cumul des valeurs des courbes (pour correspondance au 100% du cercle)
    int n_points=0;
    for( int i=0; i<d_.curves_.size(); ++i )
    {
      ccurve = (CurveStyle)d_.curves_.get(i);
      //courbes: plusieurs point a une seule coord possibles
      for( item=ccurve.curve_.getFirstChild(); item!=null; item=item.getNext())
      {
        if ( item instanceof DPoint )
        {
          DCoord coord= ((DPoint)item).getCoord( axis );
          if ( coord!=null )
          {
            v_sum += coord.getValue();
            n_points++;
          }
        } 
      }
    }

    //v_sum represente 100% du secteur.
    if ( v_sum == 0.0f )
    {
      error( "Error: can't print sectors (sum:0)" );
      return ;
    }

    int x3d = 0, y3d = 0, r3d=0, a3d=0;

    if ( _3d )
    {
      //parce que c'est a peut pres correct
      if( d >= 48 ) //ombre 4 fois moins importante que le pie chart
      {
        x3d = - 8;
        y3d = +12;
      } else {
        y3d = d / 4;
        x3d = -(int)(2.0f*y3d/3.0f);
        // on voit rien: circulez y'a rien a voir
        if( Math.abs(x3d) < 2 ) //au moins 2 pixels d'ombrage
        {
          _3d = false;
          x3d = y3d = 0;
        }
      }
      if( _3d )
      {
        r3d = (int)Math.sqrt( (double)(x3d*x3d + y3d*y3d) );
        float a = (float)Math.atan2( (float)y3d, (float)x3d ); //-PI..PI
        if( a < 0.0f ) a += 2*Math.PI ; 
        //a est dans 0..2PI 
        a3d = (int)(180*a/(Math.PI));
      }
    }

    LineAlg l3d = new LineAlg( 0, 0, x3d, y3d );

    //locating: les sector sont detectes dans l'ordre de dessin,
    //donc on les teste tous, et c'est le dernier qui gagne
    ChartLocation loc_sector=null;

DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);

    //because line separating two sector must be drawn after sector
    //otherwise fill next sector will overwrite line.
    //this array store x,y coordinates of point to draw joined to center or pie.
    int between_sectors[] = new int[n_points*2];
    int wbs=0;
    Rect r = new Rect();
    
    while( l3d.nextPoint() )
    {
      int dx3d = l3d.x();
      int dy3d = l3d.y();
      boolean top=(dx3d==x3d)&&(dy3d==y3d);

      int angle=0;
      int ray = d >> 1; 

      r.setRect( _RG.x()+(_RG.w() -d)/2 +dx3d,
                 _RG.y()+(_RG.h() -d)/2 -dy3d, d,d ); 
      if( r.w()==0 || r.h() == 0 ) continue;

      //locating: (ici pour optimiser le calcul)
      float loc_dist=-1.0f;
      int loc_angle=-1; //en 16eme de degre
      if(d_.locating())
      {
        Point rc=r.center();
        int dx=d_.lx_-rc.x, dy=d_.ly_-rc.y;
        loc_dist=(float)Math.sqrt( dx*dx + dy*dy );
        if(loc_dist<=ray)
        {
          float a = (float)Math.atan2( (float)-dy, (float)dx ); //-PI..PI
          if( a < 0.0f ) a += 2*Math.PI ; 
          //a est dans 0..2PI 
          loc_angle = (int)(360*a/(2*Math.PI));
        }
      }
      
      int ipt=0;
      for( int i=0; i<d_.curves_.size(); ++i )
      {
        ccurve = (CurveStyle)d_.curves_.get(i);
        //Courbes: plusieurs points possibles
        for( item=ccurve.curve_.getFirstChild(); item!=null; item=item.getNext() )
        {
          if ( item instanceof DPoint )
          {
            DPoint point = (DPoint)item;
            DCoord coord = point.getCoord( axis );
            if ( coord !=null )
            {
              ipt++;
              float v = coord.getValue()/v_sum; //% du disque
              int end_angle = angle+(int)(360*v); //degree;
              if( ipt >= n_points ) //last sector ?
              {
                end_angle=360;
              }

              if( d_.drawing() )
              {
                if(top)
                {
                  d_.gc_.setForeground( getBlack() ); //black
                  Color b = ccurve.style_.back_;
                  d_.gc_.setBackground( b );
                  d_.gc_.fillArc( r.x(),r.y(),r.w(),r.h(), angle, end_angle-angle );
                  d_.gc_.drawArc( r.x(),r.y(),r.w(),r.h(), angle, end_angle-angle );
                  int cx = r.centerX(), cy=r.centerY();
                  double a = (Math.PI*angle)/180.0f;
                  between_sectors[wbs++]= (int)( cx+r.w()*Math.cos( a )/2.0f );
                  between_sectors[wbs++]= (int)( cy-r.h()*Math.sin( a )/2.0f );                  
                } else {
                  Color dark = DrawUtil.lighter( ccurve.style_.back_, 0.75f );
                  d_.gc_.setForeground( dark );
                  d_.gc_.setBackground( dark );
                  d_.gc_.fillArc( r.x(),r.y(),r.w(),r.h(), angle, end_angle-angle );
                }
              }
              else //locating
              {
                if( (loc_dist <= ray)
                    &&(loc_angle>=angle)&&(loc_angle<=end_angle) )
                {
                  //memo pour connaitre le don secteur a la fin ..
                  loc_sector = new ChartLocation( DLocated.CurveSector, point, r );
                  ((DChartLocation)loc_sector.location_).setAngles( angle, end_angle );
                }
              }
              

              //affichage de la valeur + unite de l'axe
              if( top && display_values )
              {
                int atxt = angle+(end_angle-angle)/2;
                //calculer la compensation sur le rayon en mode _3d pour
                //que le texte ne recouvre pas l'ombre:
                int r3=0;
                if ( _3d )
                {
                  /* aller, j'en dis un peu plus:
                   * en 3d on va ajouter un rayon pour ne pas afficher le texte
                   * sur l'ombre, le + c'est que ce rayon depend de l'angle
                   * du texte ET de l'angle choisit pour l'ombrage, eh, eh, eh!
                   * C'est pas beau ca !
                   */
                  int da = (int)Math.abs( atxt-a3d-180 );
                  //compenser que dans la zone de l'ombre:
                  if( da < 90 ) //PI/2
                  {
                    //approx du 1er degre:
                    float k = 1.0f - da/90.f;
                    r3 = (int)(k * r3d);
                  }
                }
                float arad=(float)( 2*Math.PI-(2*Math.PI*atxt)/360 );
                int sector_ray = ray+margin_+r3 ;
                int tx = (int)(r.x()+(r.w() >>1)+sector_ray*Math.cos(arad));
                int ty = (int)(r.y()+(r.h()>>1)+sector_ray*Math.sin(arad));

                String txt = df.format( coord.getValue() ) + axis_unit;

                Point te = d_.gc_.textExtent( txt );
                int tw = te.x;
                int th = te.y;

                //on essaye de placer le texte au mieux
                     if ( atxt < 90  ) {}/*raf*/
                else if ( atxt < 180 ) { tx -= tw; }
                else if ( atxt < 270 ) { tx -= tw; ty+=th; }
                else                    {          ty+=th; };

                Rect rtxt = new Rect( tx, ty-th, tw, th );

                renderText( txt, rtxt,  DrawUtil.LEFT|DrawUtil.BOTTOM,
                             ccurve.style_, DLocated.CurvePointText,
                             point );
                if( loc_sector!=null ) throw loc_sector;
              }

              angle = end_angle;
            }
          }//if
        }//item
      }//curve
    }//3d
    
    if( d_.drawing() )
    {
      d_.gc_.setForeground( getBlack() ); //black
      int cx = r.centerX(), cy=r.centerY();
      for( int i=0; i<between_sectors.length; )
      {
        int x = between_sectors[i++];
        int y = between_sectors[i++];
        d_.gc_.drawLine( cx,cy, x,y );
      }
    }
    else //locating
    {
      if( loc_sector!=null ) throw loc_sector;
    }
  }

  /** Render sector-ish graphic */
  protected   void renderXY( Rect _RG ) throws ChartLocation
  {
    //decorations means axis titles/units/...
    boolean display_decorations = true;
    
    //search for axis "x" and "y", must exists
    DAxis x_axis=null, y_axis=null;
    for( IDItem item=d_.G_.getFirstChild(); item!=null; item=item.getNext() )
    {
      if( item instanceof DAxis )
      {
        DAxis axis = (DAxis)item;
        if( x_axis==null && "x".equals( axis.getName() ) )
        {
          x_axis = axis;
        }
        else if ( y_axis==null && "y".equals( axis.getName() ))
        {
          y_axis = axis;
        }
        if( x_axis!=null && y_axis!=null ) break;
      }
    }

    if ( x_axis==null || y_axis==null )
    {
      String txt;
      if( x_axis==null )
      {
        txt = "Error: missing 'x' axis";
      } else {
        txt = "Error: missing 'y' axis";
      }
      error( txt );
      return ;
    }

    DChartStyle x_axis_style = styleOf( x_axis );
    DChartStyle y_axis_style = styleOf( y_axis );

    d_.gc_.setFont( x_axis_style.font_ );
    int x_axis_style_h = d_.gc_.textExtent( "X" ).y+2;
    d_.gc_.setFont( y_axis_style.font_ );
    int y_axis_style_h = d_.gc_.textExtent( "X" ).y+2;
//evidement c'est sorti de la cuisse de jupiter...    

    //metrics titre ( unit ) axe des X pour les chaines/rect
    int xg,yg,hg,wg;

    xg = _RG.x();
    yg = _RG.y();
    hg = _RG.h();
    wg = _RG.w();

DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);
    if ( display_decorations )
    {  
      d_.gc_.setFont( y_axis_style.font_ );
      //compute steps width ont Y Axis...
      int wuy= onYAxis( true,  new Rect(), y_axis, DAxis.P_STEP_UNIT, 0,0 );

      //it's a worth measure.
      int wux=0;
      if ( x_axis.getProperties().hasValue(DAxis.P_STEP_UNIT))
      {
        String s = df.format( x_axis.getProperties().get(DAxis.P_MAX, +1.0f) );
        d_.gc_.setFont( x_axis_style.font_ );
        wux = d_.gc_.textExtent( s ).x;
      }

//Note: title and unit of Y axis are drawn verticaly
      int y_axis_title_unit_w = 0;
      if ( !DrawUtil.isEmpty( y_axis.getTitle())
        || !DrawUtil.isEmpty( y_axis.getUnit()) )
      {
//TODO: SWT sait pas faire le 90 !      
//TODO:pas de rotation:           y_axis_title_unit_w = y_axis_style_h ; //texte tourne a 90 degre.
        y_axis_title_unit_w = d_.gc_.textExtent( "X" ).x+2;           
      }

      wg -= y_axis_title_unit_w // titre ( unite ) axe des Y ecris en vertical
          + margin_         // marge en titre et mesure axe des Y
          + wuy             // aff. mesure axe des Y (maximum)
          + margin_         // entre aff mesure axe des Y et l'axe des Y
          + wux/2           // au pire on doit afficher la mesure maxi (axe X)
          ;
     
      xg += y_axis_title_unit_w
         + margin_
          + wuy
          + margin_
          ;

      int x_axis_title_unit_h = 0;
      if ( !DrawUtil.isEmpty( x_axis.getTitle() )
        || !DrawUtil.isEmpty( x_axis.getUnit() ) )
           x_axis_title_unit_h = y_axis_style_h;

      hg -= y_axis_style_h/2  // 1/2 hauteur aff. mesure axe Y (au cas ou)
          + margin_           // marge en axe X et affichage mesure des X
          + x_axis_style_h    // mesure des X  
          + margin_           // entre mesure des X et titre axe X
          + x_axis_title_unit_h // titre (unite) axe des X
          ;

      yg += y_axis_style_h/2 ; //1/2 nombre possible en haut axe des Y

      Rect rg = new Rect(xg,yg,wg,hg);

      //X Axis:
      TruncateAxisTextData tatdx
            = truncateHAxisText( x_axis, x_axis_style, wg, x_axis_style_h );

      //rect global de placement titre/unit X pour recenter les rect...
      Rect grtux=new Rect( xg, _RG.bottom()-margin_-x_axis_style_h, wg, x_axis_style_h );
      //centrer dans grtux
      tatdx.rtu_.moveCenter( grtux.center() );
      //donc:
      tatdx.rttl_.moveBy( tatdx.rtu_.x(), tatdx.rtu_.y() );
      tatdx.runit_.moveBy( tatdx.rtu_.x(), tatdx.rtu_.y() );

      d_.gc_.setFont( x_axis_style.font_ );
      if( d_.drawing() )
      {
        d_.gc_.setForeground( x_axis_style.fore_ );
     
        DrawUtil.drawText( d_.gc_, tatdx.rttl_, DrawUtil.CENTER, tatdx.sttl_ );
        DrawUtil.drawText( d_.gc_, tatdx.runit_, DrawUtil.CENTER, tatdx.sunit_ );

        //lines, unit, 'n dot sur l'axe des X
        d_.gc_.setForeground( getLightGray() ); 
//TBD:render
        onXAxis( rg, x_axis, DAxis.P_STEP_LINE );
        d_.gc_.setForeground( getBlack() );
        onXAxis( rg, x_axis, DAxis.P_STEP_DOT  );
        d_.gc_.setForeground ( x_axis_style.fore_  );
        d_.gc_.setFont( x_axis_style.font_ );
        onXAxis( rg, x_axis, DAxis.P_STEP_UNIT );
      }
      else //locating
      {
        if( grtux.contains( d_.lx_, d_.ly_ ) )
        {
          //inside axis area...
          if( !DrawUtil.isEmpty( tatdx.sttl_ ) && tatdx.rttl_.contains( d_.lx_, d_.ly_ ) )
          {
            //inside title area
            throw new ChartLocation( DLocated.InAxisXTitle, x_axis, tatdx.rttl_, tatdx.sttl_ );
          }
          if( !DrawUtil.isEmpty( tatdx.sunit_ ) && tatdx.runit_.contains( d_.lx_, d_.ly_ ) )
          {
            //dans la zone de titre
            throw new ChartLocation( DLocated.InAxisXUnit, x_axis,
                                      tatdx.runit_, tatdx.sunit_ );
          }
          //so, in axis area.
          throw new ChartLocation( DLocated.InAxisXArea, x_axis, 
                                    grtux, tatdx.sttl_+tatdx.sunit_ );
        }

        //locating unit step text ...
        onXAxis( rg, x_axis, DAxis.P_STEP_UNIT );
      }

      // Axe Y:
      TruncateAxisTextData tatdy
        = truncateVAxisText( y_axis, y_axis_style, xg-_RG.x(), hg );

      //rect global de placement titre/unit Y
//TODO: pas de text verticaux dans SWT.      
      //a note: a cause de la rotation on utilise la translation du QPainter
      // pour le placement correct, les calculs sont fait a origine 0,0
 //     Rect grtuy=new Rect( yg, 0, hg, y_axis_style_h );
 Rect grtuy = new Rect( 0, yg, xg-_RG.x(), hg ); //y_axis_style_h, hg );
      //centrer le rect titre+unit Y
      tatdy.rtu_.moveCenter( grtuy.center() );

      //replace title and unit rectangle (remember: drawn verticaly)
      tatdy.rttl_ .setLocation( tatdy.rtu_.x(), tatdy.rtu_.y() );
      tatdy.runit_.setLocation( tatdy.rtu_.x(), tatdy.rtu_.bottom()-tatdy.runit_.h() );

      d_.gc_.setFont( y_axis_style.font_ );
      if( d_.drawing() )
      {
//TODO: pas de texte verticaux...        
/*        d_->P_->save();
        d_->P_->translate( _RG.left(), _RG.top()+_RG.height() );
        d_->P_->rotate( -90.0f );
*/        
        d_.gc_.setForeground( y_axis_style.fore_ );

        DrawUtil.drawText( d_.gc_, tatdy.rttl_,  DrawUtil.CENTER|DrawUtil.VERTICAL, tatdy.sttl_ );
        DrawUtil.drawText( d_.gc_, tatdy.runit_, DrawUtil.CENTER|DrawUtil.VERTICAL, tatdy.sunit_  );
//        d_->P_->restore();

        //lines, unit, 'n dot sur l'axe des Y
        d_.gc_.setForeground( getLightGray() ); 
        onYAxis( false, rg, y_axis, DAxis.P_STEP_LINE, 0,0 );
        d_.gc_.setForeground( getBlack() );
        onYAxis( false, rg, y_axis, DAxis.P_STEP_DOT, 0,0  );
        d_.gc_.setForeground( y_axis_style.fore_  );
        d_.gc_.setFont( y_axis_style.font_  );
        onYAxis( false, rg, y_axis, DAxis.P_STEP_UNIT, 0,0 );
      } 
      else //locating
      {
//TODO: pas de texte verticaux, locating faux        
        //on deplace le pt en (0,0) sur grtuy et on invers x/y:
        //pour prendre en compte la rotation
        int lx=d_.lx_;
        int ly=d_.ly_;
        //rotation di -90: on inverse x/-y w/h
        grtuy.setRect( -grtuy.y(), grtuy.x(), grtuy.h(), +grtuy.w() );
        tatdy.rttl_.setRect( -tatdy.rttl_.y(), tatdy.rttl_.x(), tatdy.rttl_.h(), +tatdy.rttl_.w() );
        tatdy.runit_.setRect( -tatdy.runit_.y(), tatdy.runit_.x(), tatdy.runit_.h(), +tatdy.runit_.w() );
        //translation:
        grtuy.moveBy(  _RG.left(), _RG.top() /*+_RG.height() */);
        tatdy.rttl_.moveBy(  _RG.left(), _RG.top() /*+_RG.height() */);
        tatdy.runit_.moveBy(  _RG.left(), _RG.top() /*+_RG.height() */);
        if( grtuy.contains( lx, ly ) )
        {
          //in axis area... 
          if( !DrawUtil.isEmpty( tatdy.sttl_ ) && tatdy.rttl_.contains( lx, ly ) )
          {
            //in axis title area
            throw new ChartLocation( DLocated.InAxisYTitle, y_axis,
                                      tatdy.rttl_, tatdy.sttl_ );
          }
          if( !DrawUtil.isEmpty( tatdy.sunit_ ) && tatdy.runit_.contains( lx, ly ) )
          {
            //in unit text area
            throw new ChartLocation( DLocated.InAxisYUnit, y_axis,
                                      tatdy.runit_, tatdy.sunit_ );
          }
          //so in axis area...
          throw new ChartLocation( DLocated.InAxisYArea, y_axis, 
                                    grtuy, tatdy.sttl_+tatdy.sunit_ );
        }

        //locating unit step text ...
        onYAxis( false, rg, y_axis, DAxis.P_STEP_UNIT, 0,0 );
      }
    }

    //graph rect without axis
    Rect rg = new Rect(xg,yg,wg,hg);

    if( d_.drawing() )
    {
      d_.gc_.setForeground( getBlack() );
      d_.gc_.drawLine( xg, yg+hg, xg+wg, yg+hg ); // X axis
      d_.gc_.drawLine( xg, yg,    xg,    yg+hg ); // Y axis
    }

    //draw curves in ordre.
    float x_min = x_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
          x_max = x_axis.getProperties().get( DAxis.P_MAX, -1.0f ),
          y_min = y_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
          y_max = y_axis.getProperties().get( DAxis.P_MAX, +1.0f );
    float kx = x_max - x_min;
    float ky = y_max - y_min;
    if ( (kx==0.0f)||(ky==0.0f) )
    {
      error( "Error: can't draw curves (divison by 0)" );
      return;
    }
  
    kx = wg/kx;
    ky = hg/ky;
    //basis for histogram bar (xy curve might be bars)
    int y_bar_zero ;
    float bar_zero;

    //except if we haven't 0
    if ( y_min > 0.0f ) {
      y_bar_zero = rg.bottom(); //DAxis.MIN
        bar_zero = y_min;
    } else if ( y_max < 0.0f ) {
      y_bar_zero = rg.top(); //DAxis.MAX
        bar_zero = y_max;
    } else {
      //position of 0on Y axis:
      y_bar_zero = rg.bottom()-(int)(ky*(-y_min));
        bar_zero = 0.0f;
    }
     
    int last_x=-1, last_y=-1; 
    float  last_vy=-1;
    CurveStyle ccurve=null;
    Rect rp=new Rect();
    Rect r_bar=new Rect();

    //for locating mode, curve are checking in reverse order to handle covering.
    int ic_inc = d_.locating() ? -1 : +1;
    int ic_size = d_.curves_.size();
    for( int ic=d_.locating() ? ic_size-1 : 0;
         d_.locating()? (ic>=0) : (ic<ic_size);
         ic += ic_inc )
    {
      ccurve = (CurveStyle)d_.curves_.get(ic);      
      boolean first=true;
      IDItem last_item=null;
      DPoint point=null;
      for( IDItem item=ccurve.curve_.getFirstChild(); item!=null; item=item.getNext() )
      {
        if ( item instanceof DPoint )
        {
          point = (DPoint)item;
          DCoord cx = point.getCoord( x_axis );
          DCoord cy = point.getCoord( y_axis );
          if ( cx!=null && cy!=null )
          {
            float vx = cx.getValue();
            float vy = cy.getValue();
            //transformation.
            int px = xg+(int)(kx*(vx-x_min));
            int py = yg+hg-(int)(ky*(vy-y_min));
            if ((!first)|| ccurve.curve_.isType( DCurve.T_POINTS ) )
            {
              if( ccurve.curve_.isType( DCurve.T_LINE ) )
              {
                if( d_.drawing() )
                {
                  d_.gc_.setForeground( ccurve.style_.back_ );
                  //d_->P_->setBrush(Qt::NoBrush);
                  d_.gc_.drawLine( last_x, last_y, px,py );
                }
                else //locating
                {
                  if( DrawUtil.segmentContains( last_x, last_y, px, py, d_.lx_, d_.ly_ ) )
                  {
                    throw new ChartLocation( DLocated.CurveLine, point, new Rect(last_x,last_y,px-last_x,py-last_y));
                  }
                }
              }
              else if ( ccurve.curve_.isType( DCurve.T_POINTS ))
              {
//TBD:choix de 6x6 pour la dimension des points
                rp.setRect( px-3, py-3, 6, 6 );
                if( d_.drawing() )
                {
                  d_.gc_.setForeground( getBlack() );
                  d_.gc_.setBackground( ccurve.style_.back_ );
                  d_.gc_.fillRectangle( rp.x(), rp.y(), 6,6 );
                  d_.gc_.drawRectangle( rp.x(), rp.y(), 6,6 );
                }
                else //locating
                {
                  if( rp.contains( d_.lx_, d_.ly_ ) )
                  {
                     throw new ChartLocation( DLocated.CurvePoint, point, rp );
                  }
                }
              }
              else if ( ccurve.curve_.isType( DCurve.T_BAR ))
              {   
                float bh=last_vy-bar_zero;
                int  ibh=(int)(ky*(bh));
                if (ibh>0)ibh+=1;else if (ibh<0) ibh-=1;
                r_bar.setRect( last_x, y_bar_zero, px-last_x+1, -ibh );
                r_bar.normalize();
/* cf ci-dessus               if(r_bar.h()<0) 
                {
                     //redresser le rectangle sinon contains marche pas
                     r_bar.setRect( r_bar.x(), r_bar.y()+r_bar.height(),
                                    r_bar.width(), -r_bar.height() );
                   }*/
                if( d_.drawing() )
                {
                  d_.gc_.setForeground( getBlack() );
                  d_.gc_.setBackground( ccurve.style_.back_ );
                  d_.gc_.fillRectangle( r_bar.x(), r_bar.y(), r_bar.w(), r_bar.h() );
                  d_.gc_.drawRectangle( r_bar.x(), r_bar.y(), r_bar.w(), r_bar.h() );
                }
                else //locating
                {
                  //on s'arrete ici parce que l'on recouvre tout...
                  if( r_bar.contains( d_.lx_, d_.ly_ ) )
                  {
                    throw new ChartLocation( DLocated.CurveBar, last_item, r_bar );
                  }
                }
              }
            } 
            else
            {
              first=false;
            }
           last_x = px;
           last_y = py;
           last_vy= vy;
         }
         else
         {
error( "Warning:Point without (x,y) coord.");
          }
       }
 
       last_item=item;
     }//for 

      //last histogram bar
//TBD: c'est pas terrible, mais sinon ca veut dire que la derniere
// valeur d'une courbe histo d'a de sens que pour X (termine la valeur precedente).
     if ( !first && (ccurve.curve_.isType( DCurve.T_BAR ) ))
     {
       int px=rg.right(); //on force jusqu'a la fin du graphe
       r_bar.setRect( last_x-1, last_y, px-last_x+1, rg.bottom()-last_y+1 );
       r_bar.normalize();
       if( d_.drawing() )
       {
         d_.gc_.setForeground( getBlack() );
         d_.gc_.setBackground( ccurve.style_.back_);
         d_.gc_.fillRectangle( r_bar.x(), r_bar.y(), r_bar.w(), r_bar.h() );
         d_.gc_.drawRectangle( r_bar.x(), r_bar.y(), r_bar.w(), r_bar.h() );
       }
       else //locating
       {
         if( r_bar.contains( d_.lx_, d_.ly_ ) )
         {
           throw new ChartLocation( DLocated.CurveBar, point, r_bar );
         }
       }
     }

    }//for
  }
  
  private static Color c_black_=null;
  protected Color getBlack()
  {
    if( c_black_==null)
    {
      c_black_=new Color(null,0,0,0);
    }
    return c_black_;
  }
  
  private static Color c_light_gray_=null;
  protected Color getLightGray()
  {
    if( c_light_gray_==null )
    {
      c_light_gray_ = new Color( null, 160,160,160);
    }
    return c_light_gray_;
  }
  
  private static Color c_red_ = null;
  protected Color getRed()
  {
    if( c_red_==null )
    {
      c_red_ = new Color( null, 255,0,0);
    }
    return c_red_;
  }
  
  private static class ForceCoord implements ICoordTransformation
  {
    public float value_ = 0.0f;
    public float apply( float _f ) { return value_; }
  }
  private static ForceCoord force_coord = new ForceCoord();
  
  protected void drawBox3D( View3D _view, Box3D _bb )
  {
    Point2D p[] = new Point2D[8];
    float xn = _bb.getXMin(), xx = _bb.getXMax();
    float yn = _bb.getYMin(), yx = _bb.getYMax();
    float zn = _bb.getZMin(), zx = _bb.getZMax();
    p[0] = _view.projection( xn,yn,zn, new Point2D() );
    p[1] = _view.projection( xn,yn,zx, new Point2D() );
    p[2] = _view.projection( xx,yn,zx, new Point2D() );
    p[3] = _view.projection( xx,yn,zn, new Point2D() );
    p[4] = _view.projection( xn,yx,zx, new Point2D() );
    p[5] = _view.projection( xx,yx,zx, new Point2D() );
    p[6] = _view.projection( xx,yx,zn, new Point2D() );
    p[7] = _view.projection( xn,yx,zn, new Point2D() );

    d_.gc_.drawLine( (int)p[0].getX(),(int)p[0].getY(), (int)p[1].getX(),(int)p[1].getY());
    d_.gc_.drawLine( (int)p[0].getX(),(int)p[0].getY(), (int)p[3].getX(),(int)p[3].getY());
    d_.gc_.drawLine( (int)p[0].getX(),(int)p[0].getY(), (int)p[7].getX(),(int)p[7].getY());
    d_.gc_.drawLine( (int)p[2].getX(),(int)p[2].getY(), (int)p[1].getX(),(int)p[1].getY());
    d_.gc_.drawLine( (int)p[2].getX(),(int)p[2].getY(), (int)p[3].getX(),(int)p[3].getY());
    d_.gc_.drawLine( (int)p[2].getX(),(int)p[2].getY(), (int)p[5].getX(),(int)p[5].getY());
    d_.gc_.drawLine( (int)p[4].getX(),(int)p[4].getY(), (int)p[1].getX(),(int)p[1].getY());
    d_.gc_.drawLine( (int)p[4].getX(),(int)p[4].getY(), (int)p[5].getX(),(int)p[5].getY());
    d_.gc_.drawLine( (int)p[4].getX(),(int)p[4].getY(), (int)p[7].getX(),(int)p[7].getY());
    d_.gc_.drawLine( (int)p[6].getX(),(int)p[6].getY(), (int)p[3].getX(),(int)p[3].getY());
    d_.gc_.drawLine( (int)p[6].getX(),(int)p[6].getY(), (int)p[5].getX(),(int)p[5].getY());
    d_.gc_.drawLine( (int)p[6].getX(),(int)p[6].getY(), (int)p[7].getX(),(int)p[7].getY());
  }
  
  //identifiers for axis X,Y,Z...
  protected static final int AxisX = 1;
  protected static final int AxisY = 2;
  protected static final int AxisZ = 3;
  
  protected String axis3DName( DAxis _axis )
  {
    String title = _axis.getTitle();

    if( title != null )
    {
      title = title + _axis.getUnit();
    } else {
      title = _axis.getUnit();
    }
  
    return title ;
  }
  
  protected void renderAnAxis3D( 
      View3D _view, Box3D _bb, DAxis _axis, DChartStyle _axis_style, int _axis_type
    )
    throws ChartLocation
  {
    float x_proj = _bb.getXMin(), x_proj2 = _bb.getXMax(),
          y_proj = _bb.getYMin(), y_proj2 = _bb.getYMax(),
          z_proj = _bb.getZMin(), z_proj2 = _bb.getZMax();

    float coord_min, coord_max;
    switch( _axis_type )
    {
      case AxisX: coord_min = _bb.getXMin(); coord_max = _bb.getXMax(); break;
      case AxisY: coord_min = _bb.getYMin(); coord_max = _bb.getYMax(); break;
      case AxisZ: coord_min = _bb.getZMin(); coord_max = _bb.getZMax(); break;
      default: return ;
    }

    Point2D p0 = new Point2D(), 
            p1 = new Point2D(),
            p2 = new Point2D();

    /* 
     * Lignes des reperes
     */
    if ( d_.drawing() && _axis.getProperties().hasValue( DAxis.P_STEP_LINE ) )
    {
      float coord_step = _axis.getProperties().get( DAxis.P_STEP_LINE, 1.0f );
      if ( coord_step != 0.0f )
      {
        float coord;
        float coord_start = (float)( coord_step*Math.ceil(coord_min/coord_step) );

        d_.gc_.setForeground( getLightGray() );

                
        //ca a l'air plus complique, mais comme ca on fait le switch une
        //seule fois par for...
        switch( _axis_type )
        {
          case AxisX: for( coord = coord_start; coord < coord_max; coord += coord_step )
                      {
                        _view.projection( coord, y_proj2, z_proj, p0 );
                        _view.projection( coord, y_proj, z_proj, p1 );
                        _view.projection( coord, y_proj, z_proj2, p2 );
          
                        d_.gc_.drawLine( (int)p0.getX(), (int)p0.getY(), (int)p1.getX(), (int)p1.getY() );
                        d_.gc_.drawLine( (int)p2.getX(), (int)p2.getY(), (int)p1.getX(), (int)p1.getY() );
                      }
                      break;
          case AxisY: for( coord = coord_start; coord < coord_max; coord += coord_step )
                      {
                        _view.projection( x_proj2, coord, z_proj, p0 );
                        _view.projection( x_proj,  coord, z_proj, p1 );
                        _view.projection( x_proj,  coord, z_proj2, p2 );
          
                        d_.gc_.drawLine( (int)p0.getX(), (int)p0.getY(), (int)p1.getX(), (int)p1.getY() );
                        d_.gc_.drawLine( (int)p2.getX(), (int)p2.getY(), (int)p1.getX(), (int)p1.getY() );
                      }
                      break;
          case AxisZ: for( coord = coord_start; coord < coord_max; coord += coord_step )
                      {
                        _view.projection( x_proj2, y_proj,  coord, p0 );
                        _view.projection( x_proj,  y_proj,  coord, p1 );
                        _view.projection( x_proj,  y_proj2, coord, p2 );
          
                        d_.gc_.drawLine( (int)p0.getX(), (int)p0.getY(), (int)p1.getX(), (int)p1.getY() );
                        d_.gc_.drawLine( (int)p2.getX(), (int)p2.getY(), (int)p1.getX(), (int)p1.getY() );
                      }
                      break;
        }
      }
    }

    /*
     * Graduation (texte) de l'axe 
     */
    String axis_title = axis3DName( _axis );
    boolean have_step_unit = false;
    
    if ( _axis.getProperties().hasValue( DAxis.P_STEP_UNIT )
      || !DrawUtil.isEmpty( axis_title )
       )
    {
      have_step_unit = true;
      float d_angle = 0.0f;
      switch( _axis_type )
      {
        case AxisX:
            _view.projection( _bb.getXMin(),_bb.getYMin(),_bb.getZMin(), p0 );
            _view.projection( _bb.getXMax(),_bb.getYMin(),_bb.getZMin(), p1 );
            d_angle = 180;
          break;
        case AxisY:
             _view.projection( _bb.getXMin(),_bb.getYMin(),_bb.getZMin(), p0 );
             _view.projection( _bb.getXMin(),_bb.getYMax(),_bb.getZMin(), p1 );
             d_angle = 0;
           break;
        case AxisZ:
             _view.projection( _bb.getXMin(),_bb.getYMin(),_bb.getZMin(), p0 );
             _view.projection( _bb.getXMin(),_bb.getYMin(),_bb.getZMax(), p1 );
             d_angle = 180;
           break;
      }
       
      float angle = (float)Math.atan2( p1.getY()-p0.getY(), p1.getX()-p0.getX() ); //rad.
      angle = (float)(180*angle/Math.PI + d_angle); //degree

//SWT en est loin de ca...      QWMatrix m ;

      if( d_.drawing() )
      {
        d_.gc_.setForeground(  _axis_style.fore_ );
        d_.gc_.setFont( _axis_style.font_ );
      }
DecimalFormat df = new DecimalFormat();
df.setMaximumFractionDigits(2);      
      if ( _axis.getProperties().hasValue( DAxis.P_STEP_UNIT ) )
      {
        float coord_step = _axis.getProperties().get( DAxis.P_STEP_UNIT, +1.0f );
        if (  coord_step != 0.0f )
        {
          float coord;
//          float *cx, *cy, *cz;
          //optim: fait le switch une seule fois par for...
          /*
          switch( _axis_type )
          { 
            case AxisX:
          cx = &coord;  cy = &y_proj;  cz = &z_proj2 ; break;
            case AxisY:
          cx = &x_proj; cy = &coord;   cz = &z_proj2 ; break;
            case AxisZ:
          cx = &x_proj; cy = &y_proj2; cz = &coord ; break;
          }*/
 Rect rtxt = new Rect();
 int fh = d_.gc_.getFontMetrics().getHeight();
          for( coord = (float)(coord_step*Math.ceil(coord_min/coord_step));
               coord < coord_max;
               coord += coord_step )
          {
            switch( _axis_type )
            { 
              case AxisX: _view.projection(  coord,  y_proj,  z_proj2, p2 ); break;
              case AxisY: _view.projection(  x_proj, coord,   z_proj2, p2 ); break;
              case AxisZ: _view.projection(  x_proj, y_proj2, coord,   p2 ); break;
            }
            
//SWT?? tu fais quoi avec ca ???            
   /*         m.reset();
            m.translate( px[2], py[2] );
            m.rotate( angle );*/

            String txt = df.format( coord );
            int wt = d_.gc_.textExtent( txt ).x;
            rtxt.setRect( -wt/2, -3-fh, wt, fh );
rtxt.moveBy( (int)p2.getX(), (int)p2.getY() );            

            if( d_.drawing() )
            {
//SWT???
/*              
                d_->P_->save();
                d_->P_->setWorldMatrix( m );
                d_->P_->drawText( rtxt, Qt::AlignCenter, txt );
                d_->P_->restore();*/
                DrawUtil.drawText( d_.gc_, rtxt, DrawUtil.CENTER, txt );
              }
              else //locating
              {
/*SWT*declare forfait*                QRegion reg( rtxt );
                reg = m * reg;
                if ( reg.contains( _pt_to_locate ) )
                {
                  DChartLocation::Location location;
                  switch( _axis_type )
                  {
                    case AxisX:
                      location = DChartLocation::AxisXStepUnitText; break;
                    case AxisY:
                      location = DChartLocation::AxisYStepUnitText; break;
                    case AxisZ:
                      location = DChartLocation::AxisZStepUnitText; break;
                  }
                  d_->location_= DChartLocation( location, _axis, reg.boundingRect(), txt );
                  return ;
                }
                */
              }
            }
          }
        }
      }

      /*
       * Axis title.
       */
      if ( !DrawUtil.isEmpty( axis_title ) )
      {
        String location=null;
 Rect rtxt = new Rect();
 int fh = d_.gc_.getFontMetrics().getHeight();       
        switch( _axis_type )
        {
          case AxisX:
                _view.projection( _bb.getXMin(),y_proj,z_proj2, p0 );
                _view.projection( _bb.getXMax(),y_proj,z_proj2, p1 );
               location = DLocated.InAxisXTitle;
            break;
          case AxisY:
                _view.projection( x_proj,_bb.getYMin(),z_proj2, p0 );
                _view.projection( x_proj,_bb.getYMax(),z_proj2, p1 );
                location = DLocated.InAxisYTitle; 
            break;

          case AxisZ:
                 _view.projection( x_proj,y_proj2,_bb.getZMin(), p0 );
                 _view.projection( x_proj,y_proj2,_bb.getZMax(), p1 );
                 location = DLocated.InAxisZTitle; 
            break;
        }

        float center_x=(p0.getX()+p1.getX())/2,
              center_y=(p0.getY()+p1.getY())/2;
/*SWT*? toujours pas ?
          m.reset();
          m.translate( center_x, center_y );
          m.rotate( angle );
*/
          int wt = d_.gc_.textExtent( axis_title ).x;
          rtxt.setRect( -wt/2, -6-2*fh, wt, fh );

//TODO: pas de rotation ... donc "a plat"
//-X-8 au dessus des unites et marge...
rtxt.moveCenter( (int)center_x, (int)center_y - (have_step_unit?(fh+4):0) -4 );

          if( d_.drawing() )
          {
/*            
            d_->P_->save();
            d_->P_->setWorldMatrix( m );
            d_->P_->drawText( rtxt, Qt::AlignCenter, axis_title );
      d_->P_->restore();
*/
  DrawUtil.drawText( d_.gc_, rtxt, DrawUtil.CENTER, axis_title );      
          }
          else //locating
          {
  if( rtxt.contains( d_.lx_, d_.ly_ ))
  {
    throw new ChartLocation( location, _axis, rtxt ); 
  }
/*SWT...            
            QRegion reg( rtxt );
            reg = m * reg ;
            if( reg.contains( _pt_to_locate ) )
            {
              d_->location_= DChartLocation( location, _axis, reg.boundingRect(),
                                    axis_title );
              return;
            }
*/            
          }
        }
      }
    
   protected static final int XProj=1;
   protected static final int YProj=2;
   protected static final int ZProj=3;
   protected static final int Curve=4;
   protected static final int Finished=5;

  
  /** render XYZ curves */
  void renderXYZ( Rect _RG ) throws ChartLocation
  {
    //decorations means axis title/units...
    boolean display_decorations = true;
    boolean is_surface = d_.G_.isGraphicType( DGraphic.T_SURFACE );

    DAxis x_axis = null;
    DAxis y_axis = null;
    DAxis z_axis = null;
    int   axis_missed = 3;
    for( IDItem item = d_.G_.getFirstChild(); item!=null; item=item.getNext() )
    {
      if( item instanceof DAxis )
      {
        DAxis axis = (DAxis)item;
        if( x_axis==null && "x".equals( axis.getName() ))
        {
          x_axis = axis; axis_missed--;
        }
        else if ( y_axis==null & "y".equals( axis.getName() ))
        {
          y_axis = axis; axis_missed--;
        }
        else if ( z_axis==null & "z".equals( axis.getName() ))
        {
          z_axis = axis; axis_missed--;
        }
        if( axis_missed==0 ) break;
      }
    }

    if ( axis_missed > 0 )
    {
      String txt;
      if(x_axis==null) {
        txt = "Error: missing 'x' axis";
      } else 
      if(y_axis==null) {
        txt = "Error: missing 'y' axis";
      } else {
        txt = "Error: missing 'z' axis";
      }
      error( txt );
      return ;
    }

    DChartStyle x_axis_style = styleOf( x_axis );
    DChartStyle y_axis_style = styleOf( y_axis );
    DChartStyle z_axis_style = styleOf( z_axis );

    //garder un peu d'espace pour les texte sur les axes 
    //c'est completement choisi au hazard (ou a peu pres!).
    int max_font_h  = 0 ;
    if ( display_decorations )
    {
      d_.gc_.setFont( x_axis_style.font_ );
      max_font_h = d_.gc_.getFontMetrics().getHeight();
      d_.gc_.setFont( y_axis_style.font_ );
      int hf = d_.gc_.getFontMetrics().getHeight();
      if ( max_font_h < hf ) max_font_h = hf;
      d_.gc_.setFont( z_axis_style.font_ );
      hf = d_.gc_.getFontMetrics().getHeight();
      if ( max_font_h < hf ) max_font_h = hf;
      //prevoir le titre + les unites sur l'axe
      max_font_h = 2*max_font_h +6;
    }

    Rect gR = new Rect( _RG.left()+max_font_h, _RG.top()+max_font_h,
                        _RG.w()-2*max_font_h,  _RG.h()-max_font_h );

    //boite englobante des courbes a afficher.
    Box3D bb = new Box3D( x_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
                          x_axis.getProperties().get( DAxis.P_MAX, +1.0f ),
                          y_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
                          y_axis.getProperties().get( DAxis.P_MAX, +1.0f ),
                          z_axis.getProperties().get( DAxis.P_MIN, -1.0f ),
                          z_axis.getProperties().get( DAxis.P_MAX, +1.0f ) );

    float phi   = d_.G_.getProperties().get( DGraphic.P_XYZ_PHI,   0.3f );
    float theta = d_.G_.getProperties().get( DGraphic.P_XYZ_THETA, 0.2f );

    View3D view = new View3D( phi, theta, 0, 1.0f, bb );

    //mise a l'echelle de la vue 3D en fonction du rectangle de dessin
    view.scaleTo( gR.x(), gR.y(), gR.w(), gR.h() );

    int min_w_h = gR.w() ;
    if (min_w_h > gR.h() ) min_w_h = gR.h();

    //bounding Box
    if ( display_decorations )
    {
      if( d_.drawing() )
      {
        d_.gc_.setForeground( getLightGray() );
        drawBox3D( view, bb );
      }

      // render des axes...
      renderAnAxis3D( view, bb, x_axis, x_axis_style, AxisX );
//      if( d_->located() ) return ;
      renderAnAxis3D( view, bb, y_axis, y_axis_style, AxisY );
//      if( d_->located() ) return ;
      renderAnAxis3D( view, bb, z_axis, z_axis_style, AxisZ );
//      if( d_->located() ) return ;
    }

    //pour projection des courbes.
    float x_proj = bb.getXMin(), y_proj = bb.getYMin(), z_proj = bb.getZMin(); 
    CurveStyle ccurve =null;
    CurveStyle p_ccurve=null; //courbe precedente (connection plan)
    IDItem         item=null;
    IDItem       p_item=null;
    float       last_x=0, last_y=0;

    //memo h,s,v du fond pour les couleurs de projection des courbes...
    DChartStyle bgsty = styleOf( d_.G_ );
    Color background_color = bgsty.back_;
    HSV bhsv = new HSV( background_color );

    // check property status for curves projections
    boolean do_xproj = d_.G_.getProperties().get( DGraphic.P_XYZ_XPROJ, true);
    boolean do_yproj = d_.G_.getProperties().get( DGraphic.P_XYZ_YPROJ, true);
    boolean do_zproj = d_.G_.getProperties().get( DGraphic.P_XYZ_ZPROJ, true);

    Point2D p = new Point2D();
    
    for( int ic=0; ic<d_.curves_.size(); ++ic )
    {
      ccurve = (CurveStyle)d_.curves_.get(ic);
      //la courbe est dessinee 4 fois: 3 projections + la courbe
      int what=XProj;

      if( display_decorations
//       && d_.G_.display3DCurveProjection() 
       && d_.drawing() /*en locating on ne fait pas les projections*/
        )
      {
        //courbes projetees: couleur de la courbe en plus clair ...
        Color proj_color = DrawUtil.lighter( ccurve.style_.back_, 1.75f );
        HSV hsv = new HSV( proj_color );
        //sauf si ecart < 20, me demandez pas pourquoi, c'est une heuristique
        //finement elaboree...
        int ecart=(int)(Math.abs(bhsv.getV()-hsv.getV())+Math.abs(bhsv.getS()-hsv.getS()))>>1;
        if(ecart < 20 )
        {
          //couleur de projection trop proche du fond, on bascule
          //en couleur plus fonce..
          proj_color = DrawUtil.lighter( ccurve.style_.back_, 0.75f );
        }
        d_.gc_.setForeground( proj_color );
        what = XProj ;
      } else {
        what = Curve ;
      }
  for( ;what<Finished; )
  {
 
      //reset des projecitons
      view.resetCoordTransformation();
      switch( what )
      {
        case XProj: { if(!do_xproj) { what=YProj; continue; }
                      force_coord.value_ = x_proj ;
                      view.setXCoordTransformation( force_coord );
                      what = YProj;
                      break; }
        case YProj: { if(!do_yproj) { what=ZProj; continue; }
                      force_coord.value_ = y_proj ;
                      view.setYCoordTransformation( force_coord );
                      what = ZProj;
                      break; }
        case ZProj: { if(!do_zproj) { what=Curve; continue; }
                      force_coord.value_ = z_proj ;
                      view.setZCoordTransformation( force_coord );
                      what = Curve;
                      break; }
        case Curve: { if(d_.drawing()) d_.gc_.setForeground( ccurve.style_.back_ ); 
          what = Finished;
                      break; }
        case Finished:break;
      }//switch

      if ( is_surface )
      {
        p_item = p_ccurve!=null ? p_ccurve.curve_.getFirstChild(): null;
      }

      boolean first=true; 
      for( item = ccurve.curve_.getFirstChild(); item!=null; item=item.getNext() )
      {
        DPoint point   = (DPoint)item;
        DPoint p_point = p_item!=null ? (DPoint)p_item: null; 
        if( point!=null )
        {
          float v_x=0, v_y=0, v_z=0 ;
          DCoord c=null;
          boolean ok=true;
          if( (c=point.getCoord( x_axis))!=null ) { v_x=c.getValue(); } else { ok=false; }
          if( (c=point.getCoord( y_axis))!=null ) { v_y=c.getValue(); } else { ok=false; }
          if( (c=point.getCoord( z_axis))!=null ) { v_z=c.getValue(); } else { ok=false; }
  
          if( ok )
          {
            float p_x,p_y;
            view.projection( v_x,v_y,v_z, p );
            {
              p_x = p.getX();
              p_y = p.getY();
              if ( first )
              {
                first=false;
              }
              else
              {
                if(d_.drawing())
                {
                  d_.gc_.drawLine( (int)last_x, (int)last_y, (int)p_x, (int)p_y );
                }
                else //locating
                {
                  if( what==Finished ) //pas les projections
                  {
                    if( DrawUtil.segmentContains( (int)last_x, (int)last_y, (int)p_x, (int)p_y, d_.lx_,d_.ly_))
                    {
                      throw new ChartLocation( DLocated.CurveLine, item,
                             new Rect((int)p_x,(int)p_y, (int)(last_x-p_x),(int)(last_y-p_y) ));
                    }
                  }
                }
              }
            }
            last_x=p_x;
            last_y=p_y;

            //lien avec la courbe precedente (en mode surface)
            if ( p_point!=null )
            {
              float p_v_x=0,p_v_y=0,p_v_z=0;
              ok = true;
              ok = ok&&((c=p_point.getCoord( x_axis ))!=null); if(ok) {p_v_x=c.getValue(); }
              ok = ok&&((c=p_point.getCoord( y_axis ))!=null); if(ok) {p_v_y=c.getValue(); }
              ok = ok&&((c=p_point.getCoord( z_axis ))!=null); if(ok) {p_v_z=c.getValue(); }
              if( ok )
              {
                view.projection( p_v_x,p_v_y,p_v_z, p);
                {
                  float p_p_x = p.getX(),  p_p_y=p.getY();
                  if( d_.drawing() )
                  {
                    d_.gc_.drawLine( (int)p_x,(int)p_y, (int)p_p_x,(int)p_p_y);
                  } 
                  else // locating
                  {
                    if( what==Finished )
                    {
 //TBD en 3D(2D) plusieurs courbes peuvent se superposer...
                      if( DrawUtil.segmentContains( (int)p_x,(int)p_y, (int)p_p_x,(int)p_p_y, d_.lx_,d_.ly_ ))
                      {
                        throw new ChartLocation( DLocated.CurveLine, item, 
                            new Rect( (int)p_x,(int)p_y,(int)(p_p_x-p_x),(int)(p_p_y-p_y)));
                      } 
                    }
                  }
                }
              }
            }
          }
          if ( p_item!=null ) p_item = p_item.getNext();
        }
      }//for

  }//for:what

      p_ccurve=ccurve;
    }//for
  }

  
/*
class TESTRT_EXPORT DChartPainter
{
 protected:
  void renderHistor ( const QRect & );
  void renderBargraph ( const QRect & );

  QString valueWithUnit( float _value, DAxis *_axis ) const ;

*/
}
//Source file: C:\\Program Files\\eclipse21\\workspace\\apidoc\\com\\rational\\apidoc\\drivers\\builder\\TBuilder.java

//Source file: C:\\Program Files\\eclipse21\\workspace\\apidoc\\com\\rational\\apidoc\\drivers\\ui\\TBuilder.java

package com.rational.apidoc.drivers.builder;

import com.rational.apidoc.core.*;
import com.rational.apidoc.drawutil.DrawUtil;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.Device;
import java.util.Enumeration;
import java.util.Hashtable;
import com.rational.apidoc.parsers.DParser;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.FontMetrics;
import org.eclipse.swt.graphics.Rectangle;


public class TBuilder extends DParser 
{
   private Image brokenImage = null;
   private int alignment = DParagraph.LEFT;
   private int bottomMargin = 0;
   private int cellSpacing = 0;
   private TCell currentCell = null;
   private int currentHLine = 0;
   private int currentNLine = 0;
   private int currentNPage = 0;
   private TCell currentPage = null;
   private int currentPosition = 1;
   private int currentWLine = 0;
   private int currentXPos = 0;
   private int currentYPos = 0;
   private Font defaultFont = null;
   private Font previousFont = null;
   private Device device = null;
   private boolean drawPopup = true;
   private TCell firstCellLine = null;
   private boolean flatPopup = false;
   private Hashtable fonts = new Hashtable();
   private TCell footerCell = null;
   private GC gc = null;
   private TCell headerCell = null;
   private Hashtable imagePool = new Hashtable();
   private TCell lastCellLine = null;
   private int leftMargin = 0;
   private int lineSpacing = 0;
   private int listOffsetSpacing = 0;
   private int maxWidth = 100;
   private int minWidth = 20;
   private int pageBegin = 0;
   private int pageEnd = 0;
   private int pageHeight = 0;
   private int prefixSpacing = 0;
   private int rightMargin = 0;
   private int titleSpacing = 0;
   private int topMargin = 0;
   private boolean wordWrap = true;
   private int xoffset = 0;
   private DFooter footer = null;
   private DHeader header = null;
   private TCell firstCell = null;
   static int level = 0;
   
   private int currentNColumn = 0;
   private int currentTotalColumn = 0;
   private int currentNRow = 0;
   private DRow currentRow = null;
   private DTable currentTable = null;
   private int maxCellHeight = 0;
   
   
   private class CellInfo {
		  TCell cell = null;
		  int rowspan = 0;
   }
   
   private CellInfo[] currentRowSpan = null;
   
   private class Range 
   {
		public int start = 0;
		public int end = 0;
      
		/**
		 * @roseuid 3F0168C202A7
		 */
		public Range() 
		{
       
		}
   }
   
   /**
    * @param gc
    * @param d
    * @roseuid 3EF82BBF01A5
    */
   public TBuilder(GC gc, Device d) 
   {
      this.gc = gc;
	  this.device = d; 
            
      if (gc!=null)    
         defaultFont = gc.getFont();
      else
	     defaultFont = new Font(d, "courier", 12, SWT.NORMAL);   
   }
   
   
   private TBuilder _newBuilder(int size)
	  {
		 TBuilder b = new TBuilder(getGc(), getDevice());
		 b.setFlatPopup(getFlatPopup());
		 b.setCurrentPosition(getCurrentPosition());
		 b.setWordWrap(getWordWrap());
		 b.setMaxWidth(size);
		 b.setImagePool(getImagePool());
		 b.setFontPool(getImagePool());
		 b.setImageSearchPath(getImageSearchPath());
		 b.setCurrentNPage(getCurrentNPage());
		 return b;
	  }
   
	  private void _disposeBuilder(TBuilder b)
	  {
		 b.setImagePool(null);
		 b.setImageSearchPath(null);
		 b.setFontPool(null);
	  }
   
	  /**
	   * @param i
	   * @roseuid 3F004C9B00F9
	   */
	  private void _addListItemPrefix(IDItem i) 
	  {  
		 if (!(i.getParent() instanceof DList)) return;
   	
		 addListEntry();
		 String rval = getListStringPrefix(i.getParent());
      
		 if (rval != null)
		 {
			Point p = getGc().textExtent(rval);
			addXSpacing(-p.x);
			TCell c = addCell(p.x, p.y, i, rval.length());
			c.setText(rval);
			addXSpacing(getTitleSpacing());
		 }    
	  }
   
	  /**
	   * @param i
	   * @param text
	   * @param r
	   * @roseuid 3F0168EB0089
	   */
	  private void _buildPourTextBuffer(IDItem i, String text, Range r) 
	  {
		 if (text.length() == 0) return;
      
		 int w, wprev;
		 int count = r.end - r.start;
		 TCell item = null;
      
		 FontMetrics fm = getGc().getFontMetrics();
		 int lineSpacing = fm.getAscent()+fm.getDescent();
      
		 // rien a traiter
		 if (r.start >= r.end)
		 {
		   if (text.getBytes()[r.start] == '\n')
		   {
			   item = addCell(0, lineSpacing, i, 0);
		   }
		   else
		   {	
			   addYSpacing(lineSpacing);	
		   }
		   return;
		 }
      
		 String qs = null;
		 if (r.end == text.length())
			qs = text;
		 else
			qs = text.substring(r.start, r.end);
      
		 GC gc = getGc();
		 Point p = gc.textExtent(qs);
      
		 // Si on est en non WORD WRAP ou que le buffer rentre
		 // dans la ligne alors on ajoute ce dernier
		 // sur la ligne courante
		 if (!getWordWrap() 
			 || (getCurrentWLine() + p.x <= getMaxWidth()))
		 {
		   item = addCell(p.x, p.y, i, count);
		   item.setText(text, r.start, count); 
		   return;
		 }
         
		 // sinon on va couper le buffer sur un separateur de mot afin
		 // de la faire rentrer sur plusieurs lignes
		 wprev = 0;
      
		 Range r2 = new Range();
		 r2.start = r.start; // position dernier caractere de la sous chaine
      
      
		 while (r2.start < r.end)
		 {
			r2.end = 0;   // taille en pixel de la sous chaine
      	 
			_buildGetNextWord(qs, 
							  getMaxWidth() - getCurrentWLine(),
							  r2);
              
			count = r2.start - r.start;
			TCell cell = addCell(r2.end, lineSpacing, i, count);
			cell.setText(text, r.start, count);
			newLine();
         
			r.start = r2.start;
		 }    
	  }
   
	  /**
	   * @param b
	   * @param i
	   * @param text
	   * @roseuid 3F003ACA012D
	   */
	  private void _buildTextBuffer(IDItem i, String text) 
	  {
		 if (text.length()==0) return;
       
		 Range r = new Range();
		 r.start = 0;
		 r.end = text.indexOf("\n");
      
		 // tant que l'on a pas atteint la fin de la ligne
		 while (r.end != -1)
		 {
		   _buildPourTextBuffer(i, text, r);
      	
		   // ligne non adaptative car terminee par \n
		   // on ajoute donc une nouvelle ligne
		   newLine();
      	
		   r.start = r.end + 1;
      	
		   // recherche du \n suivant 
		   r.end = text.indexOf("\n", r.start);
		 }
      
		 if (r.start < text.length())
		 {
			r.end = text.length();
		   _buildPourTextBuffer(i, text, r);
		 }    
	  }
   
	  /**
	   * @param c
	   * @return int
	   * @roseuid 3F003121028F
	   */
	  private int _getColumnSizeForCell(DCell c) 
	  {
		  int size = 0;
		  if (currentTable.getAutoResize())  // toute la largeur du document 
		  {
				 float ws = (float)getMaxWidth() * currentTable.getWidthScale();
				 for (int i=0; i<c.getColSpan()+1; i = i + 1)
				 {
				   float factor = 1.0f;
				   int s = 0;
				   if ((currentTable.getColumnWidth() != null)
						&& (currentTable.getColumnWidth().length > currentNColumn + i))
				   {
					   factor = currentTable.getColumnWidth()[currentNColumn + i];
					   s = (int)(ws * factor);
				   }
				   else
				   {
					   s = (int)(ws / currentTotalColumn);
				   }
			  	
				   if (s < getMinWidth())
				   {
					 s = getMinWidth();
				   }
				   size = size + s;
				 }
		 }
		 else 
		 {
				size = getMaxWidth();
		 }
		 return size;    
	  }
   
	  /**
	   * @param b
	   * @param c
	   * @roseuid 3F0030E403A0
	   */
	  private void _setBuilderMarginForCell(TBuilder b) 
	  {
		  int m = getCellSpacing();
	   
	      b.setLeftMargin(m + currentTable.getBorder());	  
	      b.setTopMargin( m + currentTable.getBorder());   	  
	      b.setBottomMargin(m + currentTable.getBorder());
	      b.setRightMargin(m + currentTable.getBorder());
	  }
   
   
   /**
    * Access method for the alignment property.
    * 
    * @return   the current value of the alignment property
    */
   public int getAlignment() 
   {
      return alignment;    
   }
   
   /**
    * Sets the value of the alignment property.
    * 
    * @param aAlignment the new value of the alignment property
    */
   public void setAlignment(int aAlignment) 
   {
      alignment = aAlignment;    
   }
   
   /**
    * Access method for the bottomMargin property.
    * 
    * @return   the current value of the bottomMargin property
    */
   public int getBottomMargin() 
   {
      return bottomMargin;    
   }
   
   /**
    * Sets the value of the bottomMargin property.
    * 
    * @param aBottomMargin the new value of the bottomMargin property
    */
   public void setBottomMargin(int aBottomMargin) 
   {
      bottomMargin = aBottomMargin;    
   }
   
   /**
    * Access method for the cellSpacing property.
    * 
    * @return   the current value of the cellSpacing property
    */
   public int getCellSpacing() 
   {
      return cellSpacing;    
   }
   
   /**
    * Sets the value of the cellSpacing property.
    * 
    * @param aCellSpacing the new value of the cellSpacing property
    */
   public void setCellSpacing(int aCellSpacing) 
   {
      cellSpacing = aCellSpacing;    
   }
   
   /**
    * Access method for the currentCell property.
    * 
    * @return   the current value of the currentCell property
    */
   public TCell getCurrentCell() 
   {
      return currentCell;    
   }
   
   /**
    * Access method for the currentHLine property.
    * 
    * @return   the current value of the currentHLine property
    */
   public int getCurrentHLine() 
   {
      return currentHLine;    
   }
   
   /**
    * Access method for the currentNPage property.
    * 
    * @return   the current value of the currentNPage property
    */
   public int getCurrentNPage() 
   {
      return currentNPage;    
   }
   
   /**
    * Access method for the currentNPage property.
    * 
	* @param i
	*/
   public void setCurrentNPage(int i) {
	   currentNPage = i;
   }
   
   /**
    * Access method for the currentPage property.
    * 
    * @return   the current value of the currentPage property
    */
   public TCell getCurrentPage() 
   {
      return currentPage;    
   }
   
   /**
    * Access method for the currentPosition property.
    * 
    * @return   the current value of the currentPosition property
    */
   public int getCurrentPosition() 
   {
      return currentPosition;    
   }
   
   /**
    * Sets the value of the currentPosition property.
    * 
    * @param aCurrentPosition the new value of the currentPosition property
    */
   public void setCurrentPosition(int aCurrentPosition) 
   {
      currentPosition = aCurrentPosition;    
   }
   
   /**
    * Access method for the currentWLine property.
    * 
    * @return   the current value of the currentWLine property
    */
   public int getCurrentWLine() 
   {
      return currentWLine;    
   }
   
   /**
    * Access method for the currentXPos property.
    * 
    * @return   the current value of the currentXPos property
    */
   public int getCurrentXPos() 
   {
      return currentXPos;    
   }
   
   /**
    * Access method for the currentYPos property.
    * 
    * @return   the current value of the currentYPos property
    */
   public int getCurrentYPos() 
   {
      return currentYPos;    
   }
   
   /**
    * Access method for the defaultFont property.
    * 
    * @return   the current value of the defaultFont property
    */
   public Font getDefaultFont() 
   {
      return defaultFont;    
   }
   
   /**
    * Sets the value of the defaultFont property.
    * 
    * @param aDefaultFont the new value of the defaultFont property
    */
   public void setDefaultFont(Font aDefaultFont) 
   {
      defaultFont = aDefaultFont;    
   }
   
   /**
    * Access method for the device property.
    * 
    * @return   the current value of the device property
    */
   public Device getDevice() 
   {
      return device;    
   }
   
   /**
    * Determines if the drawPopup property is true.
    * 
    * @return   <code>true<code> if the drawPopup property is true
    */
   public boolean getDrawPopup() 
   {
      return drawPopup;    
   }
   
   /**
    * Sets the value of the drawPopup property.
    * 
    * @param aDrawPopup the new value of the drawPopup property
    */
   public void setDrawPopup(boolean aDrawPopup) 
   {
      drawPopup = aDrawPopup;    
   }
   
   /**
    * Access method for the firstCellLine property.
    * 
    * @return   the current value of the firstCellLine property
    */
   public TCell getFirstCellLine() 
   {
      return firstCellLine;    
   }
   
   /**
    * Determines if the flatPopup property is true.
    * 
    * @return   <code>true<code> if the flatPopup property is true
    */
   public boolean getFlatPopup() 
   {
      return flatPopup;    
   }
   
   /**
    * Sets the value of the flatPopup property.
    * 
    * @param aFlatPopup the new value of the flatPopup property
    */
   public void setFlatPopup(boolean aFlatPopup) 
   {
      flatPopup = aFlatPopup;    
   }
   
   /**
    * Access method for the footerCell property.
    * 
    * @return   the current value of the footerCell property
    */
   public TCell getFooterCell() 
   {
      return footerCell;    
   }
   
   /**
    * Sets the value of the footerCell property.
    * 
    * @param aFooterCell the new value of the footerCell property
    */
   public void setFooterCell(TCell aFooterCell) 
   {
      footerCell = aFooterCell;    
   }
   
   /**
    * Access method for the gc property.
    * 
    * @return   the current value of the gc property
    */
   public GC getGc() 
   {
      return gc;    
   }
   
   /**
    * Access method for the headerCell property.
    * 
    * @return   the current value of the headerCell property
    */
   public TCell getHeaderCell() 
   {
      return headerCell;    
   }
   
   /**
    * Sets the value of the headerCell property.
    * 
    * @param aHeaderCell the new value of the headerCell property
    */
   public void setHeaderCell(TCell aHeaderCell) 
   {
      headerCell = aHeaderCell;    
   }
   
   /**
    * Access method for the imagePool property.
    * 
    * @return   the current value of the imagePool property
    */
   public Hashtable getImagePool() 
   {
      return imagePool;    
   }
   
   /**
    * Sets the value of the imagePool property.
    * 
    * @param aImagePool the new value of the imagePool property
    */
   public void setImagePool(Hashtable aImagePool) 
   {
      imagePool = aImagePool;    
   }
   
   /**
	* Access method for the imagePool property.
	* 
	* @return   the current value of the imagePool property
	*/
   public Hashtable getFontPool() 
   {
		 return fonts;    
   } 
   
   /**
	* Sets the value of the imagePool property.
	* 
	* @param aImagePool the new value of the imagePool property
	*/
   public void setFontPool(Hashtable aFontPool) 
   {
		 fonts = aFontPool;    
   }
   
   /**
    * Access method for the leftMargin property.
    * 
    * @return   the current value of the leftMargin property
    */
   public int getLeftMargin() 
   {
      return leftMargin;    
   }
   
   /**
    * Sets the value of the leftMargin property.
    * 
    * @param aLeftMargin the new value of the leftMargin property
    */
   public void setLeftMargin(int aLeftMargin) 
   {
      leftMargin = aLeftMargin;    
   }
   
   /**
    * Access method for the lineSpacing property.
    * 
    * @return   the current value of the lineSpacing property
    */
   public int getLineSpacing() 
   {
      return lineSpacing;    
   }
   
   /**
    * Sets the value of the lineSpacing property.
    * 
    * @param aLineSpacing the new value of the lineSpacing property
    */
   public void setLineSpacing(int aLineSpacing) 
   {
      lineSpacing = aLineSpacing;    
   }
   
   /**
    * Access method for the listOffsetSpacing property.
    * 
    * @return   the current value of the listOffsetSpacing property
    */
   public int getListOffsetSpacing() 
   {
      return listOffsetSpacing;    
   }
   
   /**
    * Sets the value of the listOffsetSpacing property.
    * 
    * @param aListOffsetSpacing the new value of the listOffsetSpacing property
    */
   public void setListOffsetSpacing(int aListOffsetSpacing) 
   {
      listOffsetSpacing = aListOffsetSpacing;    
   }
   
   /**
    * Access method for the maxWidth property.
    * 
    * @return   the current value of the maxWidth property
    */
   public int getMaxWidth() 
   {
      return maxWidth;    
   }
   
   /**
    * Sets the value of the maxWidth property.
    * 
    * @param aMaxWidth the new value of the maxWidth property
    */
   public void setMaxWidth(int aMaxWidth) 
   {
	  maxWidth = Math.max(aMaxWidth, getMinWidth());
   }
   
   /**
    * Access method for the minWidth property.
    * 
    * @return   the current value of the minWidth property
    */
   public int getMinWidth() 
   {
      return minWidth;    
   }
   
   /**
    * Sets the value of the minWidth property.
    * 
    * @param aMinWidth the new value of the minWidth property
    */
   public void setMinWidth(int aMinWidth) 
   {
      minWidth = aMinWidth;    
   }
   
   /**
    * Determines if the modePage property is true.
    * 
    * @return   <code>true<code> if the modePage property is true
    */
   public boolean getModePage() 
   {
      return (getPageHeight()>0);    
   }
   
   /**
    * Access method for the pageBegin property.
    * 
    * @return   the current value of the pageBegin property
    */
   public int getPageBegin() 
   {
      return pageBegin;    
   }
   
   /**
    * Sets the value of the pageBegin property.
    * 
    * @param aPageBegin the new value of the pageBegin property
    */
   public void setPageBegin(int aPageBegin) 
   {
      pageBegin = aPageBegin;    
   }
   
   /**
    * Access method for the pageEnd property.
    * 
    * @return   the current value of the pageEnd property
    */
   public int getPageEnd() 
   {
      return pageEnd;    
   }
   
   /**
    * Sets the value of the pageEnd property.
    * 
    * @param aPageEnd the new value of the pageEnd property
    */
   public void setPageEnd(int aPageEnd) 
   {
      pageEnd = aPageEnd;    
   }
   
   /**
    * Access method for the pageHeight property.
    * 
    * @return   the current value of the pageHeight property
    */
   public int getPageHeight() 
   {
      return pageHeight;    
   }
   
   /**
    * Sets the value of the pageHeight property.
    * 
    * @param aPageHeight the new value of the pageHeight property
    */
   public void setPageHeight(int aPageHeight) 
   {
      pageHeight = aPageHeight;    
   }
   
   /**
    * Access method for the rightMargin property.
    * 
    * @return   the current value of the rightMargin property
    */
   public int getRightMargin() 
   {
      return rightMargin;    
   }
   
   /**
    * Sets the value of the rightMargin property.
    * 
    * @param aRightMargin the new value of the rightMargin property
    */
   public void setRightMargin(int aRightMargin) 
   {
      rightMargin = aRightMargin;    
   }
   
   /**
    * Access method for the titleSpacing property.
    * 
    * @return   the current value of the titleSpacing property
    */
   public int getTitleSpacing() 
   {
      return titleSpacing;    
   }
   
   /**
    * Sets the value of the titleSpacing property.
    * 
    * @param aTitleSpacing the new value of the titleSpacing property
    */
   public void setTitleSpacing(int aTitleSpacing) 
   {
      titleSpacing = aTitleSpacing;    
   }
   
   /**
    * Access method for the topMargin property.
    * 
    * @return   the current value of the topMargin property
    */
   public int getTopMargin() 
   {
      return topMargin;    
   }
   
   /**
    * Sets the value of the topMargin property.
    * 
    * @param aTopMargin the new value of the topMargin property
    */
   public void setTopMargin(int aTopMargin) 
   {
      topMargin = aTopMargin;    
   }
   
   /**
    * Determines if the wordWrap property is true.
    * 
    * @return   <code>true<code> if the wordWrap property is true
    */
   public boolean getWordWrap() 
   {
      return wordWrap;    
   }
   
   /**
    * Sets the value of the wordWrap property.
    * 
    * @param aWordWrap the new value of the wordWrap property
    */
   public void setWordWrap(boolean aWordWrap) 
   {
      wordWrap = aWordWrap;    
   }
   
   /**
    * Access method for the xoffset property.
    * 
    * @return   the current value of the xoffset property
    */
   public int getXoffset() 
   {
      return xoffset;    
   }
   
   /**
    * Sets the value of the xoffset property.
    * 
    * @param aXoffset the new value of the xoffset property
    */
   public void setXoffset(int aXoffset) 
   {
      xoffset = aXoffset;    
   }
   
   /**
    * Access method for the footer property.
    * 
    * @return   the current value of the footer property
    */
   public DFooter getFooter() 
   {
      return footer;    
   }
   
   /**
    * Sets the value of the footer property.
    * 
    * @param aFooter the new value of the footer property
    */
   public void setFooter(DFooter aFooter) 
   {
      footer = aFooter;    
   }
   
   /**
    * Access method for the header property.
    * 
    * @return   the current value of the header property
    */
   public DHeader getHeader() 
   {
      return header;    
   }
   
   /**
    * Sets the value of the header property.
    * 
    * @param aHeader the new value of the header property
    */
   public void setHeader(DHeader aHeader) 
   {
      header = aHeader;    
   }
   
   /**
    * Access method for the firstCell property.
    * 
    * @return   the current value of the firstCell property
    */
   public TCell getFirstCell() 
   {
      return firstCell;    
   }
   
   /**
    * Sets the value of the firstCell property.
    * 
    * @param aFirstCell the new value of the firstCell property
    */
   public void setFirstCell(TCell aFirstCell) 
   {
      firstCell = aFirstCell;    
   }
   
   /**
    * @param w
    * @param h
    * @param item
    * @param count
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF9510A011C
    */
   public TCell addCell(int w, int h, IDItem item, int count) 
   {
   	  TCell c = new TCell(); //currentPage, currentCell);
   	  addCell(w, h, item, count, c);
   	  return c;    
   }
   
   /**
    * @param w
    * @param h
    * @param item
    * @param count
    * @param c
    * @roseuid 3F00143402E1
    */
   public void addCell(int w, int h, IDItem item, int count, TCell c) 
   {   
   	   if (currentPage==null)
   	   {
		   TCell pageCell = new TCell();
		   if (firstCell == null)
		   {
				firstCell  = pageCell;
		   }
		   currentPage  = pageCell;
		   currentCell  = null;
		   currentYPos  = 0;
		   currentXPos = getXoffset();
   	   }
   	
   	   // accrochage dans l'arbre
   	   if (currentPage.getChild() == null)
	       currentPage.setChild(c);
       c.setParent(currentPage);
       c.setPrevious(currentCell);
       
       if (currentCell != null)
          currentCell.setNext(c);
       
	   if (firstCellLine == null)
	   {
			// new line
			c.setRect(currentXPos + getLeftMargin(), currentYPos + getTopMargin(), w, h);
			firstCellLine = c;
			lastCellLine = c;
   	  	
			currentWLine = currentWLine + w;
			currentHLine = h;
			currentXPos = w;
			currentNLine = currentNLine + 1;
	   }
	   else
	   {
			c.setRect(currentXPos + getLeftMargin(), currentYPos + getTopMargin(), w, h);
			currentHLine = Math.max(currentHLine, h);
			currentWLine = currentWLine + w;
			currentXPos = currentXPos + w;
	   }
       
       // position des caracteres
	   c.setPosition(currentPosition);
	   c.setNumberLine(currentNLine);
	   c.setItem(item);
	   currentPosition = currentPosition + count;   
	   
	   // mise a jour de la taille du parent
	   if (currentPage != null)
	   {
		  currentPage.getRect().height = getTopMargin()+ currentYPos + currentHLine + getBottomMargin();
	      currentPage.getRect().width  = Math.max(currentPage.getRect().width, getLeftMargin() + currentWLine + getRightMargin());
	   }
       
       currentCell = c;    
   }
   
   /**
    * @param s
    * @roseuid 3EFAF6900056
    */
   public void addXSpacing(int s) 
   {
      currentWLine = currentWLine + s;    
   }
   
   /**
    * @param s
    * @roseuid 3EFAF6BA00BA
    */
   public void addYSpacing(int s) 
   {
      currentHLine = currentHLine + s;    
   }
   
   /**
    * @param item
    * @roseuid 3EFAA5FD0218
    */
   protected void applyFont(IDItem item) 
   {
   	   Font font = null;
   	   previousFont = getGc().getFont();
	   	
	   if ((item.getStyle() != null)
			&& (item.getStyle().getFont() != null))
	   {
			IDFont f = item.getStyle().getFont();
			font = (Font)fonts.get(f.getKey());
			
			if (font == null)
			{
				font = DrawUtil.getFont(f);
				fonts.put(f.getKey(), font);
			}
	   }
	   else
	   {
	   	   font = getDefaultFont();
	   }   
	   
	   getGc().setFont(font);    
   }
   
   protected void restoreFont()
   {
   	  if (previousFont!=null)
   	     getGc().setFont(previousFont);
   }
   
   
   
   /**
    * @roseuid 3EF849AE0212
    */
   public void dispose() 
   {
   	  clear();
   	  
      // dispose all fonts
      if (fonts != null)
      {
	     for (Enumeration e = fonts.elements() ; e.hasMoreElements() ;) 
	     {
		    ((Font)e.nextElement()).dispose();
	     }
	     fonts.clear();
      }
	   
      //	reset all imagePool
      if (imagePool != null)
      {
	     for (Enumeration e = imagePool.elements() ; e.hasMoreElements() ;) 
	     {
		    ((Image)e.nextElement()).dispose();
	     }
	     imagePool.clear();
      }
   }
   
   
   /**
    * @return int
    * @roseuid 3EF8475300C0
    */
   public int getHeight() 
   {
	  int h = 0;
	  for (TCell c = getFirstCell(); c != null; c = c.getNext() )
	  {
		h = h +  c.getRect().height;
	  }
	  return h;    
   }
   
   /**
    * @param i
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EFACC7B0159
    */
   public TCell getPage(int i) 
   {
	 int j = 0;
     for (TCell c = getFirstCell(); c != null; c = c.getNext() )
	 {
	    if (i == j) return c;
		j = j + 1; 
	 }
     return  null;    
   }
   
   /**
    * @return int
    * @roseuid 3EFACC4101BA
    */
   public int getPageCount() 
   {
   	  int i = 0;
	  for (TCell c = getFirstCell(); c != null; c = c.getNext() )
	  {
		 i = i + 1;
	  }
	  return i;    
   }
   
   /**
    * @return int
    * @roseuid 3EF847650152
    */
   public int getWidth() 
   {
   	  int w = 0;
	  for (TCell c = getFirstCell(); c != null; c = c.getNext() )
	  {
		 w = Math.max(w, c.getRect().width);
	  }
      return w;    
   }
   
   /**
    * @param p
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF95FE600FE
    */
   public TCell mapPointToCell(Point p) 
   {
	  return mapPointToCell(getFirstCell(), p);    
   }
   
   /**
    * @param c
    * @param p
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF96C530167
    */
   protected TCell mapPointToCell(TCell c, Point p) 
   {
	 TCell i;
	 for (i = c.getChild(); i != null; i = i.getNext())
	 {
	   if ( ((i.getRect().y <= p.y) && (p.y <= i.getRect().y + i.getRect().height))
			&& ((i.getRect().x <= p.x) && (p.x <= i.getRect().x + i.getRect().width)))
	   {
				break;
	   }
	 }
	   
	 if (i==null) return null;
	   
	 // la cellule est trouvee si elle en contient d'autre on y va en 
	 // profondeur
	 if (i.getChild()!=null)
	 {
	 	 p.x = p.x - i.getRect().x;
	 	 p.y = p.y - i.getRect().y;
		 TCell cf = mapPointToCell(i, p);
	   	 
		 if (cf!=null) return cf;
	 }
	   
	 p.x = p.x - i.getRect().x;
	 p.y = p.y - i.getRect().y;
	 return  i;    
   }
   
   /**
    * @param pos
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF9610F0282
    */
   public TCell mapPositionToCell(int pos) 
   {
      return mapPositionToCell(getFirstCell(), pos);    
   }
   
   /**
    * @param c
    * @param pos
    * @return com.rational.apidoc.drivers.builder.TCell
    * @roseuid 3EF96C810159
    */
   protected TCell mapPositionToCell(TCell c, int pos) 
   {
	 TCell i;
     for (i = c.getChild(); i != null; i = i.getNext())
	 {
	 	if (i.getChild()!=null)
	 	{
	 		TCell j = mapPositionToCell(i, pos);
	 		if (j!=null) return j;
	 	}
	 	if ((i.getPosition() <= pos)
	 	     && (pos <= i.getPosition() + i.getLen()))
	 	{
	 		return i;
	 	}
	 }
	 return null;    
   }
   
   /**
    * @roseuid 3EF9A6E0028C
    */
   public void newLine() 
   {
      if (firstCellLine != null)
      {
      	int xoffset = getLeftMargin();
      	
      	if (getAlignment() == DParagraph.RIGHT)
      	       xoffset = xoffset + getWidth() - currentWLine;
      	else
      	if (getAlignment() == DParagraph.CENTER)
      	{
      	       xoffset = xoffset + (getWidth() - currentWLine)/2; 
      	}
      	
      	int maxDsc = 0;
      	TCell c;
      	for (c = firstCellLine; c!=null; c = c.getNext())
      	{
      		if (c.getItem() != null)
      		{
      			applyFont(c.getItem());
      			maxDsc=Math.max(maxDsc, getGc().getFontMetrics().getDescent());
				restoreFont();
      		}
      	}
      	
      	for (c = firstCellLine; c!=null; c = c.getNext())
      	{
      		int locDsc;
      		if (c.getItem()!=null)
      		{
				applyFont(c.getItem());
      			locDsc = getGc().getFontMetrics().getDescent();
				restoreFont();
      			c.setRect(xoffset + c.getRect().x, 
      			          c.getRect().y - (maxDsc-locDsc)+(currentHLine-c.getRect().height), 
      			          c.getRect().width, c.getRect().height);
      			//xoffset = xoffset + c.getRect().width;
      		}
      	}
      }
      
      if (getModePage())
      {
      	 int fh = 0;
      	 if (getFooterCell()!=null)
      	 {
			fh = fh + getFooterCell().getRect().height;
      	 }
      	 
      	 if ((currentYPos + currentHLine + fh) > getPageHeight())
      	 { 
      	 	newPage();
      	 }
      	 else
      	 {
			currentYPos = currentYPos + currentHLine;
      	 }
      }
      else
      {   
         currentYPos = currentYPos + currentHLine;
      }
      
	  currentXPos = getXoffset(); 
      
      currentHLine  = 0;
      currentWLine  = 0;
      firstCellLine = null;    
   }
   
   /**
    * @roseuid 3EFAF84300A1
    */
   public void newPage() 
   {
	  // creation de la page
      TCell pageCell = new TCell();
      if (firstCell == null)
	     firstCell  = pageCell;
      else
      {
		 currentPage.setNext(pageCell);
      }
         
      currentPage  = pageCell;
      currentNPage = currentNPage + 1;
      currentCell  = null;
      int hLine = 0;
	  
      // ajout de l'entete s'il y en a une
      if (getHeader()!=null)
      {
	    doItem(getHeader());
	    
	    if (getHeaderCell()!=null)
	    {
	       // retour a la ligne
		   currentYPos  = 0;
		   currentXPos  = getXoffset();
		   currentWLine = 0;
		   currentHLine = 0;
	    
	       addCell(getHeaderCell().getRect().width, 
						 getHeaderCell().getRect().height, 
						 getHeader(), 0, getHeaderCell());
						 
		   hLine = getHeaderCell().getRect().height;
	    }
      }   
      else
      {
	    if (firstCellLine != null)
	    {
		   firstCellLine.setPrevious(null);
		   pageCell.setChild(firstCellLine);
	    }
      }
      
      // ajout du bas de page en fin de la page
      if (getFooter()!=null)
      {
		doItem(getFooter());
      	
		if (getFooterCell()!=null)
		{
            //	retour a la ligne plus positionnement en bas
			currentYPos  = getPageHeight()-getFooterCell().getRect().height;
			currentXPos  = getXoffset();
			currentWLine = 0;
			currentHLine = 0;	
			 
			addCell(getFooterCell().getRect().width, 
					getFooterCell().getRect().height, 
					getFooter(), 0, getFooterCell());
		}     	
      }
         
      TCell c;
      for (c = firstCellLine; c!=null; c = c.getNext())
      {
	     c.setParent(pageCell);
      }    
      
	  currentYPos = hLine;
	  currentXPos = getXoffset();
      currentHLine  = 0;
      currentWLine  = 0;
      firstCellLine = null;    
   }
   
   /**
    * @roseuid 3EFAF25A020E
    */
   public void clear() 
   {
   	   // clears all pages
	   for (TCell c = getFirstCell(); c != null; )
	   {
			 TCell p = c.getNext();
			 c.clear();
			 c.setRect(0,0,0,0);
			 c.setPosition(1);
		 
			 c.setNext(null);
			 c = p;
	   }
	   
	   super.clear();
	   
       currentPosition = 1;
	   currentNLine = 0; 
	   currentXPos = 0;
	   currentYPos = 0;
	   firstCellLine = null;
	   lastCellLine = null;
	   currentCell = null;
	   currentHLine = 0;
	   currentWLine = 0;    
	   alignment = DParagraph.LEFT;
	   currentPage = null;    
	   firstCell = null;
	   currentNPage = 0;    
   }
   
   /**
    * @param sp
    * @param ep
    * @roseuid 3EF9A56C002E
    */
   public void setPageRange(int sp, int ep) 
   {
      setPageBegin(sp);
      setPageEnd(ep);    
   }
   
   /**
    * @roseuid 3F7294070388
    */
   public void dump() 
   {
   	  int i = 0;
   	  for (TCell c = firstCell; c != null; c = c.getNext())
   	  {
   	  	 i = i +1;
   	  	 System.out.println("------------------ Page (" + i + ")-------------------");
   	  	 c.dump(0);
   	  }
   }


   /**
    * @param i
    */
   public void setCurrentYPos(int i) 
   {
	   currentYPos = i;
   }

    /**
     * @return
     */
    public Image getBrokenImage() {
	    if(brokenImage == null){
		   brokenImage = new Image(getDevice(),30,30);
	    }
	    return brokenImage;
   }

   /**
     * @param image
    */
   public void setBrokenImage(Image image) {
	   brokenImage = image;
   }

   /**
	* @param c
	* @roseuid 3EFC03260070
	*/
   public void doItem(DCell c) 
   {
   	  applyFont(c);
   	
	  int size = _getColumnSizeForCell(c);
   	  
	  TBuilder b = _newBuilder(size);
	 	
	  _setBuilderMarginForCell(b);
	
	  b.doChildrenItem(c);
		
	  TCell cl = b.getFirstCell();
	  if (cl != null)
	  {
		if (currentTable.getAutoResize())
			addCell(b.getMaxWidth(), cl.getRect().height, c, 0, cl);
		else
			addCell(cl.getRect().width, cl.getRect().height, c, 0, cl);
	  	
		setCurrentPosition(b.getCurrentPosition());
	    
		// mise en place des marges
		if (currentNColumn > 0)
		{
			cl.getRect().x -= currentTable.getBorder()-1;
		}
		if (currentNRow > 0)
		{
			cl.getRect().y -= currentTable.getBorder()-1;
		}
	  }

	  // on met fin au constructeur 
	  _disposeBuilder(b);
	  
	  CellInfo cf = new CellInfo();
	  cf.rowspan = c.getRowSpan();
	  cf.cell = cl;
	  
	  if (c.getRowSpan()==0)
	  {
		maxCellHeight = Math.max(maxCellHeight, cl.getRect().height);
	  }
	  
	  currentRowSpan[currentNColumn] = cf;
	  currentNColumn = currentNColumn + c.getColSpan() + 1;
	  
	  restoreFont();    
   }
   
   /**
	* @param c
	* @roseuid 3EFC0332009F
	*/
   public void doItem(DCellText c) 
   {
   	  applyFont(c);
   	
	  int size = _getColumnSizeForCell(c);
	  
	  TBuilder b = _newBuilder (size);
	  _setBuilderMarginForCell(b);
	 
	  _buildTextBuffer(c, c.getText());	 
	 	
	  TCell cl = b.getFirstCell();
	   
	  addCell(cl.getRect().width, cl.getRect().height, c, 0, b.getFirstCell());
	  setCurrentPosition(b.getCurrentPosition());
	  
	  // on met fin au constructeur 
	  _disposeBuilder(b);	
	 
	  CellInfo cf = new CellInfo();
	  cf.rowspan = c.getRowSpan();
	  cf.cell = cl;
	  
	  if (c.getRowSpan()==0)
	  {
		  maxCellHeight = Math.max(maxCellHeight, cl.getRect().height);
	  }
	  
	  currentRowSpan[currentNColumn] = cf;
	  currentNColumn = currentNColumn + c.getColSpan() + 1;
	  
	  restoreFont();    
   }
   
   /**
	* @param d
	* @roseuid 3EFC0213029B
	*/
   public void doItem(DDocument d) 
   {
   	  applyFont(d);
   	
	  setAlignment(DParagraph.LEFT);
	  doChildrenItem(d);
	  
	  restoreFont();    
   }
   
   /**
	* @param g
	* @roseuid 3EFC03750268
	*/
   public void doItem(DGraphic g) 
   {
   	  applyFont(g);
   	
	  int w = (int)((float)getMaxWidth() * g.getWidthScale());
	  int h = (int)((float)getMaxWidth() * g.getHeightScale());
	  float ratio = g.getWidthScale()/g.getHeightScale();
	  if (ratio <= 1.0f)
	  {
		if (w < 100)
		{
			w = 100;
			h = (int)(w/ratio);
		}
		else
		if (h < 100)
		{
			h = 100;
			w = (int)(h*ratio);
      		
		}
	  }
      
	  addCell(w, h, g, 0);
	  
	  restoreFont();     
   }
   
   /**
	* @param i
	* @roseuid 3EFC02FC00CA
	*/
   public void doItem(DImage i) 
   {
	 _addListItemPrefix(i);
	  
	 // search image 
	 Image img = (Image)getImagePool().get(i.getName());
	 if (img == null)
	 {
		String imgfilename = searchImage(i.getName());
		if (imgfilename != null)
		{
		   img = new Image(getDevice(), i.getName());
		   getImagePool().put(i.getName(), img);
		}
		else
		{
		   img = (Image)getBrokenImage();
		}
	 }
  
	 Rectangle r = img.getBounds();
	 int w = (int)(r.width * i.getWidthScale());
	 int h = (int)(r.height * i.getHeightScale());
	      
	 if (getWordWrap() 
		 && (w + getCurrentWLine()) > getMaxWidth())
	 {
		 newLine();	
	 }
     
	 TCell c = addCell(w, h, i, 0);
	 c.setImage(img);    
   }
   
   
   /**
	* @param l
	* @roseuid 3EFC02AF03CC
	*/
   public void doItem(DList l) 
   {
   	  applyFont(l);
   	
	  newLine();

	  IDStyle s = l.getStyle();
	
	  addListLevel();
	  setXoffset(getXoffset() * getListOffsetSpacing());
		
	  doChildrenItem(l);
	
	  removeListLevel();
	  setXoffset(getXoffset() * getListOffsetSpacing());
	
	
	  newLine();   
	
	  restoreFont(); 
   }
   
   /**
	* @param l
	* @roseuid 3EFC02AF03CC
	*/
   public void doItem(DLine l) 
   {
	   newLine();
	
	   addCell(getMaxWidth(),
				l.getSize()+getCellSpacing()*2,l,0);
	
	   newLine();    
   }
   
   /**
	* @param pb
	* @roseuid 3F004C3A029E
	*/
   public void doItem(DPageBreak pb) 
   {    
	  if (getModePage())
		  newPage();
	  else
		  newLine();
   }
   
   /**
	* @param pc
	* @roseuid 3F004C20039B
	*/
   public void doItem(DPageCounter pc) 
   { 
   	  applyFont(pc);
   	
	  String s = new Integer(getCurrentNPage()).toString();
      
	  Point p = getGc().textExtent(s);
	  TCell c = addCell(p.x, p.y, pc, s.length());
	  c.setText(s); 
	  
	  restoreFont();   
   }
   
   /**
	* @param p
	* @roseuid 3EFC025D0388
	*/
   public void doItem(DParagraph p) 
   {
   	  applyFont(p);
   	
	  int al = getAlignment();
	  newLine();
	  setAlignment(p.getAlign());
	  
	  _addListItemPrefix(p);
	  
	  doChildrenItem(p);
	  
	  newLine();
	  setAlignment(al);
	  
	  restoreFont();    
   }
   
   /**
	* @param p
	* @roseuid 3EFC02DE0329
	*/
   public void doItem(DPopup p) 
   {
   	  applyFont(p);
   	
	  if (getDrawPopup())
	  {
		setAlignment(DParagraph.LEFT);
		doChildrenItem(p);
	  }    
	  
	  restoreFont();
   }
   
   /**
	* @param p
	* @roseuid 3EFC02CA0280
	*/
   public void doItem(DPopupLink p) 
   {
   	 applyFont(p);
   	
	 doChildrenItem(p);
	 if (!getFlatPopup()) return;
	 
	 DLinkResolver r = new DLinkResolver(p);
	 DPopup popup = (DPopup)r.getTaggedItem();
	 
	 if (popup!=null)
	 {
		TBuilder b = _newBuilder(getMaxWidth());
	 	
		int m = getCellSpacing();
		b.setTopMargin(m);
		b.setBottomMargin(m);
		b.setLeftMargin(m);
		b.setRightMargin(m);
		
		b.doChildrenItem(popup);
		
		TCell c = b.getFirstCell();
		
		addCell(c.getRect().width, c.getRect().height, popup, 0);   
		setCurrentPosition(b.getCurrentPosition());

		// on met fin au constructeur 
		_disposeBuilder(b);	
	 }    
	 
	 restoreFont();
   }
   
   /**
	* @param r
	* @roseuid 3EFC03190216
	*/
   public void doItem(DRow r) 
   {
   	  applyFont(r);
   	
	  currentNColumn = 0;
	  currentRow = r;
	  maxCellHeight = 0;
	  
	  // calcul numero de colonne de la premiere cellule de la liste
	  // a partir de la deuxieme ligne
	  if (currentNRow>0)
	  {
		for (currentNColumn=0; 
			   (currentRowSpan[currentNColumn].rowspan>0) 
				  && (currentNColumn<currentRowSpan.length); currentNColumn++);
	  }
	  
	  
	  doChildrenItem(r);
	  
	  if (currentNRow>0)
	  {
		int w = 0;
		for (int i=0; i<currentRowSpan.length; i++)
		{
		  if (currentRowSpan[i]==null) continue;
		  TCell c = currentRowSpan[i].cell;
		  
		  if (currentRowSpan[i].rowspan>0)
		  { 
			 w += c.getRect().width;
			 currentRowSpan[i].rowspan -= 1;
			 if (currentRowSpan[i].rowspan==0)
			 {
				c.getRect().height = (getCurrentYPos() + maxCellHeight) - c.getRect().y;
				currentRowSpan[i] = null;
			 }
		  }
		  else
		  {
			c.getRect().x = c.getRect().x + w;			
			c.getRect().height = maxCellHeight;
			w += c.getRect().width;
		  }
		}
	  }
	  
	  int currentYPos = getCurrentYPos();
	  newLine();
	  setCurrentYPos(maxCellHeight + currentYPos);
	  currentNRow = currentNRow + 1;
	  
	  restoreFont();    
   }
   
   /**
	* @param s
	* @roseuid 3EFC02410247
	*/
   public void doItem(DSection s) 
   {
   	  applyFont(s);
   	
	  setHeader(s.getHeader());
	  setFooter(s.getFooter());
	  
	  if (getModePage())
		  newPage();
	  else
		  newLine();
          
	  addCell(0,0,s,0);
      
	  doChildrenItem(s);
	  
	  restoreFont();    
   }
   
   /**
	* @param t
	* @roseuid 3EFC030901EB
	*/
   public void doItem(DTable t) 
   {
   	  applyFont(t);
   	
	  newLine();
	  currentTable = t;
	  currentNRow = 0;
	  currentTotalColumn = t.getColumnCount();
	  currentRowSpan = new CellInfo[t.getColumnCount()];
      
      
	  doChildrenItem(t);    
	  
	  currentTable = null;
	  currentRow = null;
	  currentRowSpan = null;
	  
	  restoreFont(); 
   }
   
   /**
	* @param t
	* @roseuid 3EFC03A001FC
	*/
   public void doItem(DTag t) 
   {
	  addCell(0,0,t,0);    
   }
   
   /**
	* @param t
	* @roseuid 3EFC02EE0142
	*/
   public void doItem(DText t) 
   {
   	  applyFont(t);
   	  
	  _addListItemPrefix(t);
	  _buildTextBuffer(t, t.getText());
	  
	  restoreFont();    
   }
   
   /**
	* @param t
	* @roseuid 3EFC02A00244
	*/
   public void doItem(DTitle t) 
   {
   	  applyFont(t);
   	
	  addTitleLevel(t.getLevel());
	  String st = getTitleStringPrefix(t);
	  
	  Point r = getGc().textExtent(st);
	  
	  addYSpacing(getTitleSpacing());
	  newLine();
	  
	  TCell c = addCell(r.x, r.y + getTitleSpacing(), t, st.length());
	  c.setText(st);
	  
	  doChildrenItem(t);
	  
	  newLine();   
	  
	  restoreFont(); 
   }
   
   /**
	* @param f
	* @roseuid 3F02AEBE025D
	*/
   public void doItem(DFooter f) 
   {
   	   applyFont(f);
   	
	   //	fabrication de l'entete et pied de base
	   TBuilder b = _newBuilder (getMaxWidth());
		
	   b.doChildrenItem(f);
		
	   setFooterCell(b.getFirstCell());
	   b.setFirstCell(null);
			
	   // on met fin au constructeur 
	   _disposeBuilder(b);
	   
	   restoreFont();    
   }
   
   /**
	* @param h
	* @roseuid 3F02AEAC03AC
	*/
   public void doItem(DHeader h) 
   {
   	   applyFont(h);
   	
	   // fabrication de l'entete et pied de base
	   TBuilder b = _newBuilder(getMaxWidth());
	 	
	   b.doChildrenItem(h);
		
	   setHeaderCell(b.getFirstCell());
	   b.setFirstCell(null);
			
	   // on met fin au constructeur 
	   _disposeBuilder(b);
	   
	   restoreFont();    
   }
   
   /**
	* @param text
	* @param w
	* @param r
	* @roseuid 3F02AF060329
	*/
   private void _buildGetNextWord(String text, int w, Range r) 
   {
	  String qs = text.substring(r.start);
	  int wt = getGc().textExtent(qs).x;
	  if (wt <= w)
	  {
		r.end = wt;
		r.start = text.length(); 
	  }
	  else
	  {
		int pos = ((w * qs.length()) / wt);
		for (int i=pos; i > 0; i = i -1)
		{
		   char c = qs.charAt(i);
		   if (!Character.isSpaceChar(c)
				&& (c!='.') && (c!=';') && (c!=',') && (c!='?')
				&& (c!=':') && (c!='(') && (c!=')') && (c!='/')
				&& (c!='[') && (c!=']') && (c!='\\'))
		   {
			  continue;
		   }

		   wt = getGc().textExtent(qs.substring(0, i)).x;
		   if (wt < w)
		   {
			  r.end = wt;
			  r.start = r.start + i;
			  return;
		   }
		}
		// on est pas arrive a decouper la chaine sur des separateurs
		// on va decouper les mots par dicotomie
		int a = qs.length() / 2;
		int b = qs.length();
		while (b-a>1)
		{
		   String s = qs.substring(0, a);
		   wt = getGc().textExtent(s).x;
		   if (wt < w)
		   {
			  a = a + (b-a)/2;
		   }
		   else
		   {
			  b = a;
			  a = a / 2;
		   }
		}
		r.end = wt;
		r.start = a + r.start;
	  }
   }

}



/*
 * Created on Sep 19, 2003
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package com.rational.apidoc.drivers.html;

import java.util.LinkedList;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;

import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.ImageLoader;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.SWTError;

import com.rational.apidoc.parsers.DParser;
import com.rational.apidoc.drivers.IWriter;
import com.rational.apidoc.core.*;
import com.rational.apidoc.examples.*;
import com.rational.apidoc.chart.ChartSnapShot;

/**
 * @author sduguet
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
public class DHtmlDriver extends DParser implements IWriter {

	private ConfParser confparser;
	private PrintStream out;

	private String dirname, filename;
	private DDocument document;
	
	private static Item default_hitem = new Item();

	/**
	 * 
	 */
	public DHtmlDriver() {
		super();
		confparser = new ConfParser();
		if (!confparser.parse()) {
			System.err.println("Html Configuration parsing failed");
			confparser = null;
			return;
		}
	}

	public void setDestDirAndFileName(String dname, String fname) {
		dirname = dname != null ? dname : ".";
		filename =
			dirname + "/" + fname + confparser.getHtmlFile().getExtension();
	}

	public PrintStream getStream() {
		return out;
	}

	public void htmlConvertDump(String string) {
		int space_count = 0;
		for (int i = 0; i < string.length(); i++) {
			char c = string.charAt(i);
			if (c == ' ') {
				space_count++;
				if (space_count == 1) {
					out.print(confparser.getConvertText().getFirstSpace());
				} else {
					out.print(confparser.getConvertText().getExtraSpace());
				}
			} else {
				space_count = 0;
				Character ch = new Character(c);
				String repl = confparser.getConvertText().getReplace(ch);
				String cstr = repl != null ? repl : ch.toString();
				try {
					byte a[] = cstr.getBytes("UTF-8");
					out.write(a, 0, a.length);
				} catch (UnsupportedEncodingException e) {
					out.print(cstr);
				}
			}
		}
	}

	public void htmlDump(String string) {
		try {
			byte a[] = string.getBytes("UTF-8");
			out.write(a, 0, a.length);
		} catch (UnsupportedEncodingException e) {
			out.print(string);
		}
	}

	public void htmlDump(int value) {
		out.print(value);
	}

	public void htmlDump(LinkedList list) {
		for (int i = 0; i < list.size(); i++) {
			String elt = (String) list.get(i);
			String value = confparser.getVariableValue(elt);
			if (value == "#") {
				htmlDump("${" + elt + "}");
			} else {
				htmlDump(value != null ? value : elt);
			}
		}
	}

	private void setTagVariable(String name, String tag, IDItem item) {
		confparser.setVariableValue(
			name,
			tag != null ? tag : "_" + item.hashCode(),
			item);
	}

	private String to2DigitsHexa(int color) {
		return Integer.toHexString((color % 256) / 16)
			+ Integer.toHexString(color % 16);
	}

	private String getColorHexa(IDColor color, String default_value) {
		if (color == null) {
			return default_value;
		}
		String result = "#";
		result += to2DigitsHexa(color.getRed());
		result += to2DigitsHexa(color.getGreen());
		result += to2DigitsHexa(color.getBlue());
		return result;
	}

	private String getFontAttrsBegin(IDStyle style) {
		IDFont font = style.getFont();
		IDColor color = style.getForeColor();
		if (font == null && color == null) {
			return "";
		}
		String result = "";
		if (font != null) {
			if (font.getBold()) {
				result += "<B>";
			}
			if (font.getUnderline() || font.getDashed()) {
				result += "<U>";
			}
			if (font.getItalic()) {
				result += "<I>";
			}
		}
		result += "<FONT COLOR=" + getColorHexa(color, "#000000");
		if (font != null && font.getSize() != -1) {
			int s = font.getSize() / 3; // size 12 -> 4 (normal size)
			if (s < 1) {
				s = 1; // 1 is the minimum
			}
			if (s > 7) {
				s = 7; // 7 is the maximum
			}
			result += " SIZE=\"" + Integer.toString(s) + "\"";
		}
		result += ">";
		if (font != null && confparser.isTTFont(font.getName())) {
			result += "<TT>";
		}

		return result;
	}

	private String getFontAttrsEnd(IDStyle style) {
		IDFont font = style.getFont();
		IDColor color = style.getForeColor();
		if (font == null && color == null) {
			return "";
		}
		String result = "";
		if (font != null && confparser.isTTFont(font.getName())) {
			result += "</TT>";
		}
		result += "</FONT>";
		if (font != null) {
			if (font.getBold()) {
				result += "</B>";
			}
			if (font.getUnderline() || font.getDashed()) {
				result += "</U>";
			}
			if (font.getItalic()) {
				result += "</I>";
			}
		}
		return result;
	}

	private void updateStyleVariables(IDItem item) {
		// PRE: item.getStyle() not null (to be checked by the caller)
		confparser.setVariableValue(
			"HTML_FONT_ATTRS_BEGIN",
			getFontAttrsBegin(item.getStyle()),
			item);
		confparser.setVariableValue(
			"HTML_FONT_ATTRS_END",
			getFontAttrsEnd(item.getStyle()),
			item);
		String fore_color =
			getColorHexa(item.getStyle().getForeColor(), "#000000");
		confparser.setVariableValue(
			"HTML_COLOR",
			"=\"" + fore_color + "\"",
			item);
		confparser.setVariableValue("HTML_FORE_COLOR", fore_color, item);
		String back_color =
			getColorHexa(item.getStyle().getBackColor(), "#ffffff");
		confparser.setVariableValue(
			"HTML_BGCOLOR",
			"=\"" + back_color + "\"",
			item);
		confparser.setVariableValue("HTML_BACK_COLOR", back_color, item);
	}

	public Item getItem(IDItem item) {
		// Perform some checks
		Item hitem = confparser.getItem(item.getClass());
		if (hitem == null) {
			System.err.println("No html parser found for " + item.getClass());
			hitem = default_hitem;
		}
		return hitem;
	}

	public void doOneItemAfter(IDItem item) {
		if (item instanceof DText) {
		} else if (item instanceof DCellText) {
		} else if (item instanceof DGraphic) {
			graphicDump((DGraphic) item);
		}
	}

	public void doItem(DDocument item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("DOCUMENT_TAG", item.getTag(), item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DSection item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("SECTION_TAG", ((DSection) item).getTag(), item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DHeader item) {
		// ignored
	}

	public void doItem(DFooter item) {
		// ignored
	}

	public void doItem(DParagraph item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"PARAGRAPH_ALIGN",
			((DParagraph) item).alignToString(((DParagraph) item).getAlign()),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DTag item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("TAG_NAME", ((DTag) item).getTag(), item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DText item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		htmlDump(hitem.getBegin());
		htmlConvertDump(item.getText());
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DImage item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"IMAGE_NAME",
			((DImage) item).getName(),
			item);
		htmlDump(hitem.getBegin());
		// no childs
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DList item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		if (((DList) item).getFormat() == DStyle.BULLET) {
			confparser.setVariableValue("LIST_TYPE", "BULLETED", item);
			confparser.setVariableValue("HTML_LIST_TYPE", "UL", item);
		} else if (((DList) item).getFormat() == DStyle.NUMBER) {
			confparser.setVariableValue("LIST_TYPE", "NUMBERED", item);
			confparser.setVariableValue("HTML_LIST_TYPE", "OL", item);
		} else {
			confparser.setVariableValue("LIST_TYPE", "", item);
			confparser.setVariableValue("HTML_LIST_TYPE", "UL", item);
		}
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DTable item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		if (((DTable) item).getAutoResize()) {
			float w = 0;
			float wc[] = ((DTable) item).getColumnWidth();
			if (wc == null || wc.length < ((DTable) item).getColumnCount()) {
				w = ((DTable) item).getWidthScale();
			} else {
				for (int i = 0; i < wc.length; i++) {
					w += wc[i];
				}
			}
			confparser.setVariableValue("TABLE_WIDTH", Float.toString(w), item);
			confparser.setVariableValue(
				"HTML_TABLE_WIDTH",
				Integer.toString((int) (w * 100.0f)),
				item);
		} else {
			float w = 0.0f;
			confparser.setVariableValue("TABLE_WIDTH", Float.toString(w), item);
			confparser.setVariableValue(
				"HTML_TABLE_WIDTH",
				Integer.toString((int) (w * 100.0f)),
				item);
		}
		confparser.setVariableValue(
			"TABLE_BORDER",
			Integer.toString(((DTable) item).getBorder()),
			item);
		confparser.setVariableValue(
			"TABLE_COLUMNS",
			Integer.toString(((DTable) item).getColumnCount()),
			item);
		confparser.setVariableValue(
			"TABLE_ROWS",
			Integer.toString(((DTable) item).getRowCount()),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DRow item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"ROW_CELLS",
			Integer.toString(((DRow) item).getCellCount()),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	private void setCellVariables(DCell item) {
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"CELL_COLSPAN",
			Integer.toString(((DCell) item).getColSpan()),
			item);
		confparser.setVariableValue(
			"HTML_CELL_COLSPAN",
			Integer.toString(((DCell) item).getColSpan() + 1),
			item);
		confparser.setVariableValue(
			"CELL_ROWSPAN",
			Integer.toString(((DCell) item).getRowSpan()),
			item);
		confparser.setVariableValue(
			"HTML_CELL_ROWSPAN",
			Integer.toString(((DCell) item).getRowSpan() + 1),
			item);
		confparser.setVariableValue(
			"CELL_COL",
			Integer.toString(((DCell) item).getColumnNumber()),
			item);
		confparser.setVariableValue(
			"CELL_ROW",
			Integer.toString(((DCell) item).getLineNumber()),
			item);
		DRow row = (DRow) item.getParent();
		DTable table = (DTable) row.getParent();
		if (table.getColumnWidth() != null
			&& table.getColumnCount() > ((DCell) item).getColumnNumber()) {
			float w = table.getColumnWidth()[((DCell) item).getColumnNumber()];
			confparser.setVariableValue("CELL_WIDTH", Float.toString(w), item);
			confparser.setVariableValue(
				"HTML_CELL_WIDTH",
				Integer.toString((int) (w * 100.0f)),
				item);
		}
	}

	public void doItem(DCell item) {
		Item hitem = getItem(item);
		setCellVariables(item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DCellText item) {
		Item hitem = getItem(item);
		setCellVariables(item);
		htmlDump(hitem.getBegin());
		htmlConvertDump(((DCellText) item).getText());
		// no childs
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DLink item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		String url =
			(((DLink) item).isInternal() ? "" : ((DLink) item).getMethod())
				+ ":"
				+ ((DLink) item).getPath();
		confparser.setVariableValue("LINK_TARGET", url, item);
		confparser.setVariableValue("HTML_LINK", url, item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DLine item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"HTML_LINE_SIZE",
			Integer.toString(((DLine) item).getSize()),
			item);
		htmlDump(hitem.getBegin());
		// no childs
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DPageBreak item) {
		// ignored
	}

	public void doItem(DPageCounter item) {
		// ignored
	}

	public void doItem(DPopup item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("POPUP_TAG", ((DPopup) item).getTag(), item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DPopupLink item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"POPUP_LINK_METHOD",
			((DPopupLink) item).typeToString(((DPopupLink) item).getType()),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DSummary item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DTitle item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		setTagVariable("TITLE_TAG", ((DTitle) item).getTag(), item);
		addTitleLevel(((DTitle) item).getLevel());
		confparser.setVariableValue(
			"TITLE_LEVEL",
			Integer.toString(getTitleLevel()),
			item);
		confparser.setVariableValue(
			"HTML_TITLE_LEVEL",
			Integer.toString(getTitleLevel()),
			item);
		confparser.setVariableValue(
			"TITLE_FULL_NUMBER",
			getTitleStringPrefix(item),
			item);
		confparser.setVariableValue(
			"TITLE_NUMBER",
			Integer.toString(titleTab[getTitleLevel()]),
			item);
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DIndex item) {
	}

	public void doItem(DIndexEntry item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		htmlDump(hitem.getBegin());
		doChildrenItem(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DTimerItemContainer item) {
		// ignored
	}

	private void graphicDump(DGraphic graphic) {
		String image_filename =
			dirname
				+ "/"
				+ confparser.getVariableValue("HTML_GRAPHIC_FILE_NAME")
				+ ".jpg";
		int w = 500, h = w;
		w *= graphic.getWidthScale();
		h *= graphic.getHeightScale();
		Rectangle r = new Rectangle(0, 0, w, h);
		Image img = ChartSnapShot.Process(r, graphic);
		ImageLoader loader = new ImageLoader();
		if (loader.data == null) {
			loader.data = new ImageData[1];
		}
		loader.data[0] = img.getImageData();
		try {
			loader.save(image_filename, SWT.IMAGE_JPEG);
		} catch (SWTException _e) {
			System.err.println(
				"Error saving image:" + image_filename + ": " + _e);
		} catch (SWTError _r) {
			System.err.println(
				"Error saving image:" + image_filename + ": " + _r);
		}
		img.dispose();
	}

	public void doItem(DGraphic item) {
		Item hitem = getItem(item);
		if (item.getStyle() != null) {
			updateStyleVariables(item);
		}
		confparser.setVariableValue(
			"GRAPHIC_TITLE",
			((DGraphic) item).getTitle(),
			item);
		confparser.setVariableValue(
			"GRAPHIC_WIDTH",
			Float.toString(((DGraphic) item).getWidthScale()),
			item);
		confparser.setVariableValue(
			"HTML_GRAPHIC_WIDTH",
			Integer.toString(
				(int) (((DGraphic) item).getWidthScale() * 100.0f)),
			item);
		confparser.setVariableValue(
			"GRAPHIC_HEIGHT",
			Float.toString(((DGraphic) item).getHeightScale()),
			item);
		confparser.setVariableValue(
			"HTML_GRAPHIC_FILE_NAME",
			"img" + Integer.toHexString(item.hashCode()),
			item);
		htmlDump(hitem.getBegin());
		graphicDump(item);
		htmlDump(hitem.getEnd());
		confparser.unsetVariables(item);
	}

	public void doItem(DAxis item) {
		// already made from the enclosing DGraphic
	}

	public void doItem(DCurve item) {
		// already made from the enclosing DGraphic
	}

	public void doItem(DPoint item) {
		// already made from the enclosing DGraphic
	}

	public void write(IDItem doc) {
		document = (DDocument) doc;

		if (confparser == null) {
			System.err.println("Html Configuration parsing failed");
			return;
		}
		try {
			out = new PrintStream(new FileOutputStream(filename));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return;
		}

		// Some variables at document level
		confparser.setVariableValue("HTML_FILE_NAME", filename, document);
		confparser.setVariableValue(
			"DOCUMENT_TITLE",
			document.getTitle(),
			document);
		confparser.setVariableValue("HTML_BGCOLOR", "=\"#ffffff\"", document);

		setTitleMaxLevel(confparser.getHtmlFile().getMaxTitleLevel() - 1);
		HtmlFile hf = confparser.getHtmlFile();
		htmlDump(hf.getBegin());
		doItem(document);
		htmlDump(hf.getEnd());

		out.close();
	}

	public static void main(String[] args) {
		System.out.println("initialize...");
		DHtmlDriver hd = new DHtmlDriver();

		System.out.println("create...");

//		Example1 ex1 = new Example1();
//		hd.setDestDirAndFileName(".", "file1");
//		System.out.println("translate...");
//		hd.write(ex1.doc);
//		hd.clear();
//
//		Example2 ex2 = new Example2();
//		hd.setDestDirAndFileName(".", "file2");
//		System.out.println("translate...");
//		hd.write(ex2.doc);
//		hd.clear();
//
//		Example3 ex3 = new Example3();
//		hd.setDestDirAndFileName(".", "file3");
//		System.out.println("translate...");
//		hd.write(ex3.doc);
//		hd.clear();
//
//		Example4 ex4 = new Example4();
//		hd.setDestDirAndFileName(".", "file4");
//		System.out.println("translate...");
//		hd.write(ex4.doc);
//		hd.clear();
//
//		Example5 ex5 = new Example5();
//		hd.setDestDirAndFileName(".", "file5");
//		System.out.println("translate...");
//		hd.write(ex5.doc);
//		hd.clear();
//
//		Example6 ex6 = new Example6();
//		hd.setDestDirAndFileName(".", "file6");
//		System.out.println("translate...");
//		hd.write(ex6.doc);
//		hd.clear();
//
		Source so = new Source();
		hd.setDestDirAndFileName(".", "source");
		System.out.println("translate...");
		hd.write(so.doc);
		hd.clear();
		
		System.out.println("done!");
	}
}
/*
 * Created on 26 juin 2003
 *
 */
package com.rational.apidoc.chart;

import org.eclipse.swt.SWT;

import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Layout;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Button;

import com.rational.apidoc.drawutil.*;
import com.rational.apidoc.core.* ;

/**
 * Example to create and render a new graphic type for ApiDoc.
 * The new graphic will draw polar curves.
 * 
 * To extends graphics type, follow these steps:
 * - define a new id for graphic type: see T_POLAR field.
 * - define an IChartRender that able to draw your graphic
 *   if render have support to location, extends IChartLocation 
 *   (thrown by your ChartRender).
 * 
 * - don't forget to register the new IChartRender using DChartRenderRegistry
 *   (see it in main() method)
 * 
 * - now you can build your document containing new graphic type, see this
 *   in BuildPolarGraphic() method.
 */
public class TestUserGraphicType
{  
   /** Define a new graphic type identifier: */   
   public static final String T_POLAR = "T_POLAR.sample.chart.apidoc.rational.com";
   
   /** First axis a Polar graphic must provide */
   public static final String AXIS_RAY = "ray-axis";
   /** Second axis a Polar graphic must provide */
   public static final String AXIS_ANGLE ="angle-axis";
   
   /** Class used to identify where is located a point on a polar graphic */
   public static class PolarLocation implements IChartLocation
   {
      /** Located on graphic title area */
      public static final String GRAPHIC_TITLE="graphic-title";
      /** Located on a segment of a curve (item is first point of segment) */
      public static final String CURVE_SEGMENT="curve-segment";
      /** located on a point of a curve (only for curve type DCurve.POINT) */
      public static final String CURVE_POINT  ="curve-point";
      
      private String id_;
      private IDItem item_;
      private Rectangle rect_;
      
      public PolarLocation( String _id, IDItem _item, Rectangle _r )
      {
        id_=_id; item_=_item; rect_=_r;
      }
      
      public String getId() { return id_; }
      public IDItem getItem() { return item_; }
      public Rectangle getRectangle() { return rect_; }
   }
   
   /** 
    * Implements an IChartRender for this new graphic type
    */   
   private static class PolarChartRender implements IChartRender
   {
     public boolean canRender(String _graphic_type)
     {
       if( T_POLAR.equals( _graphic_type )) return true;
       return false;
     }

     public void draw(GC _gc, Rectangle _R, DGraphic _g)
     {   
       render( true, 0,0, _gc, _R, _g );
     }
     
     public IChartLocation locate(int _x, int _y, GC _gc, Rectangle _R, DGraphic _G)
     {
       return render( false, _x, _y, _gc, _R, _G );
     }
     
     /** support for both draw and location modes */
     public IChartLocation render( boolean _draw, int _lx, int _ly, GC _gc, Rectangle _R, DGraphic _g)
     {   
       if( _draw )
       {
         _gc.setBackground( new Color( null, 255,255,255 ));
         _gc.fillRectangle( _R );                
       }
       else
       {
         if( !_R.contains( _lx, _ly ) ) return null; //outside graphic
       }
       
       //search axis
       DAxis r_axis = null;
       DAxis a_axis = null;
       for( IDItem item=_g.getFirstChild(); item!=null; item=item.getNext() )
       {
         if( item instanceof DAxis )
         {
           DAxis axis = (DAxis)item;
           if( AXIS_RAY.equals( axis.getName() ) ) r_axis=axis;
           else if ( AXIS_ANGLE.equals( axis.getName() ) ) a_axis=axis;
         }
       }
       if( r_axis==null || a_axis==null )
       {
         if( _draw )
         {
           Color red = new Color( null, 255,0,0 );
           String missed = (r_axis==null?"'AXIS_RAY'":"")+" "+(a_axis==null?"'AXIS_ANGLE'":"");
           _gc.setForeground( red );
           _gc.drawText( "Error: missing axis "+missed+"' in graphic item", 10, 30 );
         }
         return null;
       }
       
       final int margin = 10;
       int lx = _R.x+margin;
       int ty = _R.y+margin;
       int rx = _R.x+_R.width-margin;
       int by = _R.y+_R.height-margin;
       int cx = (lx+rx)/2;
       int cy = (ty+by)/2; 
       
       //render title if any
       String title = _g.getTitle();
       if( !DrawUtil.isEmpty( title )
         && _g.getProperties().get( DGraphic.P_SHOW_TITLE, true ) )
       {
         IDStyle style = _g.getStyle();
         Font save = _gc.getFont();
         Font font = DrawUtil.getFont( style.getFont() );
         int aln = DrawUtil.HCENTER|DrawUtil.VCENTER|DrawUtil.VERTICAL;
         if( font!=null ) _gc.setFont( font );
         int tw = DrawUtil.textExtent( _gc, title, aln ).x;
         if (_draw)
         {
           //gradient background
           _gc.setForeground( new Color(null,205,220,255));
           _gc.fillGradientRectangle( lx-1, ty, tw+3, by-ty, true );
           //title color and text
           _gc.setForeground( DrawUtil.getColor(style.getForeColor()) );
           DrawUtil.drawText(_gc, lx, ty, tw, by - ty, aln, title);
         } else {
           Rect r = DrawUtil.boundingRect( _gc, lx, ty, tw, by-ty, aln, title );
           if( r.contains( _lx,_ly ) )
           {
             return new PolarLocation( PolarLocation.GRAPHIC_TITLE, _g, r.toRectangle() );
           }
           if( _lx <= lx+tw )
           {
             //located on graphic title area (outside text),
             //no need to search in curve, but no id is defined for this...
             return null;
           }
         }

         _gc.setFont( save );
         
         lx+=tw+margin;
         cx = (lx+rx)/2;
       }
       
       //first search curve and points to compute min/max
       //an array of DCurve is stored to be able to iterate in reverse order
       //as location mode need.
       DCurve curves[] = new DCurve[ _g.getChildCount() ];
       int    n_curves=0; //because curves.length take into account DAxis too... ;-(
       float xmin=0.0f, xmax=0.0f, ymin=0.0f, ymax=0.0f;
       float ray_max=0.0f;
       boolean init=false;       
       
       for( IDItem item=_g.getFirstChild(); item!=null; item=item.getNext() )
       {
         if( !(item instanceof DCurve) ) continue;
         DCurve curve = (DCurve)item;
         curves[ n_curves++ ] = curve;
         for( IDItem item2=curve.getFirstChild(); item2!=null; item2=item2.getNext())
         {
           if(!(item2 instanceof DPoint)) continue;
           DPoint point = (DPoint)item2;
           DCoord c_angle = point.getCoord( a_axis );
           DCoord c_ray   = point.getCoord( r_axis );
           if( c_angle==null || c_ray==null )
           {
             //no error message here, skip point only
             continue;
           }
           float r = c_ray.getValue();
           float a = c_angle.getValue();
           if( r > ray_max ) ray_max = r;
           //convert to pixels
           float ptx = (float)( r*Math.cos(a) );
           float pty = (float)( r*Math.sin(a) );
           if( init )
           {
             if( ptx < xmin ) xmin=ptx; else if ( ptx > xmax ) xmax=ptx;
             if( pty < ymin ) ymin=pty; else if ( pty > ymax ) ymax=pty;
           }
           else
           {
             xmin=xmax=ptx; 
             ymin=ymax=pty;
             init=true;
           }
         }
       }
       
       float K=1.0f;
       float c_max ;
       {
         c_max = xmax;
         float x = Math.abs( xmin ); if( x > c_max) c_max=x;
         if( ymax > c_max ) c_max=ymax;
         x = Math.abs( ymin ); if ( x > c_max ) c_max=x;
         int pix_min = cx-lx < cy-ty ? cx-lx : cy-ty;
         if( c_max > 0.0f ) K = pix_min / c_max;
       }
       
       //draw axis lines
       if( _draw )
       {
         _gc.setForeground( new Color( null, 220,220,220 ));
         Rectangle r = _gc.getClipping();
         _gc.setClipping( lx, ty, rx-lx, by-ty );
         //Line Step property is set on 'ray' axis ?
         float ray_step_line = r_axis.getProperties().get( DAxis.P_STEP_LINE, 0.0f );
         if( ray_step_line >=0.0f )
         {
            for( float ray=ray_step_line; ray <= c_max; ray += ray_step_line )
            {
              int p_ray = (int)( K*ray );
              int p_ray2 = 2*p_ray;
              _gc.drawOval( cx-p_ray, cy-p_ray, p_ray2, p_ray2 );
            }
         }
         
         //Line Step property is set on 'angle' axis ?
         float ang_step_line = a_axis.getProperties().get( DAxis.P_STEP_LINE, 0.0f );
         if( ang_step_line >=0.0f )
         {
           float dpi = (float)( 2*Math.PI );
           for( float a = ang_step_line; a < dpi; a+=ang_step_line )
           {
             int px = (int)( cx + K*c_max*Math.cos(a) );
             int py = (int)( cy + K*c_max*Math.sin(a) );
             _gc.drawLine( cx, cy, px, py );
           }
         }
         //Dot Step property is set on 'angle' axis ?
         float ang_step_dot = a_axis.getProperties().get( DAxis.P_STEP_DOT, 0.0f );
         if( ang_step_dot >=0.0f )
         {
           float dpi = (float)( 2*Math.PI );
           _gc.setLineStyle( SWT.LINE_DOT );
           for( float a = ang_step_dot; a < dpi; a+=ang_step_dot )
           {
             int px = (int)( cx + K*c_max*Math.cos(a) );
             int py = (int)( cy + K*c_max*Math.sin(a) );
             _gc.drawLine( cx, cy, px, py );
           }
           _gc.setLineStyle( SWT.LINE_SOLID );
         }
         
         _gc.setForeground( new Color( null, 0,0,0 ));
         _gc.drawLine( lx, cy, rx, cy );
         _gc.drawLine( cx, ty, cx, by );
         

         _gc.setClipping( r );
       }

       
       Color def_clr = new Color( null, 0,150,0 );
       //render curves...
       // - draw in forward order
       // - locate in reverse order as the painter algorythm request.
       int crv_shift= _draw ? 0: n_curves-1 ;
       for( int icrv=0; icrv<n_curves; ++icrv )
       {
         DCurve curve = curves[ Math.abs( icrv - crv_shift) ];
         //curve type is dots or lines ?
         boolean dots = DCurve.T_POINTS.equals( curve.getType() );
         if( _draw )
         {        
           Color clr = def_clr;
           IDStyle style = curve.getStyle();
           if( style!=null )
           {
             clr = DrawUtil.getColor( style.getBackColor() );
             if( dots )
             {
               _gc.setBackground( clr );
               clr = DrawUtil.getColor( style.getForeColor() );
             }
           }
           _gc.setForeground( clr );
         }
         
         boolean draw_line=false;
         int ppx=0, ppy=0;
         DPoint last_point=null;
         for( IDItem item2=curve.getFirstChild(); item2!=null; item2=item2.getNext())
         {
           if(!(item2 instanceof DPoint)) continue;
           DPoint point = (DPoint)item2;
           DCoord c_angle = point.getCoord( a_axis );
           DCoord c_ray   = point.getCoord( r_axis );
           if( c_angle==null || c_ray==null )
           {
             //no error message here, skip point only
             draw_line=false;
             last_point=null;
             continue;
           }
           float r = c_ray.getValue();
           float a = c_angle.getValue();
           //convert to pixels
           int px = (int)( cx + K*r*Math.cos(a) );
           int py = (int)( cy + K*r*Math.sin(a) );
           if( dots )
           {
             Rectangle rect = new Rectangle( px-2, py-2, 4, 4 );
             if( _draw )
             {
               _gc.fillRectangle( rect );
               _gc.drawRectangle( rect );
             }
             else
             {
               //point located on the curve Points
               if( rect.contains(_lx, _ly ) )
               {
                 return new PolarLocation( PolarLocation.CURVE_POINT, point, rect );
               }
             }
           }
           else //other curves draw using lines
           {
             if( draw_line )
             {
               if( _draw )
               {
                 _gc.drawLine( ppx, ppy, px, py );
               }
               else
               {
                 //point near this segment ?
                 if( DrawUtil.segmentContains( ppx, ppy, px, py, _lx,_ly ) )
                 {
                   return new PolarLocation( PolarLocation.CURVE_SEGMENT, last_point,
                           new Rectangle( _lx-1,_ly-1,2,2) ); 
                 }
               }
             } else {
               draw_line=true;
             } 
             ppx=px;
             ppy=py;
             last_point=point;
           }
         }
       }  
       
       return null;     
     }

   }

/*---------------------------------------------------------------------------------------
 * Code from here up to the end of file are only made to build the test.
 * You can see how to use your graphic type in an apidoc document in the
 * BuildPolarGraphic method.
 *---------------------------------------------------------------------------------------
 */
 
   /** used only for test shell */
   public static interface IBuildGraphic
   {
     DGraphic buildGraphic();
   }   

  private static DColor black = new DColor(0,0,0);
  private static DColor blue1 = new DColor(0,0,255);
  private static DColor blue2 = new DColor(80,80,255);
  private static DColor blue3 = new DColor(150,150,255);
  private static DColor blue4 = new DColor(180,180,255);
  private static DColor green = new DColor(0,255,0);
  private static DColor red   = new DColor(255,0,0);
  private static DColor magenta = new DColor(255,0,255);
  private static DColor white = new DColor(255,255,255);

  private static DFont font1 = new DFont( "f1",null,null, 14, DFont.BOLD );
  private static DFont df    = new DFont( "df", "Arial",null,10,DFont.NORMAL );
  
  private static IDStyle sd = new DStyle( "sd", df, black, white );        
  private static IDStyle s0 = new DStyle( "s0", font1, green, white );
  private static IDStyle s1 = new DStyle( "s1", df, black, red );
  private static IDStyle s2 = new DStyle( "s2", df, black, blue2 );
  private static IDStyle s3 = new DStyle( "s3", df, black, blue4 );
  private static IDStyle s4 = new DStyle( "s4", df, black, green );
  private static IDStyle s5 = new DStyle( "s5", df, black, magenta );
  private static IDStyle s6 = new DStyle( "s6", df, black, green );
     
  private static DPoint newDPoint( DCurve _p, DAxis _x, DAxis _y, float _cx, float _cy )
  {
    DPoint pt = new DPoint();
    _p.addChild( pt );
    pt.add( new DCoord( _x, _cx ));
    pt.add( new DCoord( _y, _cy ));
    return pt;
  }
  private static DPoint newDPoint3( DCurve _p, DAxis _x, DAxis _y, DAxis _z, float _cx, float _cy, float _cz )
  {
    DPoint pt = new DPoint();
    _p.addChild( pt );
    pt.add( new DCoord( _x, _cx ));
    pt.add( new DCoord( _y, _cy ));
    pt.add( new DCoord( _z, _cz ));
    return pt;
  }
    /**
   * Graphic document builder for XY graphic 
   */
  private static class BuildXYGraphic implements IBuildGraphic
  {
    public DGraphic buildGraphic()
   {
     DGraphic  gph = new DGraphic();  
     gph.setGraphicType( DGraphic.T_XY );
     gph.setTitle( "Graphic XY" );
     gph.setStyle( s0 );
//gph.setProperties().store(DGraphic.P_SHOW_TITLE, false);     
//gph.setProperties().store(DGraphic.P_SHOW_LEGEND, false); 

     DAxis  x_axs = new DAxis("x");    gph.addChild( x_axs );
     DAxis  y_axs = new DAxis("y");    gph.addChild( y_axs );
     x_axs.setStyle( sd );
     y_axs.setStyle( sd );
     x_axs.getProperties().store( DAxis.P_STEP_UNIT, 2.0f );
     x_axs.getProperties().store( DAxis.P_STEP_DOT, 0.5f );
     y_axs.getProperties().store( DAxis.P_STEP_UNIT, 2.0f );
     y_axs.getProperties().store( DAxis.P_STEP_DOT, 0.5f );
     x_axs.setTitle( "X Axis");
     x_axs.setUnit("(x)");
     y_axs.setTitle( "Y Axis");
     y_axs.setUnit("(unit)");
     
     
     DCurve    cv0 = new DCurve();            gph.addChild( cv0 );
     cv0.setName( "Bars" );
     cv0.setType( DCurve.T_BAR );
     cv0.setStyle( s4 );
     newDPoint( cv0, x_axs, y_axs, 0, 0.9f );
     newDPoint( cv0, x_axs, y_axs, 1, 1.2f );
     newDPoint( cv0, x_axs, y_axs, 2, 1.6f );
     newDPoint( cv0, x_axs, y_axs, 3, 3.4f );
     newDPoint( cv0, x_axs, y_axs, 4, 4.6f );

     DCurve    cv1 = new DCurve();            gph.addChild( cv1 );
     cv1.setName("Line");
     cv1.setStyle( s1 );
     newDPoint( cv1, x_axs, y_axs, 0,0 );
     newDPoint( cv1, x_axs, y_axs, 1,1.2f );
     newDPoint( cv1, x_axs, y_axs, 2,3.2f );
     newDPoint( cv1, x_axs, y_axs, 3,2.4f );
     newDPoint( cv1, x_axs, y_axs, 4,0.6f );

     DCurve    cv2 = new DCurve();            gph.addChild( cv2 );
     cv2.setName( "Points" );
     cv2.setType( DCurve.T_POINTS );
     cv2.setStyle( s2 );
     newDPoint( cv2, x_axs, y_axs, 0, 1.9f );
     newDPoint( cv2, x_axs, y_axs, 1, 1.2f );
     newDPoint( cv2, x_axs, y_axs, 2, 1.1f );
     newDPoint( cv2, x_axs, y_axs, 3, 0.4f );
     newDPoint( cv2, x_axs, y_axs, 4, 0.0f );
     
     x_axs.getProperties().store( DAxis.P_MIN, 0.0f );
     x_axs.getProperties().store( DAxis.P_MAX, 5.0f );
     x_axs.getProperties().store( DAxis.P_STEP_LINE, 1.0f );
     y_axs.getProperties().store( DAxis.P_MIN, 0.0f );
     y_axs.getProperties().store( DAxis.P_MAX, 5.0f );
     y_axs.getProperties().store( DAxis.P_STEP_LINE, 1.0f );
     
     return gph;
   }
  }
  
  /**
   * Graphic document builder for Polar graphic 
   */
  private static class BuildPolarGraphic implements IBuildGraphic
  {
    public DGraphic buildGraphic()
   {
     DGraphic  gph = new DGraphic();  
     // Set my graphic type 
     gph.setGraphicType( T_POLAR );
     gph.setTitle( "Polar Graphic" );
     gph.setStyle( s0 );

     DAxis  ray_axs = new DAxis( AXIS_RAY );    gph.addChild( ray_axs );
     DAxis  angle_axs = new DAxis( AXIS_ANGLE );    gph.addChild( angle_axs );
     ray_axs.setStyle( sd );
     angle_axs.setStyle( sd );
     ray_axs.getProperties().store( DAxis.P_STEP_UNIT, 2.0f );
     ray_axs.getProperties().store( DAxis.P_STEP_DOT, 0.5f );
     angle_axs.getProperties().store( DAxis.P_STEP_UNIT, Math.PI/8 );
     angle_axs.getProperties().store( DAxis.P_STEP_DOT, Math.PI/16 );
     
     DCurve    cv0 = new DCurve();            gph.addChild( cv0 );
     cv0.setName( "Petals" );
     cv0.setType( DCurve.T_LINE );
     cv0.setStyle( s4 );
     DCurve    cv1 = new DCurve();            gph.addChild( cv1 );
     cv1.setName("Spiral");
     cv1.setType( DCurve.T_POINTS );
     cv1.setStyle( s1 );
     DCurve    cv2 = new DCurve();            gph.addChild( cv2 );
     cv2.setName("Curve 3");
     cv2.setStyle( s2 );
     DCurve    cv3 = new DCurve();            gph.addChild( cv3 );
     cv3.setName("Curve 4");
     cv3.setStyle( s5 );
     
     //populate curve using various function ( ray = f( angle ) )
     float da = (float)( Math.PI/100 );
     float r;
     float rmax=0.0f;
     int cnt=10;
     for( float a=0.0f; a<6*Math.PI; a+=da )
     {
       float ca = (float)Math.cos(a);
       float sa = (float)Math.sin(a);
       
       r = (float)( (a/(10*da))*ca*sa );
       if( r > rmax ) rmax=r;
       newDPoint( cv0, ray_axs, angle_axs, r, a );
       r = (float)( a );
       if( --cnt == 0 )
       {
         if( r > rmax ) rmax=r;
         newDPoint( cv1, ray_axs, angle_axs, r, a );
         cnt=10;
       }
       r = (float)(8*ca - 10.0f*sa/(1.1f+ca));
       if( r > rmax ) rmax=r;
       newDPoint( cv2, ray_axs, angle_axs, r, a );
       r = (float)(20+3*Math.cos(10*a)-20*Math.sin(5*a)*Math.cos(2*a));
       if( r > rmax ) rmax=r;
       newDPoint( cv3, ray_axs, angle_axs, r, a );
     }
   
     ray_axs.getProperties().store( DAxis.P_MIN, 0.0f );
     ray_axs.getProperties().store( DAxis.P_MAX, 1.1f*rmax );
     ray_axs.getProperties().store( DAxis.P_STEP_LINE, rmax/5 );
     angle_axs.getProperties().store( DAxis.P_MIN, 0.0f );
     angle_axs.getProperties().store( DAxis.P_MAX, 5.0f );
     angle_axs.getProperties().store( DAxis.P_STEP_LINE, Math.PI/8 );
     
     return gph;
   }
  }
  
  private static class RenderGraphic implements MouseListener, MouseMoveListener, PaintListener
  {
    private Canvas canvas_;
    private IChartRender render_;
    private DGraphic gph_;
    private boolean mouse_down_=false;
    private int mx_=0, my_=0;
    
    public RenderGraphic( Canvas _cvs )
    {
      canvas_=_cvs; 
      canvas_.addMouseListener( this );
      canvas_.addMouseMoveListener( this );
      canvas_.addPaintListener( this );
    }
    
    public void setGraphic( DGraphic _g )
    {
      gph_ = _g;
      render_ = DChartRenderRegistry.GetRender( _g.getGraphicType() );
      canvas_.redraw();
    }
    
    public void mouseMove(MouseEvent e)
    {
      if( render_==null || gph_==null ) return ;
      if( mouse_down_ &&( gph_.isGraphicType( DGraphic.T_XYZ )
                      || gph_.isGraphicType( DGraphic.T_SURFACE)) )
      {
        move3DView(e);
        return;
      }
          GC gc = new GC(canvas_);   
          Rectangle r = canvas_.getClientArea();
          IChartLocation loc = render_.locate( e.x,e.y, gc, r, gph_ );
//System.out.println("GOT LOC="+loc);          
          if( loc!=null && loc.getItem() instanceof DPoint )
          {
            DPoint pt = (DPoint)loc.getItem();
            DCurve crv= (DCurve)pt.getParent();
            DGraphic gph=(DGraphic)crv.getParent();
            String name = crv.getName();
            if( name==null ) name=""; else name=name+": ";
            for( IDItem item=gph.getFirstChild(); item!=null; item=item.getNext() )
            {
              if( item instanceof DAxis )
              {
                DAxis axis=(DAxis)item;
                DCoord crd= pt.getCoord( axis );
                if( crd!=null )
                {
                  name = name + Float.toString( crd.getValue() )+"  ";
                }
              }
            }
            canvas_.setToolTipText( name );
          } else {
            canvas_.setToolTipText( null );
          }
          gc.dispose();
          gc = null;
        }
    public void mouseDoubleClick(MouseEvent e){}

    public void mouseDown(MouseEvent e)
    {
      mouse_down_=true;
      mx_=e.x;
      my_=e.y;      
    }
    public void mouseUp(MouseEvent e)
    {
      mouse_down_=false;
    }     
    protected void move3DView(MouseEvent e)
    {  
      if( render_==null || gph_==null ) return ;

      float phi   = gph_.getProperties().get( DGraphic.P_XYZ_PHI, 0.3f);
      float theta = gph_.getProperties().get( DGraphic.P_XYZ_THETA, 0.2f);
      int dx = e.x -mx_;
      int dy = e.y -my_;
      
      phi   += 0.02f*dy;
      theta -= 0.02f*dx;

      gph_.getProperties().store( DGraphic.P_XYZ_PHI,  phi    );
      gph_.getProperties().store( DGraphic.P_XYZ_THETA, theta );
      mx_ = e.x;
      my_ = e.y;    
      canvas_.redraw();  
    }
    
    private Image img_;
    public void paintControl(PaintEvent e)
    {
       if( render_==null || gph_==null ) return ;
       Rectangle r = canvas_.getClientArea();
       if( img_==null || img_.getBounds().width!=r.width || img_.getBounds().height!=r.height)
       {
         if( img_!=null ) img_.dispose();
         img_ = new Image(null,r.width,r.height);
       }
       GC gc = new GC( img_ );
       render_.draw( gc, r, gph_ );
       e.gc.drawImage( img_, 0, 0 );
       gc.dispose();
    }
   }
   
   /* Layout horizontally children control, all width of them (but not the last one)
    * is limited (as the FillLayout doesn't do it).
    */
   private static class HLimLayout extends Layout
   {
     private int lim_;
     public HLimLayout(int _lim){lim_=_lim;}
     protected Point computeSize(Composite composite, int wHint, int hHint, boolean flushCache)
     {
       return new Point( 100, hHint );
     }

     protected void layout(Composite composite, boolean flushCache)
     {
       Rectangle r = composite.getClientArea();
       Control c[] = composite.getChildren();
       if( c==null ) return ;
       int x = r.x;
       for( int i=0; i<c.length; ++i )
       {
         if( (i+1)==c.length)
         {
           c[i].setBounds( x, r.y, r.x+r.width-x, r.height );
         } else {
           c[i].setBounds( r.x, r.y, lim_, r.height );
           x += lim_;
         }
       }
      
     }
   }
   
   public static void main( String[] argv )
   {
     //This is very important to register the render for new graphic type.
     DChartRenderRegistry.AddRender( new PolarChartRender() );
     
     Display disp = Display.getDefault();
     final Shell shell = new Shell(disp, SWT.DIALOG_TRIM|SWT.RESIZE );
     shell.setText( "Test ApiDoc: User Graphic" );
     shell.setLayout(new HLimLayout( 150 ));

     Composite cb = new Composite( shell, SWT.NONE );
     cb.setLayout( new FillLayout( SWT.VERTICAL ));
     Group gb = new Group( cb, SWT.NONE);
     gb.setLayout( new FillLayout(SWT.VERTICAL));
     gb.setText("Graphic Type:");
     final Button btn[] = new Button[2];
     int b=-1;

     btn[++b] = new Button( gb, SWT.RADIO ); btn[b].setText("XY (standard)");
     btn[b].setData( new BuildXYGraphic());     

     btn[++b] = new Button( gb, SWT.RADIO ); btn[b].setText("Polar (user defined)");
     btn[b].setData( new BuildPolarGraphic());     
       
     final Canvas canvas = new Canvas( shell, SWT.NO_BACKGROUND );
     final RenderGraphic rg = new RenderGraphic( canvas );
    
     //just on first button is enough
     SelectionListener listen_button= new SelectionListener(){
       public void widgetSelected(SelectionEvent e)
       {   
         for(int i=0; i<btn.length; ++i)
         {
           if( btn[i].getSelection())
           {
             DGraphic g = ((IBuildGraphic)btn[i].getData()).buildGraphic();
             rg.setGraphic( g );
             break;            
           }
         }
       }
 
       public void widgetDefaultSelected(SelectionEvent e){} //never called
     };
     
     //set selection listeners to all buttons...
     for( int i=0; i<btn.length; ++i )
     {
       btn[i].addSelectionListener( listen_button );
     }
         
     shell.setSize( 600, 500 );
     shell.open();
 
     
      while (!shell.isDisposed())
        if (!disp.readAndDispatch())
          disp.sleep();
   }
}
